#! /usr/bin/perl

use strict;
use Cwd;
use Getopt::Long;
use File::Basename;
use Socket;
use Sys::Hostname;

$| = 1;

my $MIN_JAVA_VERSION = "1.6";
my @JAVA_VER = split (/\./, $MIN_JAVA_VERSION);

my $CWD = getcwd ();

# This is the directory where the PARAMETERS, and the configure script for the
# DB, file can be found.
my $ORACLE_DIR = "src/database/oracle";

# these are the default values for each option.
my $help             = '';
my $always_ask       = '';
my $skip_questions   = '';
my $prefix           = $ENV{NETDB_PREFIX} || $ENV{NETDB_HOME} || "/usr/local/netdb";
my $clprefix         = $ENV{NETDB_CLPREFIX}                   || "";                    # $prefix is prepended to this value
my $dprefix          = $ENV{NETDB_DPREFIX}                    || "docs";                # $prefix is prepended to this value
my $jdkhome          = $ENV{NETDB_JDKHOME}                    || "/opt/jdk1.6.0";
my $antlr            = $ENV{NETDB_ANTLR}                      || antlr_search ();
my ($antlr_jar, $antlr_script); # these are used in trying to guess what the user downloaded.
my $jetty            = $ENV{NETDB_JETTY_HOME}                 || "$CWD/packages/jetty";
my $jcl              = $ENV{NETDB_JCL}                        || "$CWD/packages/commons-logging.jar";
my $parser           = $ENV{NETDB_PARSER}                     || "$CWD/packages/parser.jar";
my $field2sql        = $ENV{NETDB_FIELD2SQL}                  || "$CWD/packages/field2sql_parser.jar";
my $dnsjava          = $ENV{NETDB_DNSJAVA}                    || "$CWD/packages/dnsjava-1.4.3.jar";
my $oro              = $ENV{NETDB_ORO}                        || "$CWD/packages/oro.jar";
my $ora_lib          = $ENV{NETDB_ORA_LIB}                    || "$CWD/packages/ojdbc14.jar";
my $cc               = $ENV{NETDB_CC}                         || search_for ("gcc");
my $install          = $ENV{NETDB_INSTALL}                    || search_for ("install");
my $cp               = $ENV{NETDB_CP}                         || search_for ("cp");
my $gnu_tar          = $ENV{NETDB_GNU_TAR}                    || search_for ("tar");
my $perl             = $ENV{NETDB_PERL}                       || search_for ("perl");
my $perldoc          = $ENV{NETDB_PERLDOC}                    || search_for ("perldoc");
my $bash             = $ENV{NETDB_BASH}                       || search_for ("bash");
my $zsh              = $ENV{NETDB_ZSH}                        || search_for ("zsh");
my $sed              = $ENV{NETDB_SED}                        || search_for ("sed");
my $gzip             = $ENV{NETDB_GZIP}                       || search_for ("gzip");
my $db_host          = $ENV{NETDB_DB_HOST}                    || "";
my $db_port          = $ENV{NETDB_DB_PORT}                    || "";
my $db_instance      = $ENV{NETDB_INSTANCE}                   || "";
my $db_tablespace    = $ENV{NETDB_TABLESPACE}                 || "";
my $db_dba           = $ENV{NETDB_DBA}                        || "";
my $db_dba_pass      = $ENV{NETDB_DBA_PASS}                   || "";
my $db_schema        = $ENV{NETDB_DB_SCHEMA}                  || "NetDB";
my $db_schema_pass   = $ENV{NETDB_DB_SCHEMA_PASS}             || "";
my $db_client_user   = $ENV{NETDB_DB_USER}                    || "netdb_client";
my $db_client_pass   = $ENV{NETDB_CLIENT_PASS}                || "";
my $db_search_user   = $ENV{NETDB_SEARCH_USER}                || "netdb_search";
my $db_search_pass   = $ENV{NETDB_SEARCH_PASS}                || "";
my $db_guest_user    = $ENV{NETDB_GUEST_USER}                 || "netdb_guest";
my $db_guest_pass    = $ENV{NETDB_GUEST_PASS}                 || "";
my $rmi_server       = $ENV{NETDB_RMI_SERVER}                 || "localhost";
my $http_port        = $ENV{NETDB_HTTP_PORT}                  || "8080";
my $whois_cli_port   = $ENV{NETDB_WHOIS_CLI_PORT}             || "43";
my $whois_html_port  = $ENV{NETDB_WHOIS_HTML_PORT}            || "8080";
my $reg_port         = $ENV{NETDB_RMI_REG_PORT}               || "20005";
my $principal        = $ENV{NETDB_PRINCIPAL}                  || "";
my $keytab           = $ENV{NETDB_KEYTAB}                     || "";
my $cheat_uid        = $ENV{NETDB_CHEAT_UID}                  || "";
my $host_ip          = $ENV{NETDB_HOST_IP}                    || "127.0.0.1";
my $sel_tests        = $ENV{NETDB_ENABLE_SELENIUM_TEST}       || "false";
my $test_location    = $ENV{NETDB_TEST_LOCATION}              || "test/";               # $prefix is prepended to this value

# these are keys used for defining each option. $VARIABLE is the only required option.
my $PROMPT_FOR = "prompt_for";                   # (0|1)        The value is prompted for when running the script
my $PROMPT     = "prompt";                       # [text]       The prompt text
my $INFO       = "info";                         # [text]       info is only printed out once, before the prompt
my $ECHO       = "echo_user_input";              # (0|1)        echo (or don't) the characters the user types for this value
my $OPT_TYPE   = "long_opt_type";                # [text]       The option type, as defined by longopt
my $CHECK      = "check";                        # [code]       This block of code is run to verify the value that the user entered
my $POST_CODE  = "post_code";                    # [code]       This block of code is run once a valid value has been entered.
my $VARIABLE   = "variable";                     # [scalar ref] The value the user entered/default value is saved in the reference
my $SECTION    = "help_section";                 #              The help section should be selected from the variables below. It determines which section the variable will be printed in when displaying the help.
my $HELP       = "help_text";                    # [text]       The text printed when "?" is entered or the --help option is used.
my $ORDER      = "display_order";                # [number]     This is the order the questions will be prompted for and the order the option will be displayed in the help.  Actual order is undefined when two options have the same value.
my $SETTINGS   = "settings_variable";            # [text]       This is the variable name as defined in the SETTINGS file.
my $PARAM_VAR  = "parameter_file_variable_name"; # [text]       This is the variable name as defined in the PARAMETERS file.

# help sections
my $FLAG_SECTION   = "Options:";
my $PREFIX_SECTION = "Intallation Directories:";
my $OPTION_SECTION = "Variables:";
my @HELP_SECTIONS  = ($FLAG_SECTION, $PREFIX_SECTION, $OPTION_SECTION);

my $order = 0;
my %VAR_INFO; # declared on its own because it is referenced in the definition.

# The keys for %VAR_INFO are the command_line arguments.
%VAR_INFO =
(
    "help" =>
    {
        $PROMPT_FOR => 0,
        $VARIABLE   => \$help,
        $SECTION    => $FLAG_SECTION,
        $HELP       => "Show this help.",
        $ORDER      => $order++,
    },

    "always_ask" =>
    {
        $PROMPT_FOR => 0,
        $VARIABLE   => \$always_ask,
        $SECTION    => $FLAG_SECTION,
        $HELP       => "Always prompt for every question.",
        $ORDER      => $order++,
    },

    "prefix" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Installation directory for NetDB server",
        $ECHO       => 1,
        $SETTINGS   => "NETDB_PREFIX",
        $VARIABLE   => \$prefix,
        $POST_CODE  => sub
        {
            if ($clprefix      !~ /^\//) { $clprefix      = $prefix; }
            if ($dprefix       !~ /^\//) { $dprefix       = $prefix."/".$dprefix;        $dprefix =~ s/\/\/+/\//g; }
            if ($test_location !~ /^\//) { $test_location = $prefix."/".$test_location;  $test_location =~ s/\/\/+/\//g; }
        },
        $SECTION    => $PREFIX_SECTION,
        $HELP       => "This is the directory where the NetDB servers will be installed.  If it doesn't exist, it will be created upon installation.",
        $ORDER      => $order++,
    },

    "clprefix" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Installation directory for NetDB client",
        $ECHO       => 1,
        $SETTINGS   => "NETDB_CLPREFIX",
        $VARIABLE   => \$clprefix,
        $CHECK      => undef,
        $SECTION    => $PREFIX_SECTION,
        $HELP       => "This is the directory where the NetDB client will be installed.  If it doesn't exist, it will be created upon installation.",
        $ORDER      => $order++,
    },

    "dprefix" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Installation directory for javadoc files",
        $ECHO       => 1,
        $SETTINGS   => "DOCS_TARGET",
        $VARIABLE   => \$dprefix,
        $CHECK      => undef,
        $SECTION    => $PREFIX_SECTION,
        $HELP       => "This is the directory where the javadocs will be installed.  If it doesn't exist, it will be created upon installation.",
        $ORDER      => $order++,
    },

    "jdkhome" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Home directory of JDK version $MIN_JAVA_VERSION or higher",
        $ECHO       => 1,
        $SETTINGS   => "JAVA_HOME",
        $VARIABLE   => \$jdkhome,
        $CHECK      => \&checkJDK,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "jetty" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to the Jetty library",
        $ECHO       => 1,
        $SETTINGS   => "JETTY_JAR1",
        $VARIABLE   => \$jetty,
        $CHECK      => \&checkJetty,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "jcl" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to Jakarta Commons Logging library",
        $ECHO       => 1,
        $SETTINGS   => "JCL",
        $VARIABLE   => \$jcl,
        $CHECK      => \&checkExistFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "antlr" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to antlr directory",
        $ECHO       => 1,
        $SETTINGS   => "ANTLR_JAR",
        $VARIABLE   => \$antlr,
        $CHECK      => \&check_antlr,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "dnsjava" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to dnsjava library",
        $ECHO       => 1,
        $SETTINGS   => "DNSJAVA",
        $VARIABLE   => \$dnsjava,
        $CHECK      => \&checkExistFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "oro" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to jakarta-oro library",
        $ECHO       => 1,
        $SETTINGS   => "ORO",
        $VARIABLE   => \$oro,
        $CHECK      => \&checkExistFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "ora_lib" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to orcale library",
        $ECHO       => 1,
        $SETTINGS   => "ORA_LIB",
        $VARIABLE   => \$ora_lib,
        $CHECK      => \&checkExistFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "parser" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to the NetDB parser library",
        $ECHO       => 1,
        $SETTINGS   => "PARSER",
        $VARIABLE   => \$parser,
        $CHECK      => \&checkExistFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "field2sql" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to the NetDB field2sql library",
        $ECHO       => 1,
        $SETTINGS   => "FIELD2SQL",
        $VARIABLE   => \$field2sql,
        $CHECK      => \&checkExistFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "gnu_tar" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to gnu tar",
        $ECHO       => 1,
        $SETTINGS   => "GNUTAR",
        $VARIABLE   => \$gnu_tar,
        $CHECK      => \&checkExecutableFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "perl" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to perl",
        $ECHO       => 1,
        $SETTINGS   => "PERL",
        $VARIABLE   => \$perl,
        $CHECK      => \&checkExecutableFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "perldoc" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to perldoc",
        $ECHO       => 1,
        $SETTINGS   => "PERLDOC",
        $VARIABLE   => \$perldoc,
        $CHECK      => \&checkExecutableFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "bash" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full bath to bash",
        $ECHO       => 1,
        $SETTINGS   => "BASH",
        $VARIABLE   => \$bash,
        $CHECK      => \&checkExecutableFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "zsh" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full bath to zsh",
        $ECHO       => 1,
        $SETTINGS   => "ZSH",
        $VARIABLE   => \$zsh,
        $CHECK      => \&checkExecutableFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "sed" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to gnu sed",
        $ECHO       => 1,
        $SETTINGS   => "SED",
        $VARIABLE   => \$sed,
        $CHECK      => \&checkExecutableFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "cc" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to a C compiler",
        $ECHO       => 1,
        $SETTINGS   => "CC",
        $VARIABLE   => \$cc,
        $CHECK      => \&checkExecutableFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "cp" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to cp",
        $ECHO       => 1,
        $SETTINGS   => "CP",
        $VARIABLE   => \$cp,
        $CHECK      => \&checkExecutableFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "install" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to install",
        $ECHO       => 1,
        $SETTINGS   => "INSTALL",
        $VARIABLE   => \$install,
        $CHECK      => \&checkExecutableFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "gzip" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Full path to a zip utility",
        $ECHO       => 1,
        $SETTINGS   => "GZIP",
        $VARIABLE   => \$gzip,
        $CHECK      => \&checkExecutableFile,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "test_uid" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Default user to log in as (This should only be used for testing)",
        $ECHO       => 1,
        $SETTINGS   => "CHEAT_USERID",
        $VARIABLE   => \$cheat_uid,
        $CHECK      => undef,
        $POST_CODE  => sub
        {
            if ($cheat_uid eq undef) {
                $VAR_INFO{host_ip}->{$PROMPT_FOR} = 0; }
            else
            {
                if ($host_ip eq "127.0.0.1")
                {
                    my $hostname = hostname ();
                       $host_ip  = inet_ntoa (scalar (gethostbyname ($hostname)));
                }
            }
        },
        $SECTION    => $OPTION_SECTION,
        $HELP       => "Enabling this provides a way around authentication.  It should only be used in\n".
                       "testing.  If enabled, you can set the active user when loading the web front end by\n".
                       "specifying the user netid in the url. (http://netdb-url/login?uid=willbarber)",
        $ORDER      => $order++,
    },

    "selenium_tests" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Enable selenium tests (true|false)",
        $ECHO       => 1,
        $SETTINGS   => "ENABLE_SELENIUM_TEST",
        $VARIABLE   => \$sel_tests,
        $CHECK      => undef,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "test_location" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Location of selenium test directory",
        $ECHO       => 1,
        $SETTINGS   => "TESTS_LOCATION",
        $VARIABLE   => \$test_location,
        $CHECK      => undef,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "rmi_server" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Fully qualified host name the RMI server will be running on",
        $ECHO       => 1,
        $SETTINGS   => "RMI_SERVER",
        $VARIABLE   => \$rmi_server,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "rmi_reg_port" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Port number for rmi registry",
        $ECHO       => 1,
        $SETTINGS   => "RMI_REG_PORT",
        $VARIABLE   => \$reg_port,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "host_ip" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "This is the address the http server will attempt to bind ",
        $ECHO       => 1,
        $SETTINGS   => "HTTP_BIND_ADDRESS",
        $VARIABLE   => \$host_ip,
        $CHECK      => undef,
        $SECTION    => $OPTION_SECTION,
        $HELP       => undef,
        $ORDER      => $order++,
    },

    "http_port" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Port number for http server",
        $ECHO       => 1,
        $SETTINGS   => "HTTP_PORT",
        $VARIABLE   => \$http_port,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "whois_cli_port" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Port number for whois cli server",
        $ECHO       => 1,
        $SETTINGS   => "WHOIS_CLI_PORT",
        $VARIABLE   => \$whois_cli_port,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "whois_html_port" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Port number for whois html server",
        $ECHO       => 1,
        $SETTINGS   => "WHOIS_HTML_PORT",
        $VARIABLE   => \$whois_html_port,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "principal" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 0,
        $PROMPT     => "K5 principal to be used by the RMI server",
        $ECHO       => 1,
        $SETTINGS   => "PRINCIPAL",
        $VARIABLE   => \$principal,
        $CHECK      => undef,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "keytab" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 0,
        $PROMPT     => "Full path to the keytab file associated with the principal",
        $ECHO       => 1,
        $SETTINGS   => "KEYTAB",
        $VARIABLE   => \$keytab,
        $CHECK      => undef,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "db_host" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $INFO       => "The next several options relate to the database.  Your DBA should be able to\n".
                       "provide you with the necessary information",
        $PROMPT     => "Fully qualified host name running the Oracle listener",
        $ECHO       => 1,
        $SETTINGS   => "DB_URL",
        $VARIABLE   => \$db_host,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "db_port" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Port the Oracle listener is using",
        $ECHO       => 1,
        $VARIABLE   => \$db_port,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
    },

    "db_tablespace" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Tablespace that will be used when creating the database",
        $ECHO       => 1,
        $VARIABLE   => \$db_tablespace,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
        $PARAM_VAR  => "TABLESPACE",
    },

    "db_instance" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Instance name of the database, according to the tnsnames.ora file",
        $ECHO       => 1,
        $VARIABLE   => \$db_instance,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
        $PARAM_VAR  => "SERVER",
    },

    "dba" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Oracle account with DBA privileges.  Alternatively, 'echo' will create a script\n".
                       "which you or your DBA can review, before running it manually.",
        $ECHO       => 1,
        $VARIABLE   => \$db_dba,
        $CHECK      => \&checkAnyValue,
        $POST_CODE  => sub { $VAR_INFO{dba_pass}->{$PROMPT_FOR} = 0 if $db_dba eq "echo"; },
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
        $PARAM_VAR  => "DBA_USER",
    },

    "dba_pass" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Password for the DBA account",
        $ECHO       => 0,
        $VARIABLE   => \$db_dba_pass,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
        $PARAM_VAR  => "DBA_PASSWORD",
    },

    "db_schema" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "DB schema to be used when connected to database",
        $ECHO       => 1,
        $VARIABLE   => \$db_schema,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
        $PARAM_VAR  => "DBO_USER",
    },

    "db_schema_pass" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "DB schema password to be used when creating the database",
        $ECHO       => 0,
        $VARIABLE   => \$db_schema_pass,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
        $PARAM_VAR  => "DBO_PASSWORD",
    },

    "db_client_user" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 0,
        $PROMPT     => "User to be used by the NetDB client",
        $ECHO       => 1,
        $VARIABLE   => \$db_client_user,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
        $PARAM_VAR  => "CLIENT_USER",
    },

    "db_client_pass" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Password to be used by NetDB DB client user",
        $ECHO       => 0,
        $VARIABLE   => \$db_client_pass,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
        $PARAM_VAR  => "CLIENT_PASSWORD",
    },

    "db_search_user" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 0,
        $PROMPT     => "DB user for search operations",
        $ECHO       => 1,
        $VARIABLE   => \$db_search_user,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
        $PARAM_VAR  => "SEARCH_USER",
    },

    "db_search_pass" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Password to be used by NetDB search user",
        $ECHO       => 0,
        $VARIABLE   => \$db_search_pass,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
        $PARAM_VAR  => "SEARCH_PASSWORD",
    },

    "db_guest_user" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 0,
        $PROMPT     => "DB user for guest account",
        $ECHO       => 1,
        $VARIABLE   => \$db_guest_user,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
        $PARAM_VAR  => "GUEST_USER",
    },

    "db_guest_pass" =>
    {
        $OPT_TYPE   => "s",
        $PROMPT_FOR => 1,
        $PROMPT     => "Password to be used by NetDB guest user",
        $ECHO       => 0,
        $VARIABLE   => \$db_guest_pass,
        $CHECK      => \&checkAnyValue,
        $SECTION    => $OPTION_SECTION,
        $HELP       => "",
        $ORDER      => $order++,
        $PARAM_VAR  => "GUEST_PASSWORD",
    },
);

# create a hash, for the command line options, that longopt can understand.
my %options;
foreach my $key (keys (%VAR_INFO))
{
    my $opt_and_type = $key;

    if ($VAR_INFO{$key}->{$OPT_TYPE} ne undef) {
        $opt_and_type .= "=".$VAR_INFO{$key}->{$OPT_TYPE}; }

    $options{$opt_and_type} = $VAR_INFO{$key}->{$VARIABLE};
}

parse_SETTINGS_file ();

# --skip_questions is an undocumented option.  Using it will skip the "read the INSTALL" questions.
GetOptions (skip_questions => \$skip_questions, %options) || die "Error parsing command line options.\n";
help () if $help;

if (! $skip_questions)
{
    foreach my $question ( "Have you read the instructions in the INSTALL file?",
                           "Did you really?" )
    {
        print $question, ": ";
        my $response = <STDIN>;
           $response =~ s/(^\s*)|(\s*$)//g;

        if ($response !~ /^y(es)?$/i) {
            die "It would be a good idea to read them before continuing.\n"; }
    }
}

# Attempt to get values for the database from the PARAMETERS file.  This is done
# by sourcing the PARAMETERS file and printing the created shell variables.
if (-e "$ORACLE_DIR/PARAMETERS")
{
    print "PARAMETERS file found.  Retrieving database info from it.  The values in the\n".
          "PARAMETERS file will override any command line arguments.\n";

    my ($variable_dump, %param_mappping);
    foreach my $key (keys (%VAR_INFO))
    {
        if ($VAR_INFO{$key}->{$PARAM_VAR} ne undef)
        {
            my $param_var = $VAR_INFO{$key}->{$PARAM_VAR};

            $param_mappping{$param_var} = $VAR_INFO{$key}->{$VARIABLE};

            $variable_dump .= "echo '$param_var='\$$param_var;";
        }
    }

    foreach my $var (`source $ORACLE_DIR/PARAMETERS; $variable_dump`)
    {
        $var =~ /^(.+?)=(.*)\s*/;

        my ($name, $value) = ($1, $2);

        ${$param_mappping{$name}} = $value;
    }
}

# prompt the user for all the values.
foreach my $key (sort { $VAR_INFO{$a}->{$ORDER} <=> $VAR_INFO{$b}->{$ORDER} } keys (%VAR_INFO))
{
    my $hash_ref = $VAR_INFO{$key};

    if (! $hash_ref->{$PROMPT_FOR}) {
        next; }

    get_value ($key);

    if ($hash_ref->{$POST_CODE} ne undef) {
        $hash_ref->{$POST_CODE} (); }
}

print "\nWriting SETTINGS file ...\n";

open SETTINGS, ">SETTINGS" or die "Failed to open file SETTINGS for writting\n";
print SETTINGS <<EOF;
export NETDB_PREFIX         := $prefix
export NETDB_CLPREFIX       := $clprefix
export DOCS_TARGET          := $dprefix
export JAVA_HOME            := $jdkhome
export JAVA                 := \$(JAVA_HOME)/bin/java
export JVM                  := \$(JAVA)
export JAVAC                := \$(JAVA_HOME)/bin/javac
export JAVAH                := \$(JAVA_HOME)/bin/javah
export RMIC                 := \$(JAVA_HOME)/bin/rmic
export JAR                  := \$(JAVA_HOME)/bin/jar
export JETTY_JAR1           := $jetty/javax.servlet.jar
export JETTY_JAR2           := $jetty/org.mortbay.jetty.jar
export JCL                  := $jcl
export ANTLR                := $antlr_script
export ANTLR_JAR            := $antlr_jar
export DNSJAVA              := $dnsjava
export ORO                  := $oro
export ORA_LIB              := $ora_lib
export PARSER               := $parser
export FIELD2SQL            := $field2sql
export GNUTAR               := $gnu_tar
export PERL                 := $perl
export PERLDOC              := $perldoc
export BASH                 := $bash
export ZSH                  := $zsh
export SED                  := $sed
export CC                   := $cc
export CP                   := $cp
export INSTALL              := $install
export GZIP                 := $gzip
export CHEAT_USERID         := $cheat_uid
export ENABLE_SELENIUM_TEST := $sel_tests
export TESTS_LOCATION       := $test_location
export RMI_SERVER           := $rmi_server
export RMI_REG_PORT         := $reg_port
export HTTP_BIND_ADDRESS    := $host_ip
export HTTP_PORT            := $http_port
export WHOIS_CLI_PORT       := $whois_cli_port
export WHOIS_HTML_PORT      := $whois_html_port
export PRINCIPAL            := $principal
export KEYTAB               := $keytab
export DB_URL               := jdbc:oracle:thin:\@$db_host:$db_port:$db_instance
export DB_SCHEMA            := $db_schema
export DB_USER              := $db_client_user
export DB_PASSWD            := $db_client_pass
export DB_SEARCH_USER       := $db_search_user
export DB_SEARCH_PASSWD     := $db_search_pass
EOF

print "\nCreating lib/ directory...\n";
mkdir "lib/";

print "Creating PARAMETERS file...\n";
system ("(cd $ORACLE_DIR; ./configure --instance    \"$db_instance\"".
                                    " --tablespace  \"$db_tablespace\"".
                                    " --dba         \"$db_dba\"          --dba_password    \"$db_dba_pass\"".
                                    " --schema      \"$db_schema\"       --schema_password \"$db_schema_pass\"".
                                    " --client_user \"$db_client_user\"  --client_password \"$db_client_pass\"".
                                    " --search_user \"$db_search_user\"  --search_password \"$db_search_pass\"".
                                    " --guest_user  \"$db_guest_user\"   --guest_password  \"$db_guest_pass\")");

# get_value ( $key - the key as defined in $VAR_INFO )
sub get_value
{
    my ($key) = @_;

    my $prompt   = $VAR_INFO{$key}->{$PROMPT};
    my $variable = $VAR_INFO{$key}->{$VARIABLE};
    my $echo     = $VAR_INFO{$key}->{$ECHO};
    my $check    = $VAR_INFO{$key}->{$CHECK};
    my $info     = $VAR_INFO{$key}->{$INFO};

    my $input;
    my $value_state = 1;
    my $already_printed_info = 0;
    do
    {
        if (! ${$variable} || $always_ask || ! $value_state)
        {
            if (! $already_printed_info && $info)
            {
                $already_printed_info = 1;
                print "$info\n";
            }

            print "$prompt [${$variable}]: ";

            if (! $echo) {
                system ("stty -echo"); }

            $input = <STDIN>;
            $input =~ s/(^\s+)|(\s+$)//g;

            if ($input eq "?")
            {
                if ($VAR_INFO{$key}->{$HELP}) {
                    print "\n", $VAR_INFO{$key}->{$HELP}, "\n"; }
            }
            else {
                ${$variable} = $input if $input; }

            if (! $echo)
            {
                print "\n";
                system ("stty echo");
            }
        }

        if ($input ne "?" && $check != undef) {
            $value_state = &$check (${$variable}); }
    } while (! $value_state || $input eq "?");
}

sub parse_SETTINGS_file
{
    my %settings_mappping;
    foreach my $key (keys (%VAR_INFO))
    {
        if ($VAR_INFO{$key}->{$SETTINGS} ne undef) {
            $settings_mappping{$VAR_INFO{$key}->{$SETTINGS}} = $VAR_INFO{$key}->{$VARIABLE}; }
    }

    open (SETTINGS, "SETTINGS") || return;
    foreach my $line (<SETTINGS>)
    {
        $line =~ /export (.+?) +:= +(.+)/;
        my ($variable, $value) = ($1, $2);

        ${$settings_mappping{$variable}} = $value;
    }
    close (SETTINGS);

    if (${$settings_mappping{DB_URL}} =~ /jdbc:oracle:thin:\@(.+?):(.+?):(.+?)/) {
        ($db_host, $db_port, $db_instance) = ($1, $2, $3); }

    if (${$settings_mappping{JETTY_JAR1}} =~ /\.jar$/) {
        $jetty =~ s/(\/.*\/).+?\.jar$/\1/; }
}

sub print_check
{
    printf ("%-45s", "Checking ".$_[0]."... ");
}

sub search_for
{
    my ($command) = @_;

    foreach my $path (split (":", $ENV{PATH}))
    {
        if (-e "$path/$command") {
            return "$path/$command"; }
    }

    return "/bin/$command"            if (-e "/bin/$command");
    return "/sbin/$command"           if (-e "/sbin/$command");
    return "/usr/bin/$command"        if (-e "/usr/bin/$command");
    return "/usr/sbin/$command"       if (-e "/usr/sbin/$command");
    return "/usr/local/bin/$command"  if (-e "/usr/local/bin/$command");

    # too bad, we couldn't find it...
    return $command;
}

sub antlr_search
{
    # they have the full install, in the default location.
    if (-e "/usr/local/lib/antlr.jar" && -e "/usr/local/bin/antlr") {
        return "/usr/local/"; }

    # they have just the jar, as installed by aptitude
    if (-e "/usr/share/java/antlr.jar") {
        return "/usr/share/java/antlr.jar"; }

    # they have just the jar, and dropped it in the packages folder
    if (-e "packages/antlr.jar") {
        return "packages/antlr.jar"; }

    # who knows where they have it?
    return "/usr/local/";
}

sub checkJDK
{
    print_check ("JDK");

    checkDir ($jdkhome)                 || return;
    checkDir ("$jdkhome/bin")           || return;
    checkFile ("$jdkhome/bin/java")     || return;
    checkExecFile ("$jdkhome/bin/java") || return;

    my $found_version = '';

    open STR, "$jdkhome/bin/java -version 2>&1 | " or die "Failed to execute java.\n";

    while (<STR>)
    {
        if ($_ =~ /java version "((\d+)\.(\d+)\..*)"/)
        {
            if ($2 < $JAVA_VER[0] || $3 < $JAVA_VER[1])
            {
                print "Failed. JDK version should be $MIN_JAVA_VERSION or higher. Found $1.\n";
                close STR;
                return;
            }
            else {
                $found_version = $1; }
        }
    }

    close STR;

    if (! $found_version)
    {
        print "Failed. Could not determine JDK version.\n";
        print "If you're sure that '$jdkhome/bin/java' is a valid version of java, I can go ahead and try to use it.\n";

        my $force;
        do
        {
            print "Force the use of '$jdkhome/bin/java'? (y/n) ";
            $force = <STDIN>;
            chomp ($force);
        } while ($force !~ /^(y(es)?|no?)$/i);

        if ($force =~ /y(es)?/i) {
            return 1; }

        return 0;
    }

    print "OK. Found version $found_version\n";

    return 1;

} # end of checkJDK ()


sub checkJetty
{
    print_check ("Jetty");

    checkDir ($jetty)                          || return;
    checkFile ("$jetty/javax.servlet.jar")     || return;
    checkFile ("$jetty/org.mortbay.jetty.jar") || return;

    print "OK.\n";
    return 1;
}

sub check_antlr
{
    my $dir = shift;

    print_check ("antlr");

    if ($antlr_script eq undef) {
        $antlr_script = "/usr/local/bin/antlr.jar"; }

    # if they have the full install, they'll probably type in the base directory.
    if (-d $dir)
    {
        # however, they may just be exceeding lazy and only type in the directory to find antlr.jar
        if (-e "$dir/antlr.jar") {
            $antlr_jar    = "$dir/antlr.jar"; }
        else
        {
            $antlr_jar    = "$dir/lib/antlr.jar";
            $antlr_script = "$dir/bin/antlr";
        }
    }
    elsif ($dir =~ /\.jar/) { # if they downloaded just the jar, we probably won't find the antlr script
        $antlr_jar = $dir; }

    checkFile ($antlr_jar) || return;

# This is needed to build one of the parsers, but not NetDB
#   checkFile ($antlr_script) || return;

    print "OK\n";
    return 1;
}

sub checkExistFile
{
    my $file = shift;

    print_check (basename ($file));
    checkFile ($file) || return;
    print "OK.\n";

    return 1;
}

sub checkExecutableFile
{
    my $file = shift;

    print_check (basename ($file));

    checkFile ($file)     || return;
    checkExecFile ($file) || return;
    print "OK.\n";

    return 1;
}

sub checkExistDir
{
    my $dir = shift;

    print_check ($dir);
    checkDir ("$dir") || return;
    print "OK.\n";

    return 1;
}

sub checkDir
{
    my $dir = shift;

    if (! -e $dir)
    {
        print "Failed. Directory $dir does not exist.\n";
        return;
    }

    return 1;
}

sub checkFile
{
    my $file = shift;

    if (! -f $file)
    {
        print "Failed. File $file does not exist.\n";
        return;
    }

    return 1;
}

sub checkExecFile
{
    my $file = shift;

    if (! -x $file)
    {
        print "Failed. File $file is not executable.\n";
        return;
    }

    return 1;
}

sub checkAnyValue
{
    if ($_[0] eq undef)
    {
        print "You must enter a value.\n";
        return;
    }

    return 1;
}

sub help
{
    print "Usage: ./configure [OPTIONS]... [VAR=VALUE]...\n\n".
          "While running the script, you can type a '?' as the response to any prompt to\n".
          "display the help for that item.\n";

    foreach my $key (keys (%VAR_INFO))
    {
        if ($VAR_INFO{$key}->{$POST_CODE} ne undef) {
            $VAR_INFO{$key}->{$POST_CODE} (); }
    }

    my @sorted_keys = sort { $VAR_INFO{$a}->{$ORDER} <=> $VAR_INFO{$b}->{$ORDER} } keys (%VAR_INFO);
    foreach my $section (@HELP_SECTIONS)
    {
        print $section, "\n";

        foreach my $key (@sorted_keys)
        {
            if ($VAR_INFO{$key}->{$SECTION} eq $section)
            {
                my $printed_help = ($VAR_INFO{$key}->{$HELP} eq undef) ? $VAR_INFO{$key}->{$PROMPT} : $VAR_INFO{$key}->{$HELP};
                   $printed_help =~ s/\n/ /g;

                printf ("  --%-30s  %s", $key."=".uc ($key), $printed_help);

                if ($VAR_INFO{$key}->{$VARIABLE} =~ /scalar/i && ${$VAR_INFO{$key}->{$VARIABLE}} ne undef) {
                    print " [${$VAR_INFO{$key}->{$VARIABLE}}]"; }

                print "\n";
            }
        }
        print "\n";
    }

    exit;
}
sub print_hash
{
    my ($hash, $indent) = @_;

    foreach my $key (sort keys (%{$hash}))
    {
        print $indent,
              (($key eq undef) ? "*undef*" : $key), "  ",
              (($hash->{$key} eq undef) ? "*undef*" : $hash->{$key}), "\n";

        if ($hash->{$key} =~ /hash/i) {
            print_hash ($hash->{$key}, $indent."  "); }
        elsif ($hash->{$key} =~ /array/i) {
            print join (" :: ", @{$hash->{$key}}), "\n"; }
    }
}
