-- DNS.sql --- functions dealing with DNS names

create or replace package body DNS as

    letters constant varchar(52) := 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz';
    numbers constant varchar(10) := '0123456789';
    dash    constant varchar(1)  := '-';
    dot     constant varchar(1)  := '.';
    tilda   constant varchar(1)  := '~';

    -- check_label: return true if the input label is a valid DNS label
    function check_label(label in varchar) return boolean as
    begin

        -- DNS labels must be 63 chacters or less
        if length(label) > 63 then return(false); end if;

        -- A valid DNS label is a string containing only [-A-Za-z0-9], starting and
        -- ending with a letter or number, and containing at least one letter.

        if translate(tilda||label,tilda||letters||numbers||dash,tilda) != tilda  or
           translate(tilda||label,tilda||numbers||dash,tilda)           = tilda  or
           label like dash||'%'                                                  or
           label like '%'||dash          then return(false);
        end if;

        return(true);

    end check_label;

    -- check_name: return true if the input name is a valid fully-qualified DNS name
    function check_name(name in varchar) return boolean as
        s varchar(512) := name;
        p int := instr(s,dot);
    begin

        -- split name on dots and feed each segment to check_label

        while p > 0 loop
            if not check_label(substr(s,1,p-1)) then return(false); end if;
            s := substr(s,p+1);
            p := instr(s,dot);
        end loop;

        if length(s) > 0 and check_label(s)
            then return(true);
            else return(false);
        end if;

    end check_name;

    -- name: return the "name" part of a fully-qualified DNS name
    function name(fullname in varchar) return varchar as
    begin

        -- be reflexive with nulls and '.'
        if fullname is null or fullname = '.' then return(fullname); end if;

        -- add a trailing dot, then chop at the first dot
        return(substr(fullname||'.',1,instr(fullname||'.',dot)-1));

    end name;

    -- domain: return the "domain" part of a fully-qualified DNS name
    function domain(fullname in varchar) return varchar as
        domain varchar(512);
    begin

        -- be reflexive with nulls and '.'
        if fullname is null or fullname = '.' then return(fullname); end if;

        -- it there's a trailing dot, strip it
        domain := regexp_replace(fullname,'\.$');

        -- if there's a dot, return everything after it, otherwise return just '.'
        if instr(domain,dot) > 0
            then return(substr(domain,instr(domain,dot)+1));
            else return('.');
        end if;

    end domain;

    -- level: return the number of DNS labels in input name (assumes valid name)
    function labels(name in varchar) return int as
        s varchar(512) := name;
    begin
        -- return -1 if the input is null
        if name is null then return(-1); end if;

        -- return 0 if the input is the DNS root, '.'
        if name = '.' then return(0); end if;

        -- chop trailing dot
        if substr(name,-1,1) = '.' then s := substr(name,1,length(name)-1); end if;

        -- return number of dots plus one
        return(nvl(length(translate(s,dot||letters||numbers||dash,dot)),0)+1);

    end labels;

end DNS;
/

