-- IP.sql --- functions for manipulating IP addresses and prefixes

create or replace package body IP as

    -- convenient types for storing IP addresses
    type    IP_num    is varray(6) of number(10,0);  -- numeric (IPv4 for now, expand for IPv6)
    subtype IP_addr   is varchar(39);                -- address string
    subtype IP_prefix is varchar(43);                -- prefix string

    -- family - return IP address version
    function family(ip in varchar) return int deterministic as
    begin
        -- NULL is unacceptable
        if ip is null then
            raise_application_error(-20100, 'Invalid IP Address: NULL');
        end if;
        if instr(ip,':') = 0 then return(4); end if;
        return(6);
    end family;

    -- ipv4_to_num - verify and convert character IPv4 address to an array of integers
    --               the first 4 elements are the individual octet values and the last
    --               2 elements are the mask length and the 32-bit integer value of the
    --               IP address.  Returns false on invalid IP address.
    function ipv4_to_num ( ipaddr in varchar, ipnum out IP_num )
      return boolean deterministic as

        ip varchar(64);

        point int := 1;
        mark  int := 0;
        last  int := 1;

    begin
        -- default values
        ipnum := IP_num(0,0,0,0,32,0);

        -- strip leading and trailing spaces
        ip := trim(ipaddr);

        -- split on '/', check and set prefix length
        mark := instr(ip,'/',point);
        if mark > 0 then
            ipnum(5) := to_number(substr(ip,mark+1));
            if ipnum(5) is null or ipnum(5) < 1 or ipnum(5) > 32 then
                return(false);
            end if;
            ip := substr(ip,1,mark-1);
        end if;

        -- check and set IP address octets
        if translate(ip,'.1234567890','.') not in ('...','..','.') then
            return(false);
        end if;

        for i in 1..3 loop
            mark := instr(ip,'.',point);
            if mark = point then return(false); end if;
            if mark > 0 then
                ipnum(i) := to_number(substr(ip,point,mark-point));
                if ipnum(i) is null or ipnum(i) < 0 or ipnum(i) > 255 then
                    return(false);
                end if;
                point    := mark + 1;
                last     := i + 1;
            end if;
        end loop;

        if point <= length(ip) then
            ipnum(last) := to_number(substr(ip,point),'FM999');
            if ipnum(last) is null or ipnum(last) < 0 or ipnum(last) > 255 then
                return(false);
            end if;
        end if;

        -- calculate the integer value
        ipnum(6) := ipnum(4) + ipnum(3) * 2**8 + ipnum(2) * 2**16 + ipnum(1) * 2**24;

        return(true);

        exception
            when others then return(false);

    end ipv4_to_num;

    -- ipv4_to_bytes - break a 32-bit IP number into four 8-bit numbers
    function ipv4_to_bytes ( ip in int, ip1 out int, ip2 out int,
                                        ip3 out int, ip4 out int  )
      return boolean deterministic as

        mask int;

    begin

        mask := to_number('ff000000','xxxxxxxx');  select bitand(ip,mask) into ip1 from dual;
        mask := to_number('00ff0000','xxxxxxxx');  select bitand(ip,mask) into ip2 from dual;
        mask := to_number('0000ff00','xxxxxxxx');  select bitand(ip,mask) into ip3 from dual;
        mask := to_number('000000ff','xxxxxxxx');  select bitand(ip,mask) into ip4 from dual;

        ip1 := ip1 / 2 ** 24;
        ip2 := ip2 / 2 ** 16;
        ip3 := ip3 / 2 **  8;

        return(true);

        exception
            when others then return(false);

    end ipv4_to_bytes;

    -- ipv4_format - convert IP_num into a standard varchar IP address
    function ipv4_format
    (
      ipnum  in IP_num,
      normal in boolean,    -- return internal normalized form
      prefix in boolean     -- return full prefix, not just address
    )
      return varchar deterministic as

        output IP_prefix;
        pad_length int := 39;
        format char(5) := 'FM999';

    begin

        -- set numeric format based on 'normal'
        if normal then format := 'FM099'; end if;

        -- create IP address
        output := to_char(ipnum(1),format);
        for i in 2..4 loop
            output := output || '.' || to_char(ipnum(i),format);
        end loop;

        -- add prefix length if requested
        if prefix then output := output || '/' || to_char(ipnum(5),format); end if;

        -- right-align if requested
        if normal then
            if prefix then pad_length := 43; end if;
            output := lpad(output,pad_length);
        end if;

        return(output);

        exception
            when others then return('');

    end ipv4_format;

    -- ipv4_normalize - convert character IPv4 address/networks to canonical form
    --                  and optionally convert to network, broadcast, or netmask.
    function ipv4_normalize
    (
      ip     in varchar,
      normal in boolean,              -- return internal normalized form
      prefix in boolean,              -- return full prefix, not just address
      part   in varchar default null  -- return 'network', 'broadcast', or 'netmask'
    )
      return varchar deterministic as

        mask int;
        netcast int;
        ipnum IP_num;

    begin
        -- convert to numbers
        if not ipv4_to_num(ip,ipnum) then return(''); end if;

        -- reset ipnum to network, broadcast, or netmask if requested
        if part is not null then
            case part
                when 'network' then
                    mask :=  (2**ipnum(5) - 1) * 2**(32-ipnum(5));
                    select bitand(ipnum(6), mask)+0 into netcast from dual;
                when 'broadcast' then
                    mask := 2**(32-ipnum(5)) - 1;
                    select ipnum(6)-bitand(ipnum(6), mask)+mask into netcast from dual;
                when 'netmask' then
                    netcast := (2**ipnum(5) - 1) * 2**(32-ipnum(5));
                else
                    raise_application_error(-20095, 'Idiot programmer, bogus input: ' || part);
            end case;
            if ipnum(6) != netcast then
                ipnum(6) := netcast;
                if not ipv4_to_bytes(ipnum(6),ipnum(1),ipnum(2),ipnum(3),ipnum(4)) then
                    return('');
                end if;
            end if;
        end if;

        return(ipv4_format(ipnum,normal,prefix));

        exception
            when others then return('');

    end ipv4_normalize;

    -- normalize - return IP address in internal normalized form
    function normalize(ip in varchar) return varchar deterministic as
        ipstr IP_prefix;
    begin
        -- be reflexive with nulls
        if ip is null then return(null); end if;
        -- branch on IP version
        if family(ip) = 6 then
            raise_application_error(-20099, 'IPv6 support not implemented');
        end if;
        -- convert
        ipstr := ipv4_normalize(ip,true,(instr(ip,'/') > 0));
        if ipstr is null then
            raise_application_error(-20100, 'Invalid IP Address: ' || ip);
        end if;
        return(ipstr);
    end normalize;

    -- host - return IP address in "standard" form
    function host(ip in varchar) return varchar deterministic as
        ipstr IP_addr;
    begin
        -- be reflexive with nulls
        if ip is null then return(null); end if;
        -- branch on IP version
        if family(ip) = 6 then
            raise_application_error(-20099, 'IPv6 support not implemented');
        end if;
        -- convert
        ipstr := ipv4_normalize(ip,false,false);
        if ipstr is null then
            raise_application_error(-20100, 'Invalid IP Address: ' || ip);
        end if;
        return(ipstr);
    end host;

    -- text - return IP address in "prefix" form
    function text(ip in varchar) return varchar deterministic as
        ipstr IP_prefix;
    begin
        -- be reflexive with nulls
        if ip is null then return(null); end if;
        -- branch on IP version
        if family(ip) = 6 then
            raise_application_error(-20099, 'IPv6 support not implemented');
        end if;
        -- convert
        ipstr := ipv4_normalize(ip,false,true);
        if ipstr is null then
            raise_application_error(-20100, 'Invalid IP Address: ' || ip);
        end if;
        return(ipstr);
    end text;

    -- network - return network prefix with host bits zeroed
    function network(ip in varchar) return varchar deterministic as
        ipstr IP_prefix;
    begin
        -- be reflexive with nulls
        if ip is null then return(null); end if;
        -- branch on IP version
        if family(ip) = 6 then
            raise_application_error(-20099, 'IPv6 support not implemented');
        end if;
        -- convert
        ipstr := ipv4_normalize(ip,false,true,'network');
        if ipstr is null then
            raise_application_error(-20100, 'Invalid IP Address: ' || ip);
        end if;
        return(ipstr);
    end network;

    -- broadcast - return broadcast address
    function broadcast(ip in varchar) return varchar deterministic as
        ipstr IP_addr;
    begin
        -- be reflexive with nulls
        if ip is null then return(null); end if;
        -- branch on IP version
        if family(ip) = 6 then
            raise_application_error(-20099, 'IPv6 support not implemented');
        end if;
        -- convert
        ipstr := ipv4_normalize(ip,false,false,'broadcast');
        if ipstr is null then
            raise_application_error(-20100, 'Invalid IP Address: ' || ip);
        end if;
        return(ipstr);
    end broadcast;

    -- netmask - return the netmask
    function netmask(ip in varchar) return varchar deterministic as
        ipstr IP_addr;
    begin
        -- be reflexive with nulls
        if ip is null then return(null); end if;
        -- branch on IP version
        if family(ip) = 6 then
            raise_application_error(-20099, 'IPv6 support not implemented');
        end if;
        -- convert
        ipstr := ipv4_normalize(ip,false,false,'netmask');
        if ipstr is null then
            raise_application_error(-20100, 'Invalid IP Address: ' || ip);
        end if;
        return(ipstr);
    end netmask;

    -- masklen - return the netmask length
    function masklen(ip in varchar) return int deterministic as
        ipnum IP_num;
    begin
        -- NULL is unacceptable
        if ip is null then
            raise_application_error(-20100, 'Invalid IP Address: NULL');
        end if;
        -- branch on IP version
        if family(ip) = 6 then
            raise_application_error(-20099, 'IPv6 support not implemented');
        end if;
        -- convert to numbers
        if not ipv4_to_num(ip,ipnum) then
            raise_application_error(-20100, 'Invalid IP Address: ' || ip);
        end if;
        return(ipnum(5));
    end masklen;

    -- plus - add an integer to an IP address
    function plus(ip in varchar, addend int) return varchar deterministic as
        ipnum IP_num;
    begin
        -- NULL is unacceptable
        if ip is null then
            raise_application_error(-20100, 'Invalid IP Address: NULL');
        end if;
        -- branch on IP version
        if family(ip) = 6 then
            raise_application_error(-20099, 'IPv6 support not implemented');
        end if;
        -- convert to numbers
        if not ipv4_to_num(ip,ipnum) then
            raise_application_error(-20100, 'Invalid IP Address: ' || ip);
        end if;
        -- add
        ipnum(6) := ipnum(6) + addend;
        if ipnum(6) < 0 or ipnum(6) > 4294967295 then
            raise_application_error(-20097, 'Addition results in illegal IP address');
        end if;
        -- update ipnum and return formatted version
        if ipv4_to_bytes(ipnum(6),ipnum(1),ipnum(2),ipnum(3),ipnum(4)) then
            return(ipv4_format(ipnum,false,false));
        else
            raise_application_error(-20096, 'You can''t get here from there ...');
        end if;
    end plus;

    -- difference - calculate the difference between two IP addresses
    function difference(minuend varchar, subtrahend varchar) return int deterministic as
        x IP_num;
        y IP_num;
    begin
        -- NULL is unacceptable
        if minuend is null or subtrahend is null then
            raise_application_error(-20100, 'Invalid IP Address: NULL');
        end if;
        -- branch on IP version
        if family(minuend) = 6 or family(subtrahend) = 6 then
            raise_application_error(-20099, 'IPv6 support not implemented');
        end if;
        -- convert to numbers
        if not ipv4_to_num(minuend,x) then
            raise_application_error(-20100, 'Invalid IP Address: ' || minuend);
        end if;
        if not ipv4_to_num(subtrahend,y) then
            raise_application_error(-20100, 'Invalid IP Address: ' || subtrahend);
        end if;
        -- subtract and return result
        return(x(6) - y(6));
    end difference;

end IP;
/
