-- NetDB_AdminTeam.sql --- contains functions related to AdminTeam

create or replace package body NetDB_AdminTeam as

    -- Insert or delete an AdminTeam_Member row.
    procedure member ( p_action    in varchar, -- {'add', 'remove'}
                       p_op_uid    in int,
                       p_lock      in timestamp,
                       p_ateam_id  in int,
                       p_person_id in int )
    as
        v_person_id int := NetDB_Person.verify_or_die (p_person_id);
    begin
        locked_or_die (p_op_uid, p_lock, p_ateam_id);

        case p_action
            when 'add' then
                insert into AdminTeam_Member values (p_ateam_id, v_person_id);

            when 'remove' then
                delete from AdminTeam_Member where AdminTeam = p_ateam_id and person = v_person_id;

            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;
    exception
        when NetDB.unique_constraint_failed then
            null;
    end member;


    -- Insert or delete an AdminTeam_Orgs row.
    procedure org ( p_action   in varchar, -- {'add', 'remove'}
                    p_op_uid   in int,
                    p_lock     in timestamp,
                    p_ateam_id in int,
                    p_org_id   in int )
    as
        v_org_id int := NetDB_Organization.verify_or_die (p_org_id);
    begin
        locked_or_die (p_op_uid, p_lock, p_ateam_id);

        case p_action
            when 'add' then
                insert into AdminTeam_Orgs values (p_ateam_id, v_org_id);

            when 'remove' then
                delete from AdminTeam_Orgs where AdminTeam = p_ateam_id and organization = v_org_id;

            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;
    exception
        when NetDB.unique_constraint_failed then
            null;
    end org;


    -- Check to see if a user has a lock for a record.
    procedure locked_or_die ( p_op_uid   in int,
                              p_lock     in timestamp,
                              p_ateam_id in int )
    as
        v_lock timestamp := null;
    begin
        select locked into v_lock from AdminTeam
         where id = verify_or_die (p_ateam_id) and locked_by = p_op_uid and locked = p_lock and
               Util.datediff ('second', p_lock) < NetDB_Default.get_int ('LockTimeOut');
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_not_locked ( 'adminteam',
                                                  handle (p_ateam_id),
                                                  NetDB_User.handle (p_op_uid) );
    end locked_or_die;


    -- Lock a record.  If the current user has already locked this record, renew the lock.
    function lock_record ( p_op_uid   in int,
                           p_ateam_id in int ) return NetDB.ref_cursor
    as
        v_lock      timestamp;
        v_locked_by int;

        out_record result_set       := result_set ();
        ret_cursor NetDB.ref_cursor := null;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'AdminTeam');
        NetDB_Owner.has_access_or_die (p_op_uid, p_ateam_id, 'AdminTeam');

        select locked, locked_by into v_lock, v_locked_by from AdminTeam where id = p_ateam_id;

        -- if not locked, locked by this user, or the lock has expired, we're good to go
        if ( v_lock is null or v_locked_by = p_op_uid or
             Util.datediff ('second', v_lock) > NetDB_Default.get_int ('LockTimeOut') )
        then
               update AdminTeam set locked    = systimestamp,
                                    locked_by = p_op_uid
                where id = p_ateam_id
            returning locked into v_lock;
        else
            NetDB_Util.throw_locked_by_other ( 'adminteam',
                                                handle (p_ateam_id),
                                                NetDB_User.handle (v_locked_by),
                                                v_lock );
        end if;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'locked', 0, 'timestamp', null, v_lock, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end lock_record;


    -- Unlock a record.
    procedure unlock_record ( p_op_uid   in int,
                              p_lock     in timestamp,
                              p_ateam_id in int ) as
    begin
        locked_or_die (p_op_uid, p_lock, p_ateam_id);

        update AdminTeam set locked    = null,
                             locked_by = null
         where id = p_ateam_id and locked = p_lock and locked_by = p_op_uid;
    end unlock_record;


    -- Get the info of an admin team. (calls info_pipelined and collects the result into a cursor)
    function info (p_ateam_id in int) return NetDB.ref_cursor
    is
        v_id varchar(1024) := p_ateam_id;
    begin
        if (v_id is null) then
            v_id := 'null';
        end if;

        return NetDB_Util.collect_into_cursor ('NetDB_AdminTeam.info_pipelined (' || v_id || ')');
    end info;


    -- Get the info of an admin team.
    function info_pipelined (p_ateam_id in int) return result_set pipelined
    as
        res_row result_row := result_row (null, null, null, null, null, null, null);

        -- Admin Team info expressed as a cursor (instead of defining a bunch of variables).
        cursor ateam_info is
        select id,      name,     email,      phone,      comments,
               created, modified, created_by, modified_by
          from AdminTeam
         where id = verify_or_die (p_ateam_id);

        i ateam_info%ROWTYPE;

        person_cursor NetDB.ref_cursor;
    begin
        open ateam_info;
        fetch ateam_info into i;
        close ateam_info;

        pipe row (result_row ('key',  'object',    null, 'string',    'ateam',    null,       null));
        pipe row (result_row ('key',  'id',        null, 'string',    i.id,       null,       null));
        pipe row (result_row ('key',  'name',      null, 'string',    i.name,     null,       null));
        pipe row (result_row ('key',  'email',     null, 'string',    i.email,    null,       null));
        pipe row (result_row ('key',  'phone',     null, 'string',    i.phone,    null,       null));
        pipe row (result_row ('key',  'comment',   null, 'string',    i.comments, null,       null));
        pipe row (result_row ('key',  'created',   null, 'timestamp', null,       i.created,  null));
        pipe row (result_row ('key',  'modified',  null, 'timestamp', null,       i.modified, null));

        -- Get the created_by info.
        pipe row (result_row ('data', 'created.by', null, null, null, null, null));

        person_cursor := NetDB_Person.info (i.created_by);
        loop
            fetch person_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                     res_row.value_string, res_row.value_timestamp, res_row.value_int;

            exit when person_cursor%notfound;

            pipe row (res_row);
        end loop;
        close person_cursor;

        -- Get the modified_by info.
        pipe row (result_row ('data', 'mod.by', 0, null, null, null, null));

        person_cursor := NetDB_Person.info (i.modified_by);
        loop
            fetch person_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                     res_row.value_string, res_row.value_timestamp, res_row.value_int;

            exit when person_cursor%notfound;

            pipe row (res_row);
        end loop;
        close person_cursor;

        -- Get the groups.
        pipe row (result_row ('data', 'groups', 0, null, null, null, null));

        for r in ( select Owner.label
                     from Owner
                     join AdminTeam_Owner on AdminTeam_Owner.owner = Owner.id
                    where AdminTeam_Owner.AdminTeam = p_ateam_id )
        loop
            pipe row (result_row ('key', r.label, null, 'string', ' ', null, null));
        end loop;

        -- Get the members of the AdminTeam
        pipe row (result_row ('data', 'members', 0, null, null, null, null));

        for r in ( select *
                     from Person
                     join AdminTeam_Member on AdminTeam_Member.person = Person.id
                    where AdminTeam_Member.AdminTeam = p_ateam_id )
        loop
            pipe row (result_row ('data', r.id, 1, null, null, null, null));

            person_cursor := NetDB_Person.info (r.id);
            loop
                fetch person_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                         res_row.value_string, res_row.value_timestamp, res_row.value_int;

                exit when person_cursor%notfound;

            pipe row (res_row);
            end loop;
            close person_cursor;
        end loop;

        -- Get the organizations of the AdminTeam
        pipe row (result_row ('data', 'departments', 0, null, null, null, null));

        for r in ( select Organization.organization
                     from Organization
                     join AdminTeam_Orgs on Organization.id = AdminTeam_Orgs.organization
                    where AdminTeam_Orgs.AdminTeam = p_ateam_id )
        loop
            pipe row (result_row ('key', r.organization, 0, 'string', ' ', null, null));
        end loop;
    end info_pipelined;


    -- Delete a record.
    procedure del ( p_op_uid   in int,
                    p_lock     in timestamp,
                    p_ateam_id in int ) as
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'AdminTeam');
        NetDB_Owner.has_access_or_die (p_op_uid, p_ateam_id, 'AdminTeam');
        locked_or_die (p_op_uid, p_lock, p_ateam_id);

        NetDB_Log.ins (p_op_uid, 'delete', p_ateam_id, 'adminteam');

        delete from AdminTeam where locked = p_lock and locked_by = p_op_uid and id = p_ateam_id;

    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('adminteam', handle (p_ateam_id));
    end del;


    -- Inserts a new admin team and locks it.
    function ins ( p_op_uid   in int,
                   p_name     in varchar,
                   p_email    in varchar,
                   p_phone    in varchar,
                   p_comments in varchar := null ) return NetDB.ref_cursor
    as
        v_ateam_id     int       := 0;
        v_lock         timestamp := null;
        sys_time_stamp timestamp := systimestamp;

        out_record  result_set := result_set ();
        ret_cursor  NetDB.ref_cursor;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'AdminTeam');

        -- validate strings
        NetDB_Util.validate (p_name, 'Admin Team name', 'An Admin Team must have a name.');
        NetDB_Util.validate (p_email, 'email address', 'An Admin Team must have an email address.');
        NetDB_Util.validate (p_phone, 'phone number', 'An Admin Team must have a phone number.');
        NetDB_Util.validate (p_comments, 'comment');

        insert into AdminTeam ( name,           email,      created,        created_by,
                                phone,          comments,   modified,       modified_by,
                                locked,         locked_by )
                       values ( p_name,         p_email,    sys_time_stamp, p_op_uid,
                                p_phone,        p_comments, sys_time_stamp, p_op_uid,
                                sys_time_stamp, p_op_uid )
        returning id, locked into v_ateam_id, v_lock;

        NetDB_Log.set_action ('insert');

        out_record.extend(2);
        out_record(1) := result_row ('key', 'id',     0, 'string',    v_ateam_id, null,   null);
        out_record(2) := result_row ('key', 'locked', 0, 'timestamp', null,       v_lock, null);

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('adminteam', p_name);

        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- Update an admin team.
    procedure upd ( p_op_uid   in int,
                    p_lock     in timestamp,
                    p_ateam_id in int,
                    p_name_new in varchar,
                    p_email    in varchar,
                    p_phone    in varchar,
                    p_comments in varchar := null ) as
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'AdminTeam');
        NetDB_Owner.has_access_or_die (p_op_uid, p_ateam_id, 'AdminTeam');
        locked_or_die (p_op_uid, p_lock, p_ateam_id);

        -- validate strings
        NetDB_Util.validate (p_name_new, 'Admin Team name', 'An Admin Team must have a name.');
        NetDB_Util.validate (p_email, 'email address', 'An Admin Team must have an email address.');
        NetDB_Util.validate (p_phone, 'phone number', 'An Admin Team must have a phone number.');
        NetDB_Util.validate (p_comments, 'comment');

        update AdminTeam set name        = p_name_new,
                             email       = p_email,
                             phone       = p_phone,
                             comments    = p_comments,
                             modified    = systimestamp,
                             modified_by = p_op_uid
         where locked = p_lock and locked_by = p_op_uid and id = p_ateam_id;

        NetDB_Log.set_action ('update');
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    -- Verify that an admin team conforms to the rules. Then log the record.
    procedure verify_and_log ( p_op_uid   in int,
                               p_ateam_id in int )
    as
        v_groups int := 0;
    begin
        NetDB_Owner.has_access_or_die (p_op_uid, p_ateam_id, 'AdminTeam');

        select min (AdminTeam) into v_groups from AdminTeam_Owner where adminteam = p_ateam_id;

        NetDB_Log.ins (p_op_uid, NetDB_Log.get_action (), p_ateam_id, 'adminteam');
    exception
        when NO_DATA_FOUND then
            raise_application_error (NetDB.ex_record_requires_group,
                'The admin team "' || handle (p_ateam_id) ||
                '" does not have ownership group record.' );
    end verify_and_log;


    -- Verify that an ID and/or name combination exists.
    function verify_or_die ( p_ateam_id in int     := null,
                             p_name     in varchar := null ) return int
    as
        v_identifier varchar(256) := p_name;
        v_ateam_id   int;
    begin
        if (p_ateam_id is null and p_name is null) then
            NetDB_Util.throw_no_record_id_handle ();
        elsif (p_ateam_id is not null and p_name is not null) then
            select id
              into v_ateam_id
              from AdminTeam
             where lower (name) = lower (p_name) and id = p_ateam_id;
        elsif (p_name is not null) then
            select id into v_ateam_id from AdminTeam where lower (name) = lower (p_name);
        else -- p_ateam_id is not null
            v_identifier := p_ateam_id;

            select id into v_ateam_id from AdminTeam where id = p_ateam_id;
        end if;

        return v_ateam_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('adminteam', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('adminteam');
    end verify_or_die;


    -- Return the handle for an admin team.
    function handle (p_ateam_id in int) return varchar
    as
        v_name AdminTeam.name%TYPE := null;
    begin
        select name into v_name from AdminTeam where id = p_ateam_id;

        return v_name;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('adminteam', p_ateam_id);
    end handle;

end NetDB_AdminTeam;
/
