-- NetDB_DHCPoption.sql --- contains functions related to dhcp options

create or replace package body NetDB_DHCPoption as

    -- Verify that an option can be used with a record.
    procedure verify_application_or_die ( p_option_id in int,
                                          p_label     in varchar, -- record type
                                          p_record_id in int )
    as
        v_option_id int := verify_or_die (p_option_id);
        v_type      DHCPoption.type%TYPE := null;
    begin
        -- Make sure that the option can be used with the record type.
        select count (*) into NetDB.existence_counter from DHCPoption_Application
         where dhcpoption = v_option_id and label = lower (p_label);

        if (NetDB.existence_counter <= 0) then
            raise_application_error (NetDB.ex_record_not_found,
                'The option "' || handle (v_option_id) ||
                '" cannot be added to "' || NetDB_RecordType.display (p_label) || '" records.');
        end if;

        -- DHCP options that truly are options (marked with the type of "option")
        -- should only be defined once per record. The other parameters can be defined
        -- as many times as a user would like.
        select lower (type) into v_type from DHCPoption where id = v_option_id;

        if (v_type = 'option') then
            case lower (p_label)
                when 'dhcpservice' then
                    select count (*) into NetDB.existence_counter from DHCPservice_Options
                     where dhcpoption = v_option_id and dhcpservice = p_record_id;

                when 'interface' then
                    select count (*) into NetDB.existence_counter from Interface_DHCPoptions
                     where dhcpoption = v_option_id and interface = p_record_id;

                when 'addressspace' then
                    select count (*) into NetDB.existence_counter from IPaddressSpace_DHCPoptions
                     where dhcpoption = v_option_id and ipaddressspace = p_record_id;

                when 'network' then
                    select count (*) into NetDB.existence_counter from Network_DHCPoptions
                     where dhcpoption = v_option_id and network = p_record_id;

                else
                    NetDB_Util.throw_unknown_record_type (p_label);
            end case;

            if (NetDB.existence_counter > 0) then
                raise_application_error (NetDB.ex_record_already_exists,
                    'The option "' || handle (v_option_id) || '" is already used on the ' ||
                    NetDB_RecordType.display (p_label) ||
                    ' "' || NetDB_Util.handle (p_record_id, p_label, true) || '".');
            end if;
        end if;
    end verify_application_or_die;


    -- Add an application to an option.
    -- Unused and untested.
    procedure application_add ( p_op_uid       in int,
                                p_app_label    in varchar,
                                p_option_id    in int )
    as
        v_option_id int := verify_or_die (p_option_id, null);
    begin
        insert into DHCPoption_Application (dhcpoption, label) values (v_option_id, p_app_label);
    exception
        when NetDB.unique_constraint_failed then
            null;
        when NetDB.parent_does_not_exist then
            NetDB_Util.throw_record_does_not_exist ('application', p_app_label);
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end application_add;


    -- Remove an application to an option.
    -- Unused and untested.
    procedure application_remove ( p_op_uid       in int,
                                   p_app_label    in varchar,
                                   p_option_id    in int )
    as
        v_option_id int := verify_or_die (p_option_id, null);
    begin
        delete from DHCPoption_Application where dhcpoption = v_option_id and label = p_app_label;
    end application_remove;


    -- Add a DHCP option to a record.
    procedure add ( p_op_uid       in int,
                    p_lock         in timestamp,
                    p_record_id    in int,
                    p_record_type  in varchar, -- {DHCPservice, Interface, AddressSpace, Network}
                    p_option_id    in int     := null,
                    p_option_label in varchar := null,
                    p_value        in varchar )
    as
        v_record_id int := NetDB_Util.verify_or_die (p_record_id, null, p_record_type);
        v_option_id int := verify_or_die (p_option_id, p_option_label);
    begin
        NetDB_Util.validate (p_value, 'value', 'The ' || handle (v_option_id) || ' must have a value.');

        case lower (p_record_type)
            when 'dhcpservice' then
                NetDB_Access.has_access_or_die (p_op_uid, 'dhcpservice');
                verify_application_or_die (v_option_id, p_record_type, p_record_id);

                insert into DHCPservice_Options (dhcpservice, dhcpoption,  value)
                                         values (p_record_id, v_option_id, p_value);

            when 'interface' then
                NetDB_Util.locked_or_die (p_op_uid, p_lock, p_record_id, p_record_type);
                verify_application_or_die (v_option_id, p_record_type, p_record_id);

                insert into Interface_DHCPoptions (interface,   dhcpoption,  value)
                                           values (p_record_id, v_option_id, p_value);

            when 'addressspace' then
                NetDB_Util.locked_or_die (p_op_uid, p_lock, p_record_id, p_record_type);
                verify_application_or_die (v_option_id, p_record_type, p_record_id);

                insert into IPaddressSpace_DHCPoptions (IPaddressSpace, dhcpoption,  value)
                                                values (p_record_id,    v_option_id, p_value);

            when 'network' then
                NetDB_Util.locked_or_die (p_op_uid, p_lock, p_record_id, p_record_type);
                verify_application_or_die (v_option_id, p_record_type, p_record_id);

                insert into Network_DHCPoptions (network,     dhcpoption,  value)
                                         values (p_record_id, v_option_id, p_value);

            else
                NetDB_Util.throw_unknown_record_type (p_record_type);
        end case;
    exception
        when NetDB.unique_constraint_failed then
            null;
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end add;


    -- Remove a DHCP option from a record.
    -- p_record_type: {DHCPservice, Interface, IPAddressSpace, Network}
    procedure remove ( p_op_uid       in int,
                       p_lock         in timestamp,
                       p_record_id    in int,
                       p_record_type  in varchar,
                       p_option_id    in int     := null,
                       p_option_label in varchar := null )
    as
        v_record_id int := NetDB_Util.verify_or_die (p_record_id, null, p_record_type);
        v_option_id int := verify_or_die (p_option_id, p_option_label);
    begin
        case lower (p_record_type)
            when 'dhcpservice' then
                NetDB_Access.has_access_or_die (p_op_uid, 'dhcpservice');

                delete from DHCPservice_Options
                      where dhcpservice = p_record_id and dhcpoption = v_option_id;

            when 'interface' then
                NetDB_Util.locked_or_die (p_op_uid, p_lock, p_record_id, p_record_type);

                delete from Interface_DHCPoptions
                      where interface = p_record_id and dhcpoption = v_option_id;

            when 'addressspace' then
                NetDB_Util.locked_or_die (p_op_uid, p_lock, p_record_id, p_record_type);

                delete from IPaddressSpace_DHCPoptions
                      where IPaddressSpace = p_record_id and dhcpoption = v_option_id;

            when 'network' then
                NetDB_Util.locked_or_die (p_op_uid, p_lock, p_record_id, p_record_type);

                delete from Network_DHCPoptions
                      where network = p_record_id and dhcpoption = v_option_id;

            else
                NetDB_Util.throw_unknown_record_type (p_record_type);
        end case;
    end remove;


    -- same as list_pipelined (calls list_pipelined and collects the result into a cursor)
    function list return NetDB.ref_cursor as begin
         return NetDB_Util.collect_into_cursor ('NetDB_DHCPoption.list_pipelined ()');
    end list;


    -- returns a list of all dhcp options
    function list_pipelined return result_set pipelined
    as
        v_last_label DHCPoption.label%TYPE := null;
    begin
        for r in ( select do.label, doa.label application
                     from DHCPoption do join DHCPoption_Application doa on doa.dhcpoption = do.id
                    order by do.id )
        loop
            if (r.label != v_last_label or v_last_label is null) then
                pipe row (result_row ('data', r.label, 0, null, null, null, null));
            end if;

            pipe row (result_row ('key', r.application, null, 'string', ' ', null, null));

            v_last_label := r.label;
        end loop;
    end list_pipelined;


    procedure del ( p_op_uid    in int,
                    p_option_id in int)
    as
        v_option_id int := verify_or_die (p_option_id);
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'dhcpoption');

        delete from DHCPoption where id = p_option_id;
    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('dhcpoption', handle (p_option_id));
    end del;


    function ins ( p_op_uid   in int,
                   p_label    in varchar,
                   p_datatype in varchar,
                   p_type     in varchar := null ) return NetDB.ref_cursor
    as
        v_id int := 0;

        out_record result_set       := result_set ();
        ret_cursor NetDB.ref_cursor := null;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'dhcpoption');

        NetDB_Util.validate (p_label,    'label',     'The DHCP Option must have a label.');
        NetDB_Util.validate (p_datatype, 'data type', 'The DHCP Option must have a data type.');
        NetDB_Util.validate (p_type,     'type');

        insert into DHCPoption (datatype,   label,   type)
                        values (p_datatype, p_label, p_type)
        returning id into v_id;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'id', 0, 'string', v_id, null, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('dhcpoption', p_label);
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    procedure upd ( p_op_uid    in int,
                    p_option_id in int,
                    p_label_new in varchar,
                    p_datatype  in varchar,
                    p_type      in varchar := null )
    as
        v_option_id int := verify_or_die (p_option_id);

        v_label    DHCPoption.label%TYPE    := null;
        v_datatype DHCPoption.datatype%TYPE := null;
        v_type     DHCPoption.type%TYPE     := null;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'dhcpoption');

        NetDB_Util.validate (p_label_new, 'label',     'The DHCP Option must have a label.');
        NetDB_Util.validate (p_datatype,  'data type', 'The DHCP Option must have a data type.');
        NetDB_Util.validate (p_type,      'type');

        update DHCPoption set datatype = p_datatype,
                              label    = p_label_new,
                              type     = p_type
         where id = p_option_id;
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('dhcpoption', p_label_new);
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    -- verify that a DCHP option exists.
    function verify_or_die ( p_option_id in int     := null,
                             p_label     in varchar := null ) return int
    as
        v_identifier varchar(256) := p_label;
        v_option_id  int          := 0;
    begin
        if (p_option_id is null and p_label is null) then
            NetDB_Util.throw_no_record_id_handle ();
        elsif (p_option_id is not null and p_label is not null) then
            select id into v_option_id from DHCPoption
             where lower (label) = lower (p_label) and id = p_option_id;
        elsif (p_label is not null) then
            select id into v_option_id from DHCPoption where lower (label) = lower (p_label);
        else -- p_option_id is not null
            v_identifier := p_option_id;

            select id into v_option_id from DHCPoption where id = p_option_id;
        end if;

        return v_option_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('dhcpoption', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('dhcpoption');
    end verify_or_die;


    function handle (p_option_id in int) return varchar
    as
        v_option DHCPoption.label%TYPE := null;
    begin
        select label into v_option from DHCPoption where id = p_option_id;

        return v_option;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('dhcpoption', p_option_id);
    end handle;

end NetDB_DHCPoption;
/
