-- NetDB_DHCPservice.sql --- contains functions related to dhcp services

create or replace package body NetDB_DHCPservice as

    -- same as list_pipelined (calls list_pipelined and collects the result into a cursor)
    function list return NetDB.ref_cursor as begin
        return NetDB_Util.collect_into_cursor ('NetDB_DHCPservice.list_pipelined ()');
    end list;


    -- Generates a list of all DHCP services and stuffs it in a generic cursor.
    function list_pipelined return result_set pipelined
    as
        cursor data_cursor is
            select label, id, description from DHCPservice;

        i data_cursor%ROWTYPE;
    begin
        open data_cursor;
        loop
            fetch data_cursor into i;
            exit when data_cursor%notfound;

            pipe row (result_row ('data', i.label,  0,    null,     null,          null, null));
            pipe row (result_row ('key', 'name',    null, 'string', i.label,       null, null));
            pipe row (result_row ('key', 'id',      null, 'string', i.id,          null, null));
            pipe row (result_row ('key', 'comment', null, 'string', i.description, null, null));
        end loop;
        close data_cursor;
    end list_pipelined;

    -- Unused.
    procedure del ( p_op_uid     in int,
                    p_service_id in int )
    as
        v_service_id int := verify_or_die (p_service_id);
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'dhcpservice');

        delete from DHCPservice where id = p_service_id;
    end del;


    -- Unused .
    function ins ( p_op_uid in int,
                   p_label  in varchar,
                   p_desc   in varchar := null ) return NetDB.ref_cursor
    as
        v_id int := 0;

        out_record result_set       := result_set ();
        ret_cursor NetDB.ref_cursor := null;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'dhcpservice');

        NetDB_Util.validate (p_label, 'label');
        NetDB_Util.validate (p_desc,  'description');

        if (p_label is null or p_op_uid is null) then
            raise_application_error (NetDB.ex_bad_parameter,
                'DCHP services must have both a label and record of the user who last modified it.');
        end if;

        insert into DHCPservice (label,   description, modified_by, modified)
                         values (p_label, p_desc,      p_op_uid,    systimestamp)
        returning id into v_id;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'id', 0, 'string', v_id, null, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- Unused.
    procedure upd ( p_op_uid     in int,
                    p_service_id in int,
                    p_label_new  in varchar,
                    p_desc       in varchar := null )
    as
        v_service_id int := verify_or_die (p_service_id);

        v_label DHCPservice.label%TYPE       := null;
        v_desc  DHCPservice.description%TYPE := null;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'dhcpservice');

        NetDB_Util.validate (p_label_new, 'label');
        NetDB_Util.validate (p_desc, 'description');

        select label, description into v_label, v_desc from DHCPservice where id = p_service_id;

        update DHCPservice set label       = coalesce (p_label_new, v_label),
                               description = p_desc,
                               modified_by = p_op_uid,
                               modified    = systimestamp
         where id = p_service_id;
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    -- verify that a DCHP service exists.
    -- Unused.
    function verify_or_die ( p_service_id in int     := null,
                             p_label      in varchar := null ) return int
    as
        v_identifier varchar(256) := p_label;
        v_service_id int          := 0;
    begin
        if (p_service_id is null and p_label is null) then
            NetDB_Util.throw_no_record_id_handle ();
        elsif (p_service_id is not null and p_label is not null) then
            select id into v_service_id from DHCPservice
             where lower (label) = lower (p_label) and id = p_service_id;
        elsif (p_label is not null) then
            select id into v_service_id from DHCPservice where lower (label) = lower (p_label);
        else -- p_service_id is not null
            v_identifier := p_service_id;

            select id into v_service_id from DHCPservice where id = p_service_id;
        end if;

        return v_service_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('dhcpservice', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('dhcpservice');
    end verify_or_die;


    -- Unused and untested.
    function handle (p_service_id in int) return varchar
    as
        v_service DHCPservice.label%TYPE := null;
    begin
        select label into v_service from DHCPservice where id = p_service_id;

        return v_service;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('dhcpservice', p_service_id);
    end handle;

end NetDB_DHCPservice;
/
