-- NetDB_Domain.sql --- Domain package body.

create or replace package body NetDB_Domain as

    -- Check if a user can assign names in a domain.
    function can_use ( p_user_id   in int,
                       p_domain_id in int ) return boolean as
    begin
        return NetDB_Owner.check_access (p_user_id, p_domain_id, 'domain_assign');
    end can_use;


    -- Check to see if a user can use a domain as a normal name or MX.
    -- DOES NOT honor the all groups flag.
    function can_use_as_name ( p_user_id   in int,
                               p_domain_id in int ) return boolean as
    begin
        return NetDB_Owner.check_access (p_user_id, p_domain_id, 'domain_use', false);
    end can_use_as_name;


    function is_limited (p_domain_id in int) return boolean
    as
        v_limited int := 0;
    begin
        select limited into v_limited from Domain where id = p_domain_id;

        return Util.to_boolean (v_limited);
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('domain', p_domain_id);
    end is_limited;


    -- Check to see if a user has a lock for a record.
    procedure locked_or_die ( p_op_uid    in int,
                              p_lock      in timestamp,
                              p_domain_id in int )
    as
        v_lock timestamp := null;
    begin
        select locked into v_lock from Domain
         where id = verify_or_die (p_domain_id) and locked_by = p_op_uid and locked = p_lock and
               Util.datediff ('second', p_lock) < NetDB_Default.get_int ('LockTimeOut');
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_not_locked ( 'domain',
                                                  handle (p_domain_id),
                                                  NetDB_User.handle (p_op_uid) );
    end locked_or_die;


    -- Lock a record.
    function lock_record ( p_op_uid    in int,
                           p_domain_id in int ) return NetDB.ref_cursor
    as
        v_lock      timestamp;
        v_locked_by int;

        out_record result_set       := result_set ();
        ret_cursor NetDB.ref_cursor := null;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Domain');
        NetDB_Owner.has_access_or_die (p_op_uid, p_domain_id, 'Domain');

        select locked, locked_by into v_lock, v_locked_by from Domain where id = p_domain_id;

        -- if not locked, locked by this user, or the lock has expired, we're good to go
        if ( v_lock is null or v_locked_by = p_op_uid or
             Util.datediff ('second', v_lock) > NetDB_Default.get_int ('LockTimeOut') )
        then
               update Domain set locked    = systimestamp,
                                 locked_by = p_op_uid
                where id = p_domain_id
            returning locked into v_lock;
        else
            NetDB_Util.throw_locked_by_other ( 'domain',
                                                handle (p_domain_id),
                                                NetDB_User.handle (v_locked_by),
                                                v_lock );
        end if;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'locked', 0, 'timestamp', null, v_lock, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end lock_record;


    -- Unlock a record.
    procedure unlock_record ( p_op_uid    in int,
                              p_lock      in timestamp,
                              p_domain_id in int ) as
    begin
        locked_or_die (p_op_uid, p_lock, p_domain_id);

        update Domain
           set locked    = null,
               locked_by = null
         where id = p_domain_id and locked = p_lock and locked_by = p_op_uid;
    end unlock_record;


    -- List domains a user can use or all domains if no user id is given.
    function list (p_user_id in int := null) return NetDB.ref_cursor
    as
        v_all_groups boolean            := true;
        v_domain     Domain.domain%TYPE := null;
        v_domain_id  Domain.id%TYPE     := null;

        out_record  result_set := result_set ();
        ret_cursor  NetDB.ref_cursor;
        data_cursor NetDB.ref_cursor;
    begin
        if (p_user_id is not null) then
            v_all_groups := NetDB_User.has_all_groups (p_user_id);
        end if;

        if (not v_all_groups) then
            open data_cursor for
                  select Domain.domain, Domain.id
                    from Domain
                    join Domain_assignNameIn on Domain_assignNameIn.domain = Domain.id
                    join Users_Owner         on Domain_assignNameIn.owner  = Users_Owner.owner
                   where Users_Owner.userid = p_user_id
                order by lower (Domain.domain);
        else
            open data_cursor for
                  select domain, id from Domain order by lower (domain);
        end if;

        loop
            fetch data_cursor into v_domain, v_domain_id;
            exit when data_cursor%notfound;

            out_record.extend(1);
            out_record(out_record.COUNT) := result_row ('key', v_domain, null, 'string', v_domain_id, null, null);
        end loop;
        close data_cursor;

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end list;


    -- Get the info of a domain. (calls info_pipelined and collects the result into a cursor)
    function info (p_domain_id in int) return NetDB.ref_cursor
    as
        v_id varchar(1024) := p_domain_id;
    begin
        if (v_id is null) then
            v_id := 'null';
        end if;

        return NetDB_Util.collect_into_cursor ('NetDB_Domain.info_pipelined (' || v_id || ')');
    end info;


    -- Get the info of a domain.
    function info_pipelined (p_domain_id in int) return result_set pipelined
    as
        domain_row    Domain%ROWTYPE;
        res_row       result_row       := result_row (null, null, null, null, null, null, null);
        person_cursor NetDB.ref_cursor := null;
    begin
        select * into domain_row from Domain where id = verify_or_die (p_domain_id);

        pipe row (result_row ('key', 'object',      null, 'string',   'domain',             null,                null                 ));
        pipe row (result_row ('key', 'id',          null, 'string',    domain_row.id,       null,                null                 ));
        pipe row (result_row ('key', 'domain',      null, 'string',    domain_row.domain,   null,                null                 ));
        pipe row (result_row ('key', 'limited?',    null, 'boolean',   null,                null,                domain_row.limited   ));
        pipe row (result_row ('key', 'comment',     null, 'string',    domain_row.comments, null,                null                 ));
        pipe row (result_row ('key', 'created',     null, 'timestamp', null,                domain_row.created,  null                 ));
        pipe row (result_row ('key', 'modified',    null, 'timestamp', null,                domain_row.modified, null                 ));

        -- Get the created_by info.
        pipe row (result_row ('data', 'created.by', 0, null, null, null, null));

        person_cursor := NetDB_Person.info (domain_row.created_by);
        loop
            fetch person_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                     res_row.value_string, res_row.value_timestamp, res_row.value_int;

            exit when person_cursor%notfound;

            pipe row (res_row);
        end loop;
        close person_cursor;

        -- Get the modified_by info.
        pipe row (result_row ('data', 'mod.by', 0, null, null, null, null));

        person_cursor := NetDB_Person.info (domain_row.modified_by);
        loop
            fetch person_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                     res_row.value_string, res_row.value_timestamp, res_row.value_int;

            exit when person_cursor%notfound;

            pipe row (res_row);
        end loop;
        close person_cursor;

        -- Get the groups.
        pipe row (result_row ('data', 'groups', 0, null, null, null, null));

        for r in ( select * from Domain_Owner, Owner
                    where Domain_Owner.domain = p_domain_id and Domain_Owner.owner = Owner.id
                    order by lower (Owner.label) )
        loop
            pipe row (result_row ('key', r.label, null, 'string', ' ', null, null));
        end loop;

        -- Get the groups allowed to assign names in the domain.
        pipe row (result_row ('data', 'assign', 0, null, null, null, null));

        for r in ( select * from Domain_assignNameIn, Owner
                    where Domain_assignNameIn.domain = p_domain_id and Owner.id = Domain_assignNameIn.owner
                    order by lower (Owner.label) )
        loop
            pipe row (result_row ('key', r.label, null, 'string', ' ', null, null));
        end loop;

        -- Get the groups allowed to use the domain as a name.
        pipe row (result_row ('data', 'useAsName', 0, null, null, null, null));

        for r in ( select * from Domain_useAsName, Owner
                    where Domain_useAsName.domain = p_domain_id and Owner.id = Domain_useAsName.owner
                    order by lower (Owner.label) )
        loop
            pipe row (result_row ('key', r.label, null, 'string', ' ', null, null));
        end loop;

    end info_pipelined;


    -- Delete a record and log the action.
    procedure del ( p_op_uid    in int,
                    p_lock      in timestamp,
                    p_domain_id in int ) as
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Domain');
        NetDB_Owner.has_access_or_die (p_op_uid, p_domain_id, 'Domain');

        -- Lock verification is being deferred to NetDB_Name.del (...), which actually calls
        -- Domain.locked_or_die (...).  There's hardly a reason to check it twice.
        -- locked_or_die (p_op_uid, p_lock, p_domain_id);
        NetDB_Log.ins (p_op_uid, 'delete', p_domain_id, 'domain');

        -- Due to the way that the tables are defined, deleting from the Name table will
        -- delete from the Domain table. However, deleting from the Domain table does NOT
        -- delete from the Name table.
        NetDB_Name.del (p_op_uid, p_lock, p_domain_id, null, 'Domain');
    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('domain', handle (p_domain_id));
    end del;


    -- Insert a name, found in the Name table, into the Domain table and locks it.
    function ins ( p_op_uid  in int,
                   p_domain  in varchar,  -- Fully qualified name
                   p_limited in numeric := 1,
                   p_comment in varchar := null ) return NetDB.ref_cursor
    as
        v_domain_id    int       := NetDB_Name.verify_or_die (null, p_domain, 'Domain', true);
        v_lock         timestamp := null;
        sys_time_stamp timestamp := systimestamp;

        out_record  result_set := result_set ();
        ret_cursor  NetDB.ref_cursor;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Domain');

        if (v_domain_id is null) then
            raise_application_error (NetDB.ex_idiot_programmer,
                'The domain name "' || p_domain || '" has not been reserved.');
        end if;

        NetDB_Util.validate (p_comment, 'comment');

        insert into Domain ( id,             created_by, modified_by, created,
                             modified,       limited,    comments,    locked,         locked_by )
                    values ( v_domain_id,    p_op_uid,   p_op_uid,    sys_time_stamp,
                             sys_time_stamp, p_limited,  p_comment,   sys_time_stamp, p_op_uid )
        returning locked into v_lock;

        NetDB_Log.set_action ('insert');

        out_record.extend(2);
        out_record(1) := result_row ('key', 'id',     0, 'string',    v_domain_id, null,   null);
        out_record(2) := result_row ('key', 'locked', 0, 'timestamp', null,        v_lock, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- Update domain info.
    procedure upd ( p_op_uid     in int,
                    p_lock       in timestamp,
                    p_domain_id  in int,
                    p_domain_new in varchar,
                    p_limited    in numeric := 1,
                    p_comment    in varchar := null ) as
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Domain');
        NetDB_Owner.has_access_or_die (p_op_uid, p_domain_id, 'Domain');
        locked_or_die (p_op_uid, p_lock, p_domain_id);

        NetDB_Util.validate (p_comment, 'comment');

        -- Allow for case changes of the domain.
        NetDB_Name.update_domain_name (p_op_uid, p_lock, p_domain_id, p_domain_new);

        update Domain
           set modified_by = p_op_uid,
               modified    = systimestamp,
               limited     = p_limited,
               comments    = p_comment
         where id = p_domain_id;

        NetDB_Log.set_action ('update');
    end upd;


    -- Verify that a domain conforms to the rules and inserts the record into the log.
    procedure verify_and_log ( p_op_uid    in int,
                               p_domain_id in int )
    as
        v_groups  int := 0;
        v_name_id int := 0;
    begin
        NetDB_Owner.has_access_or_die (p_op_uid, p_domain_id, 'Domain');

        -- Verify that the name has been accepted.
        v_name_id := NetDB_Name.verify_or_die (p_domain_id, null, 'Domain', false);

        -- A domain must be in at least one group.
        select count(*) into v_groups from Domain_Owner where domain = p_domain_id;

        if (v_groups <= 0) then
            raise_application_error (NetDB.ex_record_requires_group,
                'The domain "' || handle (p_domain_id) || '" must have at least one group.');
        end if;

        NetDB_Log.ins (p_op_uid, NetDB_Log.get_action (), p_domain_id, 'domain');
    exception
        when NetDB.record_not_found then
            raise_application_error (NetDB.ex_record_not_found,
                'The name "' || handle (p_domain_id) || '" must be accepted.');
    end verify_and_log;


    -- Verify domain ID / handle or complain.
    function verify_or_die ( p_domain_id in int     := null,
                             p_domain    in varchar := null ) return int
    as
        v_identifier varchar(256) := p_domain;
        v_domain_id  int          := 0;
    begin
        if (p_domain_id is null and p_domain is null) then
            NetDB_Util.throw_no_record_id_handle ();
        elsif (p_domain_id is not null and p_domain is not null) then
            select id into v_domain_id from Domain
             where lower (domain) = lower (p_domain) and id = p_domain_id;
        elsif (p_domain is not null) then
            select id into v_domain_id from Domain where lower (domain) = lower (p_domain);
        else -- p_domain_id is not null
            v_identifier := p_domain_id;

            select id into v_domain_id from Domain where id = p_domain_id;
        end if;

        return v_domain_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('domain', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('domain');
    end verify_or_die;


    -- Get the full name for a domain.
    function handle (p_domain_id in int) return varchar
    as
        v_domain Domain.domain%TYPE := null;
    begin
        select domain into v_domain from Domain where id = p_domain_id;

        return v_domain;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('domain', p_domain_id);
    end handle;

end NetDB_Domain;
/
