-- NetDB_Location.sql --- contains functions related to Location table

create or replace package body NetDB_Location as

    -- returns list of locations restricted by p_filter
    -- if p_filter is null then all locations are shown
    function list (p_filter in varchar) return NetDB.ref_cursor
    as
        v_filter varchar(256) := translate (p_filter, '"', '''');

        out_record result_set := result_set ();
        ret_cursor NetDB.ref_cursor;
    begin
        for r in ( select name, site, code
                     from Location
                    where lower (name) like lower (nvl (v_filter, '%')) )
        loop
            out_record.extend(4);
            out_record(out_record.COUNT - 3) := result_row ('data',  r.name,    0,     null,    null,   null, null);
            out_record(out_record.COUNT - 2) := result_row ('key',  'location', null, 'string', r.name, null, null);
            out_record(out_record.COUNT - 1) := result_row ('key',  'site',     null, 'string', r.site, null, null);
            out_record(out_record.COUNT)     := result_row ('key',  'code',     null, 'string', r.code, null, null);
        end loop;

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end list;


    -- Delete a location.
    procedure del ( p_op_uid      in int,
                    p_location_id in int )
    as
        v_location_id int := verify_or_die (p_location_id);
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'location');

        delete Location where id = p_location_id;
    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('location', handle (p_location_id));
    end del;


    -- Insert a location.
    function ins ( p_op_uid   in int,
                   p_location in varchar,
                   p_site     in varchar,
                   p_code     in varchar ) return NetDB.ref_cursor
    as
        v_location_id int;

        v_location Location.name%TYPE := translate (p_location, '"', '''');
        v_site     Location.site%TYPE := translate (p_site,     '"', '''');
        v_code     Location.code%TYPE := translate (p_code,     '"', '''');

        out_record result_set := result_set ();
        ret_cursor NetDB.ref_cursor;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'location');

        NetDB_Util.validate (p_location, 'location name', 'The Location must have a name.');
        NetDB_Util.validate (p_site, 'site', 'The Location must have a site.');
        NetDB_Util.validate (p_code, 'code', 'The Location must have a code.');

        insert into Location (  name,       site,   code)
                      values (v_location, v_site, v_code)
        returning id into v_location_id;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'id', 0, 'string', v_location_id, null,   null);

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.unique_constraint_failed then
            -- The name is not unique.
            if (instr (lower (sqlerrm), 'location_name') > 0) then
                NetDB_Util.throw_record_already_exists ('location', p_location);

            -- The site and code combination is not unique.
            elsif (instr (lower (sqlerrm), 'location_site_code') > 0) then
                NetDB_Util.throw_record_already_exists ('site and code', p_site || '-' || p_code);
            end if;

        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- Update a location.
    procedure upd ( p_op_uid       in int,
                    p_location_id  in int,
                    p_location_new in varchar,
                    p_site         in varchar,
                    p_code         in varchar )
    as
        v_location_id  int                := verify_or_die (p_location_id);
        v_location_new Location.name%TYPE := translate (p_location_new, '"', '''');
        v_site         Location.site%TYPE := translate (p_site,         '"', '''');
        v_code         Location.code%TYPE := translate (p_code,         '"', '''');
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'location');

        NetDB_Util.validate (p_location_new, 'location name', 'The Location must have a name.');
        NetDB_Util.validate (p_site, 'site', 'The Location must have a site.');
        NetDB_Util.validate (p_code, 'code', 'The Location must have a code.');

        update Location set name = v_location_new,
                            site = v_site,
                            code = v_code
         where id = p_location_id;
    exception
        when NetDB.unique_constraint_failed then
            -- The name is not unique.
            if (instr (lower (sqlerrm), 'location_name') > 0) then
                NetDB_Util.throw_record_already_exists ('location', p_location_new);

            -- The site and code combination is not unique.
            elsif (instr (lower (sqlerrm), 'location_site_code') > 0) then
                NetDB_Util.throw_record_already_exists ('site and code', p_site || '-' || p_code);
            end if;

        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    -- given location name verifies that location exists and returns the location id
    function verify_or_die ( p_location_id in int     := null,
                             p_location    in varchar := null )return int
    as
        v_identifier  varchar(256) := p_location;

        v_location_id int;
        v_location    Location.name%TYPE := translate (p_location, '"', '''');
    begin
        if (p_location_id is null and p_location is null) then
            NetDB_Util.throw_no_record_id_handle ();

        elsif (p_location_id is not null and p_location is not null) then
            select id into v_location_id from Location
             where id = p_location_id and name = v_location;

        elsif (p_location is not null) then
            select id into v_location_id from Location where name = v_location;

        else -- p_location_id is not null
            v_identifier := p_location_id;

            select id into v_location_id from Location where id = p_location_id;
        end if;

        return v_location_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('location', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('location');
    end verify_or_die;


    function handle (p_location_id in int) return varchar
    as
        v_loc Location.name%TYPE := null;
    begin
        select name into v_loc from Location where id = p_location_id;

        return v_loc;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('location', p_location_id);
    end handle;

end NetDB_Location;
/
