-- NetDB_MXrecord.sql --- functions dealing with MXrecords

create or replace package body NetDB_MXrecord as

    -- Internal function used to verify that a received_by name is valid or spit out a meaningful error.
    function verify_received_by_or_die ( p_received_by_id in int     := null,
                                         p_received_by    in varchar := null ) return int
    as
        mx_receive_id int;
    begin
        begin
            mx_receive_id := NetDB_Name.verify_or_die (p_received_by_id, p_received_by, 'A', true);
        exception
            when NetDB.record_not_found then
                mx_receive_id := NetDB_Name.verify_or_die (p_received_by_id, p_received_by, '%', true);

                raise_application_error (NetDB.ex_bad_MX_target,
                    '"' || NetDB_Name.fullname (mx_receive_id) || '" is an invalid MX target. ' ||
                    'MX''s can only point to: Nodes, Networks, Interfaces and IP Addresses.');
        end;

        return mx_receive_id;
    exception
        when NetDB.record_not_found then
            -- If the receive name was not found, reformat the error message to something
            -- the user can understand. (The name "[name]" does not exist as an A name.)
            -- If the exception comes from someplace else, this is functionally a reraise.
            raise_application_error ( sqlcode,
                NetDB_Util.strip_ora_code (replace ( sqlerrm,
                                                    'with the label "%"',
                                                    'as an A name' )) );
    end verify_received_by_or_die;


    procedure del ( p_op_uid         in int,
                    p_lock           in timestamp,
                    p_sent_to_id     in int     := null,
                    p_sent_to        in varchar := null,
                    p_received_by_id in int     := null,
                    p_received_by    in varchar := null )
    as
        mx_send_id    int := NetDB_Name.verify_or_die (p_sent_to_id, p_sent_to, 'mx', true);
        mx_receive_id int := verify_received_by_or_die (p_received_by_id, p_received_by);
    begin
        NetDB_Name.locked_or_die (p_op_uid, p_lock, mx_receive_id);

        delete from MXrecord where sent_to = mx_send_id and received_by = mx_receive_id;
    end del;


    procedure ins ( p_op_uid         in int,
                    p_lock           in timestamp,
                    p_sent_to_id     in int     := null,
                    p_sent_to        in varchar := null,
                    p_received_by_id in int     := null,
                    p_received_by    in varchar := null,
                    pref             in int )
    as
        mx_send_id    int := NetDB_Name.verify_or_die (p_sent_to_id, p_sent_to, 'mx', true);
        mx_receive_id int := verify_received_by_or_die (p_received_by_id, p_received_by);
    begin
        NetDB_Name.locked_or_die (p_op_uid, p_lock, mx_receive_id);
        NetDB_Owner.has_access_or_die (p_op_uid, mx_send_id, 'mx');

        insert into MXrecord (sent_to,    received_by,   preference)
                      values (mx_send_id, mx_receive_id, pref);
    exception
        when NetDB.check_constraint_failed then
            raise_application_error (NetDB.ex_bad_parameter,
                'MX preferences must be between 0 and 65,535.');

        when NetDB.unique_constraint_failed then
            null;
    end ins;


    -- Update the preference for an MX record.
    procedure upd ( p_op_uid         in int,
                    p_lock           in timestamp,
                    p_sent_to_id     in int     := null,
                    p_sent_to        in varchar := null,
                    p_received_by_id in int     := null,
                    p_received_by    in varchar := null,
                    p_new_pref       in int )
    as
        mx_send_id     int := NetDB_Name.verify_or_die (p_sent_to_id, p_sent_to, 'mx', true);
        mx_receive_id  int := verify_received_by_or_die (p_received_by_id, p_received_by);

        send_domain    Name.fullname%TYPE;
        send_domain_id int;

        no_domain_access exception;
    begin
        NetDB_Name.locked_or_die (p_op_uid, p_lock, mx_receive_id);

        begin
            send_domain    := NetDB_Name.fullname (mx_send_id);
            send_domain_id := NetDB_Domain.verify_or_die (null, send_domain);

            if (not NetDB_Domain.can_use_as_name (p_op_uid, send_domain_id)) then
                raise no_domain_access;
            end if;
        exception
            when NetDB.record_not_found then
                send_domain    := DNS.domain (NetDB_Name.fullname (mx_send_id));
                send_domain_id := NetDB_Domain.verify_or_die (null, send_domain);

                if (not NetDB_Domain.can_use (p_op_uid, send_domain_id)) then
                    raise no_domain_access;
                end if;
        end;

        update MXrecord set preference = p_new_pref
         where sent_to = mx_send_id and received_by = mx_receive_id;
    exception
        when no_domain_access then
            raise_application_error (NetDB.ex_no_access_to_record,
                'The user "' || NetDB_User.handle (p_op_uid) ||
                '" cannot modify MX priorities in the domain "' || send_domain || '".');
    end upd;

end NetDB_MXrecord;
/
