-- NetDB_Model.sql --- contains functions related to Model table

create or replace package body NetDB_Model as

    function make_and_model (p_model_id in int) return varchar
    as
        make_and_model varchar(256);
    begin
        select make || ' ' || modelname into make_and_model from Model where id = p_model_id;

        return make_and_model;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('model', p_model_id);
    end make_and_model;


    -- Replaces "old model" with "new model" and updates any nodes that had "old model".
    procedure supplant ( p_op_uid        in int,
                         p_model_id_old  in int,
                         p_make_new      in varchar,
                         p_modelname_new in varchar )
    as
        v_old  int            := verify_or_die (p_model_id_old);
        v_make Make.make%TYPE := NetDB_Make.verify_or_die (p_make_new);
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Model');

        update Node set model = verify_or_die (null, p_make_new, p_modelname_new)
         where model = p_model_id_old;

        delete from Model where id = p_model_id_old;
    end supplant;


    function list (p_filter in varchar := null) return NetDB.ref_cursor
    as
        v_key varchar(256);

        out_record  result_set := result_set ();
        ret_cursor  NetDB.ref_cursor;
    begin
        for r in ( select id, make, modelname, modeltype from Model
                    where modelname like coalesce (p_filter, '%') )
        loop
            -- form key
            v_key := r.make || ' ' || r.modelname;
            if (r.modeltype is not null) then
                v_key := v_key || ' : ' || r.modeltype;
            end if;

            out_record.extend(5);
            out_record(out_record.COUNT - 4) := result_row ('data',  v_key,     0,  null,    null,        null, null);
            out_record(out_record.COUNT - 3) := result_row ('key',  'model', null, 'string', r.modelname, null, null);
            out_record(out_record.COUNT - 2) := result_row ('key',  'type',  null, 'string', r.modeltype, null, null);
            out_record(out_record.COUNT - 1) := result_row ('key',  'make',  null, 'string', r.make,      null, null);
            out_record(out_record.COUNT)     := result_row ('key',  'id',    null, 'string', r.id,        null, null);
        end loop;

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end list;


    procedure del ( p_op_uid   in int,
                    p_model_id in int )
    as
        v_model_id int := verify_or_die (p_model_id);
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Model');

        delete from Model where id = p_model_id;
    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('Model', make_and_model (p_model_id));
    end del;


    function ins ( p_op_uid    in int,
                   p_make      in varchar,
                   p_modelname in varchar,
                   p_modeltype in varchar := null ) return NetDB.ref_cursor
    as
        v_model_id int;

        v_make      Make.make%TYPE := NetDB_Make.verify_or_die (p_make);
        v_modeltype Model.modeltype%TYPE;

        out_record result_set := result_set ();
        ret_cursor NetDB.ref_cursor;
    begin
        -- people with node access are allowed to create new models.
        if (not NetDB_Access.check_access (p_op_uid, 'Model')) then
            NetDB_Access.has_access_or_die (p_op_uid, 'Node');

            if (p_modeltype is not null) then
                raise_application_error (NetDB.ex_no_access_to_record,
                    'The user "' ||  NetDB_User.handle (p_op_uid) ||
                    '" does not have sufficient access rights to edit the model type.');
            end if;
        end if;

        NetDB_Util.validate (p_modelname, 'model name', 'A model must have a name');

        if (p_modeltype is not null) then
            v_modeltype := NetDB_ModelType.verify_or_die (p_modeltype);
        end if;

        insert into Model (make,   modelname,   modeltype,   modified_by, modified)
                   values (p_make, p_modelname, v_modeltype, p_op_uid,    systimestamp)
        returning id into v_model_id;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'id', 0, 'string', v_model_id, null, null);

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.unique_constraint_failed then
            raise_application_error (NetDB.ex_record_already_exists,
                'The Model "' || p_modelname || '" already exists with the Make "' || p_make || '".');

        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    procedure upd ( p_op_uid        in int,
                    p_model_id      in int,
                    p_make_new      in varchar,
                    p_modelname_new in varchar,
                    p_modeltype_new in varchar := null)
    as
        v_existing_id int;

        v_make      Make.make%TYPE := NetDB_Make.verify_or_die (p_make_new);
        v_modeltype Model.modeltype%TYPE;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Model');
        NetDB_Util.validate (p_modelname_new, 'model name', 'A model must have a name');

        if (p_modeltype_new is not null) then
            v_modeltype := NetDB_ModelType.verify_or_die (p_modeltype_new);
        end if;

        update Model set make        = p_make_new,
                         modelname   = p_modelname_new,
                         modeltype   = v_modeltype,
                         modified_by = p_op_uid,
                         modified    = systimestamp
         where id = p_model_id;
    exception
        when NetDB.unique_constraint_failed then
            raise_application_error (NetDB.ex_record_already_exists,
                'The Model "' || p_modelname_new || '" already exists with the Make "' || p_make_new || '".');

        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    -- Verify model ID / handle (a combination of make and model) or complain.
    function verify_or_die ( p_model_id in int     := null,
                             p_make     in varchar := null,
                             p_model    in varchar := null ) return int
    as
        v_identifier varchar(256) := p_make || ' ' || p_model;
        v_model_id   int          := 0;
    begin
        if (p_model_id is null and p_make is null and p_model is null) then
            NetDB_Util.throw_no_record_id_handle ();

        elsif ( (p_make is not null and p_model is     null) or
                (p_make is     null and p_model is not null) )
        then
            raise_application_error (NetDB.ex_bad_parameter,
                'The make and model must both be non null.');

        elsif (p_model_id is not null and p_make is not null and p_model is not null) then
            select id into v_model_id from Model
             where lower (make)      = lower (p_make)  and
                   lower (modelname) = lower (p_model) and
                   id                = p_model_id;

        elsif (p_make is not null and p_model is not null) then
            select id into v_model_id from Model
             where lower (make) = lower (p_make) and lower (modelname) = lower (p_model);

        else -- p_model_id is not null
            v_identifier := p_model_id;

            select id into v_model_id from Model where id = p_model_id;
        end if;

        return v_model_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('model', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('model');
    end verify_or_die;

end NetDB_Model;
/
