-- NetDB_Node.sql --- contains functions related to nodes

create or replace package body NetDB_Node as

    -- given node id returns node types
    function list_types (p_node_id in int) return node_type_result pipelined as
    begin
        for r in (select type from Node_NodeType where node = p_node_id)
        loop
            pipe row (node_type_row (r.type));
        end loop;
    end list_types;


    -- Add or remove OSes that are on a node.
    procedure OS ( p_action  in varchar,  -- { add, remove }
                   p_op_uid  in int,
                   p_lock    in timestamp,
                   p_node_id in int     := null,
                   p_node    in varchar := null,
                   p_os      in varchar,
                   p_version in varchar )
    as
        v_node_id int         := verify_or_die (p_node_id, p_node);
        v_os      varchar(64) := NetDB_OS.verify_or_die (p_os);
      --v_os      OS.os%TYPE  := NetDB_OS.verify_or_die (p_os); -- Try it and see
    begin
        locked_or_die (p_op_uid, p_lock, v_node_id);

        case lower (p_action)
            when 'add' then
                insert into Node_OS (node, os, version) values (v_node_id, p_os, p_version);
            when 'remove' then
                delete from Node_OS where node = v_node_id and os = p_os;
            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;
    exception
        when NetDB.unique_constraint_failed then
            null;
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end OS;


    -- Add or remove administrators relating to a node.
    procedure admin ( p_action    in varchar,  -- { add, remove }
                      p_op_uid    in int,
                      p_lock      in timestamp,
                      p_node_id   in int     := null,
                      p_node      in varchar := null,
                      p_ateam_id  in varchar := null,
                      p_ateam     in varchar := null,
                      p_person_id in int     := null,
                      p_regid     in varchar := null )
    as
        v_node_id   int := verify_or_die (p_node_id, p_node);
        v_person_id int := null;
        v_ateam_id  int := null;
    begin
        locked_or_die (p_op_uid, p_lock, v_node_id);

        if (p_ateam_id is not null or p_ateam is not null) then
            v_ateam_id := NetDB_AdminTeam.verify_or_die (p_ateam_id, p_ateam);
        end if;

        if (p_person_id is not null or p_regid is not null) then
            v_person_id := NetDB_Person.verify_or_die (p_person_id, p_regid);
        end if;

        case lower (p_action)
            when 'add' then
                begin
                    if (v_ateam_id  is not null) then
                        insert into Node_Admin (node, adminteam) values (v_node_id, v_ateam_id);
                    end if;
                exception
                    when NetDB.unique_constraint_failed then
                        null;
                end;

                begin
                    if (v_person_id is not null) then
                        insert into Node_Admin (node, person) values (v_node_id, v_person_id);
                    end if;
                exception
                    when NetDB.unique_constraint_failed then
                        null;
                end;

            when 'remove' then
                if (v_ateam_id  is not null) then
                    delete from Node_Admin where node = v_node_id and adminteam = v_ateam_id;
                end if;

                if (v_person_id is not null) then
                    delete from Node_Admin where node = v_node_id and person = v_person_id;
                end if;

            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;
    end admin;


    -- Add or remove users listed as using a node.
    procedure user ( p_action    in varchar,  -- { add, remove }
                     p_op_uid    in int,
                     p_lock      in timestamp,
                     p_node_id   in int     := null,
                     p_node      in varchar := null,
                     p_person_id in int     := null,
                     p_regid     in varchar := null )
    as
        v_node_id   int := verify_or_die (p_node_id, p_node);
        v_person_id int := NetDB_Person.verify_or_die (p_person_id, p_regid);
    begin
        locked_or_die (p_op_uid, p_lock, v_node_id);

        case lower (p_action)
            when 'add' then
                insert into Node_User (node, person) values (v_node_id, v_person_id);
            when 'remove' then
                delete from Node_User where node = v_node_id and person = v_person_id;
            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;
    exception
        when NetDB.unique_constraint_failed then
            null;
    end user;


    -- Add or remove types relating to a node.
    procedure type ( p_action  in varchar,  -- { add, remove }
                     p_op_uid  in int,
                     p_lock    in timestamp,
                     p_node_id in int     := null,
                     p_node    in varchar := null,
                     p_type    in varchar )
    as
        v_node_id int := verify_or_die (p_node_id, p_node);
    begin
        NetDB_Access.has_access_or_die (p_op_uid, p_type, v_node_id);
        locked_or_die (p_op_uid, p_lock, v_node_id);

        case lower (p_action)
            when 'add' then
                insert into Node_NodeType (node, type) values (v_node_id, p_type);
            when 'remove' then
                delete from Node_NodeType where node = v_node_id and type = p_type;
            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;
    exception
        when NetDB.unique_constraint_failed then
            null;

        when NetDB.check_constraint_failed then
            raise_application_error (NetDB.ex_record_not_found,
                '"' || p_type || '" is not a valid node type.');

        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end type;


    -- Add or remove tags relating to a node.
    procedure tag ( p_action  in varchar, -- { add, remove }
                    p_op_uid  in int,
                    p_lock    in timestamp,
                    p_node_id in int,
                    p_node    in varchar,
                    p_label   in varchar,
                    p_value   in varchar )
    as
        v_node_id int := verify_or_die (p_node_id, p_node);
    begin
        NetDB_Util.validate (p_label, 'label');
        NetDB_Util.validate (p_value, 'value');

        locked_or_die (p_op_uid, p_lock, v_node_id);

        case lower (p_action)
            when 'add' then
                insert into NodeTag (node, label, value) values (v_node_id, p_label, p_value);
            when 'remove' then
                delete from NodeTag where node = v_node_id and label = p_label;
            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;
    exception
        when NetDB.unique_constraint_failed then
            raise_application_error (NetDB.ex_record_already_exists,
                'The label "' || p_label || '" already exists on the node "' ||
                handle (v_node_id) || '".');

        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end tag;


    -- Check to see if a user has a lock for a record.
    procedure locked_or_die ( p_op_uid  in int,
                              p_lock    in timestamp,
                              p_node_id in int )
    as
        v_lock timestamp := null;
    begin
        select locked into v_lock from Node
         where id = p_node_id and locked_by = p_op_uid and locked = p_lock and
               Util.datediff ('second', p_lock) < NetDB_Default.get_int ('LockTimeOut');
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_not_locked ( 'node',
                                                  handle (p_node_id),
                                                  NetDB_User.handle (p_op_uid) );
    end locked_or_die;


    -- Lock a record.
    function lock_record ( p_op_uid  in int,
                           p_node_id in int ) return NetDB.ref_cursor
    as
        v_node_id   int := verify_or_die (p_node_id);
        v_lock      timestamp;
        v_locked_by int;

        out_record result_set       := result_set ();
        ret_cursor NetDB.ref_cursor := null;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Node', v_node_id);
        NetDB_Owner.has_access_or_die (p_op_uid, v_node_id, 'Node');

        select locked, locked_by into v_lock, v_locked_by from Node where id = p_node_id;

        -- if not locked, locked by this user, or the lock has expired, we're good to go
        if ( v_lock is null or v_locked_by = p_op_uid or
             Util.datediff ('second', v_lock) > NetDB_Default.get_int ('LockTimeOut') )
        then
               update Node set locked    = systimestamp,
                               locked_by = p_op_uid
                where id = p_node_id
            returning locked into v_lock;
        else
            NetDB_Util.throw_locked_by_other ( 'node',
                                                handle (p_node_id),
                                                NetDB_User.handle (v_locked_by),
                                                v_lock );
        end if;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'locked', 0, 'timestamp', null, v_lock, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end lock_record;


    -- Unlock a record.
    procedure unlock_record ( p_op_uid  in int,
                              p_lock    in timestamp,
                              p_node_id in int ) as
    begin
        locked_or_die (p_op_uid, p_lock, p_node_id);

        update Node set locked    = null,
                        locked_by = null
         where id = p_node_id and locked = p_lock and locked_by = p_op_uid;
    end unlock_record;


    -- returns node info
    function info ( p_node_id in int,
                    p_user_id in int := null ) return NetDB.ref_cursor
    as
        v_id varchar(1024) := p_node_id;
    begin
        if (v_id is null) then
            v_id := 'null';
        end if;

        return NetDB_Util.collect_into_cursor ('NetDB_Node.info_pipelined (' || v_id || ',' || p_user_id || ')');
    end info;


    function info_pipelined ( p_node_id in int,
                              p_user_id in int := null ) return result_set pipelined
    as
        v_node_fullname Name.fullname%TYPE;

        res_row result_row := result_row (null, null, null, null, null, null, null);

        ip_cursor     NetDB.ref_cursor := null;
        person_cursor NetDB.ref_cursor := null;
        intfc_cursor  NetDB.ref_cursor := null;

        -- Node info expressed as a cursor (instead of defining a bunch of variables).
        cursor node_info is
        select Organization.organization dept, Location.name location,    Node.room,
               Node.state,                     IPaddressSpace.prefix_txt, Model.make,
               Model.modelname,                Node.comments,             Node.expiration,
               Node.modified,                  Node.created,              Node.created_by,
               Node.modified_by,               Location.site,             Location.code
          from Node
          left join IPaddressSpace on Node.ipaddressspace = IPaddressSpace.id
               join Organization   on Organization.id     = Node.organization
               join Location       on Location.id         = Node.location
               join Model          on Model.id            = Node.model
         where Node.id = verify_or_die (p_node_id);

         i node_info%ROWTYPE;
    begin
        open node_info;
        fetch node_info into i;
        close node_info;

        pipe row (result_row ('key', 'object',     null, 'string',    'node',       null,         null));
        pipe row (result_row ('key', 'id',         null, 'string',    p_node_id,    null,         null));
        pipe row (result_row ('key', 'dept',       null, 'string',    i.dept,       null,         null));
        pipe row (result_row ('key', 'location',   null, 'string',    i.location,   null,         null));
        pipe row (result_row ('key', 'site',       null, 'string',    i.site,       null,         null));
        pipe row (result_row ('key', 'code',       null, 'string',    i.code,       null,         null));
        pipe row (result_row ('key', 'room',       null, 'string',    i.room,       null,         null));
        pipe row (result_row ('key', 'state',      null, 'string',    i.state,      null,         null));
        pipe row (result_row ('key', 'make',       null, 'string',    i.make,       null,         null));
        pipe row (result_row ('key', 'model',      null, 'string',    i.modelname,  null,         null));
        pipe row (result_row ('key', 'addr_space', null, 'string',    i.prefix_txt, null,         null));
        pipe row (result_row ('key', 'comment',    null, 'string',    i.comments,   null,         null));
        pipe row (result_row ('key', 'exp.date',   null, 'timestamp', null,         i.expiration, null));
        pipe row (result_row ('key', 'modified',   null, 'timestamp', null,         i.modified,   null));
        pipe row (result_row ('key', 'created',    null, 'timestamp', null,         i.created,    null));

        if ( p_user_id is not null and
             NetDB_Owner.check_access (p_user_id, p_node_id, 'Node') and
             NetDB_Access.check_access (p_user_id, 'Node', p_node_id) )
        then
            pipe row (result_row ('key', 'can_modify?', null, 'boolean', null, null, 1));
        else
            pipe row (result_row ('key', 'can_modify?', null, 'boolean', null, null, 0));
        end if;

        -- now created.by
        pipe row (result_row ('data', 'created.by', 0, null, null, null, null));

        person_cursor := NetDB_Person.info (i.created_by);
        loop
            fetch person_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                     res_row.value_string, res_row.value_timestamp, res_row.value_int;

            exit when person_cursor%notfound;

            pipe row (res_row);
        end loop;
        close person_cursor;

        -- now mod.by
        pipe row (result_row ('data', 'mod.by', 0, null, null, null, null));

        person_cursor := NetDB_Person.info (i.modified_by);
        loop
            fetch person_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                     res_row.value_string, res_row.value_timestamp, res_row.value_int;

            exit when person_cursor%notfound;

            pipe row (res_row);
        end loop;
        close person_cursor;

        -- groups
        pipe row (result_row ('data', 'groups', 0, null, null, null, null));

        for r in ( select Owner.label
                     from Node_Owner join Owner on Owner.id = Node_Owner.owner
                    where Node_Owner.node = p_node_id )
        loop
            pipe row (result_row ('key', r.label, null, 'string', ' ', null, null));
        end loop;

        -- oses
        pipe row (result_row ('data', 'oses', 0, null, null, null, null));

        for r in (select os, version from Node_OS where node = p_node_id)
        loop
            pipe row (result_row ('key', r.os, null, 'string', nvl(r.version, ' '), null, null));
        end loop;

        -- types
        pipe row (result_row ('data', 'types', 0, null, null, null, null));

        for r in ( select Node_NodeType.type, NodeType.rank, NodeType.extkey
                     from Node_NodeType
                     join NodeType on NodeType.type = Node_NodeType.type
                    where node = p_node_id )
        loop
            pipe row (result_row ('data', r.type,  1,     null,     null,     null, null));
            pipe row (result_row ('key', 'extkey', null, 'string',  r.extkey, null, null));
            pipe row (result_row ('key', 'type',   null, 'string',  r.type,   null, null));
            pipe row (result_row ('key', 'rank',   null, 'integer', null,     null, r.rank));
       end loop;

        -- tags
        pipe row (result_row ('data', 'tags', 0, null, null, null, null));

        for r in (select label, value from NodeTag where node = p_node_id)
        loop
            pipe row (result_row ('key', r.label, null, 'string', nvl (r.value, ' '), null, null));
        end loop;

        -- users
        pipe row (result_row ('data', 'users', 0, null, null, null, null));

        for r in (select person from Node_User where node = p_node_id)
        loop
            pipe row (result_row ('data', r.person, 1, null, null, null, null));

            person_cursor := NetDB_Person.info (r.person);
            loop
                fetch person_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                         res_row.value_string, res_row.value_timestamp, res_row.value_int;

                exit when person_cursor%notfound;

                pipe row (res_row);
            end loop;
            close person_cursor;
        end loop;

        -- admins
        pipe row (result_row ('data', 'admins', 0, null, null, null, null));

        pipe row (result_row ('data', 'ateams', 1, null, null, null, null));
        for r in (select adminteam from Node_Admin where node = p_node_id and adminteam is not null)
        loop
            for d in (select name, phone, email from AdminTeam where id = r.adminteam)
            loop
                pipe row (result_row ('data', d.name, 2, null, null, null, null));
                pipe row (result_row ('key', 'phone', null, 'string', d.phone, null, null));
                pipe row (result_row ('key', 'email', null, 'string', d.email, null, null));
            end loop;
        end loop;

        pipe row (result_row ('data', 'admins', 1, null, null, null, null));
        for r in (select person from Node_Admin where node = p_node_id and person is not null)
        loop
            pipe row (result_row ('data', r.person, 2, null, null, null, null));

            person_cursor := NetDB_Person.info (r.person);
            loop
                fetch person_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                         res_row.value_string, res_row.value_timestamp, res_row.value_int;

                exit when person_cursor%notfound;

                pipe row (res_row);
            end loop;
            close person_cursor;
        end loop;

        -- ipc addresses
        ip_cursor := NetDB_IPaddress.info (null, p_node_id);
        loop
            fetch ip_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                 res_row.value_string, res_row.value_timestamp, res_row.value_int;

            exit when ip_cursor%notfound;
            if( ip_cursor%rowcount = 1 ) then
                pipe row (result_row ('data', 'ipc_addresses', 0, null, null, null, null));
            end if;

            pipe row (res_row);
        end loop;
        close ip_cursor;

        -- node names
        for r in ( select *
                    from table (cast (NetDB_Name.names (null, null, p_node_id, null) as result_set)) )
        loop
            pipe row (result_row ( r.type, r.key, r.hash_level,
                                   r.value_type, r.value_string, r.value_timestamp, r.value_int ));
        end loop;

        -- interfaces
        pipe row (result_row ('data', 'interfaces', 0, null, null, null, null));

        for intfc in (select id from Interface where node = p_node_id)
        loop
            intfc_cursor := NetDB_Interface.info (intfc.id);

            loop
                fetch intfc_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                        res_row.value_string, res_row.value_timestamp, res_row.value_int;

                exit when intfc_cursor%notfound;

                res_row.hash_level := res_row.hash_level + 1;
                pipe row (res_row);
            end loop;
            close intfc_cursor;
        end loop;

        -- MX slaves
        v_node_fullname := lower (handle (p_node_id));

        for r in ( select rownum, r_by.node id, s_to.fullname sent_to_name, r_by.fullname rec_by_name
                     from MXrecord mx
                     join Name s_to on s_to.id = mx.sent_to
                     join Name r_by on r_by.id = mx.received_by
                    where mx.sent_to = ( select id from Name where dnslabel    = 'MX' and
                                                                   fullname_lc = v_node_fullname ) and
                          r_by.fullname_lc != v_node_fullname )
        loop
            if (r.rownum = 1) then
                pipe row (result_row ('data', 'mxSlaves',     0, null, null, null, null));
                pipe row (result_row ('data',  r.id,          1, null, null, null, null));
                pipe row (result_row ('data',  r.rec_by_name, 2, null, null, null, null));
            end if;

            pipe row (result_row ('key', r.sent_to_name, null, 'string', ' ', null, null));
        end loop;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('node', handle (p_node_id));
    end info_pipelined;


    procedure del ( p_op_uid  in int,
                    p_lock    in timestamp,
                    p_node_id in int )
    as
        v_node_id int := verify_or_die (p_node_id);
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Node', p_node_id);
        NetDB_Owner.has_access_or_die (p_op_uid, p_node_id, 'Node');

        locked_or_die (p_op_uid, p_lock, p_node_id);
        NetDB_Log.ins (p_op_uid, 'delete', p_node_id, 'node');

        delete from Node where id = p_node_id;
    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('node', handle (p_node_id));
    end del;


    function ins ( p_op_uid          in int,
                   p_org             in varchar,
                   p_location        in varchar,
                   p_room            in varchar,
                   p_state           in varchar,
                   p_make            in varchar,
                   p_model           in varchar,
                   p_comment         in varchar   := null,
                   p_expire_date     in timestamp := null,
                   p_dhcpservice     in varchar   := null,
                   p_ipadddressspace in varchar   := null ) return NetDB.ref_cursor
    as
        v_state      State.state%TYPE := NetDB_State.verify_or_die (p_state);
        v_org_id     int := NetDB_Organization.verify_or_die (null, p_org);
        v_loc_id     int := NetDB_Location.verify_or_die (null, p_location);
        v_model_id   int := NetDB_Model.verify_or_die (null, p_make, p_model);
        v_service_id int := null;
        v_space_id   int := null;

        v_node_id    int       := null;
        v_lock       timestamp := null;

        sys_time_stamp timestamp := systimestamp;

        out_record  result_set := result_set ();
        ret_cursor  NetDB.ref_cursor;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Node');

        NetDB_Util.validate (p_room, 'room', 'A room number must be specified.');
        NetDB_Util.validate (p_comment, 'comment');

        if (p_dhcpservice is not null) then
            v_service_id := NetDB_DHCPservice.verify_or_die (null, p_dhcpservice);
        end if;

        if (p_ipadddressspace is not null) then
            v_space_id := NetDB_IPaddressSpace.verify_or_die (null, p_ipadddressspace);
        end if;

        NetDB_State.has_access_or_die (p_op_uid, v_state);

        insert into Node ( organization, location,       room,           state,          model,
                           comments,     expiration,     dhcpservice,    ipaddressspace, created_by,
                           modified_by,  created,        modified,       locked_by,      locked )
                  values ( v_org_id,     v_loc_id,       p_room,         v_state,        v_model_id,
                           p_comment,    p_expire_date,  v_service_id,   v_space_id,     p_op_uid,
                           p_op_uid,     sys_time_stamp, sys_time_stamp, p_op_uid,       sys_time_stamp )
        returning id, locked into v_node_id, v_lock;

        NetDB_Log.set_action ('insert');

        out_record.extend(2);
        out_record(1) := result_row ('key', 'id',     0, 'string',    v_node_id, null,   null);
        out_record(2) := result_row ('key', 'locked', 0, 'timestamp', null,      v_lock, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    procedure upd ( p_op_uid          in int,
                    p_lock            in timestamp,
                    p_node_id         in int,
                    p_org             in varchar,
                    p_location        in varchar,
                    p_room            in varchar,
                    p_state           in varchar,
                    p_make            in varchar,
                    p_model           in varchar,
                    p_comment         in varchar   := null,
                    p_expire_date     in timestamp := null,
                    p_dhcpservice     in varchar   := null,
                    p_ipadddressspace in varchar   := null )
    as
        v_node_id    int := verify_or_die (p_node_id);
        v_state      State.state%TYPE := NetDB_State.verify_or_die (p_state);
        v_org_id     int := NetDB_Organization.verify_or_die (null, p_org);
        v_loc_id     int := NetDB_Location.verify_or_die (null, p_location);
        v_model_id   int := NetDB_Model.verify_or_die (null, p_make, p_model);
        v_service_id int := null;
        v_space_id   int := null;

        node_row Node%ROWTYPE;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Node', p_node_id);
        NetDB_Owner.has_access_or_die (p_op_uid, p_node_id, 'Node');

        locked_or_die (p_op_uid, p_lock, p_node_id);

        NetDB_Util.validate (p_room, 'room', 'A room number must be specified.');
        NetDB_Util.validate (p_comment, 'comment');

        if (p_dhcpservice is not null) then
            v_service_id := NetDB_DHCPservice.verify_or_die (null, p_dhcpservice);
        end if;

        if (p_ipadddressspace is not null) then
            v_space_id := NetDB_IPaddressSpace.verify_or_die (null, p_ipadddressspace);
        end if;

        select * into node_row from Node where id = p_node_id;

        if (node_row.state != p_state) then
            NetDB_State.has_access_or_die (p_op_uid, node_row.state);
            NetDB_State.has_access_or_die (p_op_uid, v_state);
        end if;

        update Node set organization   = v_org_id,
                        location       = v_loc_id,
                        room           = p_room,
                        state          = v_state,
                        model          = v_model_id,
                        comments       = p_comment,
                        expiration     = p_expire_date,
                        DHCPservice    = v_service_id,
                        IPaddressSpace = v_space_id,
                        modified_by    = p_op_uid,
                        modified       = systimestamp
         where id = p_node_id;

        NetDB_Log.set_action ('update');
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    procedure verify_and_log ( p_op_uid  in int,
                               p_node_id in int ) as
    begin
        NetDB_Owner.has_access_or_die (p_op_uid, p_node_id, 'Node');

        -- A node must have a name
        select count (*) into NetDB.existence_counter from Name where node = p_node_id;
        if (NetDB.existence_counter < 1) then
            raise_application_error (NetDB.ex_record_requires_name,
                'Nodes must have at least one name.');
        end if;

        -- A node must have at least one group.
        select count (*) into NetDB.existence_counter from Node_Owner where node = p_node_id;

        if (NetDB.existence_counter < 1) then
            raise_application_error (NetDB.ex_record_requires_group,
                'The node "' || handle (p_node_id) || '" must have only one group.');
        end if;

        -- A node must have at least one OS.
        select count (*) into NetDB.existence_counter from Node_OS where node = p_node_id;

        if (NetDB.existence_counter < 1) then
            raise_application_error (NetDB.ex_node_must_have_OS,
                'The node "' || handle (p_node_id) || '" must have at least one OS.');
        end if;

        -- A node must have at least one Admin.
        select count (*) into NetDB.existence_counter from Node_Admin where node = p_node_id;

        if (NetDB.existence_counter < 1) then
            raise_application_error (NetDB.ex_node_must_have_admin,
                'The node "' || handle (p_node_id) || '" must have at least one admin.');
        end if;

        -- If a node has IPC addresses, they all must have a name.
        select count (*) into NetDB.existence_counter from IPaddress
          left join Name on Name.ipaddress = IPaddress.id
         where IPaddress.ipcnode = p_node_id and Name.name is null;

        if (NetDB.existence_counter > 0) then
            raise_application_error (NetDB.ex_record_requires_name, 'All IPC addresses must have a name.');
        end if;

        -- if the node is a template, make sure there are no interfaces
        select count (*) into NetDB.existence_counter
          from Node_NodeType
          join Interface on Interface.node = Node_NodeType.node
         where Node_NodeType.node = p_node_id and Node_NodeType.type = 'Template';

        if (NetDB.existence_counter > 0) then
            raise_application_error (NetDB.ex_bad_interface, 'Template nodes cannot have interfaces.');
        end if;

        -- if the node is not an advanced node, make sure there are no named IP addresses on
        -- the interface, and that the interface doesn't have a name.
        select count (*) into NetDB.existence_counter
          from Node_NodeType
         where node = p_node_id and type = 'Advanced';

        if (NetDB.existence_counter = 0) then
            select count (*) into NetDB.existence_counter
              from Interface
              join IPaddress       on IPaddress.interface  = Interface.id
         left join Name iface_name on iface_name.interface = Interface.id
         left join Name ip_name    on ip_name.ipaddress    = IPaddress.id
             where Interface.node = p_node_id and coalesce (ip_name.fullname, iface_name.fullname) is not null;

            if (NetDB.existence_counter > 0) then
                raise_application_error (NetDB.ex_bad_interface,
                    'Only advanced nodes can have named IP addresses and named interfaces.');
            end if;
        end if;

        NetDB_Log.ins (p_op_uid, NetDB_Log.get_action (), p_node_id, 'node');
    end verify_and_log;


    -- Verify node ID / handle or complain.
    function verify_or_die ( p_node_id  in int     := null,
                             p_fullname in varchar := null ) return int
    as
        v_identifier varchar(256) := p_fullname;
        v_node_id    int;
    begin
        if (p_node_id is null and p_fullname is null) then
            NetDB_Util.throw_no_record_id_handle ();

        elsif (p_node_id is not null and p_fullname is not null) then
            select related_node into v_node_id from Name
             where related_node = p_node_id and fullname_lc = lower (p_fullname);

        elsif (p_fullname is not null) then
            select related_node into v_node_id from Name
             where related_node is not null and fullname_lc = lower (p_fullname);

        else -- p_node_id is not null
            v_identifier := p_node_id;

            select id into v_node_id from Node where id = p_node_id;
        end if;

        return v_node_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('node', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('node');
    end verify_or_die;


    function handle (p_node_id in int) return varchar
    as
        v_name Name.fullname%TYPE := null;
    begin
        select fullname into v_name from Name
         where node = p_node_id and rownum = 1
         order by fullname;

        return v_name;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('node', p_node_id);
    end handle;

end NetDB_Node;
/
