-- NetDB_NodeType.sql --- contains functions related to node types

create or replace package body NetDB_NodeType as

    -- List node types. If a user ID argument is provided, list only
    -- the node types the user is able to access.
    function list (p_user_id in int := null) return NetDB.ref_cursor
    as
        out_record  result_set := result_set ();
        ret_cursor  NetDB.ref_cursor;

        type_row  NodeType%ROWTYPE;
        the_query NetDB.ref_cursor;
    begin
        if (p_user_id is null or NetDB_User.has_all_records (p_user_id)) then
            open the_query for
                select * from NodeType;
        else
            open the_query for
                select nt.*
                  from NodeType          nt
             left join Users_AccessGroup ua on ua.accessgroup = nt.accessgroup
                 where ua.userid = p_user_id or nt.accessgroup is null;
        end if;

        loop
            fetch the_query into type_row;

            exit when the_query%notfound;

            out_record.extend(8);
            out_record(out_record.COUNT - 7) := result_row ('data', type_row.type, 0,    'string',  null,                 null, null);
            out_record(out_record.COUNT - 6) := result_row ('key', 'type',         null, 'string',  type_row.type,        null, null);
            out_record(out_record.COUNT - 5) := result_row ('key', 'extkey',       null, 'string',  type_row.extkey,      null, null);
            out_record(out_record.COUNT - 4) := result_row ('key', 'plural',       null, 'string',  type_row.plural,      null, null);
            out_record(out_record.COUNT - 3) := result_row ('key', 'abbrev',       null, 'string',  type_row.abbrev,      null, null);
            out_record(out_record.COUNT - 2) := result_row ('key', 'rank',         null, 'integer', null,                 null, type_row.rank);
            out_record(out_record.COUNT - 1) := result_row ('key', 'description',  null, 'string',  type_row.description, null, null);
            out_record(out_record.COUNT)     := result_row ('key', 'access',       null, 'string',  type_row.accessgroup, null, null);
        end loop;

        close the_query;

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end list;


    -- verify_or_die wrapped for boolean return
    function exists (p_type in varchar) return boolean
    as
        v_found NodeType.type%TYPE;
    begin
        v_found := verify_or_die (p_type);

        return true;
    exception
        when NetDB.record_not_found then
            return false;
    end exists;


    -- Verify that p_type is a node type or complain.
    function verify_or_die (p_type in varchar) return varchar
    as
        v_found NodeType.type%TYPE;
    begin
        if (p_type is null) then
            NetDB_Util.throw_no_record_id_handle ();
        else
            select type into v_found from NodeType where lower (p_type) = lower (type);
        end if;

        return v_found;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('NodeType', p_type);
    end verify_or_die;


    -- Unused and untested.
    procedure del ( p_op_uid in int,
                    p_type   in varchar ) as
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'NodeType');

        delete from NodeType where lower (type) = lower (p_type);
    end del;


    -- Unused and untested.
    procedure ins ( p_op_uid       in int,
                    p_type         in varchar,
                    p_extkey       in varchar,
                    p_plural       in varchar,
                    p_abbrev       in varchar,
                    p_rank         in int,
                    p_desc         in varchar := null,
                    p_access_group in int     := null ) as
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'NodeType');

        NetDB_Util.validate (p_type, 'type', 'The Node Type must be named.');
        NetDB_Util.validate (p_desc, 'description');

        insert into NodeType (  type,   extkey,   plural,   rank,   description,   accessgroup)
                      values (p_type, p_extkey, p_plural, p_rank, p_desc,        p_access_group);
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- Unused and untested.
    procedure upd ( p_op_uid       in int,
                    p_type         in varchar,
                    p_type_new     in varchar := null,
                    p_extkey       in varchar,
                    p_plural       in varchar,
                    p_abbrev       in varchar,
                    p_rank         in int,
                    p_desc         in varchar := null,
                    p_access_group in int     := null ) as
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'NodeType');

        NetDB_Util.validate (p_desc, 'description');

        update NodeType set type        = coalesce (p_type_new, type),
                            extkey      = p_extkey,
                            plural      = p_plural,
                            abbrev      = p_abbrev,
                            rank        = p_rank,
                            description = p_desc,
                            accessgroup = p_access_group
         where lower (type) = lower (p_type);
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;

end NetDB_NodeType;
/
