-- NetDB_Organization.sql --- contains functions related to Organization table

create or replace package body NetDB_Organization as

    procedure consultant ( p_action    in varchar,  -- { insert, update, delete }
                           p_op_uid    in int,
                           p_person_id in int,
                           p_org_id    in int     := null,
                           p_org_name  in varchar := null,
                           p_primary   in int     := 0 )
    as
        v_action    varchar(16) := lower (p_action);

        v_org_id    int := NetDB_Organization.verify_or_die (p_org_id, p_org_name);
        v_person_id int := NetDB_Person.verify_or_die (p_person_id, null);
    begin
        if (v_action in ('insert', 'update') and not p_primary in (0, 1)) then
            raise_application_error (NetDB.ex_bad_parameter,
                '"Primary" flag value must be either 0 or 1.');
        end if;

        NetDB_Access.has_access_or_die (p_op_uid, 'systemmanagement');

        case v_action
            when 'insert' then
                insert into Consultant (person,      organization, primary)
                                values (v_person_id, v_org_id,     p_primary);

            when 'update' then
                update Consultant set primary = p_primary
                 where person = p_person_id and organization = v_org_id;

            when 'delete' then
                delete from Consultant where person = p_person_id and organization = v_org_id;

            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;
    exception
        when NetDB.unique_constraint_failed then
            null;
    end consultant;


    -- Replaces "old organization" with "new organization". This
    -- includes moving any LNA's between organizations
    function supplant ( p_op_uid in int,
                        p_id_old in int,
                        p_id_new in int ) return NetDB.ref_cursor
    as
        old_id int := verify_or_die (p_id_old);
        new_id int := verify_or_die (p_id_new);

        out_record result_set := result_set ();
        ret_cursor NetDB.ref_cursor;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'organization');

        for r in (select * from AdminTeam_Orgs where organization = old_id)
        loop
            begin
                update AdminTeam_Orgs set organization = new_id
                 where adminteam = r.adminteam and organization = r.organization;
            exception
                when NetDB.unique_constraint_failed then
                    delete from AdminTeam_Orgs
                     where adminteam = r.adminteam and organization = r.organization;
            end;
        end loop;

        for r in (select * from LNA where organization = old_id)
        loop
            begin
                update LNA set organization = new_id
                 where lna = r.lna and organization = r.organization;
            exception
                when NetDB.unique_constraint_failed then
                    delete from LNA where lna = r.lna and organization = r.organization;
            end;
        end loop;

        update Node set organization = new_id where organization = old_id;

        delete from Organization where id = old_id;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'action', 0, 'string', 'replace', null, null);

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end supplant;


    -- returns list of departments restricted by p_filter
    -- if p_filter is null then all locations are shown
    function list (p_filter in varchar := null) return NetDB.ref_cursor
    as
        out_record  result_set := result_set ();
        ret_cursor  NetDB.ref_cursor;
    begin
        for r in ( select * from Organization
                    where lower (organization) like lower (nvl (p_filter, '%')) )
        loop
            out_record.extend(4);
            out_record(out_record.COUNT - 3) := result_row ('data', r.organization,    0,  null,    null,           null, null);
            out_record(out_record.COUNT - 2) := result_row ('key', 'name',          null, 'string', r.organization, null, null);
            out_record(out_record.COUNT - 1) := result_row ('key', 'id',            null, 'string', r.id,           null, null);
            out_record(out_record.COUNT)     := result_row ('key', 'RegID',         null, 'string', r.org_regid,    null, null);
        end loop;

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end list;


    function info (p_org_id in int) return NetDB.ref_cursor as begin
        return NetDB_Util.collect_into_cursor ('NetDB_Organization.info_pipelined (' || p_org_id || ')');
    end info;


    function info_pipelined (p_org_id in int) return result_set pipelined
    as
        org_row Organization%ROWTYPE;
        res_row result_row   := result_row (null, null, null, null, null, null, null);

        person_cursor NetDB.ref_cursor := null;
    begin
        select * into org_row from Organization where id = p_org_id;

        pipe row (result_row ('key', 'object', null, 'string', 'department',         null, null));
        pipe row (result_row ('key', 'name',   null, 'string', org_row.organization, null, null));
        pipe row (result_row ('key', 'id',     null, 'string', to_char (org_row.id), null, null));
        pipe row (result_row ('key', 'RegID',  null, 'string', org_row.org_regid,    null, null));

        pipe row (result_row ('data', 'consultants', 0, null, null, null, null));
        for r in (select person, primary from Consultant where Consultant.organization = org_row.id)
        loop
            pipe row (result_row ('data', to_char(r.person), 1,     null,     null, null, null));
            pipe row (result_row ('key', 'primary?',         null, 'boolean', ' ',  null, r.primary));

            person_cursor := NetDB_Person.info (r.person);
            loop
                fetch person_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                         res_row.value_string, res_row.value_timestamp, res_row.value_int;
                exit when person_cursor%notfound;

                pipe row (res_row);
            end loop;
            close person_cursor;
        end loop;
    end info_pipelined;


    procedure del ( p_op_uid in int,
                    p_org_id in int )
    as
        v_org_id int := verify_or_die (p_org_id);
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'organization');

        delete from Organization where id = p_org_id;
    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('organization', handle (p_org_id));
    end del;


    function ins ( p_op_uid in int,
                   p_name   in varchar,
                   p_reg_id in varchar ) return NetDB.ref_cursor
    as
        v_org_id int;

        out_record result_set := result_set ();
        ret_cursor NetDB.ref_cursor;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'organization');
        NetDB_Util.validate (p_name, 'department name', 'A department must have a name.');

        insert into Organization (org_regid, organization)
                          values (p_reg_id, p_name)
        returning id into v_org_id;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'id', 0, 'string', v_org_id, null,   null);

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('organization', p_name);

        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- Performs update
    function upd ( p_op_uid   in int,
                   p_org_id   in int,
                   p_name_new in varchar,
                   p_reg_id   in varchar ) return NetDB.ref_cursor
    as
        out_record result_set := result_set ();
        ret_cursor NetDB.ref_cursor;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'organization');
        NetDB_Util.validate (p_name_new, 'department name', 'A department must have a name.');

        update Organization set organization = p_name_new,
                                org_regid    = p_reg_id
         where id = p_org_id;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'action', 0, 'string', 'update', null, null);

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('organization', p_name_new);

        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    -- given organization (a.k.a department) id and/or name verifies that it
    -- exists and returns its id
    function verify_or_die ( p_org_id in int     := null,
                             p_name   in varchar := null ) return int
    as
        v_identifier varchar(256) := p_name;
        v_org_id     int;
    begin
        if (p_org_id is null and p_name is null) then
            NetDB_Util.throw_no_record_id_handle ();

        elsif (p_org_id is not null and p_name is not null) then
            select id into v_org_id from Organization
             where lower (organization) = lower (p_name) and id = p_org_id;

        elsif (p_name is not null) then
            select id into v_org_id from Organization where lower (organization) = lower (p_name);

        else -- p_org_id is not null
            v_identifier := v_org_id;

            select id into v_org_id from Organization where id = p_org_id;
        end if;

        return v_org_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('organization', v_identifier);
    end verify_or_die;


    function handle (p_org_id in int) return varchar
    as
        v_name Organization.organization%TYPE;
    begin
        select organization into v_name from Organization where id = p_org_id;

        return v_name;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('organization', p_org_id);
    end handle;

end NetDB_Organization;
/
