-- NetDB_Owner.sql --- contains functions related to ownership security group.

create or replace package body NetDB_Owner as

    -- Add a group to "own" a record.
    procedure add ( p_op_uid      in int,
                    p_lock        in timestamp,
                    p_owner_id    in int     := null,
                    p_name        in varchar := null,
                    p_record_id   in int,
                    p_record_type in varchar )
    as
        v_owner_id int := verify_or_die (p_owner_id, p_name);
    begin
        NetDB_Access.has_access_or_die (p_op_uid, p_record_type);

        -- The op_user must be the member of the group in order to add the group to records.
        if (not is_member_of_group (p_op_uid, v_owner_id)) then
            raise_application_error (NetDB.ex_not_a_member_of_group,
                'The user "' || NetDB_User.handle (p_op_uid) || '" cannot add the group "' ||
                handle (v_owner_id) || '" to records.');
        end if;

        NetDB_Util.locked_or_die (p_op_uid, p_lock, p_record_id, p_record_type);

        case lower (p_record_type)
            when 'adminteam' then
                insert into AdminTeam_Owner values (p_record_id, v_owner_id);

            when 'domain' then
                insert into Domain_Owner values (p_record_id, v_owner_id);

            when 'domain_assign' then
                insert into Domain_assignNameIn values (p_record_id, v_owner_id);

            when 'domain_use' then
                insert into Domain_useAsName values (p_record_id, v_owner_id);

            when 'addressspace' then
                insert into IPaddressSpace_Owner values (p_record_id, v_owner_id);

            when 'network' then
                insert into Network_Owner values (p_record_id, v_owner_id);

            when 'node' then
                insert into Node_Owner values (p_record_id, v_owner_id);

            when 'resname' then
                insert into ResName_Owner values (p_record_id, v_owner_id);

            when 'user' then
                insert into Users_Owner values (p_record_id, v_owner_id);

            else
                NetDB_Util.throw_unknown_record_type (p_record_type);
        end case;
    exception
        when NetDB.unique_constraint_failed then
            null;
    end add;


    -- Remove a group to "own" a record.
    procedure remove ( p_op_uid      in int,
                       p_lock        in timestamp,
                       p_owner_id    in int     := null,
                       p_name        in varchar := null,
                       p_record_id   in int,
                       p_record_type in varchar )
    as
        v_owner_id int := verify_or_die (p_owner_id, p_name);
    begin
        NetDB_Access.has_access_or_die (p_op_uid, p_record_type);
        NetDB_Util.locked_or_die (p_op_uid, p_lock, p_record_id, p_record_type);

        if (not is_member_of_group (p_op_uid, v_owner_id)) then
            raise_application_error (NetDB.ex_not_a_member_of_group,
                'The user "' || NetDB_User.handle (p_op_uid) || '" cannot remove the group "' ||
                handle (v_owner_id) || '" from records.');
        end if;

        case  lower (p_record_type)
            when 'adminteam' then
                delete from AdminTeam_Owner where adminteam = p_record_id and owner = v_owner_id;

            when 'domain' then
                delete from Domain_Owner where domain = p_record_id and owner = v_owner_id;

            when 'domain_assign' then
                delete from Domain_assignNameIn where domain = p_record_id and owner = v_owner_id;

            when 'domain_use' then
                delete from Domain_useAsName where domain = p_record_id and owner = v_owner_id;

            when 'addressspace' then
                delete from IPaddressSpace_Owner where ipaddressspace = p_record_id and owner = v_owner_id;

            when 'network' then
                delete from Network_Owner where network = p_record_id and owner = v_owner_id;

            when 'node' then
                delete from Node_Owner where node  = p_record_id and owner = v_owner_id;

            when 'resname' then
                delete from ResName_Owner where resname = p_record_id and owner = v_owner_id;

            when 'user' then
                delete from Users_Owner where userid = p_record_id and owner = v_owner_id;

            else
                NetDB_Util.throw_unknown_record_type (p_record_type);
        end case;

        if (not check_access (p_op_uid, p_record_id, p_record_type)) then
            raise_application_error (NetDB.ex_no_access_to_record,
                'Removal of the Group "' || handle (v_owner_id) ||
                '" results in the user "' || NetDB_User.handle (p_op_uid) ||
                '" losing access to the '|| NetDB_RecordType.display (p_record_type) ||
                ' "' || NetDB_Util.handle (p_record_id, p_record_type, true) || '".');
        end if;
    end remove;


    -- Check if a user is a member of a group with access to a record.
    function check_access ( p_user_id              in int,
                            p_record_id            in int,
                            p_record_type          in varchar,
                            p_check_all_group_flag in boolean := true,
                            p_check_active_state   in boolean := true ) return boolean
    as
        v_user_id    int := NetDB_User.verify_or_die (p_user_id);
        v_record_id  int := NetDB_Util.verify_or_die (p_record_id, null, p_record_type);

        IP_row   IPaddress%ROWTYPE;
        name_row Name%ROWTYPE;

        -- Assume that there are no MX received by names.
        no_mx_entries boolean := true;
    begin
        if (p_check_active_state   and not NetDB_User.is_active (p_user_id))  then return false; end if;
        if (p_check_all_group_flag and NetDB_User.has_all_groups (p_user_id)) then return true;  end if;

        case lower (p_record_type)
            when 'address' then
                select * into IP_row from IPaddress where id = p_record_id;

                if (IP_row.ipcnode is not null) then
                    return check_access (p_user_id, IP_row.ipcnode, 'Node', p_check_all_group_flag);
                elsif (IP_row.interface is not null) then
                    return check_access (p_user_id, IP_row.interface, 'Interface', p_check_all_group_flag);
                elsif (IP_row.ipaddressspace is not null) then
                    return check_access (p_user_id, IP_row.ipaddressspace, 'AddressSpace', p_check_all_group_flag);
                end if;

            when 'adminteam' then
                select count (*) into NetDB.existence_counter
                  from AdminTeam_Owner natural join Users_Owner
                 where userid = p_user_id and adminteam = p_record_id;

            when 'domain' then
                select count (*) into NetDB.existence_counter
                  from Domain_Owner natural join Users_Owner
                 where userid = p_user_id and domain = p_record_id;

            when 'domain_assign' then
                select count (*) into NetDB.existence_counter
                  from Domain_assignNameIn natural join Users_Owner
                 where userid = p_user_id and domain = p_record_id;

            when 'domain_use' then
                select count (*) into NetDB.existence_counter
                  from Domain_useAsName natural join Users_Owner
                 where userid = p_user_id and domain = p_record_id;

            when 'interface' then
                return check_access (p_user_id, NetDB_Interface.get_parent_id (p_record_id), 'Node', p_check_all_group_flag);

            when 'addressspace' then
                select count (*) into NetDB.existence_counter
                  from IPaddressSpace_Owner natural join Users_Owner
                 where userid = p_user_id and ipaddressspace = p_record_id;

            when 'mx' then
                -- If the MX shares a name with an "A" name, verify that the user
                -- has access to the record which the "A" name refers.
                begin
                    select * into name_row from Name
                     where dnslabel = 'A' and
                           fullname = (select fullname from Name where id = p_record_id);

                    return check_access ( p_user_id,
                                          coalesce ( name_row.ipaddress,
                                                     name_row.interface,
                                                     name_row.node,
                                                     name_row.network ),
                                          name_row.label,
                                          p_check_all_group_flag );
                exception
                    when NO_DATA_FOUND then null;
                end;

                -- The MX does not share a name. Verify the user has access to any
                -- of the records which the received_by names refer.
                for r in ( select Name.*
                             from Name
                             join MXrecord mx on mx.received_by = Name.id
                            where mx.sent_to = (select id from Name where id = p_record_id and label = 'MX') )
                loop
                    no_mx_entries := false;

                    if ( check_access ( p_user_id,
                                        coalesce ( r.ipaddress,
                                                   r.interface,
                                                   r.node,
                                                   r.network ),
                                        r.label,
                                        p_check_all_group_flag ) )
                    then
                        return true;
                    end if;
                end loop;

                -- If no were names checked, we must give the user access. Otherwise, there were
                -- names checked, and the user does not have access.
                return no_mx_entries;

            when 'network' then
                select count (*) into NetDB.existence_counter
                  from Network_Owner natural join Users_Owner
                 where userid = p_user_id and network = p_record_id;

            when 'node' then
                select count (*) into NetDB.existence_counter
                  from Node_Owner natural join Users_Owner
                 where userid = p_user_id and node = p_record_id;

            when 'resname' then
                select count (*) into NetDB.existence_counter
                  from ResName_Owner natural join Users_Owner
                 where userid = p_user_id and resname = p_record_id;

            when 'user' then
                select count (*) into NetDB.existence_counter
                  from Users_Owner uo1
                  join Users_Owner uo2 on uo2.owner = uo1.owner
                 where uo1.userid = p_user_id and uo2.userid = p_record_id;

            else
                NetDB_Util.throw_unknown_record_type (p_record_type);
        end case;

        return Util.to_boolean (NetDB.existence_counter);
    end check_access;


    -- A convenient wrapper for check_access.
    -- If the user doesn't have access, raise NetDB.ex_no_access_to_record.
    procedure has_access_or_die ( p_user_id              in int,
                                  p_record_id            in int,
                                  p_record_type          in varchar,
                                  p_check_all_group_flag in boolean := true,
                                  p_check_active_state   in boolean := true ) as
    begin
        if (not check_access (p_user_id, p_record_id, p_record_type, p_check_all_group_flag, p_check_active_state)) then
            raise_application_error (NetDB.ex_no_access_to_record,
                'The user "' || NetDB_User.handle (p_user_id) ||
                '" does not have access to the ' || NetDB_RecordType.display (p_record_type) ||
                ' "' || NetDB_Util.handle (p_record_id, p_record_type, true) || '".');
        end if;
    end has_access_or_die;


    -- Check to see if a user is a member of a group.
    function is_member_of_group ( p_user_id  in int,
                                  p_owner_id in int ) return boolean as
    begin
        if (NetDB_User.has_all_groups (p_user_id)) then
            return true;
        end if;

        select count (*) into NetDB.existence_counter from Users_Owner
         where userid = p_user_id and owner = p_owner_id;

        return Util.to_boolean (NetDB.existence_counter);
    end is_member_of_group;


    -- Checks to see if a user has a lock for a record.
    procedure locked_or_die ( p_op_uid   in int,
                              p_lock     in timestamp,
                              p_owner_id in int )
    as
        v_lock timestamp;
    begin
        select locked into v_lock from Owner
         where id = verify_or_die (p_owner_id) and locked_by = p_op_uid and locked = p_lock and
               Util.datediff ('second', p_lock) < NetDB_Default.get_int ('LockTimeOut');
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_not_locked ( 'owner',
                                                  handle (p_owner_id),
                                                  NetDB_User.handle (p_op_uid) );
    end locked_or_die;


    -- Locks Group/Owner record.
    function lock_record ( p_op_uid   in int,
                           p_owner_id in int ) return NetDB.ref_cursor
    as
        v_lock      timestamp;
        v_locked_by int;

        out_record result_set       := result_set ();
        ret_cursor NetDB.ref_cursor := null;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Owner');

        select locked, locked_by
          into v_lock, v_locked_by
          from Owner
         where id = verify_or_die (p_owner_id);

        -- if not locked, locked by this user, or the lock has expired, we're good to go
        if ( v_lock is null or v_locked_by = p_op_uid or
             Util.datediff ('second', v_lock) > NetDB_Default.get_int ('LockTimeOut') )
        then
               update Owner set locked    = systimestamp,
                                locked_by = p_op_uid
                where id = p_owner_id
            returning locked into v_lock;
        else
            NetDB_Util.throw_locked_by_other ( 'owner',
                                                handle (p_owner_id),
                                                NetDB_User.handle (v_locked_by),
                                                v_lock );
        end if;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'locked', 0, 'timestamp', null, v_lock, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end lock_record;


    -- Unlock a record.
    procedure unlock_record ( p_op_uid   in int,
                              p_lock     in timestamp,
                              p_owner_id in int ) as
    begin
        locked_or_die (p_op_uid, p_lock, p_owner_id);

        update Owner set locked    = null,
                         locked_by = null
         where id = p_owner_id and locked = p_lock and locked_by = p_op_uid;
    end unlock_record;


    -- List all the groups a user is a member of. Output can be restricted by use of filter.
    function list ( p_filter  in varchar := null,
                    p_user_id in int ) return NetDB.ref_cursor
    as
        v_all_groups   boolean                := false;
        v_key          Owner.label%TYPE       := null;
        v_value_string Owner.description%TYPE := null;

        out_record  result_set := result_set ();
        ret_cursor  NetDB.ref_cursor;
        data_cursor NetDB.ref_cursor;
    begin
        if (not NetDB_User.is_active (p_user_id)) then
            raise_application_error (NetDB.ex_inactive_user,
                'NetDB access for "' || NetDB_User.handle (p_user_id) || '" has been deactivated.');
        else
            v_all_groups := NetDB_User.has_all_groups (p_user_id);
        end if;

        if (not v_all_groups) then
            open data_cursor for
                select label, nvl (description, ' ')
                  from Owner
                  join Users_Owner on Users_Owner.owner = Owner.id
                 where Users_Owner.userid = p_user_id and lower (label) like nvl (lower (p_filter), '%')
              order by lower (label);
        else
            open data_cursor for
                select label, nvl (description, ' ')
                  from Owner
                 where lower (label) like nvl (lower (p_filter), '%')
              order by lower (label);
        end if;

        loop
            fetch data_cursor into v_key, v_value_string;
            exit when data_cursor%notfound;

            out_record.extend(1);
            out_record(out_record.COUNT) := result_row ('key', v_key, null, 'string', v_value_string, null, null);
        end loop;

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end list;


    function info (p_owner_id in int) return NetDB.ref_cursor
    as
        v_id varchar(1024) := p_owner_id;
    begin
        if (v_id is null) then
            v_id := 'null';
        end if;

        return NetDB_Util.collect_into_cursor ('NetDB_Owner.info_pipelined (' || v_id || ')');
    end info;


    -- Provides information about a group.
    function info_pipelined (p_owner_id in int) return result_set pipelined
    as
        res_row result_row := result_row (null, null, null, null, null, null, null);

        -- Owner info expressed as a cursor (instead of defining a bunch of variables).
        cursor owner_info is
        select label, description, created, modified, created_by, modified_by
          from Owner
         where id = verify_or_die (p_owner_id);

        i owner_info%ROWTYPE;

        person_cursor NetDB.ref_cursor;
    begin
        open owner_info;
        fetch owner_info into i;
        close owner_info;

        pipe row (result_row ('key', 'object',   null, 'string',   'group',        null,       null));
        pipe row (result_row ('key', 'id',       null, 'string',    p_owner_id,    null,       null));
        pipe row (result_row ('key', 'name',     null, 'string',    i.label,       null,       null));
        pipe row (result_row ('key', 'comment',  null, 'string',    i.description, null,       null));
        pipe row (result_row ('key', 'created',  null, 'timestamp', null,          i.created,  null));
        pipe row (result_row ('key', 'modified', null, 'timestamp', null,          i.modified, null));

        -- created.by person info
        pipe row (result_row ('data', 'created.by', 0, null, null,null, null));

        person_cursor := NetDB_Person.info (i.created_by);
        loop
            fetch person_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                     res_row.value_string, res_row.value_timestamp, res_row.value_int;

            exit when person_cursor%notfound;

            pipe row (res_row);
        end loop;
        close person_cursor;

        -- mod.by person info
        pipe row (result_row ('data', 'mod.by', 0, null, null, null, null));

        person_cursor := NetDB_Person.info (i.modified_by);
        loop
            fetch person_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                     res_row.value_string, res_row.value_timestamp, res_row.value_int;

            exit when person_cursor%notfound;

            pipe row (res_row);
        end loop;
        close person_cursor;
    end info_pipelined;


    -- Delete a record.
    procedure del ( p_op_uid   in int,
                    p_lock     in timestamp ,
                    p_owner_id in int ) as
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Owner');
        locked_or_die (p_op_uid, p_lock, p_owner_id);
        NetDB_Log.ins (p_op_uid, 'delete', p_owner_id, 'owner');

        delete from Owner where locked = p_lock and locked_by = p_op_uid and id = p_owner_id;
    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('owner', handle (p_owner_id));
    end del;


    -- Insert a group and locks it.
    function ins ( p_op_uid      in int,
                   p_name        in varchar,
                   p_description in varchar := null ) return NetDB.ref_cursor
    as
        v_owner_id     int       := 0;
        v_lock         timestamp := null;
        sys_time_stamp timestamp := systimestamp;

        out_record  result_set := result_set ();
        ret_cursor  NetDB.ref_cursor;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Owner');

        NetDB_Util.validate (p_name, 'group name', 'A group must have a name.');
        NetDB_Util.validate (p_description, 'comment');

           insert into Owner
                  ( label,     description,   created_by, modified_by, created,         modified,
                    locked_by, locked )
           values ( p_name,    p_description, p_op_uid,   p_op_uid,    sys_time_stamp,  sys_time_stamp,
                    p_op_uid,  sys_time_stamp )
        returning id, locked into v_owner_id, v_lock;

        NetDB_Log.ins (p_op_uid, 'insert', v_owner_id, 'owner');

        out_record.extend(2);
        out_record(1) := result_row ('key', 'id',     0, 'string',    v_owner_id, null,   null);
        out_record(2) := result_row ('key', 'locked', 0, 'timestamp', null,       v_lock, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('Group', p_name);

        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- Update the info for a group.
    procedure upd ( p_op_uid      in int,
                    p_lock        in timestamp,
                    p_owner_id    in int,
                    p_name_new    in varchar,
                    p_description in varchar ) as
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'Owner');
        locked_or_die (p_op_uid, p_lock, p_owner_id);

        NetDB_Util.validate (p_name_new, 'group name', 'A group must have a name.');
        NetDB_Util.validate (p_description, 'comment');

        update Owner set label       = p_name_new,
                         description = p_description,
                         modified    = systimestamp,
                         modified_by = p_op_uid
         where id = p_owner_id;

        NetDB_Log.ins (p_op_uid, 'update', p_owner_id, 'owner');
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('Group', p_name_new);

        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    -- Verify group ID / handle or complain.
    function verify_or_die ( p_owner_id in int     := null,
                             p_name     in varchar := null ) return int
    as
        v_identifier varchar(256) := p_name;
        v_owner_id   int;
    begin
        if (p_owner_id is null and p_name is null) then
            NetDB_Util.throw_no_record_id_handle ();

        elsif (p_owner_id is not null and p_name is not null) then
            select id into v_owner_id from Owner
             where lower (label) = lower (p_name) and id = p_owner_id;

        elsif (p_name is not null) then
            select id into v_owner_id from Owner where lower (label) = lower (p_name);

        else -- p_owner_id is not null
            v_identifier := p_owner_id;

            select id into v_owner_id from Owner where id = p_owner_id;
        end if;

        return v_owner_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('owner', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('owner');
    end verify_or_die;


    -- Get a group name.
    function handle (p_owner_id in int) return varchar
    as
        v_name Owner.label%TYPE := null;
    begin
        select label into v_name from Owner where id = p_owner_id;

        return v_name;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('owner', p_owner_id);
    end handle;

end NetDB_Owner;
/
