-- NetDB_SS.sql --- function related to a simple search

create or replace package body NetDB_SS as

    -- find IP address closest to that name using the information associated with a name.
    function closest_ip_address( p_ipaddr_id       in     int,
                                 p_intfc           in     int,
                                 p_node_id         in     int,
                                 p_aliasee         in     int,
                                 p_related_node_id in     int,
                                 p_name_type       in out varchar ) return IP.address
    as
        v_ipaddr IP.prefix; -- in case of template nodes this vairable must be able to hold ip prefix
        v_intfc  int := p_intfc;
    begin
        case
            when p_ipaddr_id is not null then
                select ipaddress_txt, interface into v_ipaddr, v_intfc
                  from IPaddress where id = p_ipaddr_id;

                if v_intfc is not null then
                    p_name_type := 'ip_addr_name';
                    v_intfc     := null;
                else
                    p_name_type := 'ipc_addr_name';
                end if;

            when p_intfc is not null then
                select min(ipaddress) into v_ipaddr
                  from IPaddress where interface = p_intfc;

                if v_ipaddr is not null then
                    select ipaddress_txt into v_ipaddr
                      from IPaddress where ipaddress = v_ipaddr;
                end if;
                p_name_type := 'intfc_name';

            else
                v_ipaddr := null;
        end case;

        if v_ipaddr is null then
            select min(ipaddress) into v_ipaddr
            from (select ipaddress.ipaddress from IPaddress
                   where ipcnode = p_related_node_id
                      union
                  select ipaddress.ipaddress
                    from Interface
                    left join IPaddress on Interface.id = ipaddress.interface
                   where Interface.node = p_related_node_id) ips;

            if v_ipaddr is not null then
                 select ipaddress_txt into v_ipaddr
                   from IPaddress where ipaddress = v_ipaddr;
            end if;

            if p_node_id is not null then
                p_name_type := 'node_name';
            elsif p_aliasee is not null then
                p_name_type := 'alias_name';
            end if;
        end if;

        -- see if node has template ip address space associated with it
        if v_ipaddr is null then
        begin
            select IPaddressSpace.prefix_txt into v_ipaddr
              from IPaddressSpace join Node on Node.IPaddressSpace = IPaddressSpace.id
              where Node.id = p_related_node_id;
        exception
            when no_data_found then v_ipaddr := null;
        end;
        end if;

        return(v_ipaddr);
    end closest_ip_address;


    -- performs simple search for admin teams
    function get_ateams( p_filter in varchar := null ) return NetDB.ref_cursor
    as
        ret_cursor NetDB.ref_cursor;
    begin
        open ret_cursor for
            select 'key' type, name key, null hash_level, 'string' value_type, ' ' value_string
              from AdminTeam where lower(name) like lower(nvl(p_filter, '%'));

        return ret_cursor;
    end get_ateams;


    -- performs simple search for domains
    function get_domains( p_filter in varchar := null ) return NetDB.ref_cursor
    as
        ret_cursor NetDB.ref_cursor;
        find_exact int          := null;
        v_filter   varchar(256) := '%';
    begin
        if p_filter is null then
            raise_application_error(NetDB.ex_bad_parameter, 'Please provide a search string.');
        end if;

        -- If the filter includes a dot, assume the domain being searched for is
        -- fully qualified.  Strip trailing dots because they're never stored in
        -- the Domain table.  If the filter is all dots, leave one so it matches
        -- the root domain.

        if instr(p_filter, '.') > 0 then
            v_filter := nvl(rtrim(p_filter, '.'), '.');

        -- Okay, there's no dot, so append '.%' to find the string in any domain
        -- (unless the search is for all domains, then the added dot would cause
        -- the search miss top-level domains).

        elsif p_filter != '%' then
            find_exact := 1;
            v_filter   := p_filter || '.%';

        end if;

        open ret_cursor for
            select 'key'    type,
                    domain  key,
                    null    hash_level,
                   'string' value_type,
                   ' '      value_string
             from Domain
            where (lower(domain) like lower(v_filter)) or
                  (lower(domain) like lower(p_filter) and find_exact is not null);

        return ret_cursor;
    end get_domains;


    -- performs simple search for ownership groups
    function get_groups( p_filter in varchar := null ) return NetDB.ref_cursor
    as
        ret_cursor NetDB.ref_cursor;
    begin
        open ret_cursor for
            select 'key' type, label key, null hash_level, 'string' value_type, ' ' value_string
              from Owner where lower(label) like lower(nvl(p_filter, '%'));

        return ret_cursor;
    end get_groups;


    -- main function for simple searches related to networks
    -- actual searches performed by other functions
    function get_networks( p_search_param1 in varchar := null, -- Name or starting IP
                           p_search_param2 in varchar := null, -- ending IP
                           p_search_type   in varchar,         -- {name, IP}
                           p_op_uid        in int     := null ) return NetDB.ref_cursor
    as
        ret_cursor NetDB.ref_cursor;
    begin
        case
            when p_search_type = 'name' then
                open ret_cursor for
                    select type, key, hash_level, value_type, value_string
                      from table(get_networks_by_name(p_search_param1, p_op_uid));

            when p_search_type = 'ip' then
                open ret_cursor for
                    select type, key, hash_level, value_type, value_string
                      from table(get_networks_by_ip(p_search_param1, p_search_param2, p_op_uid));

            else
                raise_application_error(NetDB.ex_bad_parameter,
                    'Unknown search type "' || p_search_type || '".');
        end case;

        return ret_cursor;
    end get_networks;


    -- Given IP1 and possibly IP2 performs simple search for networks.
    -- If both IPs are given then assume IP range search, otherwise
    -- only IP1 is used for string search.
    function get_networks_by_ip( p_ip1    in IP.address := null,
                                 p_ip2    in IP.address := null,
                                 p_op_uid in int        := null ) return ss_result pipelined
    as
        v_dd_id        int := NetDB_User.def_domain(p_op_uid);
        v_network_id   int;
        v_ip           IP.prefix;
        v_ip_txt       IP.prefix;
        v_main_sql     varchar(2048);

        ips_cursor NetDB.ref_cursor;
    begin
        if p_ip1 is null then return; end if;

        v_main_sql :=
            'select IPaddress.ipaddress, IPaddress.ipaddress_txt, IPaddressSpace.network ' ||
             ' from IPaddress'                                                             ||
             ' join IPaddressSpace on IPaddressSpace.id = IPaddress.ipaddressspace '       ||
             'where IPaddress.ipaddressspace is not null and ';

        if p_ip2 is null then -- IP string
            v_main_sql := v_main_sql || 'IPaddress.ipaddress_txt like ''' || p_ip1 || '''';
        else -- IP range
            v_main_sql := v_main_sql ||
                          'IPaddress.ipaddress between ''' || IP.normalize(p_ip1) ||
                          ''' and ''' || IP.normalize(p_ip2) || '''';
        end if;

        v_main_sql := v_main_sql || ' union ' ||
            'select IPaddressSpace.prefix, IPaddressSpace.prefix_txt, IPaddressSpace.network ' ||
              'from IPaddressSpace where ';

        if p_ip2 is null then -- IP string
            v_main_sql := v_main_sql || 'IPaddressSpace.address_txt like ''' || p_ip1 || '''';

        else -- IP range
            v_main_sql := v_main_sql ||
                          'IPaddressSpace.address between ''' || IP.normalize(p_ip1) ||
                          ''' and ''' || IP.normalize(p_ip2) || '''';
        end if;

        open ips_cursor for v_main_sql;
        loop
            fetch ips_cursor into v_ip, v_ip_txt, v_network_id;
            exit when ips_cursor%NOTFOUND;

            pipe row(ss_row('data', v_ip_txt, 0,     null,    null));
            pipe row(ss_row('key', 'ip',      null, 'string', v_ip_txt));

            -- pull out IP name
            if instr(v_ip_txt, '/' ) = 0 then
                for r in ( select decode(Name.domain, v_dd_id, Name.name, Name.fullname) display_name
                             from Name join IPaddress on Name.ipaddress = IPaddress.id
                            where IPaddress.ipaddress = v_ip order by Name.fullname_lc )
                loop
                    pipe row(ss_row('key', 'dhcp_name', null, 'string', r.display_name));
                    exit; -- we need only first row
                end loop;
            end if;

            -- pull out network name
            for r in ( select fullname, decode(domain, v_dd_id, name, fullname) display_name
                         from Name where network = v_network_id order by fullname_lc )
            loop
                pipe row(ss_row('key', 'handle',    null, 'string', r.fullname    ));
                pipe row(ss_row('key', 'net_name',  null, 'string', r.display_name));
                exit;  -- we need only first row
            end loop;

            -- get network comment
            for n in ( select comments from Network where id = v_network_id ) loop
                pipe row(ss_row('key', 'comment',  null, 'string', n.comments));
            end loop;
        end loop;
        close ips_cursor;

        return;
    end get_networks_by_ip;


    -- performs search for networks by name
    function get_networks_by_name( p_filter in varchar := null,
                                   p_op_uid in int     := null ) return ss_result pipelined
    as
        v_dd_id      int := NetDB_User.def_domain(p_op_uid);
        v_ipaddr     int;
        v_ipaddress  IP.address;
        v_network_id int;
        v_filter     varchar(512) := lower(p_filter);
        v_name_type  varchar(9);
    begin
        -- do not do search if filter is not given
        if p_filter is null then return; end if;

        -- If no domain is given, search in all domains.
        if instr(v_filter, '.') = 0 then
            v_filter := trim(v_filter) || '.%';
        end if;

        for r in ( select Name.id,       Name.name,    Name.domain,
                          Name.fullname, Name.aliasee, Name.ipaddress, Name.network
                     from Name
                    where Name.fullname_lc like v_filter and
                          Name.dnslabel != 'MX' and
                          Name.related_node is null )
        loop
            v_ipaddr     := r.ipaddress;
            v_network_id := r.network;

            if    v_ipaddr     is not null then v_name_type := 'dhcp_name';
            elsif v_network_id is not null then v_name_type := 'net_name';
            end if;

            if r.aliasee is not null then
                -- follow aliasee to the 'A' name info
                select ipaddress, network into v_ipaddr, v_network_id
                  from Name
                 where aliasee is null
                 start with id = r.id connect by prior aliasee = id;

                -- at this point we know to which name alias is pointing
                if    v_ipaddr     is not null then v_name_type := 'dhcp_name';
                elsif v_network_id is not null then v_name_type := 'net_name';
                end if;
            end if;

            if v_ipaddr is not null then
            begin
                select IPaddressSpace.network into v_network_id
                  from IPaddress
                  join IPaddressSpace on IPaddress.ipaddressspace = IPaddressSpace.id
                 where IPaddress.id = v_ipaddr;

            exception
                when no_data_found then v_network_id := null;
            end;
            end if;

            if v_network_id is not null then

                -- at this point we have valid info to pass
                pipe row(ss_row('data', r.name, 0, null, null));

                if v_ipaddr is not null then
                    select ipaddress into v_ipaddress from IPaddress where id = v_ipaddr;
                    pipe row(ss_row('key', 'ip',  null, 'string', IP.host(v_ipaddress)));
                else -- name whenever found is a network name and we need to pull first address space
                    for n in ( select prefix_txt from IPaddressSpace
                                where network = v_network_id order by address )
                    loop
                        pipe row(ss_row('key', 'ip',  null, 'string', n.prefix_txt));
                        exit; -- we need only first prefix
                    end loop;
                end if;

                if r.domain = v_dd_id then
                    pipe row(ss_row('key', v_name_type,  null, 'string', r.name));
                else
                    pipe row(ss_row('key', v_name_type,  null, 'string', r.fullname));
                end if;


                -- pull out network name
                for n in ( select fullname, decode(domain, v_dd_id, name, fullname) display_name
                             from Name where network = v_network_id order by fullname_lc )
                loop
                    pipe row(ss_row('key', 'handle',  null, 'string', n.fullname));

                    if v_name_type = 'dhcp_name' then
                        pipe row(ss_row('key', 'net_name',  null, 'string', n.display_name));
                    end if;
                    exit; -- we need only one name
                end loop;

                -- get network comment
                for n in ( select comments from Network where id = v_network_id ) loop
                    pipe row(ss_row('key', 'comment',  null, 'string', n.comments));
                end loop;
            end if;
        end loop;

        return;
    end get_networks_by_name;


    -- main function for simple searches related to nodes
    -- actual searches performed by other functions
    function get_nodes( p_search_param1 in varchar := null, -- name, hw address, starting IP
                        p_search_param2 in varchar := null, -- ending IP
                        p_search_type   in varchar,         -- { name, IP, hw }
                        p_op_uid        in int     := null ) return NetDB.ref_cursor
    as
        ret_cursor NetDB.ref_cursor;
    begin
        case
            when p_search_type = 'name' then
                open ret_cursor for
                    select type, key, hash_level, value_type, value_string
                      from table(get_nodes_by_name(lower(p_search_param1), p_op_uid));

            when p_search_type = 'hw' then
                open ret_cursor for
                    select type, key, hash_level, value_type, value_string
                      from table(get_nodes_by_hw(p_search_param1, p_op_uid));

            when p_search_type = 'ip' then
                open ret_cursor for
                    select type, key, hash_level, value_type, value_string
                      from table(get_nodes_by_ip(p_search_param1, p_search_param2, p_op_uid));

            else
                raise_application_error(NetDB.ex_bad_parameter,
                    'Unknown search type "' || p_search_type || '".');
        end case;

        return ret_cursor;
    end get_nodes;


    -- Simple search for node by hardware address.
    function get_nodes_by_hw( p_hw     in varchar,
                              p_op_uid in int := null ) return ss_result pipelined
    as
        v_dd_id     int := NetDB_User.def_domain(p_op_uid); -- default domain
        v_id        int;                                    -- id of the node
        v_intfc     int;
        v_node_name Name.fullname%TYPE;
        v_handle    Name.fullname%TYPE;
        v_name      Name.fullname%TYPE;
        v_name_type varchar(15);
        v_ipaddr    IP.address;
    begin
        if p_hw is null then return; end if;

        select id, node into v_intfc, v_id
          from Interface where hw_address = lower(p_hw);

        if v_id is null then return; end if; -- nothing found

        pipe row(ss_row('data', v_id, 0, null, null));

        -- get proper node name
        for r in ( select decode(domain, v_dd_id, name, fullname) node_name, fullname handle
                     from Name where node = v_id order by fullname_lc )
        loop
            v_node_name := r.node_name;
            v_handle    := r.handle;
            exit; -- we need only first row
        end loop;

        -- see if interface has a name
        for r in ( select decode(domain, v_dd_id, name, fullname) name
                     from Name where interface = v_intfc order by fullname_lc )
        loop
            v_name := r.name;
            exit; -- we need only first row
        end loop;

        if v_name is null then
            v_name      := v_node_name;
            v_name_type := 'node_name';
        else
            v_name_type := 'intfc_name';
        end if;

        -- get relevant IP address
        begin
            select min(ipaddress) into v_ipaddr from IPaddress where interface = v_intfc;
        exception
            when no_data_found then null; -- catch and ignore exception
        end;

        if v_ipaddr is null then
        begin
            select min(ipaddress) into v_ipaddr
              from (select ipaddress
                      from IPaddress
                      join Interface on Interface.id = IPaddress.interface
                     where Interface.node = v_id
                        union
                    select ipaddress from IPaddress where ipcnode = v_id);
        exception
            when no_data_found then null; -- catch and ignore exception
        end;
        end if;

        pipe row(ss_row('key', 'name',      null, 'string',       v_name     ));
        pipe row(ss_row('key', 'handle',    null, 'string', lower(v_handle)  ));
        pipe row(ss_row('key', 'node_name', null, 'string',       v_node_name));
        pipe row(ss_row('key', 'NodeID',    null, 'string',       v_id       ));
        pipe row(ss_row('key', 'name_type', null, 'string',       v_name_type));

        if v_ipaddr is not null then
            pipe row(ss_row('key', 'ip', null, 'string', IP.host(v_ipaddr)));
        end if;

        -- node types
        pipe row(ss_row('data', 'node_type', 1, null, null));
        for r in ( select NodeType.Abbrev type, NodeType.Rank rank
                     from Node_NodeType join NodeType on Node_NodeType.type = NodeType.type
                    where Node_NodeType.node = v_id ) loop
            pipe row(ss_row('key', r.type, null, 'string', to_char(r.rank)));
        end loop;

        return;

    exception
        when no_data_found then return;
    end get_nodes_by_hw;


    -- Given IP1 and possibly IP2 performs simple search for nodes.
    -- If both IPs are given then assume IP range search, otherwise
    -- only IP1 is used for string search.
    function get_nodes_by_ip( p_ip1    in IP.address := null,
                              p_ip2    in IP.address := null,
                              p_op_uid in int        := null ) return ss_result pipelined
    as
        v_dd_id     int := NetDB_User.def_domain(p_op_uid);
        v_ip        IP.address;
        v_ip_txt    IP.address;
        v_intfc     int;
        v_node_id   int;
        v_handle    Name.fullname%TYPE;
        v_node_name Name.fullname%TYPE;
        v_name      Name.fullname%TYPE := null;
        v_name_type varchar(15);
        v_ip_type   varchar(8);
        v_main_sql  varchar(2048);

        ips_cursor NetDB.ref_cursor;
    begin
        if p_ip1 is null then return; end if;

        v_main_sql :=
            'select IPaddress.interface, '                                   ||
            '       IPaddress.ipaddress, '                                   ||
            '       IPaddress.ipaddress_txt, '                               ||
            '       nvl(IPaddress.ipcnode, Interface.node) node, '           ||
            '       case when IPaddress.ipcnode is not null then ''ipc_ip''' ||
            '            else ''intfc_ip'' '                                 ||
            '       end case '                                               ||
            'from IPaddress '                                                ||
            'left join Interface on IPaddress.interface = Interface.id '     ||
            'where ( ipcnode is not null or interface is not null) and ';

        if p_ip2 is null then -- IP string
            v_main_sql := v_main_sql || 'ipaddress_txt like ''' || p_ip1 || '''';

        else -- IP range
            v_main_sql := v_main_sql || 'ipaddress between ''' ||
                          IP.normalize(p_ip1) || ''' and ''' || IP.normalize(p_ip2) || '''';
        end if;

        open ips_cursor for v_main_sql;
        loop
            fetch ips_cursor into v_intfc, v_ip, v_ip_txt, v_node_id, v_ip_type;
            exit when ips_cursor%NOTFOUND;

            pipe row(ss_row('data', v_ip_txt, 0, null, null));

            for r in ( select decode(domain, v_dd_id, name, fullname) node_name, fullname handle
                         from Name where node = v_node_id order by fullname_lc )
            loop
                v_node_name := r.node_name;
                v_handle    := r.handle;
                exit; -- we need only first row
            end loop;

            pipe row(ss_row('key', 'handle',      null, 'string', v_handle  ));
            pipe row(ss_row('key', 'node_name',   null, 'string', v_node_name));
            pipe row(ss_row('key', 'NodeID',      null, 'string', v_node_id  ));
            pipe row(ss_row('key', 'ip',          null, 'string', v_ip_txt   ));

            if v_intfc is null then
                pipe row(ss_row('key', 'ip_type', null, 'string', 'ipc_ip'));
            else
                pipe row(ss_row('key', 'ip_type', null, 'string', 'intfc_ip'));
            end if;

            -- name
            for r in ( select decode(domain, v_dd_id, name, fullname) name
                         from Name
                         join IPaddress on Name.IPaddress = IPaddress.id
                        where IPaddress.IPaddress = v_ip order by Name.fullname_lc )
            loop
                v_name := r.name;
                exit; -- we need only first row
            end loop;

            if v_name is not null then
                if v_ip_type = 'ipc_ip' then
                    v_name_type := 'ipc_addr_name';
                else
                    v_name_type := 'ip_addr_name';
                end if;
            else
                -- IP does not have a name
                -- try to pick name from an interface
                for r in ( select decode(domain, v_dd_id, name, fullname) name
                             from Name where interface = v_intfc order by fullname_lc )
                loop
                    v_name := r.name;
                    exit; -- we need only first row
                end loop;

                if v_name is null then
                    -- use node name if intfc does not have one
                    v_name      := v_node_name;
                    v_name_type := 'node_name';
                else
                    v_name_type := 'intfc_name';
                end if;
            end if;

            pipe row(ss_row('key', 'name',      null, 'string', v_name));
            pipe row(ss_row('key', 'name_type', null, 'string', v_name_type));

            v_name := null; -- reseting this variable

            -- node department
            for r in ( select Organization.organization
                         from Node join Organization on Node.organization = Organization.id
                        where Node.id = v_node_id )
            loop
                pipe row(ss_row('key', 'department', null, 'string', trim(r.organization)));
            end loop;

            -- node make and model
            for r in ( select Model.make || ' ' || Model.modelname make_and_model
                         from Model join Node on Node.model = Model.id
                        where Node.id = v_node_id )
            loop
                pipe row(ss_row('key', 'make_and_model', null, 'string', trim(r.make_and_model)));
                exit; -- we need only first one
            end loop;

            -- node os
            for r in ( select os from Node_OS where node = v_node_id order by lower(os) )
            loop
                pipe row(ss_row('key', 'os', null, 'string', trim(r.os)));
                exit; -- we need only first one
            end loop;

            -- node types
            pipe row(ss_row('data', 'node_type', 1, null, null));
            for r in ( select NodeType.Abbrev type, NodeType.Rank rank
                     from Node_NodeType join NodeType on Node_NodeType.type = NodeType.type
                    where Node_NodeType.node = v_node_id ) loop
                pipe row(ss_row('key', r.type, null, 'string', to_char(r.rank)));
            end loop;
        end loop;
        close ips_cursor;

        return;
    end get_nodes_by_ip;


    -- performs search for nodes by name
    function get_nodes_by_name( p_filter in varchar := null,
                                p_op_uid in int     := null ) return ss_result pipelined
    as
        v_dd_id     int := NetDB_User.def_domain(p_op_uid); -- id of default domain
        v_name      Name.fullname%TYPE;
        v_ipaddr    IP.prefix;
        v_ipaddr_id int;
        v_name_type varchar(15);
        v_filter    varchar(512)  := p_filter;
    begin
        -- do not do search if filter is not given
        if p_filter is null then return; end if;

        -- If no domain is given, search in all domains.
        if instr(v_filter, '.') = 0 then
            v_filter := trim(v_filter) || '.%';
        end if;

        for n in
          ( select max(names.id)              id,
                   max(names.name)            name,
                   max(names.ipaddress)       ipaddress,
                   max(names.interface)       interface,
                   max(names.node)            node,
                   max(names.aliasee)         aliasee,
                   max(names.related_node_id) related_node_id,
                   min(node_name.fullname)    handle,
                   min(decode(node_name.domain, v_dd_id,
                              node_name.name,   node_name.fullname)) node_name,
                   min(decode(aliasee_name.domain, v_dd_id,
                              aliasee_name.name,   aliasee_name.fullname)) alias_of
              from (select id,
                           decode(domain, v_dd_id, name, fullname) name,
                           name.fullname_lc,
                           name.ipaddress,
                           name.interface,
                           name.node,
                           name.aliasee,
                           related_node related_node_id
                      from Name
                     where Name.fullname_lc like v_filter and Name.dnslabel != 'MX'
                     order by name.fullname_lc) names
              join Name node_name    on node_name.node  = related_node_id
         left join Name aliasee_name on aliasee_name.id = names.aliasee
             where related_node_id is not null
             group by names.id
             order by max(names.fullname_lc) )
        loop
            v_name_type := 'unknown';
            v_ipaddr_id := n.ipaddress;
            v_ipaddr    := closest_ip_address(v_ipaddr_id,       n.interface,
                                              n.node,            n.aliasee,
                                              n.related_node_id, v_name_type);

            pipe row(ss_row('data', n.handle || ' ' || n.name,    0,     null, null));
            pipe row(ss_row('key', 'name',                     null, 'string', n.name));
            pipe row(ss_row('key', 'handle',                   null, 'string', n.handle));
            pipe row(ss_row('key', 'node_name',                null, 'string', n.node_name));
            pipe row(ss_row('key', 'NodeID',                   null, 'string', n.related_node_id));
            pipe row(ss_row('key', 'name_type',                null, 'string', v_name_type));

            if v_ipaddr is not null then
                pipe row(ss_row('key', 'ip', null, 'string', v_ipaddr));
            end if;

            if n.alias_of is not null then
                pipe row(ss_row('key', 'alias_of', null, 'string', n.alias_of));
            end if;

            -- node department
            for r in ( select Organization.organization
                         from Node join Organization on Node.organization = Organization.id
                        where Node.id = n.related_node_id )
            loop
                pipe row(ss_row('key', 'department', null, 'string', trim(r.organization)));
            end loop;

            -- node make and model
            for r in ( select Model.make || ' ' || Model.modelname make_and_model
                         from Model join Node on Node.model = Model.id
                        where Node.id = n.related_node_id )
            loop
                pipe row(ss_row('key', 'make_and_model', null, 'string', trim(r.make_and_model)));
                exit; -- we need only first one
            end loop;

            -- node os
            for r in ( select os from Node_OS where node = n.related_node_id order by lower(os) )
            loop
                pipe row(ss_row('key', 'os', null, 'string', trim(r.os)));
                exit; -- we need only first one
            end loop;

            -- node types
            pipe row(ss_row('data', 'node_type', 1, null, null));
            for r in ( select NodeType.Abbrev type, NodeType.Rank rank
                     from Node_NodeType join NodeType on Node_NodeType.type = NodeType.type
                    where Node_NodeType.node = n.related_node_id ) loop
                pipe row(ss_row('key', r.type, null, 'string', to_char(r.rank)));
            end loop;
        end loop;

        -- now mxes
        for n in
          ( select decode(sent_to.domain, v_dd_id,
                          sent_to.name,   sent_to.fullname) name,
                   received_by.fullname_lc  handle,
                   received_by.related_node node_id,
                   decode(received_by.domain, v_dd_id,
                          received_by.name,   received_by.fullname) received_by,
                   MXrecord.preference,
                   received_by.fullname received_by_fullname
              from MXrecord
              join Name sent_to     on sent_to.id     = MXrecord.sent_to
              join Name received_by on received_by.id = MXrecord.received_by
             where sent_to.fullname_lc like v_filter )
        loop
            pipe row(ss_row('data', n.handle || ' ' || n.name,    0,     null,  null));
            pipe row(ss_row('key', 'name',                     null, 'string',  n.name));
            pipe row(ss_row('key', 'handle',                   null, 'string',  n.handle));
            pipe row(ss_row('key', 'node_name',                null, 'string',  n.received_by || ' (' || n.preference || ')'));
            pipe row(ss_row('key', 'received_by',              null, 'string',  n.received_by ));
            pipe row(ss_row('key', 'received_by_fullname',     null, 'string',  n.received_by_fullname ));
            pipe row(ss_row('key', 'preference',               null, 'string',  n.preference));
            pipe row(ss_row('key', 'NodeID',                   null, 'string',  n.node_id));
            pipe row(ss_row('key', 'name_type',                null, 'string', 'MX'));

            v_ipaddr    := closest_ip_address(null, null, null, null, n.node_id, v_name_type);
            if v_ipaddr is not null then
                pipe row(ss_row('key', 'ip', null, 'string', v_ipaddr));
            end if;

            -- node types
            pipe row(ss_row('data', 'node_type', 1, null, null));
            for r in ( select NodeType.Abbrev type, NodeType.Rank rank
                     from Node_NodeType join NodeType on Node_NodeType.type = NodeType.type
                    where Node_NodeType.node = n.node_id ) loop
                pipe row(ss_row('key', r.type, null, 'string', to_char(r.rank)));
            end loop;
        end loop;

        return;
    end get_nodes_by_name;


    -- performs simple search for users
    -- it calls get_users_by_name(...) to perform actual search
    function get_users( p_filter in varchar := null ) return NetDB.ref_cursor
    as
        ret_cursor NetDB.ref_cursor;
    begin
        open ret_cursor for
            select type, key, hash_level, value_type, value_string
              from table(get_users_by_name(p_filter));

        return ret_cursor;
    end get_users;


    -- performs actual search for a users
    function get_users_by_name ( p_filter in varchar := null ) return ss_result pipelined as
    begin
        for r in ( select Person.netid netid,
                          Person.name name,
                          Person.person_regid regid,
                          Person.id id
                     from Users join Person on Users.id = Person.id
                    where lower(Person.netid) like lower(nvl(p_filter, '%'))
                       or lower(Person.name)  like lower(nvl(p_filter, '%')) )
        loop
            pipe row(ss_row('data',
                             r.netid || ' ' || r.name || ' ' || r.regid || ' ' || r.id,
                             0, null, null));
            pipe row(ss_row('key', 'name',  null, 'string', r.name));
            pipe row(ss_row('key', 'NetID', null, 'string', r.netid));
            pipe row(ss_row('key', 'RegID', null, 'string', r.regid));

        end loop;

        return;
    end get_users_by_name;

end NetDB_SS;
/
