-- NetDB_State.sql --- State package body.

create or replace package body NetDB_State as

    -- Checks if the user has access to a particular state.
    function check_access ( p_user_id in int,
                            p_state   in varchar ) return boolean
    as
        v_state   State.state%TYPE := verify_or_die (p_state);
        state_row State%ROWTYPE;

        access_group AccessGroup.label%TYPE;
    begin
        select * into state_row from State where lower (state) = lower (v_state);

        if (state_row.accessgroup is null) then return true; end if;

        access_group := NetDB_Access.handle (state_row.accessgroup);

        return NetDB_Access.check_access (p_user_id, access_group, null, true);
    end check_access;


    -- A convenient wrapper for check_access.
    -- If the user doesn't have access, raise NetDB.ex_no_access_to_record_type.
    procedure has_access_or_die ( p_user_id in int,
                                  p_state   in varchar )
    as
        v_state State.state%TYPE := verify_or_die (p_state);
    begin
        if (not check_access (p_user_id, v_state)) then
            raise_application_error (NetDB.ex_no_access_to_record_type,
                'The user "' || NetDB_User.handle (p_user_id) ||
                '" does not have access to the state "' || v_state || '".');
        end if;
    end has_access_or_die;


    -- returns list of states, a user has access to, restricted by p_filter
    -- if p_filter is null then all states, the user can access, are shown
    function list ( p_filter  in varchar := null,
                    p_user_id in int     := null ) return NetDB.ref_cursor
    as
        out_record  result_set := result_set ();
        ret_cursor  NetDB.ref_cursor;
        data_cursor NetDB.ref_cursor;

        state_name  State.state%TYPE;
        rank        State.rank%TYPE;
        deny_dhcp   State.denydhcp%TYPE;
        deny_dns    State.denydns%TYPE;
        description State.description%TYPE;
        access_id   State.accessgroup%TYPE;
        access_name AccessGroup.label%TYPE;
    begin
        if (p_user_id is null) then
            open data_cursor for
                select State.*, AccessGroup.label agroup
                  from State
                  left join AccessGroup on State.accessgroup = AccessGroup.id
                 where lower (state) like lower (nvl (p_filter, '%'));
        else
            -- The logic behind this query is a little obtuse:
            --
            -- 1) select all states that match the filter
            -- 2) select all states the user DOESN'T have access to
            -- 3) subtract 2 from 1
            --
            -- Note: this query DOES NOT honor the "all records" flag.

            open data_cursor for
                select State.*, AccessGroup.label agroup
                  from State
                  left join AccessGroup on State.accessgroup = AccessGroup.id
                 where lower (state) like lower (nvl (p_filter, '%'))
                    minus
                select State.*, AccessGroup.label agroup
                  from State
                  left join AccessGroup on State.accessgroup = AccessGroup.id
                 where accessgroup is not null and
                       accessgroup not in (select accessgroup from users_accessgroup where userid = p_user_id);
        end if;

        loop
            fetch data_cursor into state_name, rank, deny_dhcp, deny_dns, description, access_id, access_name;
            exit when data_cursor%notfound;

            out_record.extend(7);
            out_record(out_record.COUNT - 6) := result_row ('data',  state_name,      0,  null,     null,         null, null);
            out_record(out_record.COUNT - 5) := result_row ('key',  'name',        null, 'string',  state_name,   null, null);
            out_record(out_record.COUNT - 4) := result_row ('key',  'rank',        null, 'integer', null,         null, rank);
            out_record(out_record.COUNT - 3) := result_row ('key',  'denyDHCP',    null, 'boolean', null,         null, deny_dhcp);
            out_record(out_record.COUNT - 2) := result_row ('key',  'denyDNS',     null, 'boolean', null,         null, deny_dns);
            out_record(out_record.COUNT - 1) := result_row ('key',  'description', null, 'string',  description,  null, null);
            out_record(out_record.COUNT)     := result_row ('key',  'access',      null, 'string',  access_name,  null, null);
        end loop;

        close data_cursor;

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end list;


    -- Delete a state.
    procedure del ( p_op_uid in int,
                    p_state  in varchar )
    as
        v_state State.state%TYPE := verify_or_die (p_state);
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'state');
        delete State where lower (state) = lower (p_state);
    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('State', p_state);
    end del;


    -- Insert a state.
    procedure ins ( p_op_uid       in int,
                    p_state        in varchar,
                    p_rank         in int,
                    p_denyDHCP     in int     := 0,
                    p_denyDNS      in int     := 0,
                    p_description  in varchar := null,
                    p_access_group in varchar := null )
    as
        v_access_group int;
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'state');
        NetDB_Util.validate (p_state, 'state', 'States must be named.');
        NetDB_Util.validate (p_description, 'description');

        if (p_access_group is not null) then
            v_access_group := NetDB_Access.verify_or_die (null, p_access_group);
        end if;

        insert into State (  state,   rank,   denyDHCP,   denyDNS,   description,   accessgroup)
                   values (p_state, p_rank, p_denyDHCP, p_denyDNS, p_description, v_access_group);
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('state', p_state);
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- Update a state.
    procedure upd ( p_op_uid       in int,
                    p_state        in varchar,
                    p_state_new    in varchar,
                    p_rank         in int,
                    p_denyDHCP     in int     := 0,
                    p_denyDNS      in int     := 0,
                    p_description  in varchar := null,
                    p_access_group in varchar := null )
    as
        v_access_group int;
        v_state State.state%TYPE := verify_or_die (p_state);
    begin
        NetDB_Access.has_access_or_die (p_op_uid, 'state');
        NetDB_Util.validate (p_state_new, 'state', 'States must be named.');
        NetDB_Util.validate (p_description, 'description');

        if (p_access_group is not null) then
            v_access_group := NetDB_Access.verify_or_die (null, p_access_group);
        end if;

        update State set state       = p_state_new,
                         rank        = p_rank,
                         denyDHCP    = p_denyDHCP,
                         denyDNS     = p_denyDNS,
                         description = p_description,
                         accessgroup = v_access_group
         where lower (state) = lower (p_state);
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('state', p_state_new);
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    function verify_or_die (p_state in varchar) return varchar
    as
        v_state State.state%TYPE;
    begin
        select state into v_state from State where lower (state) = lower (p_state);

        return v_state;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('state', p_state);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('state');
    end verify_or_die;

end NetDB_State;
/
