-- NetDB_Util.sql --- utility functions specific to NetDB

create or replace package body NetDB_Util as

    -- sysmaint_check - cry foul if system maintained value is input
    procedure sysmaint_check(col in varchar, value varchar) as
    begin
        if (value is not null) then
            raise_application_error(NetDB.ex_system_maintained, 'The column "' || trim(col) ||
              '" is maintained by the system and cannot be inserted or updated.');
        end if;
    end sysmaint_check;


    -- validate() - check string validity
    --              optionally complain if the string is null
    procedure validate(string varchar, thing varchar := 'input string', null_error varchar := null)
    as begin

        if thing is null then
            raise_application_error(NetDB.ex_idiot_programmer, '"thing" must not be null.');
        end if;

        if string is null and null_error is not null then
            raise_application_error(NetDB.ex_bad_parameter, null_error);

        elsif regexp_instr(string, '[^[:print:]]') != 0 then
            raise_application_error(NetDB.ex_bad_parameter,
              'The ' || thing || ' contains non-printable characters. New lines are in the set of non-printable characters.');
        end if;

    end validate;


    procedure locked_or_die ( p_op_uid      in int,
                              p_lock        in timestamp,
                              p_record_id   in int,
                              p_record_type in varchar )
    as
        v_type varchar(32) := lower (p_record_type);
    begin
        case
            -- Due to the way that _Util.locked_or_die (...) is used in the code, it will get called
            -- with record_types that don't have a locking mechanism. In those cases, we don't
            -- actually have to check anything, but we also don't want to raise an exception.
            when v_type = 'mx'           then null;
            when v_type = 'resname'      then null;

            when v_type = 'address'      then NetDB_IPaddress.locked_or_die (p_op_uid, p_lock, p_record_id);
            when v_type = 'adminteam'    then NetDB_AdminTeam.locked_or_die (p_op_uid, p_lock, p_record_id);
            when v_type in ('domain', 'domain_assign', 'domain_use')
                                         then NetDB_Domain.locked_or_die (p_op_uid, p_lock, p_record_id);
            when v_type = 'interface'    then NetDB_Interface.locked_or_die (p_op_uid, p_lock, p_record_id);
            when v_type = 'addressspace' then NetDB_IPaddressSpace.locked_or_die (p_op_uid, p_lock, p_record_id);
            when v_type = 'name'         then NetDB_Name.locked_or_die (p_op_uid, p_lock, p_record_id);
            when v_type = 'network'      then NetDB_Network.locked_or_die (p_op_uid, p_lock, p_record_id);
            when v_type = 'node'         then NetDB_Node.locked_or_die (p_op_uid, p_lock, p_record_id);
            when v_type = 'owner'        then NetDB_Owner.locked_or_die (p_op_uid, p_lock, p_record_id);
            when v_type = 'user'         then NetDB_User.locked_or_die (p_op_uid, p_lock, p_record_id);
            else
                throw_unknown_record_type (p_record_type);
        end case;
    end locked_or_die;


    function handle ( p_record_id   in int,
                      p_record_type in varchar,
                      user_readable in boolean := false ) return varchar
    as
        v_type   varchar(32)  := lower (p_record_type);
        v_handle varchar(256) := null;
    begin
        case
            when v_type = 'access'       then v_handle := NetDB_Access.handle (p_record_id);
            when v_type = 'address'      then v_handle := NetDB_IPaddress.handle (p_record_id, user_readable);
            when v_type = 'adminteam'    then v_handle := NetDB_AdminTeam.handle (p_record_id);
            when v_type = 'dhcpoption '  then v_handle := NetDB_DHCPoption.handle (p_record_id);
            when v_type = 'dhcpservice'  then v_handle := NetDB_DHCPservice.handle (p_record_id);
            when v_type in ('domain', 'domain_assign', 'domain_use')
                                         then v_handle := NetDB_Domain.handle (p_record_id);
            when v_type = 'interface'    then v_handle := NetDB_Interface.handle (p_record_id);
            when v_type = 'addressspace' then v_handle := NetDB_IPaddressSpace.handle (p_record_id, user_readable);
            when v_type = 'location'     then v_handle := NetDB_Location.handle (p_record_id);
            when v_type = 'mx'           then v_handle := NetDB_Name.fullname (p_record_id);
            when v_type = 'network'      then v_handle := NetDB_Network.handle (p_record_id);
            when v_type = 'node' or NetDB_NodeType.exists (v_type)
                                         then v_handle := NetDB_Node.handle (p_record_id);
            when v_type = 'organization' then v_handle := NetDB_Organization.handle (p_record_id);
            when v_type = 'owner'        then v_handle := NetDB_Owner.handle (p_record_id);
            when v_type = 'user'         then v_handle := NetDB_User.handle (p_record_id);
            else
                v_handle := p_record_id;
        end case;

        return v_handle;
    end handle;


    function verify_or_die ( p_record_id    in int     := null,
                             p_handle       in varchar := null,
                             p_record_type  in varchar,
                             p_name_type    in varchar := null, -- only needed when searching for a name.
                             check_reserved in int     := null ) return int
    as
        v_type      varchar(32) := lower (p_record_type);
        v_record_id int := 0;
    begin
        case
            when v_type = 'alias'        then v_record_id := verify_or_die (p_record_id, p_handle, 'name', 'A', check_reserved);
            when v_type = 'access'       then v_record_id := NetDB_Access.verify_or_die (p_record_id, p_handle);
            when v_type = 'address'      then v_record_id := NetDB_IPaddress.verify_or_die (p_record_id, p_handle);
            when v_type = 'adminteam'    then v_record_id := NetDB_AdminTeam.verify_or_die (p_record_id, p_handle);
            when v_type = 'dhcpoption'   then v_record_id := NetDB_DHCPoption.verify_or_die (p_record_id, p_handle);
            when v_type = 'dhcpservice'  then v_record_id := NetDB_DHCPservice.verify_or_die (p_record_id, p_handle);
            when v_type in ('domain', 'domain_assign', 'domain_use')
                                         then v_record_id := NetDB_Domain.verify_or_die (p_record_id, p_handle);
            when v_type = 'interface'    then v_record_id := NetDB_Interface.verify_or_die (p_record_id, p_handle);
            when v_type = 'addressspace' then v_record_id := NetDB_IPaddressSpace.verify_or_die (p_record_id, p_handle);
            when v_type = 'location'     then v_record_id := NetDB_Location.verify_or_die (p_record_id, p_handle);
            when v_type = 'make'         then v_record_id := NetDB_Make.verify_or_die (p_handle);

            -- MX insertion ends up here, while the name is still in a reserved state. We need to
            -- force checking of reserved names, while still allowing to check for MX's in either
            -- state. Thus, we default check_reserved to null, above, and use an nvl (...) call here.
            when v_type = 'mx'           then v_record_id := verify_or_die (p_record_id, p_handle, 'name', p_name_type, nvl (check_reserved, 1));

            when v_type = 'name'         then v_record_id := NetDB_Name.verify_or_die (p_record_id, p_handle, p_name_type, Util.to_boolean (check_reserved));
            when v_type = 'network'      then v_record_id := NetDB_Network.verify_or_die (p_record_id, p_handle);
            when v_type = 'node' or NetDB_NodeType.exists (v_type)
                                         then v_record_id := NetDB_Node.verify_or_die (p_record_id, p_handle);
            when v_type = 'organization' then v_record_id := NetDB_Organization.verify_or_die (p_record_id, p_handle);
            when v_type = 'os'           then v_record_id := NetDB_OS.verify_or_die (p_handle);
            when v_type = 'owner'        then v_record_id := NetDB_Owner.verify_or_die (p_record_id, p_handle);
            when v_type = 'person'       then v_record_id := NetDB_Person.verify_or_die (p_record_id, p_handle);
            when v_type = 'resname'      then v_record_id := p_record_id;
            when v_type = 'state'        then v_record_id := NetDB_State.verify_or_die (p_handle);
            when v_type = 'user'         then v_record_id := NetDB_User.verify_or_die (p_record_id, p_handle);
            else
                throw_unknown_record_type (p_record_type);
        end case;

        return v_record_id;
    end verify_or_die;


    -- fullname: given name and domain will assemble fully qualified domain name
    -- will return null if ether of them are null.
    function fullname ( p_name   in varchar,
                        p_domain in varchar ) return varchar as
    begin
        if (p_name is null or p_domain is null) then
            return null;
        end if;

        return ltrim (rtrim (p_name)) || '.' || ltrim (rtrim (p_domain));
    end fullname;


    -- This function will run "full_function_text" and collect the results into a cursor.
    -- full_function_text is expected to be a function call (with arguments) that returns some sort,
    -- of aggregate result.  For instance, a pipelined function.
    --
    -- e.g.: some_cursor := collect_into_cursor ('NetDB_Node.info_pipelined (1)');
    --
    function collect_into_cursor (full_function_text in varchar) return NetDB.ref_cursor
    is
        munged_function varchar(2048) := full_function_text;
        ret_cursor      NetDB.ref_cursor;
    begin
        -- no injection
        munged_function := replace (munged_function, ';');

        -- Assume missing variables are supposed to be null.
        munged_function := regexp_replace (munged_function, '\(\s*,', '(NULL,', 1, 0);
        munged_function := regexp_replace (munged_function, ',\s*\)', ',NULL)', 1, 0);

        while (regexp_instr (munged_function, ',\s*,') > 0)
        loop
            munged_function := regexp_replace (munged_function, ',\s*,',  ',NULL,', 1, 0);
        end loop;

        open ret_cursor for 'select * from table (' || munged_function || ')';
        return ret_cursor;
    end collect_into_cursor;


    -- Raises NetDB.ex_bad_parameter with the appropriate message.
    procedure throw_no_record_id_handle as begin
        raise_application_error (NetDB.ex_bad_parameter, 'Record handle or ID must be supplied.');
    end throw_no_record_id_handle;


    -- Raises NetDB.ex_bad_parameter with the appropriate message.
    procedure throw_unknown_record_type (p_type in varchar) as begin
        raise_application_error (NetDB.ex_bad_parameter, 'Unknown record type "' || p_type || '".');
    end throw_unknown_record_type;


    -- Raises NetDB.ex_bad_parameter with the appropriate message.
    procedure throw_unknown_action (p_action in varchar) as begin
        raise_application_error (NetDB.ex_bad_parameter, 'Unknown action "' || p_action || '".');
    end throw_unknown_action;


    -- Raises NetDB.ex_single_record_expected with the appropriate message.
    procedure throw_single_record_expected (p_record_type in varchar) as begin
        raise_application_error (NetDB.ex_single_record_expected,
            'Multiple ' || NetDB_RecordType.plural (p_record_type) ||
            ' have the requested ID/handle.');
    end throw_single_record_expected;


    -- Raises NetDB.ex_record_in_use with the appropriate message.
    procedure throw_still_in_use ( p_record_type   in varchar,
                                   p_record_handle in varchar ) as
    begin
        raise_application_error (NetDB.ex_record_in_use,
            'Unable to delete the ' || NetDB_RecordType.display (p_record_type) || ' "' ||
             p_record_handle || '" because it is still in use.');
    end throw_still_in_use;


    -- Raises NetDB.ex_record_not_found with the appropriate message.
    procedure throw_record_does_not_exist ( p_record_type   in varchar,
                                            p_record_handle in varchar ) as
    begin
        raise_application_error (NetDB.ex_record_not_found,
            'The ' || NetDB_RecordType.display (p_record_type) ||' "' || p_record_handle ||
            '" does not exist.');
    end throw_record_does_not_exist;


    -- Raises NetDB.ex_record_already_exists with the appropriate message.
    procedure throw_record_already_exists ( p_record_type   in varchar,
                                            p_record_handle in varchar ) as
    begin
        raise_application_error (NetDB.ex_record_already_exists,
            'The ' || NetDB_RecordType.display (p_record_type) ||' "' || p_record_handle ||
            '" already exists.');
    end throw_record_already_exists;


    -- Raises NetDB.ex_record_not_locked with the appropriate message.
    procedure throw_record_not_locked ( p_record_type   in varchar,
                                        p_record_handle in varchar,
                                        p_user_handle   in varchar ) as
    begin
        raise_application_error (NetDB.ex_record_not_locked,
            'The ' || NetDB_RecordType.display (p_record_type) || ' "' || p_record_handle ||
            '" is not locked by the user "' || p_user_handle || '".');
    end throw_record_not_locked;


    -- Raises NetDB.ex_record_locked_by_other with the appropriate message.
    procedure throw_locked_by_other ( p_record_type   in varchar,
                                      p_record_handle in varchar,
                                      p_user_handle   in varchar,
                                      p_lock          in timestamp ) as

        -- first minute after the lock has expired
        available timestamp
          := p_lock + NUMTODSINTERVAL(NetDB_Default.get_int('LockTimeOut')+59,'second');

    begin
        raise_application_error (NetDB.ex_record_locked_by_other,
            'The ' || NetDB_RecordType.display (p_record_type) || ' "' || p_record_handle ||
            '" is locked by "' || p_user_handle || '" until ' || to_char(available,'FMHH:') ||
            to_char(available, 'MI AM') || '.');

    end throw_locked_by_other;


    -- Raises NetDB.ex_too_many_characters with the appropriate message.
    -- The parameter is expected to be the sqlerrm generated from the ORA-12899 exception.
    procedure throw_field_too_long (oracle_error_message in varchar)
    as
        -- 512 characters is the current maximum length of oracle error
        -- messages (at least this is what the documentation claims...)
        bad_col  varchar(512);
        max_size int;
    begin
        -- Extract the maximum length from the error message
        bad_col  := regexp_replace (regexp_substr (lower (oracle_error_message), 'maximum: +[0-9]+'), '[^0-9]');
        max_size := to_number (bad_col);

        -- Extract the column with the problem.
        bad_col  := regexp_replace (lower (oracle_error_message), '^.*value too large for column | \(.+\)|"');
        bad_col  := regexp_replace (bad_col, '^(.*\.)+');

        -- User-ize the column name.
        case bad_col
            when 'comments' then bad_col := 'comment';
            else null;
        end case;

        raise_application_error (NetDB.ex_too_many_characters,
            'The maximum length for the "' || bad_col || '" field is ' || max_size || ' characters.');
    end throw_field_too_long;


    -- Remove the ORA-XXXX that is prepended to sqlerrm.
    function strip_ora_code (error in varchar) return varchar as begin
        return regexp_replace (error, 'ora-[0-9]+: ', '', 1, 1, 'i');
    end strip_ora_code;

end NetDB_Util;
/
