set serveroutput on size 100000;
savepoint start_ateam_tests;

declare
    test_number int          := 1;
    err_msg     varchar(256) := null;
    v_lock      timestamp    := null;
    v_count     int          := 0;
    test_id     int          := 0;

    test_cursor NetDB.ref_cursor;
    result      result_row := result_row (null, null, null, null, null, null, null);

    -- values for insert/update tests.
    test_name      AdminTeam.name%TYPE     := 'test admin team name';
    new_test_name  AdminTeam.name%TYPE     := 'new test admin team name';
    test_email     AdminTeam.email%TYPE    := 'email@address.com';
    test_phone     AdminTeam.phone%TYPE    := '555-555-5555';
    a_comment      AdminTeam.comments%TYPE := 'a comment';

    verified_id int;

    unrecoverable_error EXCEPTION;
begin
    select count (*) into v_count from AdminTeam where lower (name) like test_name or lower (name) like new_test_name;

    if (v_count > 0) then
        err_msg := 'The test admin team "' || test_name || '" exists. If it is not a valid ' ||
                   'admin team, remove the entry and run the tests again. If it is valid, the ' ||
                   'tests must be modified.';
        raise unrecoverable_error;
    end if;

    NetDB_Test.test_info ('AdminTeam', test_number);
    NetDB_Test.test_locking ('AdminTeam', test_number);

    NetDB_Test.revoke_access (NetDB_Test.user_1.id, 'AdminTeam');

    select id into test_id from AdminTeam where rownum = 1;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert an Admin Team without permissions.';

        test_cursor := NetDB_AdminTeam.ins (NetDB_Test.user_1.id, test_name, test_email, test_phone, a_comment);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update an Admin Team without permissions.';

        NetDB_AdminTeam.upd (NetDB_Test.user_1.id, systimestamp, test_id, test_name, test_email, test_phone, a_comment);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete an Admin Team without permissions.';

        NetDB_AdminTeam.del (NetDB_Test.user_1.id, systimestamp, test_id);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    NetDB_Test.grant_access (NetDB_Test.user_1.id, 'AdminTeam');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert an Admin Team.';

        test_cursor := NetDB_AdminTeam.ins (NetDB_Test.user_1.id, test_name, test_email, test_phone, a_comment);
        result      := NetDB_Test.get_generic_row (test_cursor);
        test_id     := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        verified_id := NetDB_AdminTeam.verify_or_die (test_id, test_name);

        if (verified_id = test_id) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update an Admin Team with out access to the record.';

        NetDB_AdminTeam.upd (NetDB_Test.user_1.id, v_lock, test_id, new_test_name, test_email, test_phone, a_comment);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Add user access to an Admin Team.';

        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, test_id, 'AdminTeam');

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update an Admin Team.';

        NetDB_AdminTeam.upd (NetDB_Test.user_1.id, v_lock, test_id, new_test_name, test_email, test_phone, a_comment);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Add a user to an Admin Team.';

        NetDB_AdminTeam.member ('add', NetDB_Test.user_1.id, v_lock, test_id, NetDB_Test.user_1.id);
        NetDB_AdminTeam.member ('add', NetDB_Test.user_1.id, v_lock, test_id, NetDB_Test.user_2.id);

        select count (*) into v_count from AdminTeam_Member
         where adminteam = test_id and person in (NetDB_Test.user_1.id, NetDB_Test.user_2.id);

        if (v_count = 2) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Add a user, who is already a member, to an Admin Team.';

        NetDB_AdminTeam.member ('add', NetDB_Test.user_1.id, v_lock, test_id, NetDB_Test.user_1.id);

        select count (*) into v_count from AdminTeam_Member
         where adminteam = test_id and person in (NetDB_Test.user_1.id, NetDB_Test.user_2.id);

        if (v_count = 2) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Remove a user from an Admin Team.';

        NetDB_AdminTeam.member ('remove', NetDB_Test.user_1.id, v_lock, test_id, NetDB_Test.user_2.id);

        select count (*) into v_count from AdminTeam_Member
         where adminteam = test_id and person in (NetDB_Test.user_1.id, NetDB_Test.user_2.id);

        if (v_count = 1) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Add an organization to an Admin Team.';

        NetDB_AdminTeam.org ('add', NetDB_Test.user_1.id, v_lock, test_id, NetDB_Test.dept_1.id);

        select count (*) into v_count from AdminTeam_Orgs
         where adminteam = test_id and organization = NetDB_Test.dept_1.id;

        if (v_count = 1) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Add an organization, who is already a member, to an Admin Team.';

        NetDB_AdminTeam.org ('add', NetDB_Test.user_1.id, v_lock, test_id, NetDB_Test.dept_1.id);

        select count (*) into v_count from AdminTeam_Orgs
         where adminteam = test_id and organization = NetDB_Test.dept_1.id;

        if (v_count = 1) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Remove an organization from an Admin Team.';

        NetDB_AdminTeam.org ('remove', NetDB_Test.user_1.id, v_lock, test_id, NetDB_Test.dept_1.id);

        select count (*) into v_count from AdminTeam_Orgs
         where adminteam = test_id and organization = NetDB_Test.dept_1.id;

        if (v_count = 0) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete an Admin Team.';

        NetDB_AdminTeam.del (NetDB_Test.user_1.id, v_lock, test_id);
        verified_id := NetDB_AdminTeam.verify_or_die (test_id);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.record_not_found then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;
exception
    when unrecoverable_error then
        dbms_output.put_line ('ERROR: All remaining tests aborted. An unrecoverable error occured: ' || err_msg);
    when others then
        dbms_output.put_line ('ERROR: Unhandled exception "' || sqlcode || '" with message "' || sqlerrm || '"');
        dbms_output.put_line (err_msg);
end;
/

rollback to start_ateam_tests;
