set serveroutput on size 100000;
savepoint start_network_tests;

declare
    test_number int          := 1;
    err_msg     varchar(256) := null;
    v_lock      timestamp    := null;
    v_count     int          := 0;
    v_reserved  timestamp    := null;

    test_cursor NetDB.ref_cursor;
    result      result_row := result_row (null, null, null, null, null, null, null);

    -- Test values
    test_name_rec Name%ROWTYPE;
    test_network  Name.name%TYPE := 'test-1234509876-network';
    test_net_id   int := null;
    test_name_id  int := null;
    location_id   int := null;
    serv_row      DHCPService%ROWTYPE;

    unrecoverable_error EXCEPTION;
begin
    select id into location_id from location    where rownum = 1;
    select *  into serv_row    from DHCPService where rownum = 1;

    select count (*) into v_count from network where id = test_net_id;

    if (v_count > 0) then
        err_msg := 'The test network "' || test_net_id || '" exists. If it is not a valid ' ||
                   'network, remove the entry and run the tests again. If it is valid, the ' ||
                   'tests must be modified.';
        raise unrecoverable_error;
    end if;

    NetDB_Test.test_info ('Network', test_number);
    NetDB_Test.test_locking ('Network', test_number);

    NetDB_Test.revoke_access (NetDB_Test.user_1.id, 'Network');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert a network without permission.';

        test_cursor  := NetDB_Name.reserve (NetDB_Test.user_1.id, test_network, NetDB_Test.domain_1.domain, 'Network');
        result       := NetDB_Test.get_generic_row (test_cursor);
        test_name_id := result.value_string;
        result       := NetDB_Test.get_generic_row (test_cursor);
        v_reserved   := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        select * into test_name_rec from name where name = test_network;

        test_cursor := NetDB_Network.ins (NetDB_Test.user_1.id, null, 'comment');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update a network without permission.';

        NetDB_Network.upd (NetDB_Test.user_1.id, systimestamp, NetDB_Test.network_1.id);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete an network without permission.';

        NetDB_Network.del (NetDB_Test.user_1.id, systimestamp, NetDB_Test.network_1.id);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    NetDB_Test.grant_access (NetDB_Test.user_1.id, 'Network');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert a network.';

        test_cursor := NetDB_Network.ins (NetDB_Test.user_1.id, null, 'comment');
        result      := NetDB_Test.get_generic_row (test_cursor);
        test_net_id := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, test_name_id, null, 'Network', test_net_id, null, v_lock);
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, test_net_id, 'Network');

        NetDB_Network.verify_and_log (NetDB_Test.user_1.id, test_net_id);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update a network.';

        NetDB_Network.upd (NetDB_Test.user_1.id, v_lock, test_net_id, null, 'new comment');

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Add a location to a network.';

        NetDB_Network.location ('add', NetDB_Test.user_1.id, v_lock, test_net_id, location_id);

        select count (*) into v_count from Network_Location
         where network = test_net_id and location = location_id;

        if (v_count > 0) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Add a location to a network that already exists.';

        NetDB_Network.location ('add', NetDB_Test.user_1.id, v_lock, test_net_id, location_id);

        select count (*) into v_count from Network_Location
         where network = test_net_id and location = location_id;

        if (v_count > 0) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Remove a location from a network.';

        NetDB_Network.location ('remove', NetDB_Test.user_1.id, v_lock, test_net_id, location_id);

        select count (*) into v_count from Network_Location
         where network = test_net_id and location = location_id;

        if (v_count = 0) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete an network.';

        NetDB_Network.del (NetDB_Test.user_1.id, v_lock, test_net_id);

        test_net_id := NetDB_Network.verify_or_die (test_net_id);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.record_not_found then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert a network with a service.';

        test_cursor  := NetDB_Name.reserve (NetDB_Test.user_1.id, test_network, NetDB_Test.domain_1.domain, 'Network');
        result       := NetDB_Test.get_generic_row (test_cursor);
        test_name_id := result.value_string;
        result       := NetDB_Test.get_generic_row (test_cursor);
        v_reserved   := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        select * into test_name_rec from name where name = test_network;

        test_cursor := NetDB_Network.ins (NetDB_Test.user_1.id, serv_row.id, 'comment');
        result      := NetDB_Test.get_generic_row (test_cursor);
        test_net_id := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, test_name_id, null, 'Network', test_net_id, null, v_lock);
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, test_net_id, 'Network');

        NetDB_Network.verify_and_log (NetDB_Test.user_1.id, test_net_id);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update a network with a service.';

        rollback to start_network_tests;

        NetDB_Test.grant_access (NetDB_Test.user_1.id, 'Network');

        test_cursor  := NetDB_Name.reserve (NetDB_Test.user_1.id, test_network, NetDB_Test.domain_1.domain, 'Network');
        result       := NetDB_Test.get_generic_row (test_cursor);
        test_name_id := result.value_string;
        result       := NetDB_Test.get_generic_row (test_cursor);
        v_reserved   := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        select * into test_name_rec from name where name = test_network;

        test_cursor := NetDB_Network.ins (NetDB_Test.user_1.id, null, 'comment');
        result      := NetDB_Test.get_generic_row (test_cursor);
        test_net_id := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, test_name_id, null, 'Network', test_net_id, null, v_lock);
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, test_net_id, 'Network');

        NetDB_Network.verify_and_log (NetDB_Test.user_1.id, test_net_id);

        NetDB_Network.upd (NetDB_Test.user_1.id, v_lock, test_net_id, serv_row.id, 'comment');
        NetDB_Network.verify_and_log (NetDB_Test.user_1.id, test_net_id);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

exception
    when unrecoverable_error then
        dbms_output.put_line ('ERROR: All remaining tests aborted. An unrecoverable error occured: ' || err_msg);
    when others then
        dbms_output.put_line ('ERROR: Unhandled exception "' || sqlcode || '" with message "' || sqlerrm || '"');
        dbms_output.put_line (err_msg);
end;
/

rollback to start_network_tests;
