set serveroutput on size 100000;
savepoint start_org_tests;

declare
    test_number int          := 1;
    err_msg     varchar(256) := null;
    v_count     int          := 0;

    test_cursor NetDB.ref_cursor;
    result      result_row := result_row (null, null, null, null, null, null, null);

    -- values for insert and update.
    test_org   varchar(256) := 'Test organization';
    test_org_2 varchar(256) := 'New test organization';
    test_reg   varchar(256) := 'testorgregid';
    test_id    int := null;
    test_id_2  int := null;

    ateam_1   AdminTeam.name%TYPE     := 'ateam-one';
    ateam_2   AdminTeam.name%TYPE     := 'ateam-two';
    a_comment AdminTeam.comments%TYPE := 'a comment';
    email     AdminTeam.email%TYPE    := 'email@address.com';
    phone     AdminTeam.phone%TYPE    := '555-555-5555';

    team_id_1   int;
    team_id_2   int;
    team_lock_1 timestamp;
    team_lock_2 timestamp;
    user_lock_1 timestamp;
    user_lock_2 timestamp;

    old_consult int;
    new_consult int;
    old_LNA     int;
    new_LNA     int;
    old_ateam   int;
    new_ateam   int;

    existing_org Organization%ROWTYPE;
    verified_id  int;

    unrecoverable_error EXCEPTION;
begin
    select count (*) into v_count from Organization
     where organization in (test_org, test_org_2);

    if (v_count > 0) then
        err_msg := 'The test organization "' || test_org || '" exists. If it is not a valid ' ||
                   'organization, remove the entry and run the tests again. If it is valid, the ' ||
                   'tests must be modified.';
        raise unrecoverable_error;
    end if;

    select * into existing_org from Organization where rownum = 1;
    NetDB_Test.revoke_access (NetDB_Test.user_1.id, 'Organization');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert an organization without permission';

        test_cursor := NetDB_Organization.ins (NetDB_Test.user_1.id, test_org, test_reg);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update an organization without permission.';

        test_cursor := NetDB_Organization.upd (NetDB_Test.user_1.id, existing_org.id, test_org_2, test_reg);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete an organization without permission.';

        NetDB_Organization.del (NetDB_Test.user_1.id, existing_org.id);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    NetDB_Test.grant_access (NetDB_Test.user_1.id, 'Organization');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert an organization';

        test_cursor := NetDB_Organization.ins (NetDB_Test.user_1.id, test_org, test_reg);
        result  := NetDB_Test.get_generic_row (test_cursor);
        test_id := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        verified_id := NetDB_Organization.verify_or_die (test_id, test_org);

        if (verified_id = test_id) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update an organization.';

        test_cursor := NetDB_Organization.upd (NetDB_Test.user_1.id, test_id, test_org_2, test_reg);
        NetDB_Test.empty_cursor (test_cursor);

        verified_id := NetDB_Organization.verify_or_die (test_id, test_org_2);

        if (verified_id = test_id) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Rename an organization to something that exists.';

        test_cursor := NetDB_Organization.upd (NetDB_Test.user_1.id, test_id, NetDB_Test.dept_1_name, test_reg);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.record_already_exists then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete an organization.';

        NetDB_Organization.del (NetDB_Test.user_1.id, test_id);
        verified_id := NetDB_Organization.verify_or_die (test_id);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.record_not_found then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Replace an organization.';

        NetDB_Test.grant_access (NetDB_Test.user_1.id, 'adminteam');
        NetDB_Test.add_to_group (NetDB_Test.user_1.id, NetDB_Test.group_2.id);

        -- create admin team
        test_cursor := NetDB_AdminTeam.ins (NetDB_Test.user_1.id, ateam_1, email, phone, a_comment);
        result      := NetDB_Test.get_generic_row (test_cursor);
        team_id_1   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        team_lock_1 := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Owner.add (NetDB_Test.user_1.id, team_lock_1, NetDB_Test.group_1.id, null, team_id_1, 'AdminTeam');
        NetDB_Owner.add (NetDB_Test.user_1.id, team_lock_1, NetDB_Test.group_2.id, null, team_id_1, 'AdminTeam');

        NetDB_AdminTeam.verify_and_log (NetDB_Test.user_1.id, team_id_1);

        -- create admin team
        test_cursor := NetDB_AdminTeam.ins (NetDB_Test.user_1.id, ateam_2, email, phone, a_comment);
        result      := NetDB_Test.get_generic_row (test_cursor);
        team_id_2   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        team_lock_2 := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Owner.add (NetDB_Test.user_1.id, team_lock_2, NetDB_Test.group_1.id, null, team_id_2, 'AdminTeam');
        NetDB_Owner.add (NetDB_Test.user_1.id, team_lock_2, NetDB_Test.group_2.id, null, team_id_2, 'AdminTeam');

        NetDB_AdminTeam.verify_and_log (NetDB_Test.user_1.id, team_id_2);

        -- create organization
        test_cursor := NetDB_Organization.ins (NetDB_Test.user_1.id, test_org, test_reg);
        result  := NetDB_Test.get_generic_row (test_cursor);
        test_id := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        -- create organization
        test_cursor := NetDB_Organization.ins (NetDB_Test.user_1.id, test_org_2, test_reg);
        result    := NetDB_Test.get_generic_row (test_cursor);
        test_id_2 := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Organization.consultant ('insert', NetDB_Test.user_1.id, NetDB_Test.person_1.id, test_id);
        NetDB_Organization.consultant ('insert', NetDB_Test.user_1.id, NetDB_Test.person_2.id, test_id,   null, 1);
        NetDB_Organization.consultant ('insert', NetDB_Test.user_1.id, NetDB_Test.person_2.id, test_id_2, null, 1);

        NetDB_AdminTeam.org ('add', NetDB_Test.user_1.id, team_lock_1, team_id_1, test_id);
        NetDB_AdminTeam.org ('add', NetDB_Test.user_1.id, team_lock_1, team_id_1, test_id_2);
        NetDB_AdminTeam.org ('add', NetDB_Test.user_1.id, team_lock_2, team_id_2, test_id);

        user_lock_1 := NetDB_Test.force_lock (NetDB_Test.user_1.id, NetDB_Test.user_1.id, 'user');
        user_lock_2 := NetDB_Test.force_lock (NetDB_Test.user_1.id, NetDB_Test.user_2.id, 'user');
        NetDB_User.org ('insert', NetDB_Test.user_1.id, user_lock_1, NetDB_Test.user_1.id, test_id);
        NetDB_User.org ('insert', NetDB_Test.user_1.id, user_lock_2, NetDB_Test.user_2.id, test_id,   1);
        NetDB_User.org ('insert', NetDB_Test.user_1.id, user_lock_2, NetDB_Test.user_2.id, test_id_2, 1);

        test_cursor := NetDB_Organization.supplant (NetDB_Test.user_1.id, test_id, test_id_2);
        NetDB_Test.empty_cursor (test_cursor);

        select count (*) into old_consult from Consultant     where organization = test_id;
        select count (*) into new_consult from Consultant     where organization = test_id_2;
        select count (*) into old_LNA     from LNA            where organization = test_id;
        select count (*) into new_LNA     from LNA            where organization = test_id_2;
        select count (*) into old_ateam   from AdminTeam_Orgs where organization = test_id;
        select count (*) into new_ateam   from AdminTeam_Orgs where organization = test_id_2;

        if ( old_consult = 0 and old_LNA = 0 and old_ateam = 0 and
             new_consult = 1 and new_LNA = 2 and new_ateam = 2 )
        then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;
exception
    when unrecoverable_error then
        dbms_output.put_line ('ERROR: All remaining tests aborted. An unrecoverable error occured: ' || err_msg);
    when others then
        dbms_output.put_line ('ERROR: Unhandled exception "' || sqlcode || '" with message "' || sqlerrm || '"');
        dbms_output.put_line (err_msg);
end;
/

rollback to start_org_tests;
