/************************************************************************
 *
 * DBCall.java - this is an abstract class which sole use is to be a
 *               parent for any db call.
 *
 * $Id: DBCall.java,v 1.1.1.1 2006-06-30 23:38:08 priimak Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2001 Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.lib;

import stanford.netdb.utils.*;
import java.text.*;
import java.util.*;
import java.lang.*;
import java.sql.*;

/**
 * This is an abstract class which sole use is to be a parent for any
 * class representing db call. Database call is usually done by NetDBConnection,
 * while instances of DBCall encupsulate information neccesary to perform
 * database call. If child class implements ComplexDBCall then it has to perform
 * database call itself in doIt(NetDBConnection con) method. It is usually the
 * case when DBCall actually does several database calls. In any case methods
 * of NetDBConnection are used to perform the actual call.
 */
abstract class DBCall extends Object {
    public static final int FALSE       = 0;
    public static final int OFF         = 0;
    public static final int TRUE        = 1;
    public static final int ON          = 1;

    protected stanford.netdb.utils.Logger logger;

    protected Object ref_cursor =
        new OutParam(oracle.jdbc.driver.OracleTypes.CURSOR);


    public DBCall(stanford.netdb.utils.Logger logger) {
        this.logger = logger;
    }

    /**
     * In case XXX_PROC_CALL db call done by calling first prepareCall(...) and
     * then filling/registering up parameters marked by '?' mark. Those parameters
     * are teken from this array. Note, that it is up to the child class to
     * initiate properly this array. Also, note that each type of the value should
     * correspond to the appropreate sql type in a db call. For example if sql type
     * in a parameter of stored procedure is varchar(xxx) then corresponding element
     * of array 'values' should be instance of String. If it is Null of, let say,
     * type timestamp then it shold be Null(java.sql.Types.TIMESTAMP).
     * If it is output parameter then it should be OutParam(java.sql.Types.XXX)
     *
     * @see stanford.netdb.utils.Null
     * @see stanford.netdb.lib.OutParam
     */
    protected Object[] values = null;

    /* Constants to be returned by getType() */
    public static final int UPDATE                  = 1;
    public static final int SELECT                  = 2;
    public static final int WARN_PROC_CALL          = 3;
    public static final int SIMPLE_PROC_CALL        = 4;
    public static final int PREP_PROC_CALL          = 5;
    public static final int ORA_FUN_IGNORE_RET_CALL = 6;
    public static final int ORA_FUN_CALL            = 7;
    public static final int ORA_PROC_CALL           = 8;

    /** Method getType() in its default incarnation returns this value. */
    protected int db_call_type = ORA_FUN_CALL;

    /** Returns the SQL to be executed */
    public abstract String getSQL();

    /**
     * Should returns one of the following
     * DBCall.{UPDATE, PROC_CALL, SIMPLE_PROC_CALL, SELECT, FULL_SEARCH_SELECT}.
     * Default is SIMPLE_PROC_CALL
     */
    public int getType() {
        return db_call_type;
    }

    /** Returns reference to values array */
    public Object[] getValues() {
        return values;
    }

    /** Returns the object which is to be set up for procedure call */
    public final Object get(int index) {
        if( index >= values.length )
            return null;
        return values[index];
    }

    /**
     * Returns index of the result set which contains error code(s).
     * It is common convention in NetDB stored procedures that first result set
     * contain error code(s) if something went wrong, it is not the case for all
     * things you may want to do. For example for simple select this method should
     * return -1. Indicates which result set contain error message, numbering
     * from 1 for most cases it is 1 but may be different like, for example for
     * pNode 'info' when it is 2
     */
    public int errorResultSetIndex() {
        return 1;
    }

    /**
     * Processes result set coming out of NetDBConnection.callDB(...) and
     * builds appropriate Hashtable structure. It is up to the child class to
     * implement this method in the proper way.
     */
    public Hashtable processResultSet(Hashtable res) throws Exception {
        return res;
    }

    /** Begin transaction. Note that statement will be closed at the end. */
    protected void BeginTransaction(Connection con)
        throws SQLException {
        con.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
        con.setAutoCommit(false);
    }

    /** Commit transaction. Note that statement will be closed at the end. */
    protected void CommitTransaction(Connection con)
        throws SQLException {
        try {
            con.commit();
        } finally {
            try { con.close(); } catch(Exception ex) {}
        }
    }

    /** Rollback transaction. Note that statement will be closed at the end. */
    protected void RollbackTransaction(Connection con)
        throws SQLException {
        try {
            con.rollback();
        } finally {
            try { con.close(); } catch(Exception ex) {}
        }
    }

    public Object newRefCursor() {
        return new OutParam(oracle.jdbc.driver.OracleTypes.CURSOR);
    }

    /** Used for debugging */
    protected static void printRawData(Hashtable result_set) throws java.sql.SQLException {
        Vector rs = (Vector)result_set.get("data");
        ResultSetMetaData md = (ResultSetMetaData)result_set.get("meta_data");
        for( int j = 1; j <= md.getColumnCount(); j++ )
            System.out.print("<"+md.getColumnName(j)+"> ");
        System.out.println("");
        for( int j = 0; j < rs.size(); j++ ) {
            Vector row = (Vector)rs.elementAt(j);
            for( int k = 0; k < row.size(); k++ )
                System.out.print("["+row.elementAt(k).toString().trim()+"] ");
            System.out.println("");
        }
    }

    public String toString() {
        String sql = new String(getSQL());
        if( sql.indexOf("?") < 0 )
            return sql;
        int i = 0;
        int index = -1;
        int start_index = 0;
        String retSQL = "";
        while( (index = sql.indexOf("?", start_index)) > -1 ) {
            Object value = values[i];
            if( value instanceof String )
                value = "'"+values[i]+"'";
            else if( value instanceof Null )
                value = "Null";
            else if( value == null )
                value = "null";
            retSQL += sql.substring(start_index, index)+value.toString();
            start_index = index+1;
            i++;
        }
        retSQL += sql.substring(start_index);
        return retSQL;
    }

    protected void printSQL() {
        System.out.println(toString());
    }

    protected String parseDate(String date) throws ParseException {
        if( date != null && date.trim().length() > 0 ) {
            date = date.replace('-', '/');
            DateFormat df = DateFormat.getDateInstance(DateFormat.SHORT);
            df.setLenient(true);
            return (new java.sql.Timestamp((df.parse(date).getTime()))).toString();
        } else
            return date;
    }

    protected Object o2v(Object obj) {
        return o2n(obj, Types.VARCHAR);
    }

    protected Object o2n(Object obj, int type) {
        if( obj == null )
            return new Null(type);
        else
            return obj;
    }

    protected Hashtable get_hash_or_create(Hashtable parent, Object key) {
        return Utils.get_hash_or_create(parent, key);
    }

    // given procedure name builds sql to be passed to prepareStatement(...)
    // parameters are build by looking into values[] array
    protected String proc_call_sql(String proc_name) {
        return proc_call_sql((proc_name.startsWith("$(SCHEMA)")?"":"$(SCHEMA)")+proc_name,
                             db_call_type==ORA_PROC_CALL?values.length:(values.length - 1));
    }

    // given procedure name and amount of parameters builds sql
    // to be passed to prepareStatement(...)
    protected String proc_call_sql(String proc_name, int parameters) {
        String dbschema = System.getProperty(Constants.DBSCHEMA_KEY);
        dbschema = (dbschema==null||dbschema.trim().length()==0)?"":(dbschema.trim()+".");
        proc_name = proc_name.replaceAll("\\$\\(SCHEMA\\)", dbschema)+(parameters>0?"(":"");
        String sql = "{call ";
        if( db_call_type != ORA_PROC_CALL )
            sql += "? := ";
        sql += proc_name;
        for( int i = 0; i < parameters; i++ )
            sql += "?"+(i+1<parameters?",":"");

        // now append closing brackets, one for each opening bracket
        int start = 0;
        for(;;) {
            start = proc_name.indexOf('(', start);
            if( start == -1 )
                break;
            start++;
            sql += ")";
        }

        return sql+"}";
    }

    protected void checkErrors(String msgHeader, Hashtable result) throws NetDBException {
        Vector error_set = (Vector)result.get("errors");
        if( error_set != null && error_set.size() > 0 ) {
            Vector errors = new Vector();
            for( int j = 0; j < error_set.size(); j++ )
                errors.addElement(error_set.elementAt(j).toString().trim());
            throw new NetDBException(msgHeader, errors);
        }
    }
}
