/************************************************************************
 *
 * DefDirProvider.java -
 *
 * $Id: DefDirProvider.java,v 1.2 2006-09-26 17:29:46 priimak Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2003, Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.lib;

import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.*;
import stanford.netdb.utils.*;

class DefDirProvider implements DirectoryProvider {

    private static final int PORT = 9751;

    private static Logger common_logger = null;
    private Logger logger = null;

    public static void main(String[] cmd) {
        try {
            if( cmd.length < 2 )
                System.out.println("Usage: $ DefDirProvider (-n NetID|-r RegID|-l Lookup)");
            else {
                DefDirProvider dir = new DefDirProvider();
                dir.init(new DefaultLogger());
                if( cmd[0].equals("-n") )
                    Utils.printHash(dir.getByUID(cmd[1]));
                else if( cmd[0].equals("-r") )
                    Utils.printHash(dir.getByKey(cmd[1]));
                else if( cmd[0].equals("-l") )
                    Utils.printHash(dir.lookup(cmd[1]));
                else
                    System.out.println("Usage: $ DefDirProvider (-n NetID|-r RegID|-l Lookup)");
            }
        } catch(Exception ex) {
            ex.printStackTrace();
        }
    }

    public DefDirProvider() {}

    public boolean init(Logger logger) {
        common_logger = logger;
        return true;
    }

    public void setLogger(Logger logger) {
        this.logger = logger;
    }

    public boolean connect(NetDBpi netdb) {
        return true;
    }

    public boolean disconnect() {
        return true;
    }

    public Hashtable lookup(String searchString) throws Exception {
        return doLookup(searchString, "lookup");
    }

    public Hashtable getByKey(String key) throws Exception {
        return doLookup(key, "byKey");
    }

    public Hashtable getByUID(String uid) throws Exception {
        return doLookup(uid, "byUID");
    }

    public Hashtable verify(Hashtable person)
        throws Exception, IllegalArgumentException, DirectoryProvider.DirectoryException {
        if( person == null )
            throw new IllegalArgumentException("Argument must not be null");

        Object regid = person.get("RegID");
        if( regid == null )
            throw new IllegalArgumentException("Persons hashtale must contain key 'RegID'.");

        if( !(regid instanceof String) )
            throw new IllegalArgumentException("Value of 'RegID' is not a String.");

        String regID = ((String)regid).trim();
        if( regID.length() == 0 )
            throw new IllegalArgumentException("Key 'RegID' must have non empty value.");

        Hashtable retval = getByKey(regID);
        if( retval.size() == 0 ) {
            Hashtable return_person = new Hashtable();
            return_person.put("warning", "Did not find person in a directory.");
            return return_person;
        }

        if( retval.size() > 1 ) {
            Hashtable return_person = new Hashtable();
            return_person.put("warning", "Found more then one person in a directory.");
            return return_person;
        }

        return Utils.nHash(retval.get(retval.keys().nextElement()));
    }

    private Hashtable doLookup(String searchString, String sType) throws Exception {

        String filter = "()";
        if( sType.equals("byUID") )
            filter = "(|(suSUNetID="+searchString.trim()+")(uid="+searchString.trim()+"))";
        else if( sType.equals("byKey") )
            filter = "(suRegID="+searchString.trim()+")";
        else if( sType.equals("lookup") ) {
            filter = makeLookupFilter(searchString.trim());
        } else
            throw new Exception("Illegal search type");

        Hashtable people = dispatchCall(filter);
        // prefill 'object' -> 'person'
        Enumeration enm = people.keys();
        while( enm.hasMoreElements() ) {
            Object person = people.get(enm.nextElement());
            if( person instanceof Hashtable ) {
                ((Hashtable)person).put("object", "person");
            }
        }

        return people;
    } // end of doLookup(...)

    private Hashtable dispatchCall(String filter) throws Exception {
        Socket connection = null;
        Hashtable result = new Hashtable();
        try {
            connection = new Socket(InetAddress.getByName("127.0.0.1"), PORT);
            ObjectInputStream is = new ObjectInputStream(connection.getInputStream());
            ObjectOutputStream os = new ObjectOutputStream(connection.getOutputStream());
            os.writeObject(filter);
            result = (Hashtable)is.readObject();

        } catch (Exception ex) {
            error(ex.getMessage());
            throw ex;

        } finally {
            if( connection != null )
                try {
                    connection.close();
                } catch (Exception ex) {
                }
        }
        return result;
    }

    private String makeLookupFilter(String s) {

        String first, last;

        if (s.indexOf("(") != -1 || s.indexOf(")") != -1) return("");

        // rotate about the first comma if necessay
        if (s.indexOf(",") != -1) s = s.substring(s.indexOf(",")+1,s.length())
                                    + " " + s.substring(0,s.indexOf(","));

        // trim
        s = s.trim();

        // return filter for 'fisrt last' or just 'string'
        if (s.indexOf(" ") != -1) {
            first = s.substring(0,s.indexOf(" "));
            last  = s.substring(s.indexOf(" "),s.length()).trim();
            return("(|(&(sn=" +last+ ")(gn=" +first+ "))" +
                   "(displayName=" +first+ " " +last+ ")" +
                   "(suOtherName=" +first+ " " +last+ "))");
        } else
            return("(|(uid=" +s+ ")(suSUNetID=" +s+ ")(suGeneralID=" +s+ ")(sn=" +s+ "))");
    }

    private void error(Object msg) {
        if( logger == null ) {
            synchronized( common_logger ) {
                common_logger.error(msg);
            }
        } else
            logger.error(msg);
    }

    /*
    private void debug(Object msg) {
        if( logger == null ) {
            synchronized( common_logger ) {
                common_logger.debug(msg);
            }
        } else
            logger.debug(msg);
    }
    */
}
