/************************************************************************
 *
 * DirectoryProvider.java -
 *
 * $Id: DirectoryProvider.java,v 1.1.1.1 2006-06-30 23:38:09 priimak Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2002, Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.lib;

import java.lang.*;
import java.util.*;
import stanford.netdb.utils.*;

/**
 * Any directory providers must inherit from this interface and provide methods listed bellow
 */
public interface DirectoryProvider {

    /**
     * You would place here something like reading config files if your library reads any.
     * It is called in only once, on start up of netdb, because of this, passed logger should be stored in
     * some static variable.
     *
     * @return boolean call status, i.e. true if called succeeded and false if not
     * @throws Exception on unexpected condition, throwing this should indicate internal bug in you libraryy
     * @throws DirectoryProvider.DirectoryException should be thrown on possible unexpected condition like
     * directory server been down or alike.
     */
    public boolean init(Logger logger) throws Exception, DirectoryProvider.DirectoryException;


    /**
     * Once instance of DirectoryProvider obtained logger which is set in init NetDB will call this method
     * with intent to reset logger to a new value, which should be used instead of the one passed to
     * DirectoryProvider by calling init(...). Note, that init(...) called only once when netdb if loaded.
     */
    public void setLogger(Logger logger);

    /**
     * Before doing any directory access netdb calls connect(), then one of lookup methods and
     * finally disconnect(). Calling instance of netdb is passed to the method, which can be used
     * you want to call netdb methods inside of lookup(...), getByKey(...), verify(...) etc.
     *
     * @return boolean call status, i.e. true if called succeeded and false if not
     * @throws Exception on unexpected condition, throwing this should indicate internal bug in you libraryy
     * @throws DirectoryProvider.DirectoryException should be thrown on possible unexpected condition like
     * directory server been down or alike.
     */
    public boolean connect(NetDBpi netdb) throws Exception, DirectoryProvider.DirectoryException;

    /**
     * Called right after look up is performed.
     *
     * @return boolean call status, i.e. true if called succeeded and false if not
     * @throws Exception on unexpected condition, throwing this should indicate internal bug in you libraryy
     * @throws DirectoryProvider.DirectoryException should be thrown on possible unexpected condition like
     * directory server been down or alike.
     */
    public boolean disconnect() throws Exception, DirectoryProvider.DirectoryException;

    /**
     * General look up. This is the only look up which is called in netdb.
     *
     * @param searchString well, it is a search string
     * @return Hashtable of following structure<br>
     * <pre>
     * (Hashtable)
     * %key%  ->  (Hashtable) &lt;PERSON&gt;
     *            'object'     -> 'person'
     *            'name'       -> (String) - display name
     *            'RegID'      -> (String) - key in a registry
     *            'active?'    -> (Boolean)
     *            'department' -> (String)
     *            'lastname'   -> (String) - could be used for sorting
     *            'phone'      -> (String)
     *            'email'      -> (String)
     *            'NetID'      -> (String) - NetID ( SUNetID in Stanford )
     *            'DirID'      -> (String)
     * </pre>
     *
     * @throws Exception on unexpected condition, throwing this should indicate internal bug in you libraryy
     * @throws DirectoryProvider.DirectoryException should be thrown on possible unexpected condition like
     * directory server been down or alike.
     */
    public Hashtable lookup(String searchString) throws Exception, DirectoryProvider.DirectoryException;

    /**
     * Look up by Registry id.
     *
     * @param key Registry id
     * @return Hashatble - same as {@link #lookup(String)}
     * @throws Exception on unexpected condition, throwing this should indicate internal bug in you libraryy
     * @throws DirectoryProvider.DirectoryException should be thrown on possible unexpected condition like
     * directory server been down or alike.
     */
    public Hashtable getByKey(String key) throws Exception, DirectoryProvider.DirectoryException;

    /**
     * Look up by Net id.
     *
     * @param uid registry id
     * @return Hashatble - same as {@link #lookup(String)}
     * @throws Exception on unexpected condition, throwing this should indicate internal bug in you libraryy
     * @throws DirectoryProvider.DirectoryException should be thrown on possible unexpected condition like
     * directory server been down or alike.
     */
    public Hashtable getByUID(String uid) throws Exception, DirectoryProvider.DirectoryException;

    /*
     * This method is called by NetDB.update(...) right before updating person record. The purpose of
     * this method is to provide flexibility in giving ability for some particular instalations to
     * acually set arbitrary person data, which is the case when verify(...) returns the same person
     * without doing anything. Note, that directory provider used in Stanford will use 'RigID' value
     * from person hashtable to acually perform directory lookup and return found person hash.
     *
     * @param person Persons hashtable, see {@link #lookup(String)}
     * @return Persons hashtble (see {@link #lookup(String)}) with possible key 'warning' which in turn
     * will be returned by NetDB.update(...) call. If nothing found it is expected that return value
     * would either empty Hashtable or at most containing 'warning' message.
     */
    public Hashtable verify(Hashtable person)
        throws Exception, IllegalArgumentException, DirectoryProvider.DirectoryException;

    public class DirectoryException extends Exception {

        public DirectoryException() {
            super();
        }

        public DirectoryException(String message) {
            super(message);
        }

        public DirectoryException(String message, Throwable cause) {
            super(message, cause);
        }

        public DirectoryException(Throwable cause) {
            super(cause);
        }
    }
}
