/************************************************************************
 *
 * DomainModifyCommand.java - db call pDomain 'update'
 *
 * $Id: DomainModifyCommand.java,v 1.4 2008-03-15 02:56:01 jlidtke Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2001 Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.lib;

import stanford.netdb.utils.*;
import java.util.*;
import java.lang.*;
import java.sql.*;

/** db call pDomain 'update' */
class DomainModifyCommand extends DBCall implements ComplexDBCall {

    private Hashtable oldDomain;
    private Hashtable newDomain;
    private int id;
    private String action;
    private Timestamp lock;
    private int procedure;
    private int op_uid;

    public DomainModifyCommand(stanford.netdb.utils.Logger logger,
                               Hashtable oldDomain, Hashtable newDomain,
                               Integer op_uid) {
        super(logger);
        this.oldDomain = oldDomain;
        this.newDomain = newDomain;
        this.op_uid    = op_uid;

        Integer id_domain = null;
        if( oldDomain != null ) {
            action = "update";
            procedure = Function_Map.UPDATE;
            id_domain = new Integer((String)oldDomain.get("id"));
            id = id_domain.intValue();

            values = new Object[6];
            values[0] = op_uid;
            values[1] = lock = (Timestamp)newDomain.get("locked");
            values[2] = id;
            values[3] = newDomain.get("domain");
            values[4] = new Short( ((Boolean)newDomain.get("limited?")).booleanValue()?"1":"0" );
            values[5] = newDomain.get("comment");
        } else {
            action = "insert";
            procedure = Function_Map.INSERT;
            this.oldDomain = Utils.hash("groups", new Hashtable());

            values = new Object[5];
            values[0] = ref_cursor;
            values[1] = op_uid;
            values[2] = newDomain.get("domain");
            values[3] = new Short( ((Boolean)newDomain.get("limited?")).booleanValue()?"1":"0" );
            values[4] = newDomain.get("comment");
        }

        db_call_type = Function_Map_Controller.get_function_type (Function_Map.PACKAGE_DOMAIN, procedure);
    }

    public String getSQL() {
        return proc_call_sql(Function_Map_Controller.get_function_name (Function_Map.PACKAGE_DOMAIN, procedure));
    }

    public Hashtable doIt(NetDBConnection con) throws NetDBException, Exception {
        BeginTransaction(con.con);
        boolean commit_finally = false;
        try {
            logger.info("modifying domain "+oldDomain.get("domain"));
            logger.debug("do sigle value attributes");
            printSQL();
            Hashtable result = con.callDB(this);

            // check for errors
            Vector error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            if( action.equals("insert") ) {
                id   = Integer.parseInt((String)result.get("id"));
                lock = (Timestamp)result.get("locked");

                // now accept domain name
                DBCall call = new NameAcceptCommand(logger,
                                                    new Integer(id), "domain",
                                                    new Integer(id), null,
                                                    lock, null, op_uid);
                result = con.callDB(call);

                // check for errors
                error_set = (Vector)result.get("errors");
                if( error_set != null && error_set.size() > 0 )
                    return result;
            }

            // domain related groups
            result = update_domain_groups(con,
                                          Utils.nHash(oldDomain.get("useAsName")),
                                          Utils.nHash(newDomain.get("useAsName")),
                                          ObjectGroupCommand.DOMAIN_USE);

            if( result != null ) return result;

            result = update_domain_groups(con,
                                          Utils.nHash(oldDomain.get("assign")),
                                          Utils.nHash(newDomain.get("assign")),
                                          ObjectGroupCommand.DOMAIN_ASSIGN);

            if( result != null ) return result;

            // now groups
            Hashtable newGroups = (Hashtable)newDomain.get("groups");
            Hashtable oldGroups = (Hashtable)oldDomain.get("groups");
            Vector groups_to_insert = new Vector();
            Vector groups_to_delete = new Vector();
            Vector groups_to_keep = new Vector();
            Utils.keysIntersect(oldGroups, newGroups, groups_to_insert,
                                groups_to_delete, groups_to_keep);

            // insert groups_to_insert
            logger.debug("inserting groups_to_insert");
            ObjectGroupCommand cmd =
                new ObjectGroupCommand(logger, ObjectGroupCommand.ADD_GROUPS,
                                       id, ObjectGroupCommand.DOMAIN,
                                       lock, groups_to_insert, op_uid);
            result = cmd.doIt(con);

            // check for errors
            error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            // delete groups_to_delete
            logger.debug("deleting groups_to_delete");
            cmd =
                new ObjectGroupCommand(logger, ObjectGroupCommand.REMOVE_GROUPS,
                                       id, ObjectGroupCommand.DOMAIN,
                                       lock, groups_to_delete, op_uid);
            result = cmd.doIt(con);

            // check for errors
            error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            // now do check
            logger.debug("do check");
            DBCall verify = new VerifyCommand(logger, VerifyCommand.DOMAIN,
                                              new Integer(id), op_uid);

            result = con.callDB(verify);

            // check for errors
            error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            // unlock domain
            DBCall call = new DomainUnlockCommand(logger, new Integer(id), lock, op_uid);
            result = call.processResultSet(con.callDB(call));

            // check for errors
            error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            ((NetDBii)logger).unregister_lock(lock, new Integer(Constants.DOMAIN));

            commit_finally = true;

        } catch(Exception ex) {
            ex.printStackTrace();
            try {
                logger.debug("Rollback Transaction");
                RollbackTransaction(con.con);
            } catch(Exception exx) { }
            throw ex;

        } finally {
            if( commit_finally )
                try {
                    logger.info("modifying domain "+oldDomain.get("domain")+" done");
                    CommitTransaction(con.con);
                } catch(Exception exx) { }
            else
                try {
                    logger.info("modifying domain "+oldDomain.get("domain")+" failed");
                    logger.debug("Rollback Transaction");
                    RollbackTransaction(con.con);
                } catch(Exception exx) { }
        }
        return new Hashtable();
    }

    private Hashtable update_domain_groups(NetDBConnection con, Hashtable oldGroups,
                                           Hashtable newGroups, String which)
        throws NetDBException, InternalException, Exception
    {
        Vector groups_to_insert = new Vector();
        Vector groups_to_delete = new Vector();
        Vector groups_to_keep = new Vector();
        Utils.keysIntersect(oldGroups, newGroups, groups_to_insert,
                            groups_to_delete, groups_to_keep);

        // insert groups_to_insert
        logger.debug("inserting groups_to_insert");
        ObjectGroupCommand cmd =
            new ObjectGroupCommand(logger, ObjectGroupCommand.ADD_GROUPS,
                                   id, which, lock, groups_to_insert,
                                   op_uid);

        Hashtable result = cmd.doIt(con);

        // check for errors
        Vector error_set = (Vector)result.get("errors");
        if( error_set != null && error_set.size() > 0 )
            return result;

        // delete groups_to_delete
        logger.debug("deleting groups_to_delete");
        cmd =
            new ObjectGroupCommand(logger, ObjectGroupCommand.REMOVE_GROUPS,
                                   id, which, lock, groups_to_delete,
                                   op_uid);

        result = cmd.doIt(con);

        // check for errors
        error_set = (Vector)result.get("errors");
        if( error_set != null && error_set.size() > 0 )
            return result;

        return null;
    }
}
