/************************************************************************
 *
 * NetDBConnection.java - This class represents database connection.
 *
 * $Id: NetDBConnection.java,v 1.4 2008-03-15 02:56:10 jlidtke Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2001 Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.lib;

import stanford.netdb.utils.*;
import java.util.*;
import java.lang.*;
import java.sql.*;

/**
 * This class represents database connection.
 */
class NetDBConnection {

    // actual database connection
    public Connection con = null;

    private stanford.netdb.utils.Logger logger;
    private String db_url;
    private Properties db_props;

    public NetDBConnection(stanford.netdb.utils.Logger logger,
                           Connection con, String db_url,
                           Properties db_props)
    {
        this.logger = logger;
        this.con = con;
        this.db_url = db_url;
        this.db_props = db_props;
    }

    /**
     * releases connection, the idea is that for pool of connections,
     * it places connection back in a pool of available connections.
     * Note that connection pool is not implemented yet and may never be.
     */
    public void releaseConnection() throws Exception {
        con.close();
        con = null;
        System.gc();
    }

    public Hashtable callDB(DBCall call)
        throws Exception, SQLException, InternalException {
        return callDB(call, null);
    }

    /**
     * performs actual database call and stuff result set(s) into Hashtable
     * *** HAVE TO WRITE ABOUT FORMAT OF THE HASH ***
     */
    public Hashtable callDB(DBCall call, String searchID)
        throws Exception, SQLException, InternalException {
        boolean   sql_timing    = System.getProperty("enable.sql.timing", "").equals("true");
        Statement stmt          = null;
        CallableStatement cstmt = null;
        PreparedStatement pstmt = null;
        Hashtable result        = new Hashtable();
        ResultSet res           = null;
        Vector warnings         = new Vector();
        Object[] values = call.getValues();
        int callType = call.getType();
        Vector errors_set = new Vector();
        try {
            if( con == null || con.isClosed() )
                con = DriverManager.getConnection(db_url, db_props);
            switch( callType ) {
            case DBCall.SELECT:
                stmt = con.createStatement();
                if( searchID != null )
                    ((NetDBii)logger).storeSearch(searchID, stmt, con);

                long start_time = Calendar.getInstance().getTime().getTime();

                if( System.getProperty(Constants.DBSCHEMA_KEY) != null )
                    stmt.execute("ALTER SESSION SET CURRENT_SCHEMA="+
                            System.getProperty(Constants.DBSCHEMA_KEY).trim());
                if( call instanceof FullSearchCommand ) {
                    String sql = ((FullSearchCommand)call).getSearchSQL();
                    logger.debug(sql);
                    res = stmt.executeQuery(sql);
                } else
                    res = stmt.executeQuery(call.getSQL());

                if( sql_timing )
                    logger.info("["+call.toString()+"] completed in "+
                                ((Calendar.getInstance().getTime().getTime()-start_time)/1000.0)+" sec.");

                SQLWarning warning = res.getWarnings();

                // stuff warnings
                result.put("warnings", warnings);
                if( warning != null ) {
                    warnings.addElement(warning);
                    while( (warning = warning.getNextWarning()) != null )
                        warnings.addElement(warning);
                }

                // stuff metadata
                ResultSetMetaData metadata = res.getMetaData();
                Vector column_names = new Vector(metadata.getColumnCount());
                for( int i = 0; i < metadata.getColumnCount(); i++ ) {
                    column_names.addElement(metadata.getColumnName(i+1));
                }
                result.put("column_names", column_names);
                boolean no_metadata = false;
                int columns = metadata.getColumnCount();
                Vector resultSet = new Vector();
                result.put("resultSet", resultSet);
                while( res.next() ) {
                    Vector row = new Vector();
                    resultSet.addElement(row);
                    for( int i = 1; i <= columns; i++ ) {
                        Object cell = res.getObject(i);
                        if( cell instanceof ResultSet ) {
                            try {
                                no_metadata = true;
                                Hashtable local_result = new Hashtable();
                                Vector errors = new Vector();
                                Vector warning_set = new Vector();
                                local_result.put("errors", errors);
                                local_result.put("warnings", warning_set);
                                pickHash(((ResultSet)cell), local_result, errors, warning_set);
                                row.addElement(Utils.copyHash(local_result));
                            } finally {
                                try { ((ResultSet)cell).close(); } catch (Exception exc) { }
                            }
                        } else
                            row.addElement(cell);
                    }
                }
                if( !no_metadata ) {
                    result.put("metadata", metadata);
                }
                break;

            case DBCall.UPDATE:
                stmt = con.createStatement();
                if( searchID != null )
                    ((NetDBii)logger).storeSearch(searchID, stmt, con);
                start_time = Calendar.getInstance().getTime().getTime();
                int rowcount = stmt.executeUpdate(call.getSQL());
                if( sql_timing )
                    logger.info("["+call.getSQL()+"] completed in "+
                                ((Calendar.getInstance().getTime().getTime()-start_time)/1000.0)+" sec.");
                result = new Hashtable();
                result.put("rowcount", new Integer(rowcount));
                break;

            case DBCall.SIMPLE_PROC_CALL:
            case DBCall.WARN_PROC_CALL:
            case DBCall.PREP_PROC_CALL:
            case DBCall.ORA_FUN_IGNORE_RET_CALL:
            case DBCall.ORA_FUN_CALL:
            case DBCall.ORA_PROC_CALL:
                logger.debug(call.toString());
                if( callType == DBCall.SIMPLE_PROC_CALL ||
                    callType == DBCall.ORA_FUN_IGNORE_RET_CALL ||
                    callType == DBCall.ORA_FUN_CALL ) {
                    pstmt = con.prepareCall(call.getSQL());
                    cstmt = (CallableStatement)pstmt;
                } else
                    pstmt = con.prepareStatement(call.getSQL());

                try {
                    if( searchID != null )
                        ((NetDBii)logger).storeSearch(searchID, pstmt, con);
                    for( int i = 0; i < values.length; i++ ) {
                        if( values[i] == null )
                            throw new InternalException("NetDBConnection::callDB(...) "+
                                                        "values["+i+"] is null");
                        if      ( values[i] instanceof Null )
                            pstmt.setNull(i+1, ((Null)values[i]).getType());
                        else if ( values[i] instanceof OutParam )
                            cstmt.registerOutParameter(i+1, ((OutParam)values[i]).getType());
                        else if ( values[i] instanceof String )
                            pstmt.setString(i+1, (String)values[i]);
                        else if ( values[i] instanceof Integer )
                            pstmt.setInt(i+1, ((Integer)values[i]).intValue());
                        else if ( values[i] instanceof Short )
                            pstmt.setShort(i+1, ((Short)values[i]).shortValue());
                        else if ( values[i] instanceof Timestamp )
                            pstmt.setTimestamp(i+1, (Timestamp)values[i]);
                        else
                            throw new InternalException("NetDBConnection::callDB(...) "+
                                                        "unsupported sql type");
                    }
                } catch (Exception ex) {
                    ex.printStackTrace();
                    throw ex;
                }

                // actually execute sql
                boolean fatalError = false; // will be set to true if there is SQLException on res.next()
                boolean hasResults = true;
                try {
                    start_time = Calendar.getInstance().getTime().getTime();
                    for(int i = 1; i < 16; i++) { // try 15 times
                        try {
                            hasResults = pstmt.execute();
                            break;
                        } catch(SQLException ex) {
                            if( ex.getErrorCode() != 8177 )
                                throw ex;

                            // If we are here then we got oracle error
                            // "ORA-08177: can't serialize access for this transaction"
                            // Then we try to restart it

                            logger.debug("ORA-08177. Retrying on "+i);
                        }
                    }
                    if( sql_timing )
                        logger.info("["+call.toString()+"] completed in "+
                                    ((Calendar.getInstance().getTime().getTime()-start_time)/1000.0)+" sec.");

                    // go over result sets
                    if( hasResults ) {
                        Vector result_sets = new Vector();
                        result.put("result_sets", result_sets);

                        res = pstmt.getResultSet();

                        int counter = 1; // counter of result sets
                        boolean hasErrors = false;
                        while( res != null ) {
                            Vector result_set_data = new Vector();
                            ResultSetMetaData result_set_meta_data = res.getMetaData();

                            warning = res.getWarnings();
                            if( warning != null ) {
                                warnings.addElement(warning);
                                while( (warning = warning.getNextWarning()) != null )
                                    warnings.addElement(warning);
                            }

                            int clcount = result_set_meta_data.getColumnCount();
                            boolean hasData = false; // will set to true if there are non null data
                            try {
                                while( res.next() ) {
                                    Vector row = new Vector();
                                    result_set_data.addElement(row);
                                    for( int i = 0; i < clcount; i++ ) {
                                        Object obj = res.getObject(i+1);
                                        if( obj != null )
                                            hasData = true;
                                        row.addElement(obj==null?new Null():obj);
                                    }
                                }

                            } catch(SQLException exx) {
                                errors_set.addElement(exx.getMessage().trim());
                                fatalError = true;
                            }

                            if( ( counter == call.errorResultSetIndex() && hasData ) || hasErrors ) {
                                hasErrors = true;
                                if( hasData ) {
                                    // go over result set and collect errors
                                    for( int i = 0; i < result_set_data.size(); i++ )
                                        errors_set.addElement( ((Vector)result_set_data.elementAt(i)).elementAt(0) );
                                }

                            } else if( counter != call.errorResultSetIndex() ) {
                                // regular result set
                                Hashtable result_set = new Hashtable();
                                result_set.put("data", result_set_data);
                                result_set.put("meta_data", result_set_meta_data);
                                result_sets.addElement(result_set);
                            }

                            try { res.close(); } catch(Exception exx) { }
                            if( fatalError || ( !pstmt.getMoreResults() && (pstmt.getUpdateCount() == -1) ) )
                                break;
                            else
                                res = pstmt.getResultSet();
                            counter++;
                        }

                    } else if( callType == DBCall.ORA_FUN_CALL )
                    {
                        //ResultSet rs = (ResultSet)cstmt.getObject(1);
                        //pickHash(rs, result, errors_set, warnings);
                        Object proc_result = null;

                        try { proc_result = cstmt.getObject(1); }
                        catch (Exception ex) {}

                        if (proc_result != null)
                        {
                            try
                            {
                                ResultSet rs = (ResultSet)cstmt.getObject(1);
                                pickHash(rs, result, errors_set, warnings);
                            }
                            // if the result isn't a ResultSet, save there's no processing to
                            // be done. Just save the value
                            catch (ClassCastException ex) {
                                result.put (NetDB.RETURN_VALUE, proc_result); }
                        }
                    }

                } catch(SQLException ex) {
                    ex.printStackTrace();
                    if( !sql_session_timeout(errors_set, ex.getErrorCode()) ) {
                        for(;;) {
                            logger.debug("ex.getMessage() = ["+ex.getMessage()+"]");
                            String msg = ex.getMessage().trim();
                            if( msg.startsWith("ORA") )
                                msg = msg.substring(msg.indexOf(":")+2, msg.indexOf("\n"));
                            errors_set.addElement(msg);
                            ex = ex.getNextException();
                            if( ex == null )
                                break;
                        }
                    }

                    fatalError = true;

                } catch(Exception ex) {
                    ex.printStackTrace();
                    logger.debug("ex.getMessage() = ["+ex.getMessage()+"]");
                    fatalError = true;
                    errors_set.addElement(ex.getMessage().trim());
                }

                if( warnings.size() > 0 )
                    result.put("warnings", warnings);

                if( !fatalError && values != null ) {
                    // now collect out params
                    for( int i = 0; i < values.length; i++ ) {
                        if ( values[i] instanceof OutParam ) {
                            switch( ((OutParam)values[i]).getType() ) {
                            case oracle.jdbc.driver.OracleTypes.NUMERIC:
                            case oracle.jdbc.driver.OracleTypes.TINYINT:
                                values[i] = new Integer(cstmt.getInt(i+1));
                                break;
                            case oracle.jdbc.driver.OracleTypes.VARCHAR:
                                values[i] = cstmt.getString(i+1);
                                break;
                            case oracle.jdbc.driver.OracleTypes.TIMESTAMP:
                                values[i] = cstmt.getTimestamp(i+1);
                                break;
                            case oracle.jdbc.driver.OracleTypes.BIT:
                                values[i] = new Boolean(cstmt.getBoolean(i+1));
                                break;
                            case oracle.jdbc.driver.OracleTypes.CURSOR:
                                break;
                            default:
                                throw new InternalException("NetDBConnection::callDB(...) "+
                                                            "unsupported sql type");
                            }
                        }
                    }
                }
                break;

            default:
                throw new Exception("illegal db call type");
            }

        } catch (SQLException ex) {
            if( !sql_session_timeout(errors_set, ex.getErrorCode()) ) {
                String msg = Utils.nStr(ex.getMessage()).trim();
                if( msg.matches("^ORA-\\d+:.*") ) {
                    msg = msg.replaceAll("^ORA-\\d+:.", "");
                }
                errors_set.addElement(msg);
            }

        } finally {
            // if there is a result set, and it was ignored, we'll have to close it.
            try { ((ResultSet)cstmt.getObject(1)).close(); } catch(Exception ex) {}

            // try to close res, stmt and cstmt
            try { res.close();   } catch(Exception ex) {}
            try { stmt.close();  } catch(Exception ex) {}
            try { pstmt.close(); } catch(Exception ex) {}
            try { cstmt.close(); } catch(Exception ex) {}
        }

        if( errors_set.size() > 0 )
            result.put("errors", errors_set);

        return result;
    } // end of callDB(...)

    // generic picker
    public void pickHash(ResultSet rs, Hashtable result,
                         Vector errors_set, Vector warnings )
        throws java.sql.SQLException
    {
        ResultSetMetaData rsmd = rs.getMetaData();
        int cols = rsmd.getColumnCount();

        int[] colw = new int[cols];
        String[] coln = new String[cols];
        int totalLen = 0;

        for( int j = 0; j < cols; j++ ) {
            coln[j] = rsmd.getColumnName(j+1).trim();
            colw[j] = coln[j].length();
        }

        Vector levels = new Vector();
        Hashtable branch = result; // root branch
        levels.addElement(branch);
        int lastLevel = 0;

        while (rs.next()) {
            String type = rs.getString("TYPE");
            if( type.equals("data") ) { // make up new hash
                String key = rs.getString("KEY");
                int hashLevel = rs.getInt("HASH_LEVEL");
                if( hashLevel > -1 ) {
                    branch = new Hashtable();
                    Hashtable parent = (Hashtable)levels.elementAt(hashLevel);
                    parent.put(key, branch);
                    if( levels.size() == hashLevel+1 )
                        levels.addElement(branch);
                    else
                        levels.setElementAt(branch, hashLevel+1);
                }

                // add extra columns to a current branch hash
                for( int j = 0; j < cols; j++ ) {
                    if( !coln[j].equals("TYPE")            &&
                        !coln[j].equals("KEY")             &&
                        !coln[j].equals("HASH_LEVEL")      &&
                        !coln[j].equals("VALUE_TYPE")      &&
                        !coln[j].equals("VALUE_STRING")    &&
                        !coln[j].equals("VALUE_TIMESTAMP") &&
                        !coln[j].equals("VALUE_INT") ) {
                        Object value = rs.getObject(j+1);
                        if( value != null ) {

                            if( value instanceof String )
                                value = ((String)value).trim();

                            Hashtable target = branch;
                            if( hashLevel == -1 )
                                target = result;

                            if( value instanceof oracle.sql.DATE )
                                target.put(coln[j].trim().toLowerCase(),
                                           ((oracle.sql.DATE)value).timestampValue() );
                            else if( value instanceof oracle.sql.TIMESTAMP )
                                target.put(coln[j].trim().toLowerCase(),
                                           ((oracle.sql.TIMESTAMP)value).timestampValue() );
                            else
                                target.put(coln[j].trim().toLowerCase(),
                                           value);
                        }
                    }
                }

            } else if( type.equals("key") ) { // fill in the key -> value pair in a current hash
                String key = rs.getString("KEY");
                String valueType = rs.getString("VALUE_TYPE");
                if(      valueType.equals("string") ) {
                    if( rs.getString("VALUE_STRING") != null )
                        branch.put(key, rs.getString("VALUE_STRING").trim());

                } else if( valueType.equals("integer") )
                        branch.put(key, new Integer(rs.getInt("VALUE_INT")));

                else if( valueType.equals("boolean") )
                    branch.put(key, new Boolean(rs.getInt("VALUE_INT")==0?false:true));

                else if( valueType.equals("timestamp") ) {
                    if( rs.getTimestamp("VALUE_TIMESTAMP") != null )
                        branch.put(key, rs.getTimestamp("VALUE_TIMESTAMP"));
                }

            } else if( type.equals("error") )
                errors_set.addElement(((String)Utils.n2s(rs.getString("VALUE_STRING"))).trim());

            else if( type.equals("warning") )
                warnings.addElement(((String)Utils.n2s(rs.getString("VALUE_STRING"))).trim());
        }

        // Now that we've sucked all the data out of the cursor, close it...
        rs.close();
    } // end of pickHash(...)

    private boolean sql_session_timeout(Vector errors_set, int code) {
        if( code != 2399 && code != 2393 )
            return false;

        // we are here if session timed out for some reason
        errors_set.
            addElement("Your search has exceeded currently available resource limits.  "+
                       "The results may be incomplete.  Try again later or refine "+
                       "your search.");

        return true;
    } // end of sql_session_timeout(...)
}
