/************************************************************************
 *
 * NodeInsertCommand.java - db call NetDB_Node.master 'insert'
 *
 * $Id: NodeInsertCommand.java,v 1.5 2008-04-24 05:32:08 jlidtke Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2002, Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.lib;

import stanford.netdb.utils.*;
import stanford.netdb.lib.*;
import java.util.*;
import java.lang.*;
import java.sql.*;
import su.netdb.parser.*;

/** complex db call NetDB_Node.master 'insert' */
class NodeInsertCommand extends DBCall implements ComplexDBCall {

    public static final Parser parser = new Parser();

    private Hashtable newNode;
    private NetDBii netdb;
    private Timestamp lock;
    private int op_uid;
    private Integer idNode;

    private Hashtable acceptedIPs = new Hashtable();
    private Hashtable acceptedNames = new Hashtable();

    private static final int OP_USER_INDEX               = 1;  // p_op_uid
    private static final int ORGANIZATION_INDEX          = 2;  // p_org
    private static final int LOCATION_NAME_INDEX         = 3;  // p_location
    private static final int ROOM_INDEX                  = 4;  // p_room
    private static final int STATE_INDEX                 = 5;  // p_state
    private static final int MAKE_INDEX                  = 6;  // p_make
    private static final int MODEL_INDEX                 = 7;  // p_model
    private static final int COMMENT_INDEX               = 8;  // p_comment
    private static final int EXPIRATION_INDEX            = 9;  // p_expire_date
    private static final int DHCP_SERVICE_INDEX          = 10; // p_dhcpservice
    private static final int AS_INDEX                    = 11; // p_ipadddressspace

    public NodeInsertCommand(NetDBii netdb, Hashtable newNode, Integer op_uid)
        throws Exception
    {
        super((stanford.netdb.utils.Logger)netdb);

        this.netdb = netdb;
        this.newNode = Utils.copyHash(newNode);
        this.op_uid  = op_uid;
        db_call_type = Function_Map_Controller.get_function_type (Function_Map.PACKAGE_NODE, Function_Map.INSERT);

        values = new Object[12];
        values[0]                                = ref_cursor;
        values[OP_USER_INDEX]                    = op_uid;
        values[ORGANIZATION_INDEX]               = ((String)this.newNode.get("dept")).trim();
        values[MODEL_INDEX]                      = ((String)this.newNode.get("model")).trim();

        if( Utils.nHash(newNode.get("types")).get("Template") != null &&
            newNode.get("addr_space") != null &&
            newNode.get("addr_space").toString().trim().length() > 0 )
            values[AS_INDEX] = ((String)newNode.get("addr_space")).trim();
        else
            values[AS_INDEX] = o2v((String)newNode.get("addr_space"));

        values[LOCATION_NAME_INDEX]              = ((String)this.newNode.get("location")).trim();
        values[ROOM_INDEX]                       = ((String)this.newNode.get("room")).trim();
        values[MAKE_INDEX]                       = ((String)this.newNode.get("make")).trim();
        values[STATE_INDEX]                      = ((String)this.newNode.get("state")).trim();

        Hashtable types                          = (Hashtable)this.newNode.get("types");
        values[EXPIRATION_INDEX]                 = o2n (this.newNode.get("exp.date"), Types.TIMESTAMP);

        if( this.newNode.get("comment") != null )
        {
            String comment = ((String)this.newNode.get("comment")).trim();
            if( comment.length() > 255 )
                comment = comment.substring(0, 254);
            values[COMMENT_INDEX] = comment;
        }
        else
            values[COMMENT_INDEX] = new Null (Types.VARCHAR);

        values[DHCP_SERVICE_INDEX] = new Null (Types.VARCHAR);
    }

    public String getSQL() {
        return proc_call_sql(Function_Map_Controller.get_function_name (Function_Map.PACKAGE_NODE, Function_Map.INSERT));
    }

    public Hashtable doIt(NetDBConnection con) throws NetDBException, Exception {
        boolean commit_finally = false;
        Hashtable res = new Hashtable();
        Vector warningsVector = new Vector();
        Vector final_ip_names_to_cancel = new Vector();
        try {
            logger.info("inserting node...");
            // insert person entries for users and admins
            Hashtable adminsH  = Utils.nHash(newNode.get("admins"));
            Hashtable ateams   = Utils.nHash(adminsH.get("ateams"));
            Hashtable admins   = Utils.nHash(adminsH.get("admins"));
            Hashtable users    = Utils.nHash(newNode.get("users"));

            // now start inserting node
            BeginTransaction(con.con);
            Hashtable result = processResultSet(con.callDB(this));
            checkErrors("Failed to insert node", result);

            // get node id
            idNode = new Integer((String)result.get("id"));
            lock = (Timestamp)result.get("locked");

            // insert node types
            Enumeration enm = Utils.nHash(newNode.get("types")).keys();
            while( enm.hasMoreElements() ) {
                DBCall call =
                    new NodeTypeCommand(logger, NodeTypeCommand.INSERT_ACTION, lock, idNode,
                                        (String)enm.nextElement(), op_uid);
                checkErrors("Failed to insert node type", call.processResultSet(con.callDB(call)));
            }

            // insert node groups
            Hashtable groups = Utils.nHash(newNode.get("groups"));
            Vector groups_to_insert = new Vector(groups.keySet());
            ComplexDBCall cmd =
                new ObjectGroupCommand(logger, ObjectGroupCommand.ADD_GROUPS, idNode.intValue(),
                                       ObjectGroupCommand.NODE, lock, groups_to_insert, op_uid);
            checkErrors("Failed to insert groups", cmd.doIt(con));

            // insert node oses
            enm = Utils.nHash(newNode.get("oses")).keys();
            while( enm.hasMoreElements() ) {
                DBCall call = new NodeOSCommand(logger, NodeOSCommand.INSERT_ACTION,
                                                idNode, lock, (String)enm.nextElement(), op_uid);
                checkErrors("Failed to insert os", con.callDB(call));
            }

            // insert node tags
            Hashtable tags = Utils.nHash(newNode.get("tags"));
            enm = tags.keys();
            while( enm.hasMoreElements() ) {
                String key = (String)enm.nextElement();
                if( key.trim().length() == 0 )
                {
                    NetDBException ex = new NetDBException("Failed to insert node tag");
                    ex.add ("Empty node tags are not allowed");
                    throw ex;
                }
                DBCall call = new NodeTagsCommand(logger, NodeTagsCommand.INSERT_ACTION,
                                                  lock, idNode, key, (String)tags.get(key), op_uid);
                checkErrors("Failed to insert node tag", con.callDB(call));
            }

            // insert ateams
            enm = ateams.keys();
            while( enm.hasMoreElements() ) {
                String ateamStr = ((String)enm.nextElement()).trim();
                DBCall call = new NodeAdminCommand(logger, NodeAdminCommand.INSERT_ACTION, lock, idNode,
                                                   ateamStr, null, op_uid);
                logger.debug("Inserting ateam ["+ateamStr+"]");
                checkErrors("Failed to insert admin team", con.callDB(call));
            }

            // insert admins, i.e. persons
            enm = admins.keys();
            while( enm.hasMoreElements() ) {
                Hashtable admin = (Hashtable)admins.get(enm.nextElement());
                DBCall call = new NodeAdminCommand(logger, NodeAdminCommand.INSERT_ACTION, lock, idNode,
                                                   null, new Integer((String)admin.get("id")), op_uid);
                logger.debug("Inserting admin ["+admin.get("id")+" : "+admin.get("NetID")+"]");
                checkErrors("Failed to insert admin", con.callDB(call));
            }

            // insert users
            enm = users.keys();
            while( enm.hasMoreElements() ) {
                Hashtable user = (Hashtable)users.get(enm.nextElement());
                DBCall call = new NodeUserCommand(logger,NodeAdminCommand.INSERT_ACTION, lock, idNode,
                                                  new Integer((String)user.get("id")), op_uid);
                logger.debug("Inserting user ["+user.get("id")+" : "+user.get("NetID")+"]");
                checkErrors("Failed to insert user", con.callDB(call));
            }

            // insert interfaces
            Hashtable interfaces = Utils.nHash(newNode.get("interfaces"));
            Object[] intfs_ids = interfaces.keySet().toArray();
            for( int i = 0; i < intfs_ids.length; i++ ) {
                String intfcID = (String)intfs_ids[i];
                Hashtable intfcH = (Hashtable)interfaces.get(intfcID);
                String hw = (String)intfcH.get("hw");
                if( hw != null && hw.trim().length() == 0 )
                    hw = null;
                if( hw != null ) {
                    Hashtable parseRes = parser.parse(hw);
                    String hwStr = null;
                    try {
                        hwStr = (String)parseRes.get("hw");
                    } catch (Exception exx) { }

                    if( hwStr == null )
                        throw new Exception("Invalid hardware address format");
                }

                DBCall call = new InterfaceInsertCommand(logger, lock, idNode,
                                                         (Boolean)intfcH.get("use-dhcp?"),
                                                         (Boolean)intfcH.get("roaming?"),
                                                         hw, NetDB.ETHERNET_TYPE,
                                                         ((String)Utils.n2s(intfcH.get("comment"))).trim(),
                                                         op_uid);
                logger.debug("Inserting interface #"+intfcID);
                result = con.callDB(call);
                checkErrors("Failed to insert interface #"+intfcID, result);
                String newIntfcID = (String)result.get("id");
                logger.debug("intfc #"+intfcID+" => "+newIntfcID);

                // reset interface id
                interfaces.put(newIntfcID, interfaces.remove(intfcID));

                // insert dhcp options
                Hashtable dhcp_options = Utils.nHash(intfcH.get("dhcp-options"));
                if( dhcp_options.size() > 0 ) {
                    cmd = new DHCPOptionsActionCommand(logger, DHCPOptionsActionCommand.INSERT_ACTION,
                                  lock, new Integer(newIntfcID), Constants.INTERFACE, dhcp_options, op_uid);
                    checkErrors("Failed to insert dhcp options for interface #"+intfcID, cmd.doIt(con));
                }
            }

            // now ip addresses
            Hashtable ip_parents = new Hashtable(); // used for resetting ip addresses
            int next_parent_key = 0;

            Hashtable ips_to_insert = new Hashtable();

            Hashtable newIPs = Utils.nHash(newNode.get("ipc_addresses"));
            enm = newIPs.keys();
            while( enm.hasMoreElements() ) {
                String key = (String)enm.nextElement();
                String ipStr = key.trim();
                Hashtable parseRes = parser.parse(ipStr);
                if( parseRes.get("ip_low") == null )
                    throw new Exception("Invalid ip address format: ["+ipStr+"]");
                if( parseRes.get("ip_high") == null )
                    throw new Exception("Invalid ip address format: ["+ipStr+"]");
                Hashtable record = new Hashtable();
                record.put("key", key);
                Boolean active = (Boolean)((Hashtable)newIPs.get(key)).get("active?");
                if( active == null )
                    active = new Boolean(false);
                record.put("active?", active);

                record.put("parent key", "key "+(next_parent_key++));
                record.put("ip key", key);
                ip_parents.put(record.get("parent key"), newIPs);

                ips_to_insert.put("ipc "+ipStr, record);

            }

            // now ips from interfaces

            // first ips from interfaces in a newNode
            interfaces = Utils.nHash(newNode.get("interfaces"));
            enm = interfaces.keys();
            while( enm.hasMoreElements() ) {
                String intfcKey = (String)enm.nextElement();
                Hashtable intfc = (Hashtable)interfaces.get(intfcKey);
                Hashtable ips = Utils.nHash(intfc.get("ips"));
                Enumeration enm2 = ips.keys();
                while( enm2.hasMoreElements() ) {
                    String key = (String)enm2.nextElement();
                    String ipStr = key.trim();
                    Hashtable parseRes = parser.parse(ipStr);
                    if( parseRes.get("ip_low") == null )
                        throw new Exception("Invalid ip address format: ["+ipStr+"]");
                    if( parseRes.get("ip_high") == null )
                        throw new Exception("Invalid ip address format: ["+ipStr+"]");
                    Hashtable record = new Hashtable();
                    record.put("key", key);
                    record.put("intfcKey", intfcKey);
                    Boolean active = (Boolean)((Hashtable)ips.get(key)).get("active?");
                    if( active == null )
                        active = new Boolean(false);
                    record.put("active?", active);

                    record.put("parent key", "key "+(next_parent_key++));
                    record.put("ip key", key);
                    ip_parents.put(record.get("parent key"), ips);

                    ips_to_insert.put("intfc "+intfcKey+" "+ipStr, record);
                }
            }

            // now insert ips_to_insert
            enm = ips_to_insert.keys();
            while( enm.hasMoreElements() ) {
                Hashtable ip = (Hashtable)ips_to_insert.get(enm.nextElement());
                String ipStr = ((String)ip.get("key")).trim();
                String intfcID = (String)ip.get("intfcKey");
                logger.debug("reserving ip "+ipStr);
                String origIpStr = ipStr;
                Hashtable reserveResult = netdb.reserveIP(con, null, ipStr, 1);
                result = (Hashtable)reserveResult.get(reserveResult.keys().nextElement());
                if( result.get("errors") != null ) {
                    Vector msgs = new Vector();
                    msgs.addElement("Failed to reserve ip address "+ipStr);
                    throw new NetDBException("Failed to reserve ip address", msgs);
                }

                Hashtable new_warnings = Utils.nHash(result.get("warnings"));
                if( new_warnings.size() > 0 ) {
                    Enumeration enm1 = new_warnings.keys();
                    while( enm1.hasMoreElements() )
                        warningsVector.addElement(((String)enm1.nextElement()).trim());
                }

                Hashtable ip_addresses = Utils.nHash(result.get("ip_addresses"));
                if( ip_addresses.size() > 0 ) {
                    ipStr = ((String)ip_addresses.keys().nextElement()).trim();
                    DBCall call;
                    if( intfcID != null ) {
                        call = new IPAddressAcceptCommand(logger, ipStr, (Boolean)ip.get("active?"),
                                                          new Integer(intfcID), lock,
                                                          IPAddressAcceptCommand.INTERFACE_TYPE,
                                                          (String)ip.get("PTRpref"), op_uid);
                        logger.debug("accepting ip in intfc "+intfcID+":"+ipStr);
                    } else {
                        call = new IPAddressAcceptCommand(logger, ipStr, (Boolean)ip.get("active?"),
                                                          idNode, lock,
                                                          IPAddressAcceptCommand.IPC_TYPE,
                                                          (String)ip.get("PTRpref"), op_uid);
                        logger.debug("accepting ip in ipc node "+idNode+":"+ipStr);
                    }
                    checkErrors("Failed to insert ip address #"+ipStr, con.callDB(call));
                    acceptedIPs.put(ipStr, "");

                    // reset ip address
                    Hashtable ip_parent = Utils.nHash(ip_parents.get(Utils.n2s(ip.get("parent key"))));
                    Hashtable ipH = Utils.nHash(ip_parent.remove(ip.get("ip key")));
                    ipH.put("ip", ipStr);
                    ip_parent.put(ipStr, ipH);
                    if( !ipStr.equals(origIpStr) ) {
                        // we have to put new name
                        Hashtable names = Utils.nHash(ipH.get("names"));
                        if( names.size() == 1 ) { // do it only if there is one name
                            String nameKey = (String)names.keys().nextElement();
                            Hashtable nameH = (Hashtable)names.remove(nameKey);
                            String oldNameFqdn = ((String)nameH.get("name")).trim()+"."+((String)nameH.get("domain")).trim();

                            // check if name is auto-generated and then generate new one
                            Hashtable generated_name = netdb.generateHexName(origIpStr, 0);
                            if( (((String)generated_name.get("name")).trim()+"."+((String)generated_name.get("domain")).trim()).
                                toLowerCase().equals(oldNameFqdn.toLowerCase()) ) {

                                // record old name for cancellation
                                final_ip_names_to_cancel.addElement(oldNameFqdn);

                                // now generate new name and place it there
                                generated_name = netdb.generateHexName(ipStr, 0);
                                names.put(generated_name.get("name")+"."+generated_name.get("domain"), generated_name);
                            }
                        }
                    }
                }
            }

            // now names
            Hashtable sortResult = Utils.sortOutNames(null, newNode);
            Hashtable names_to_insert = Utils.nHash(sortResult.get("names_to_insert"));

            // insert names_to_insert
            boolean names = true;
            for( int i = 0; i < 2; i++ ) {
                // it should be done it two passes, first names and then (on second pass) aliases and mxes
                enm = names_to_insert.keys();
                while( enm.hasMoreElements() ) {
                    String key = (String)enm.nextElement();
                    // first go over parenting names,
                    // i.e. names which may have aliases and/or mx's
                    Hashtable name = (Hashtable)names_to_insert.get(key);
                    if( (names && name.get("refs") == null) ||
                        (!names && name.get("refs") != null) ) {
                        continue;
                    }

                    String nameType = (String)name.get("type");
                    if( nameType.equals("node_name") )
                        nameType = Constants.NODE_NAME_TYPE;
                    else if( nameType.equals("alias") )
                        nameType = Constants.NAME_ALIAS_TYPE;
                    else if( nameType.equals("mx") )
                        nameType = Constants.MX_TYPE;
                    else if( nameType.equals("ip_name") || nameType.equals("ipc_name") )
                        nameType = Constants.ADDRESS_NAME_TYPE;
                    else if( nameType.equals("intfc_name") )
                        nameType = Constants.INTERFACE_NAME_TYPE;

                    String fqdnName = name.get("name")+"."+name.get("domain");
                    Hashtable reservedName = netdb.reserveName(con, fqdnName, nameType);
                    name.put("name", Utils.name((String)reservedName.get("name")));
                    name.put("domain", Utils.domain((String)reservedName.get("name"),
                                                    (String)netdb.getUser().get("default_domain")));
                    DBCall call = null;
                    if( nameType.equals(Constants.NODE_NAME_TYPE) )
                        call = new NameAcceptCommand(logger, new Integer((String)reservedName.get("id")), null,
                                                     nameType, idNode, null, lock, null, op_uid);

                    else if( nameType.equals(Constants.NAME_ALIAS_TYPE) )
                        call = new NameAcceptCommand(logger, new Integer((String)reservedName.get("id")), null,
                                                      nameType, null, (String)name.get("parent"),
                                                      lock, null, op_uid);
                    else if( nameType.equals(Constants.MX_TYPE) )
                        // Note that the fact that reservedName.get("pref") is a String representation of integer
                        // unsured in NetDB.update(...) by calling method Utils.validate(...)
                        call = new NameAcceptCommand(logger, new Integer((String)reservedName.get("id")), null,
                                                     nameType, null, (String)name.get("parent"), lock,
                                                     new Integer((String)name.get("pref")), op_uid);

                    else if( nameType.equals(Constants.ADDRESS_NAME_TYPE) )
                        call = new NameAcceptCommand(logger, new Integer((String)reservedName.get("id")), null, nameType,
                                                     null, (String)name.get("ip"), lock, null, op_uid);

                    else if( nameType.equals(Constants.INTERFACE_NAME_TYPE) )
                        call = new NameAcceptCommand(logger, new Integer((String)reservedName.get("id")), null, nameType,
                                                     new Integer((String)name.get("intfs_id")), null, lock,
                                                     null, op_uid);

                    else
                        throw new Exception("Invalid nameType");

                    if( call != null )
                        checkErrors("Failed to accept name ["+fqdnName+"]", call.processResultSet(con.callDB(call)));

                    Hashtable nameH = new Hashtable();
                    nameH.put("fqdn", ((String)name.get("name")).trim().toLowerCase()+"."+((String)name.get("domain")).trim().toLowerCase());
                    nameH.put("type", new String(nameType));
                    acceptedNames.put(nameH, "");
                }
                names = false; // second pass goes over aliases and mxes
            }

            // verify whole stuff
            logger.debug("do check");
            DBCall verify = new VerifyCommand(logger, VerifyCommand.NODE, idNode, op_uid);
            result = con.callDB(verify);

            // check for errors
            Vector error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            commit_finally = true;

            if( warningsVector.size() > 0 )
                res.put("warnings", warningsVector);

            for( int i = 0; i < final_ip_names_to_cancel.size(); i++ ) {
                try {
                    ((NetDBii)logger).
                        cancelName(con,
                                   (String)final_ip_names_to_cancel.elementAt(i),
                                   Constants.ADDRESS_NAME_TYPE);
                } catch (Exception ex) {
                    logger.debug("Failed to cancel name ["+final_ip_names_to_cancel.elementAt(i)+"]");
                }
            }

            return res;

        } finally {
            if( commit_finally ) {
                try {
                    logger.info("modifying node id "+idNode+" done");
                    // Unlock the node now that we're done.
                    try {
                        DBCall call =
                            new NodeUnlockCommand(logger, idNode.intValue(), lock, op_uid);
                        call.processResultSet(con.callDB(call));
                    } catch (Exception exxx) { }
                    logger.debug("CommitTransaction");
                    CommitTransaction(con.con);
                } catch(Exception exx) { }
                // remove accepted ips from list of reserved
                Enumeration enm = acceptedIPs.keys();
                while( enm.hasMoreElements() )
                    try {
                        netdb.unregister_reserved_ip(((String)enm.nextElement()).trim());
                    } catch (Exception ex) {}
                // remove accepted names from list of reserved
                //Enumeration enm = acceptedNames.keys();
                //while( enm.hasMoreElements() )
                //    netdb.removeReservedName((Hashtable)enm.nextElement());
            } else
                try {
                    logger.info("modifying node id "+idNode+" failed");
                    logger.debug("RollbackTransaction");
                    RollbackTransaction(con.con);
                } catch(Exception exx) { }
        }
    } // end of doIt()
}
