/************************************************************************
 *
 * NodeModifyCommand.java -
 *
 * $Id: NodeModifyCommand.java,v 1.6 2008-04-24 05:32:08 jlidtke Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2002, Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.lib;

import stanford.netdb.utils.*;
import stanford.netdb.lib.*;
import java.util.*;
import java.lang.*;
import java.sql.*;
import su.netdb.parser.*;

/** complex db call pNetwork 'update' */
class NodeModifyCommand extends DBCall implements ComplexDBCall {

    public static final Parser parser = new Parser();

    private Hashtable oldNode;
    private Hashtable newNode;
    private NetDBii   netdb;
    private Integer   idNode;
    private Timestamp lock;
    private int       op_uid;

    public  Hashtable acceptedIPs   = new Hashtable();
    public  Hashtable acceptedNames = new Hashtable();

    public boolean commit_finally   = false;

    private static final int OP_USER_INDEX       = 0;  // p_op_uid
    private static final int LOCKED_INDEX        = 1;  // p_lock
    private static final int ID_NODE_INDEX       = 2;  // p_node_id
    private static final int ORGANIZATION_INDEX  = 3;  // p_org
    private static final int LOCATION_NAME_INDEX = 4;  // p_location
    private static final int ROOM_INDEX          = 5;  // p_room
    private static final int STATE_INDEX         = 6;  // p_state
    private static final int MAKE_INDEX          = 7;  // p_make
    private static final int MODEL_INDEX         = 8;  // p_model
    private static final int COMMENT_INDEX       = 9;  // p_comment
    private static final int EXPIRATION_INDEX    = 10; // p_expire_date
    private static final int DHCP_SERVICE_INDEX  = 11; // p_dhcpservice
    private static final int AS_INDEX            = 12; // p_ipadddressspace

    public NodeModifyCommand(NetDBii netdb, Hashtable oldNode, Hashtable newNode, Integer op_uid)
        throws Exception
    {
        super((stanford.netdb.utils.Logger)netdb);

        db_call_type = Function_Map_Controller.get_function_type (Function_Map.PACKAGE_NODE, Function_Map.UPDATE);

        this.netdb   = netdb;
        this.oldNode = Utils.copyHash(oldNode);
        this.newNode = Utils.copyHash(newNode);
        this.op_uid  = op_uid;
        lock         = (Timestamp)this.newNode.get("locked");

        values = new Object[13];
        values[OP_USER_INDEX]          = op_uid;
        values[ID_NODE_INDEX] = idNode = new Integer((String)this.newNode.get("id"));
        values[ORGANIZATION_INDEX]     = ((String)this.newNode.get("dept")).trim();
        values[MODEL_INDEX]            = ((String)this.newNode.get("model")).trim();
        if( Utils.nHash(newNode.get("types")).get("Template") != null )
            values[AS_INDEX] = (((newNode.get("addr_space")==null)||(newNode.get("addr_space").toString().trim().length()==0))?"":((String)newNode.get("addr_space")).trim());
        else
            values[AS_INDEX] = o2v((String)newNode.get("addr_space"));
        values[LOCATION_NAME_INDEX]    = ((String)this.newNode.get("location")).trim();
        values[ROOM_INDEX]             = ((String)this.newNode.get("room")).trim();
        values[STATE_INDEX]            = ((String)this.newNode.get("state")).trim();
        values[MAKE_INDEX]             = ((String)this.newNode.get("make")).trim();
        values[LOCKED_INDEX]           = lock;

        Hashtable types                = (Hashtable)this.newNode.get("types");
        values[EXPIRATION_INDEX]       = o2n (this.newNode.get("exp.date"), Types.TIMESTAMP);

        if( this.newNode.get("comment") != null )
        {
            String comment = ((String)this.newNode.get("comment")).trim();
            if( comment.length() > 255 )
                comment = comment.substring(0, 254);
            values[COMMENT_INDEX] = comment;
        }
        else
            values[COMMENT_INDEX] = new Null (Types.VARCHAR);

        values[DHCP_SERVICE_INDEX] = new Null (Types.VARCHAR);
    }

    public String getSQL() {
        return proc_call_sql(Function_Map_Controller.get_function_name (Function_Map.PACKAGE_NODE, Function_Map.UPDATE));
    }

    public Hashtable doIt(NetDBConnection con) throws NetDBException, Exception {

        // transactions are handled outside of this call if parent class is Node2ModifyCommand.
        boolean handle_transactions_locally = true;
        try {
            StackTraceElement[] st = (new Throwable()).getStackTrace();
            if( st[1].getClassName().equals("stanford.netdb.lib.Node2ModifyCommand") )
                handle_transactions_locally = false;
        } catch (Exception ex) { }


        Hashtable res = new Hashtable();
        Vector warningsVector = new Vector();
        Vector final_ip_names_to_cancel = new Vector();
        try {
            logger.info("modifying node id "+idNode);
            Hashtable newAdminsH = Utils.nHash(newNode.get("admins"));
            Hashtable oldAdminsH = Utils.nHash(oldNode.get("admins"));

            Hashtable newATeams = Utils.nHash(newAdminsH.get("ateams"));
            Hashtable oldATeams = Utils.nHash(oldAdminsH.get("ateams"));
            Hashtable newAdmins = Utils.nHash(newAdminsH.get("admins"));
            Hashtable oldAdmins = Utils.nHash(oldAdminsH.get("admins"));
            Hashtable newUsers = Utils.nHash(newNode.get("users"));
            Hashtable oldUsers = Utils.nHash(oldNode.get("users"));

            // now we are updating node
            if( handle_transactions_locally ) BeginTransaction(con.con);
            checkErrors("Failed to update node", processResultSet(con.callDB(this)));

            // update node types
            Hashtable oldTypes = Utils.nHash(oldNode.get("types"));
            Hashtable newTypes = Utils.nHash(newNode.get("types"));
            Vector types_to_insert = new Vector();
            Vector types_to_delete = new Vector();
            Vector types_to_keep = new Vector();
            Utils.keysIntersect(oldTypes, newTypes, types_to_insert, types_to_delete, types_to_keep);
            if( types_to_delete.size() > 0 )
                logger.debug("deleting node types to delete");
            for( int i = 0; i < types_to_delete.size(); i++ ) {
                DBCall call =
                    new NodeTypeCommand(logger, NodeTypeCommand.DELETE_ACTION, lock, idNode,
                                        (String)types_to_delete.elementAt(i), op_uid);
                checkErrors("Failed to delete node type", call.processResultSet(con.callDB(call)));
            }

            if( types_to_insert.size() > 0 )
                logger.debug("inserting node types to insert");
            for( int i = 0; i < types_to_insert.size(); i++ ) {
                DBCall call =
                    new NodeTypeCommand(logger, NodeTypeCommand.INSERT_ACTION, lock, idNode,
                                        (String)types_to_insert.elementAt(i), op_uid);
                checkErrors("Failed to insert node type", call.processResultSet(con.callDB(call)));
            }

            // update groups
            Hashtable oldGroups = Utils.nHash(oldNode.get("groups"));
            Hashtable newGroups = Utils.nHash(newNode.get("groups"));
            Vector groups_to_insert = new Vector();
            Vector groups_to_delete = new Vector();
            Vector groups_to_keep = new Vector();
            Utils.keysIntersect(oldGroups, newGroups, groups_to_insert, groups_to_delete, groups_to_keep);

            // insert groups_to_insert
            if( groups_to_insert.size() > 0 )
                logger.debug("inserting groups_to_insert");
            ComplexDBCall cmd =
                new ObjectGroupCommand(logger,
                                       ObjectGroupCommand.ADD_GROUPS, idNode.intValue(),
                                       ObjectGroupCommand.NODE, lock, groups_to_insert, op_uid);
            checkErrors("Failed to insert groups", cmd.doIt(con));

            if( groups_to_delete.size() > 0 )
                logger.debug("deleting groups_to_delete");
            cmd = new ObjectGroupCommand(logger,
                                         ObjectGroupCommand.REMOVE_GROUPS, idNode.intValue(),
                                         ObjectGroupCommand.NODE, lock, groups_to_delete, op_uid);
            checkErrors("Failed to delete groups", cmd.doIt(con));

            // modify node oses
            Hashtable oldOSes = Utils.nHash(oldNode.get("oses"));
            Hashtable newOSes = Utils.nHash(newNode.get("oses"));
            Vector oses_to_insert = new Vector();
            Vector oses_to_delete = new Vector();
            Vector oses_to_keep   = new Vector();
            Utils.keysIntersect(oldOSes, newOSes, oses_to_insert, oses_to_delete, oses_to_keep);

            for( int i = 0; i < oses_to_insert.size(); i++ ) {
                DBCall call = new NodeOSCommand(logger, NodeOSCommand.INSERT_ACTION,
                                                idNode, lock, (String)oses_to_insert.elementAt(i), op_uid);
                checkErrors("Failed to insert os", con.callDB(call));
            }

            for( int i = 0; i < oses_to_delete.size(); i++ ) {
                DBCall call = new NodeOSCommand(logger, NodeOSCommand.DELETE_ACTION,
                                                idNode, lock, (String)oses_to_delete.elementAt(i), op_uid);
                checkErrors("Failed to delete os", con.callDB(call));
            }

            // modify node tags
            Hashtable oldTags = Utils.nHash(oldNode.get("tags"));
            Hashtable newTags = Utils.nHash(newNode.get("tags"));
            Hashtable oldTagsR = new Hashtable();
            Hashtable newTagsR = new Hashtable();
            Enumeration enm = oldTags.keys();
            while( enm.hasMoreElements() ) {
                String key = (String)enm.nextElement();
                String value = (String)oldTags.get(key);
                Hashtable tag = new Hashtable();
                tag.put("key", key);
                tag.put("value", value);
                oldTagsR.put(key+value, tag);
            }
            enm = newTags.keys();
            while( enm.hasMoreElements() ) {
                String key = (String)enm.nextElement();
                String value = (String)newTags.get(key);
                Hashtable tag = new Hashtable();
                tag.put("key", key);
                tag.put("value", value);
                newTagsR.put(key+value, tag);
            }
            Vector tags_to_insert = new Vector();
            Vector tags_to_delete = new Vector();
            Vector tags_to_keep   = new Vector();
            Utils.keysIntersect(oldTagsR, newTagsR, tags_to_insert, tags_to_delete, tags_to_keep);
            for( int i = 0; i < tags_to_delete.size(); i++ ) {
                Hashtable tag = (Hashtable)oldTagsR.get(tags_to_delete.elementAt(i));
                DBCall call = new NodeTagsCommand(logger, NodeTagsCommand.DELETE_ACTION,
                                                  lock, idNode, (String)tag.get("key"), null, op_uid);
                con.callDB(call);
            }

            for( int i = 0; i < tags_to_insert.size(); i++ ) {
                Hashtable tag = (Hashtable)newTagsR.get(tags_to_insert.elementAt(i));
                String key = (String)tag.get("key");
                if( key.trim().length() == 0 )
                    throwSimpleNetDBException("Failed to insert node tag", "Empty node tags are not allowed");
                DBCall call = new NodeTagsCommand(logger, NodeTagsCommand.INSERT_ACTION,
                                                  lock, idNode, key, (String)tag.get("value"), op_uid);
                checkErrors("Failed to insert node tag", con.callDB(call));
            }

            // update ateams
            // first calculate intersection
            Vector ateams_to_insert = new Vector();
            Vector ateams_to_delete = new Vector();
            Vector ateams_to_keep   = new Vector();
            Utils.keysIntersect(oldATeams, newATeams, ateams_to_insert, ateams_to_delete, ateams_to_keep);

            // insert ateams_to_insert
            for( int i = 0; i < ateams_to_insert.size(); i++ ) {
                DBCall call =
                    new NodeAdminCommand(logger, NodeAdminCommand.INSERT_ACTION, lock, idNode,
                                         ((String)ateams_to_insert.elementAt(i)).trim(), null, op_uid);
                logger.debug("Inserting ateam ["+((String)ateams_to_insert.elementAt(i)).trim()+"]");
                checkErrors("Failed to insert admin team", con.callDB(call));
            }

            // update admins, i.e. persons
            // re-shuffle it by reg id
            Hashtable oldAdminsR = new Hashtable();
            Hashtable newAdminsR = new Hashtable();
            enm = oldAdmins.keys();
            while( enm.hasMoreElements() )
                oldAdminsR.put(((Hashtable)oldAdmins.get(enm.nextElement())).get("id"), "");

            enm = newAdmins.keys();
            while( enm.hasMoreElements() )
                newAdminsR.put(((Hashtable)newAdmins.get(enm.nextElement())).get("id"), "");

            // first calculate intersection
            Vector admins_to_insert = new Vector();
            Vector admins_to_delete = new Vector();
            Vector admins_to_keep   = new Vector();
            Utils.keysIntersect(oldAdminsR, newAdminsR, admins_to_insert, admins_to_delete, admins_to_keep);

            // insert admins_to_insert
            for( int i = 0; i < admins_to_insert.size(); i++ ) {
                DBCall call =
                    new NodeAdminCommand(logger, NodeAdminCommand.INSERT_ACTION, lock, idNode, null,
                                         new Integer((String)admins_to_insert.elementAt(i)), op_uid);
                logger.debug("Inserting admin ["+((String)admins_to_insert.elementAt(i)).trim()+"]");
                checkErrors("Failed to insert admin", con.callDB(call));
            }

            // delete admins_to_delete
            for( int i = 0; i < admins_to_delete.size(); i++ ) {
                DBCall call =
                    new NodeAdminCommand(logger, NodeAdminCommand.DELETE_ACTION, lock, idNode, null,
                                         new Integer((String)admins_to_delete.elementAt(i)), op_uid);
                logger.debug("Deleting admin ["+((String)admins_to_delete.elementAt(i)).trim()+"]");
                Hashtable result = con.callDB(call);
                checkErrors("Failed to delete admin", result);
            }

            // delete ateams to delete
            for( int i = 0; i < ateams_to_delete.size(); i++ ) {
                DBCall call =
                    new NodeAdminCommand(logger, NodeAdminCommand.DELETE_ACTION, lock, idNode,
                                         ((String)ateams_to_delete.elementAt(i)).trim(), null, op_uid);
                logger.debug("Deleting ateam ["+((String)ateams_to_delete.elementAt(i)).trim()+"]");
                Hashtable result = con.callDB(call);
                checkErrors("Failed to delete admin team", result);
            }

            // update users, i.e. persons
            // re-shuffle it by reg id
            Hashtable oldUsersR = new Hashtable();
            Hashtable newUsersR = new Hashtable();
            enm = oldUsers.keys();
            while( enm.hasMoreElements() )
                oldUsersR.put(((Hashtable)oldUsers.get(enm.nextElement())).get("id"), "");

            enm = newUsers.keys();
            while( enm.hasMoreElements() )
                newUsersR.put(((Hashtable)newUsers.get(enm.nextElement())).get("id"), "");

            // first calculate intersection
            Vector users_to_insert = new Vector();
            Vector users_to_delete = new Vector();
            Vector users_to_keep   = new Vector();
            Utils.keysIntersect(oldUsersR, newUsersR, users_to_insert, users_to_delete, users_to_keep);

            // delete users_to_delete
            for( int i = 0; i < users_to_delete.size(); i++ ) {
                DBCall call =
                    new NodeUserCommand(logger, NodeAdminCommand.DELETE_ACTION, lock, idNode,
                                        new Integer((String)users_to_delete.elementAt(i)), op_uid);
                logger.debug("Deleting user ["+((String)users_to_delete.elementAt(i)).trim()+"]");
                Hashtable result = con.callDB(call);
                checkErrors("Failed to delete user", result);
            }

            // insert users_to_insert
            for( int i = 0; i < users_to_insert.size(); i++ ) {
                DBCall call =
                    new NodeUserCommand(logger, NodeAdminCommand.INSERT_ACTION, lock, idNode,
                                        new Integer((String)users_to_insert.elementAt(i)), op_uid);
                logger.debug("Inserting user ["+((String)users_to_insert.elementAt(i)).trim()+"]");
                checkErrors("Failed to insert user", con.callDB(call));
            }

            // interfaces
            Vector interfaces_to_insert = new Vector();
            Vector interfaces_to_delete = new Vector();
            Vector interfaces_to_update   = new Vector();
            Hashtable oldInterfaces = Utils.nHash(oldNode.get("interfaces"));
            Hashtable newInterfaces = Utils.nHash(newNode.get("interfaces"));
            intfcsIntersect(oldInterfaces, newInterfaces,
                            interfaces_to_insert, interfaces_to_delete, interfaces_to_update);

            // delete interfaces_to_delete
            for( int i = 0; i < interfaces_to_delete.size(); i++ ) {
                String intfcID = (String)interfaces_to_delete.elementAt(i);
                DBCall call =
                    new InterfaceDeleteCommand(logger, lock,
                                               new Integer(intfcID), op_uid);
                logger.debug("Deleting interface #"+intfcID);
                checkErrors("Failed to delete interface #"+intfcID, con.callDB(call));

                // we have to remove interface from the hash becouse names and ip addresses
                // associated with this interface automatically removed when interfaces is deleted
                // and we don't want to cancel previously removed names and ips.
                Utils.nHash(oldNode.get("interfaces")).remove(intfcID);
            }

            // insert interfaces_to_insert
            for( int i = 0; i < interfaces_to_insert.size(); i++ ) {
                String intfcID = (String)interfaces_to_insert.elementAt(i);
                Hashtable intfcH = (Hashtable)newInterfaces.get(intfcID);
                String hw = (String)intfcH.get("hw");
                if( hw != null && hw.trim().length() == 0 )
                    hw = null;
                if( hw != null ) {
                    Hashtable parseRes = parser.parse(hw);
                    String hwStr = null;
                    try {
                        hwStr = (String)parseRes.get("hw");
                    } catch (Exception exx) { }

                    if( hwStr == null )
                        throw new Exception("Invalid hardware address format");
                }

                DBCall call =
                    new InterfaceInsertCommand(logger, lock, idNode,
                                               (Boolean)intfcH.get("use-dhcp?"),
                                               (Boolean)intfcH.get("roaming?"),
                                               hw, NetDB.ETHERNET_TYPE,
                                               ((String)Utils.n2s(intfcH.get("comment"))).trim(),
                                               op_uid);
                logger.debug("Inserting interface #"+intfcID);
                Hashtable result = con.callDB(call);
                checkErrors("Failed to insert interface #"+intfcID, result);
                String newIntfcID = (String)result.get("id");
                logger.debug("intfc #"+intfcID+" => "+newIntfcID);

                // reset interface id
                newInterfaces.put(newIntfcID, newInterfaces.remove(intfcID));

                // insert dhcp options
                Hashtable dhcp_options = Utils.nHash(intfcH.get("dhcp-options"));
                if( dhcp_options.size() > 0 ) {
                    cmd = new DHCPOptionsActionCommand(logger, DHCPOptionsActionCommand.INSERT_ACTION,
                                  lock, new Integer(newIntfcID), Constants.INTERFACE, dhcp_options, op_uid);
                    checkErrors("Failed to insert dhcp options for interface #"+newIntfcID, cmd.doIt(con));
                }
            }

            // update interfaces_to_update
            for( int i = 0; i < interfaces_to_update.size(); i++ ) {
                String intfcID = (String)interfaces_to_update.elementAt(i);
                Hashtable intfcH = (Hashtable)newInterfaces.get(intfcID);

                String hw = (String)intfcH.get("hw");
                if( hw != null && hw.trim().length() == 0 )
                    hw = null;
                if( hw != null ) {
                    Hashtable parseRes = parser.parse(hw);
                    String hwStr = null;
                    try {
                        hwStr = (String)parseRes.get("hw");
                    } catch (Exception exx) { }

                    if( hwStr == null )
                        throw new Exception("Invalid hardware address format");
                }

                DBCall call =
                    new InterfaceModifyCommand(logger, lock, new Integer(intfcID),
                                               (Boolean)intfcH.get("use-dhcp?"),
                                               (Boolean)intfcH.get("roaming?"),
                                               hw, NetDB.ETHERNET_TYPE,
                                               ((String)Utils.n2s(intfcH.get("comment"))).trim(),
                                               op_uid);
                logger.debug("Modifying interface #"+intfcID);
                checkErrors("Failed to modify interface #"+intfcID, con.callDB(call));

                // delete old dhcp options
                Hashtable dhcp_options = Utils.nHash(((Hashtable)oldInterfaces.get(intfcID)).get("dhcp-options"));
                if( dhcp_options.size() > 0 ) {
                    cmd = new DHCPOptionsActionCommand(logger, DHCPOptionsActionCommand.DELETE_ACTION,
                                  lock, new Integer(intfcID), Constants.INTERFACE, dhcp_options, op_uid);
                    checkErrors("Failed to delete dhcp options for interface #"+intfcID, cmd.doIt(con));
                }

                // insert new dhcp options
                dhcp_options = Utils.nHash(intfcH.get("dhcp-options"));
                if( dhcp_options.size() > 0 ) {
                    cmd = new DHCPOptionsActionCommand(logger, DHCPOptionsActionCommand.INSERT_ACTION,
                                  lock, new Integer(intfcID), Constants.INTERFACE, dhcp_options, op_uid);
                    checkErrors("Failed to insert dhcp options for interface #"+intfcID, cmd.doIt(con));
                }

            }

            // now ip addresses
            Hashtable ip_parents = new Hashtable(); // used for resetting ip addresses
            int next_parent_key = 0;

            Hashtable ips_to_delete = new Hashtable();
            Hashtable ips_to_insert = new Hashtable();
            Hashtable ips_to_update = new Hashtable();

            Hashtable newIPs = Utils.nHash(newNode.get("ipc_addresses"));
            enm = newIPs.keys();
            while( enm.hasMoreElements() ) {
                String key = (String)enm.nextElement();
                String ipStr = key.trim();
                Hashtable parseRes = parser.parse(ipStr);
                if( parseRes.get("ip_low") == null )
                    throw new Exception("Invalid ip address format: ["+ipStr+"]");
                if( parseRes.get("ip_high") == null )
                    throw new Exception("Invalid ip address format: ["+ipStr+"]");
                Hashtable record = new Hashtable();
                record.put("key", key);
                Boolean active = (Boolean)((Hashtable)newIPs.get(key)).get("active?");
                if( active == null )
                    active = new Boolean(false);
                record.put("active?", active);

                record.put("parent key", "key "+(next_parent_key++));
                record.put("ip key", key);
                ip_parents.put(record.get("parent key"), newIPs);

                ips_to_insert.put("ipc "+ipStr, record);

            }

            Hashtable oldIPs = Utils.nHash(oldNode.get("ipc_addresses"));
            enm = oldIPs.keys();
            while( enm.hasMoreElements() ) {
                String key = (String)enm.nextElement();
                String ipStr = key.trim();
                Hashtable parseRes = parser.parse(ipStr);
                if( parseRes.get("ip_low") == null )
                    throw new Exception("Invalid ip address format: ["+ipStr+"]");
                if( parseRes.get("ip_high") == null )
                    throw new Exception("Invalid ip address format: ["+ipStr+"]");
                String theKey = "ipc "+ipStr;
                Boolean active = (Boolean)((Hashtable)oldIPs.get(key)).get("active?");
                if( active == null ) active = Boolean.FALSE;

                if( ips_to_insert.get(theKey) == null ) {
                    ips_to_delete.put(theKey, ipStr);
                    // we remove it from oldNode becouse deleting it also deletes associated
                    // with it names and we don't want to attempt to delete these names latter
                    // when they already gone.
                    oldIPs.remove(key);

                } else {
                    Hashtable newIP = (Hashtable)ips_to_insert.remove(theKey);
                    Boolean newActive = (Boolean)newIP.get("active?");
                    if( newActive.booleanValue() != active.booleanValue() )
                        ips_to_update.put(theKey, newIP);
                }
            }

            // now ips from interfaces

            // first ips from interfaces in a newNode
            Hashtable interfaces = Utils.nHash(newNode.get("interfaces"));
            enm = interfaces.keys();
            while( enm.hasMoreElements() ) {
                String intfcKey = (String)enm.nextElement();
                Hashtable intfc = (Hashtable)interfaces.get(intfcKey);
                Hashtable ips = Utils.nHash(intfc.get("ips"));
                Enumeration enm2 = ips.keys();
                while( enm2.hasMoreElements() ) {
                    String key = (String)enm2.nextElement();
                    String ipStr = key.trim();
                    Hashtable parseRes = parser.parse(ipStr);
                    if( parseRes.get("ip_low") == null )
                        throw new Exception("Invalid ip address format: ["+ipStr+"]");
                    if( parseRes.get("ip_high") == null )
                        throw new Exception("Invalid ip address format: ["+ipStr+"]");
                    Hashtable record = new Hashtable();
                    record.put("key", key);
                    record.put("intfcKey", intfcKey);
                    Hashtable oldIP = (Hashtable)ips.get(key);
                    Boolean active = (Boolean)oldIP.get("active?");

                    record.put("active?", active!=null?active:Boolean.FALSE);
                    record.put("PTRpref", Utils.nStr(oldIP.get("PTRpref")));
                    record.put("parent key", "key "+(next_parent_key++));
                    record.put("ip key", key);
                    ip_parents.put(record.get("parent key"), ips);

                    ips_to_insert.put("intfc "+intfcKey+" "+ipStr, record);
                }
            }

            // second ips from interfaces in a oldNode
            interfaces = Utils.nHash(oldNode.get("interfaces"));
            enm = interfaces.keys();
            while( enm.hasMoreElements() ) {
                String intfcKey = (String)enm.nextElement();
                Hashtable intfc = (Hashtable)interfaces.get(intfcKey);
                Hashtable ips = Utils.nHash(intfc.get("ips"));
                Enumeration enm2 = ips.keys();
                while( enm2.hasMoreElements() ) {
                    String key = (String)enm2.nextElement();
                    String ipStr = key.trim();
                    Hashtable parseRes = parser.parse(ipStr);
                    if( parseRes.get("ip_low") == null )
                        throw new Exception("Invalid ip address format: ["+ipStr+"]");
                    if( parseRes.get("ip_high") == null )
                        throw new Exception("Invalid ip address format: ["+ipStr+"]");
                    Hashtable oldIP = (Hashtable)ips.get(key);
                    Boolean active = (Boolean)oldIP.get("active?");
                    if( active == null )
                        active = new Boolean(false);
                    String theKey = "intfc "+intfcKey+" "+ipStr;
                    if( ips_to_insert.get(theKey) == null ) {
                        ips_to_delete.put(theKey, ipStr);

                        // we remove it from oldNode becouse deleting it also deletes names
                        // associated with it and we don't want to attempt to delete
                        // these names latter when they are already gone.
                        ips.remove(key);
                    } else {
                        Hashtable newIP = (Hashtable)ips_to_insert.remove(theKey);
                        Boolean newActive = (Boolean)newIP.get("active?");
                        if( newActive.booleanValue() != active.booleanValue() ||
                            !Utils.nStr(newIP.get("PTRpref")).equals(Utils.nStr(oldIP.get("PTRpref"))) )
                            ips_to_update.put(theKey, newIP);
                    }
                }
            }

            // now delete ips_to_delete
            enm = ips_to_delete.keys();
            while( enm.hasMoreElements() ) {
                String ipStr = (String)ips_to_delete.get(enm.nextElement());
                logger.debug("deleting ip "+ipStr);
                int id = ((NetDB)logger).verify_or_die (NetDB.IP, ipStr);
                DBCall call = new IPAddressDeleteCommand(logger, lock, id, op_uid);
                checkErrors("Failed to delete ip address #"+ipStr, con.callDB(call));
            }

            // now insert ips_to_insert
            enm = ips_to_insert.keys();
            while( enm.hasMoreElements() ) {
                Hashtable ip = (Hashtable)ips_to_insert.get(enm.nextElement());
                String ipStr = ((String)ip.get("key")).trim();
                String intfcID = (String)ip.get("intfcKey");
                logger.debug("reserving ip "+ipStr);
                String origIpStr = ipStr;
                Hashtable reserveResult = netdb.reserveIP(con, null, ipStr, 1);
                Hashtable result = (Hashtable)reserveResult.get(reserveResult.keys().nextElement());
                if( result.get("errors") != null ) {
                    Vector msgs = new Vector();
                    msgs.addElement("Failed to reserve ip address "+ipStr);
                    throw new NetDBException("Failed to reserve ip address", msgs);
                }

                Hashtable new_warnings = Utils.nHash(result.get("warnings"));
                if( new_warnings.size() > 0 ) {
                    Enumeration enm1 = new_warnings.keys();
                    while( enm1.hasMoreElements() )
                        warningsVector.addElement(((String)enm1.nextElement()).trim());
                }
                Hashtable ip_addresses = Utils.nHash(result.get("ip_addresses"));
                if( ip_addresses.size() > 0 ) {
                    ipStr = ((String)ip_addresses.keys().nextElement()).trim();

                    DBCall call;
                    if( intfcID != null ) {
                        call = new IPAddressAcceptCommand(logger, ipStr, (Boolean)ip.get("active?"),
                                                          new Integer(intfcID), lock,
                                                          IPAddressAcceptCommand.INTERFACE_TYPE,
                                                          (String)ip.get("PTRpref"), op_uid);
                        logger.debug("accepting ip in intfc "+intfcID+":"+ipStr);
                    } else {
                        call = new IPAddressAcceptCommand(logger, ipStr, (Boolean)ip.get("active?"),
                                                          idNode, lock,
                                                          IPAddressAcceptCommand.IPC_TYPE,
                                                          (String)ip.get("PTRpref"), op_uid);
                        logger.debug("accepting ip in ipc node "+idNode+":"+ipStr);
                    }
                    checkErrors("Failed to insert ip address #"+ipStr, con.callDB(call));
                    acceptedIPs.put(ipStr, "");

                    // reset ip address
                    Hashtable ip_parent = Utils.nHash(ip_parents.get(Utils.n2s(ip.get("parent key"))));
                    Hashtable ipH = Utils.nHash(ip_parent.remove(ip.get("ip key")));
                    ipH.put("ip", ipStr);
                    ip_parent.put(ipStr, ipH);
                    if( !ipStr.equals(origIpStr) ) {
                        // we have to put new name
                        Hashtable names = Utils.nHash(ipH.get("names"));
                        if( names.size() == 1 ) { // do it only if there is one name
                            String nameKey = (String)names.keys().nextElement();
                            Hashtable nameH = (Hashtable)names.remove(nameKey);
                            String oldNameFqdn = ((String)nameH.get("name")).trim()+"."+((String)nameH.get("domain")).trim();

                            // check if name is auto-generated and then generate new one
                            Hashtable generated_name = netdb.generateHexName(origIpStr, 0);
                            if( (((String)generated_name.get("name")).trim()+"."+((String)generated_name.get("domain")).trim()).
                                toLowerCase().equals(oldNameFqdn.toLowerCase()) ) {

                                // record old name for cancellation
                                final_ip_names_to_cancel.addElement(oldNameFqdn);

                                // now generate new name and place it there
                                generated_name = netdb.generateHexName(ipStr, 0);
                                names.put(generated_name.get("name")+"."+generated_name.get("domain"), generated_name);
                            }
                        }
                    }
                }
            }

            // now update ips_to_update
            enm = ips_to_update.keys();
            while( enm.hasMoreElements() ) {
                Hashtable ip = (Hashtable)ips_to_update.get(enm.nextElement());
                String ipStr = ((String)ip.get("key")).trim();
                String intfcID = (String)ip.get("intfcKey");
                logger.debug("updating ip "+intfcID+":"+ipStr);
                int id = ((NetDB)logger).verify_or_die (NetDB.IP, ipStr);
                DBCall call =
                    new IPAddressUpdateCommand(logger, lock, id, (Boolean)ip.get("active?"),
                                               (String)ip.get("PTRpref"), op_uid);
                checkErrors("Failed to update ip address #"+ipStr, con.callDB(call));
            }

            // now names
            Hashtable sortResult = Utils.sortOutNames(oldNode, newNode);
            Hashtable names_to_delete = Utils.nHash(sortResult.get("names_to_delete"));
            Hashtable names_to_insert = Utils.nHash(sortResult.get("names_to_insert"));
            Hashtable names_to_update = Utils.nHash(sortResult.get("names_to_update"));

            // now delete names_to_delete
            enm = names_to_delete.keys();
            while( enm.hasMoreElements() ) {
                String key = (String)enm.nextElement();
                Hashtable name = (Hashtable)names_to_delete.get(key);

                String nameType = (String)name.get("type");
                if( nameType.equals("node_name") )
                    nameType = Constants.NODE_NAME_TYPE;
                else if( nameType.equals("alias") )
                    nameType = Constants.NAME_ALIAS_TYPE;
                else if( nameType.equals("mx") )
                    nameType = Constants.MX_TYPE;
                else if( nameType.equals("ip_name") || nameType.equals("ipc_name") )
                    nameType = Constants.ADDRESS_NAME_TYPE;
                else if( nameType.equals("intfc_name") )
                        nameType = Constants.INTERFACE_NAME_TYPE;

                DBCall call;
                if( nameType.equals(Constants.MX_TYPE) ) {
                    call = new MXDeleteCommand(logger, name.get("name")+"."+name.get("domain"),
                                               (String)name.get("parent"), lock, op_uid);
                } else
                    call = new NameDeleteCommand(logger, (String)name.get("name"),
                                                 (String)name.get("domain"), nameType, lock, op_uid);
                String fqdnName = name.get("name")+"."+name.get("domain");
                logger.debug("Deleting name "+fqdnName);
                checkErrors("Failed to delete name ["+fqdnName+"]",
                            con.callDB(call));
            }

            // insert names_to_insert
            boolean names = true;
            for( int i = 0; i < 2; i++ ) {
                // it should be done it two passes, first names and then (on second pass) aliases and mxes
                enm = names_to_insert.keys();
                while( enm.hasMoreElements() ) {
                    String key = (String)enm.nextElement();
                    // first go over parenting names,
                    // i.e. names which may have aliases and/or mx's
                    Hashtable name = (Hashtable)names_to_insert.get(key);
                    if( (names && name.get("refs") == null) ||
                        (!names && name.get("refs") != null) ) {
                        continue;
                    }

                    String nameType = (String)name.get("type");
                    if( nameType.equals("node_name") )
                        nameType = Constants.NODE_NAME_TYPE;
                    else if( nameType.equals("alias") )
                        nameType = Constants.NAME_ALIAS_TYPE;
                    else if( nameType.equals("mx") )
                        nameType = Constants.MX_TYPE;
                    else if( nameType.equals("ip_name") || nameType.equals("ipc_name") )
                        nameType = Constants.ADDRESS_NAME_TYPE;
                    else if( nameType.equals("intfc_name") )
                        nameType = Constants.INTERFACE_NAME_TYPE;

                    String fqdnName = name.get("name")+"."+name.get("domain");
                    Hashtable reservedName = netdb.reserveName(con, fqdnName, nameType);
                    name.put("name", Utils.name((String)reservedName.get("name")));
                    name.put("domain", Utils.domain((String)reservedName.get("name"),
                                                    (String)netdb.getUser().get("default_domain")));
                    DBCall call = null;
                    if( nameType.equals(Constants.NODE_NAME_TYPE) )
                        call = new NameAcceptCommand(logger, new Integer((String)reservedName.get("id")), null,
                                                     nameType, idNode, null, lock, null, op_uid);

                    else if( nameType.equals(Constants.NAME_ALIAS_TYPE) )
                        call = new NameAcceptCommand(logger, new Integer((String)reservedName.get("id")), null,
                                                     nameType, null, (String)name.get("parent"),
                                                     lock, null, op_uid);

                    else if( nameType.equals(Constants.MX_TYPE) )
                        // Note that the fact that reservedName.get("pref") is a String representation of integer
                        // unsured in NetDB.update(...) by calling method Utils.validate(...)
                        call = new NameAcceptCommand(logger, new Integer((String)reservedName.get("id")), null,
                                                     nameType, null, (String)name.get("parent"), lock,
                                                     new Integer((String)name.get("pref")), op_uid);

                    else if( nameType.equals(Constants.ADDRESS_NAME_TYPE) )
                        call = new NameAcceptCommand(logger, new Integer((String)reservedName.get("id")), null, nameType,
                                                     null, (String)name.get("ip"), lock, null, op_uid);


                    else if( nameType.equals(Constants.INTERFACE_NAME_TYPE) )
                        call = new NameAcceptCommand(logger, new Integer((String)reservedName.get("id")), null, nameType,
                                                     new Integer((String)name.get("intfs_id")), null, lock,
                                                     null, op_uid);

                    else
                        throw new Exception("Invalid nameType");

                    if( call != null )
                        checkErrors("Failed to accept name ["+fqdnName+"]", call.processResultSet(con.callDB(call)));

                    Hashtable nameH = new Hashtable();
                    nameH.put("fqdn", ((String)name.get("name")).trim().toLowerCase()+"."+((String)name.get("domain")).trim().toLowerCase());
                    nameH.put("type", new String(nameType));
                    acceptedNames.put(nameH, "");
                }
                names = false; // second pass goes over aliases and mxes
            }

            // update names_to_update
            enm = names_to_update.keys();
            while( enm.hasMoreElements() ) {
                String key = (String)enm.nextElement();
                Hashtable name = (Hashtable)names_to_update.get(key);
                String nameType = (String)name.get("type");

                DBCall call = null;
                if( nameType.equals("mx") )
                    call = new MXModifyCommand (logger, name.get("name")+"."+name.get("domain"),
                                                (String)name.get("parent"), lock, op_uid,
                                                new Integer((String)name.get("pref")));

                String fqdnName = name.get("name")+"."+name.get("domain");
                logger.debug("Modifying name "+fqdnName);

                if( call != null )
                    checkErrors("Failed to modify name ["+fqdnName+"]", con.callDB(call));
           }

            // verify whole stuff
            logger.debug("do check");
            DBCall verify = new VerifyCommand(logger, VerifyCommand.NODE, idNode, op_uid);
            Hashtable result = con.callDB(verify);

            // check for errors
            Vector error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            // unlock node
            DBCall call = new NodeUnlockCommand(logger, idNode.intValue(),
                                             (Timestamp)this.newNode.get("locked"), op_uid);
            checkErrors("Failed to unlock node", call.processResultSet(con.callDB(call)));

            ((NetDBii)logger).unregister_lock(lock, new Integer(Constants.NODE));

            commit_finally = true;

            if( warningsVector.size() > 0 )
                res.put("warnings", warningsVector);

            for( int i = 0; i < final_ip_names_to_cancel.size(); i++ ) {
                try {
                    ((NetDBii)logger).
                        cancelName(con,
                                   (String)final_ip_names_to_cancel.elementAt(i),
                                   Constants.ADDRESS_NAME_TYPE);
                } catch (Exception ex) {
                    logger.debug("Failed to cancel name ["+final_ip_names_to_cancel.elementAt(i)+"]");
                }
            }

            return res;

        } finally {
            if( commit_finally ) {
                try {
                    logger.info("modifying node id "+idNode+" done");
                    if( handle_transactions_locally ) {
                        logger.debug("CommitTransaction");
                        CommitTransaction(con.con);
                    }
                } catch(Exception exx) { }
                // remove accepted ips from list of reserved
                if( handle_transactions_locally ) {
                    Enumeration enm = acceptedIPs.keys();
                    while( enm.hasMoreElements() )
                        try {
                            netdb.unregister_reserved_ip(((String)enm.nextElement()).trim());
                        } catch (Exception ex) {}
                    // remove accepted names from list of reserved
                    //Enumeration enm = acceptedNames.keys();
                    //while( enm.hasMoreElements() )
                    //    netdb.removeReservedName((Hashtable)enm.nextElement());
                }
            } else
                try {
                    logger.info("modifying node id "+idNode+" failed");
                    if( handle_transactions_locally ) {
                        logger.debug("RollbackTransaction");
                        RollbackTransaction(con.con);
                    }
                } catch(Exception exx) { }
        }
    } // end of doIt()

    public static void intfcsIntersect(Hashtable oldIntfcs, Hashtable newIntfcs,
                                       Vector keys_to_insert,
                                       Vector keys_to_delete,
                                       Vector keys_to_keep) {
        // first fill up keys_to_keep and keys_to_insert
        Enumeration enm = newIntfcs.keys();
        while( enm.hasMoreElements() ) {
            String newKey = (String)enm.nextElement();
            if( oldIntfcs.containsKey(newKey) )
                keys_to_keep.addElement(newKey);
            else
                keys_to_insert.addElement(newKey);
        }

        // now fill up key_to_delete
        enm = oldIntfcs.keys();
        while( enm.hasMoreElements() ) {
            String oldKey = (String)enm.nextElement();
            if( !newIntfcs.containsKey(oldKey) )
                keys_to_delete.addElement(oldKey);
        }
    } // end of keysIntersect

    private void throwSimpleNetDBException(String msgHeader, String msg) throws NetDBException {
        Vector errors = new Vector();
        errors.addElement(msg);
        throw new NetDBException(msgHeader, errors);
    }
}
