/************************************************************************
 *
 * UserInsertCommand.java - db call pUser 'insert'
 *
 * $Id: UserInsertCommand.java,v 1.3 2008-03-15 02:56:19 jlidtke Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2004, Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.lib;

import stanford.netdb.utils.*;
import java.lang.*;
import java.util.*;
import java.sql.*;

/** db call pUser 'update' */
class UserInsertCommand extends DBCall implements ComplexDBCall {

    private Hashtable newUser;

    private static final int OP_USER_INDEX        = 1;  // p_op_uid
    private static final int PERSON_ID_INDEX      = 2;  // p_person_id
    private static final int DEF_GROUP_INDEX      = 3;  // p_def_group
    private static final int DEF_DOMAIN_INDEX     = 4;  // p_def_domain
    private static final int DEF_ADDR_SPACE_INDEX = 5;  // p_def_prefix
    private static final int ACTIVE_INDEX         = 6;  // p_active
    private static final int ALL_OBJ_INDEX        = 7;  // p_all_records
    private static final int ALL_GROUPS_INDEX     = 8;  // p_all_groups
    private static final int COMMENT_INDEX        = 9;  // p_comment

    public UserInsertCommand(stanford.netdb.utils.Logger logger, Hashtable newUser, Integer op_uid) {
        super(logger);
        this.newUser = newUser;
        db_call_type = Function_Map_Controller.get_function_type (Function_Map.PACKAGE_USER, Function_Map.INSERT);

        values = new Object[10];
        values[0]                    = ref_cursor;
        values[OP_USER_INDEX]        = op_uid;
        values[PERSON_ID_INDEX]      = new Integer((String)newUser.get("id"));
        values[DEF_GROUP_INDEX]      = newUser.get("default_group");
        values[DEF_DOMAIN_INDEX]     = newUser.get("default_domain");
        values[DEF_ADDR_SPACE_INDEX] = o2v(newUser.get("default_addr_space"));
        values[ACTIVE_INDEX]         = new Short( ((Boolean)newUser.get("active?")).booleanValue()?(short)1:(short)0 );
        values[ALL_OBJ_INDEX]        = new Short( ((Boolean)newUser.get("all_objects?")).booleanValue()?(short)1:(short)0 );
        values[ALL_GROUPS_INDEX]     = new Short( ((Boolean)newUser.get("all_groups?")).booleanValue()?(short)1:(short)0 );
        values[COMMENT_INDEX]        = o2v(newUser.get("comment"));
    }

    public String getSQL() {
        return proc_call_sql(Function_Map_Controller.get_function_name (Function_Map.PACKAGE_USER, Function_Map.INSERT));
    }

    public Hashtable doIt(NetDBConnection con) throws NetDBException, Exception {
        logger.info("Inserting user: "+newUser.get("NetID"));
        BeginTransaction(con.con);
        boolean commit_finally = false;
        Integer uID = null;
        Timestamp lock = null;
        try {
            logger.debug("do sigle value attributes");
            Hashtable result = con.callDB(this);

            // check for errors
            Vector error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            uID = new Integer((String)result.get("id"));
            int id = uID.intValue();
            lock = (Timestamp)result.get("locked");

            // insert groups
            Hashtable newGroups = Utils.nHash(newUser.get("groups"));

            Vector groups_to_insert = new Vector();
            Enumeration enm = newGroups.keys();
            while( enm.hasMoreElements() )
                groups_to_insert.addElement(enm.nextElement());

            // insert groups_to_insert
            logger.debug("inserting groups_to_insert");
            ComplexDBCall cmd =
                new ObjectGroupCommand(logger, ObjectGroupCommand.ADD_GROUPS, id,
                                       ObjectGroupCommand.USER, lock,
                                       groups_to_insert, (Integer)values[OP_USER_INDEX]);
            result = cmd.doIt(con);

            // check for errors
            error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            // insert departements
            Hashtable newDepartments = Utils.nHash(newUser.get("departments"));

            Vector deps_to_insert = new Vector();
            enm = newDepartments.keys();
            while( enm.hasMoreElements() )
                deps_to_insert.addElement(enm.nextElement());

            logger.debug("inserting deps_to_insert");
            for( int i = 0; i < deps_to_insert.size(); i++ ) {
                String depName = (String)deps_to_insert.elementAt(i);
                Hashtable dep = (Hashtable)newDepartments.get(depName);
                DBCall call =
                    new UserDepartmentCommand(logger, UserDepartmentCommand.INSERT_ACTION,
                                              uID, depName,
                                              ((Boolean)dep.get("official?")).booleanValue(),
                                              lock, (Integer)values[OP_USER_INDEX]);
                result = call.processResultSet(con.callDB(call));

                // check for errors
                error_set = (Vector)result.get("errors");
                if( error_set != null && error_set.size() > 0 )
                    return result;
            }

            // handle records
            Hashtable newRecords = Utils.nHash(newUser.get("records"));

            Vector recs_to_insert = new Vector();
            enm = newRecords.keys();
            while( enm.hasMoreElements() )
                recs_to_insert.addElement(enm.nextElement());

            logger.debug("insering recs_to_insert");
            for( int i = 0; i < recs_to_insert.size(); i++ ) {
                DBCall call =
                    new UserAccessCommand(logger, UserAccessCommand.INSERT_ACTION,
                                          uID, (String)recs_to_insert.elementAt(i),
                                          lock, (Integer)values[OP_USER_INDEX]);
                result = call.processResultSet(con.callDB(call));

                // check for errors
                error_set = (Vector)result.get("errors");
                if( error_set != null && error_set.size() > 0 )
                    return result;
            }

            logger.debug("do check");
            DBCall verify = new VerifyCommand(logger, VerifyCommand.USER,
                                              new Integer(id), (Integer)values[OP_USER_INDEX]);

            result = con.callDB(verify);

            // check for errors
            error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            commit_finally = true;

        } catch(Exception ex) {
            ex.printStackTrace();
            try {
                logger.info("Rollback Transaction");
                RollbackTransaction(con.con);
            } catch(Exception exx) { }
            throw ex;

        } finally {
            if( commit_finally )
                try {
                    logger.info("modifying user "+newUser.get("NetID")+" done");
                    // Unlock the user now that we're done.
                    try {
                        DBCall call =
                            new UserUnlockCommand(logger, uID, lock, (Integer)values[OP_USER_INDEX]);
                        call.processResultSet(con.callDB(call));
                    } catch (Exception exxx) { }
                    CommitTransaction(con.con);
                } catch(Exception exx) { }
            else
                try {
                    logger.info("modifying user "+newUser.get("NetID")+" failed");
                    logger.debug("Rollback Transaction");
                    RollbackTransaction(con.con);
                } catch(Exception exx) { }
        }
        return new Hashtable();
    }
}
