/************************************************************************
 *
 * UserModifyCommand.java - db call pUser 'update'
 *
 * $Id: UserModifyCommand.java,v 1.2 2008-03-15 02:56:20 jlidtke Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2002, Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.lib;

import stanford.netdb.utils.*;
import java.lang.*;
import java.util.*;
import java.sql.*;

/** db call pUser 'update' */
class UserModifyCommand extends DBCall implements ComplexDBCall {

    private Hashtable oldUser;
    private Hashtable newUser;
    private Timestamp lock;

    private static final int OP_USER_INDEX        = 0; // p_op_uid
    private static final int LOCKED_INDEX         = 1; // p_lock
    private static final int ID_INDEX             = 2; // p_user_id
    private static final int DEF_GROUP_INDEX      = 3; // p_def_group
    private static final int DEF_DOMAIN_INDEX     = 4; // p_def_domain
    private static final int DEF_ADDR_SPACE_INDEX = 5; // p_def_prefix
    private static final int ACTIVE_INDEX         = 6; // p_active
    private static final int ALL_OBJ_INDEX        = 7; // p_all_records
    private static final int ALL_GROUPS_INDEX     = 8; // p_all_groups
    private static final int COMMENT_INDEX        = 9; // p_comment

    public UserModifyCommand(stanford.netdb.utils.Logger logger, Hashtable oldUser, Hashtable newUser, Integer op_uid)
        throws Exception {
        super(logger);
        this.oldUser = oldUser;
        this.newUser = newUser;
        db_call_type = Function_Map_Controller.get_function_type (Function_Map.PACKAGE_USER, Function_Map.UPDATE);


        values = new Object[10];
        values[OP_USER_INDEX]           = op_uid;
        if (oldUser.get("id") == null)
            values[ID_INDEX]            = new Integer((String)oldUser.get("id"));
        else
            values[ID_INDEX]            = ((NetDB)logger).verify_or_die (NetDB.USER, (String)oldUser.get ("NetID"));
        values[DEF_GROUP_INDEX]         = newUser.get("default_group");
        values[DEF_DOMAIN_INDEX]        = newUser.get("default_domain");

        String defAddrSpace = (String)newUser.get("default_addr_space");
        values[DEF_ADDR_SPACE_INDEX]    = defAddrSpace != null?defAddrSpace:"";

        values[ACTIVE_INDEX]            = new Short( ((Boolean)newUser.get("active?")).booleanValue()?(short)1:(short)0 );
        values[ALL_OBJ_INDEX]           = new Short( ((Boolean)newUser.get("all_objects?")).booleanValue()?(short)1:(short)0 );
        values[ALL_GROUPS_INDEX]        = new Short( ((Boolean)newUser.get("all_groups?")).booleanValue()?(short)1:(short)0 );

        values[COMMENT_INDEX] = o2v((String)newUser.get("comment"));
        values[LOCKED_INDEX]  = lock    = (Timestamp)oldUser.get("locked");
    }

    public String getSQL() {
        return proc_call_sql(Function_Map_Controller.get_function_name (Function_Map.PACKAGE_USER, Function_Map.UPDATE));
    }

    public Hashtable doIt(NetDBConnection con) throws NetDBException, Exception {
        logger.info("Modifying user: "+newUser.get("NetID"));
        BeginTransaction(con.con);
        boolean commit_finally = false;
        try {
            //
            int id = Integer.parseInt((String)oldUser.get("id"));

            logger.debug("do sigle value attributes");
            Hashtable result = con.callDB(this);

            // check for errors
            Vector error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            // update groups
            Hashtable oldGroups = Utils.nHash(oldUser.get("groups"));
            Hashtable newGroups = Utils.nHash(newUser.get("groups"));

            Vector groups_to_insert = new Vector();
            Vector groups_to_delete = new Vector();
            Vector groups_to_keep = new Vector();
            Utils.keysIntersect(oldGroups, newGroups, groups_to_insert, groups_to_delete, groups_to_keep);

            // insert groups_to_insert
            logger.debug("inserting groups_to_insert");
            ComplexDBCall cmd =
                new ObjectGroupCommand(logger, ObjectGroupCommand.ADD_GROUPS, id,
                                       ObjectGroupCommand.USER, lock, groups_to_insert,
                                       (Integer)values[OP_USER_INDEX]);
            result = cmd.doIt(con);

            // check for errors
            error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            logger.debug("deleting groups_to_delete");
            cmd = new ObjectGroupCommand(logger, ObjectGroupCommand.REMOVE_GROUPS, id,
                                         ObjectGroupCommand.USER,
                                         lock, groups_to_delete, (Integer)values[OP_USER_INDEX]);
            result = cmd.doIt(con);

            // check for errors
            error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;


            // update departements
            Hashtable oldDepartments = Utils.nHash(oldUser.get("departments"));
            Hashtable newDepartments = Utils.nHash(newUser.get("departments"));

            Vector deps_to_insert = new Vector();
            Vector deps_to_delete = new Vector();
            Vector deps_to_keep = new Vector();
            Utils.keysIntersect(oldDepartments, newDepartments, deps_to_insert, deps_to_delete, deps_to_keep);

            logger.debug("inserting deps_to_insert");
            for( int i = 0; i < deps_to_insert.size(); i++ ) {
                String depName = (String)deps_to_insert.elementAt(i);
                Hashtable dep = (Hashtable)newDepartments.get(depName);
                DBCall call =
                    new UserDepartmentCommand(logger, UserDepartmentCommand.INSERT_ACTION,
                                              (Integer)values[ID_INDEX], depName,
                                              ((Boolean)dep.get("official?")).booleanValue(),
                                              lock, (Integer)values[OP_USER_INDEX]);
                result = call.processResultSet(con.callDB(call));

                // check for errors
                error_set = (Vector)result.get("errors");
                if( error_set != null && error_set.size() > 0 )
                    return result;
            }

            logger.debug("deleting deps_to_delete");
            for( int i = 0; i < deps_to_delete.size(); i++ ) {
                String depName = (String)deps_to_delete.elementAt(i);
                Hashtable dep = (Hashtable)oldDepartments.get(depName);
                DBCall call =
                    new UserDepartmentCommand(logger, UserDepartmentCommand.DELETE_ACTION,
                                              (Integer)values[ID_INDEX], depName,
                                              ((Boolean)dep.get("official?")).booleanValue(),
                                              lock, (Integer)values[OP_USER_INDEX]);
                result = call.processResultSet(con.callDB(call));

                // check for errors
                error_set = (Vector)result.get("errors");
                if( error_set != null && error_set.size() > 0 )
                    return result;
            }

            logger.debug("updating deps_to_keep");
            for( int i = 0; i < deps_to_keep.size(); i++ ) {
                String depName = (String)deps_to_keep.elementAt(i);
                Hashtable dep = (Hashtable)newDepartments.get(depName);
                Hashtable oldDep = (Hashtable)oldDepartments.get(depName);
                if( ((Boolean)dep.get("official?")).booleanValue() != ((Boolean)oldDep.get("official?")).booleanValue() ) {
                    // update only if official status changed
                    DBCall call =
                        new UserDepartmentCommand(logger, UserDepartmentCommand.UPDATE_ACTION,
                                                  (Integer)values[ID_INDEX], depName,
                                                  ((Boolean)dep.get("official?")).booleanValue(),
                                                  lock,(Integer)values[OP_USER_INDEX]);
                    result = call.processResultSet(con.callDB(call));

                    // check for errors
                    error_set = (Vector)result.get("errors");
                    if( error_set != null && error_set.size() > 0 )
                        return result;
                }
            }

            // handle records
            Hashtable oldRecords = Utils.nHash(oldUser.get("records"));
            Hashtable newRecords = Utils.nHash(newUser.get("records"));

            Vector recs_to_insert = new Vector();
            Vector recs_to_delete = new Vector();
            Vector recs_to_keep = new Vector();
            Utils.keysIntersect(oldRecords, newRecords, recs_to_insert, recs_to_delete, recs_to_keep);

            logger.debug("insering recs_to_insert");
            for( int i = 0; i < recs_to_insert.size(); i++ ) {
                DBCall call = new UserAccessCommand(logger, UserAccessCommand.INSERT_ACTION, (Integer)values[ID_INDEX],
                                                    (String)recs_to_insert.elementAt(i),
                                                    lock, (Integer)values[OP_USER_INDEX]);
                result = call.processResultSet(con.callDB(call));

                // check for errors
                error_set = (Vector)result.get("errors");
                if( error_set != null && error_set.size() > 0 )
                    return result;
            }

            logger.debug("deleting recs_to_delete");
            for( int i = 0; i < recs_to_delete.size(); i++ ) {
                DBCall call = new UserAccessCommand(logger, UserAccessCommand.DELETE_ACTION, (Integer)values[ID_INDEX],
                                                    (String)recs_to_delete.elementAt(i),
                                                    lock, (Integer)values[OP_USER_INDEX]);
                result = call.processResultSet(con.callDB(call));

                // check for errors
                error_set = (Vector)result.get("errors");
                if( error_set != null && error_set.size() > 0 )
                    return result;
            }

            // do check
            logger.debug("do check");
            DBCall verify = new VerifyCommand(logger, VerifyCommand.USER,
                                              new Integer(id), (Integer)values[OP_USER_INDEX]);

            result = con.callDB(verify);

            // check for errors
            error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            DBCall call = new UserUnlockCommand(logger, ((Integer)values[ID_INDEX]),
                                                lock, (Integer)values[OP_USER_INDEX]);
            result = call.processResultSet(con.callDB(call));

            // check for errors
            error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            ((NetDBii)logger).unregister_lock(lock, new Integer(Constants.USER));

            commit_finally = true;

        } catch(Exception ex) {
            ex.printStackTrace();
            try {
                logger.info("Rollback Transaction");
                RollbackTransaction(con.con);
            } catch(Exception exx) { }
            throw ex;

        } finally {
            if( commit_finally )
                try {
                    logger.info("modifying user "+newUser.get("NetID")+" done");
                    CommitTransaction(con.con);
                } catch(Exception exx) { }
            else
                try {
                    logger.info("modifying user "+newUser.get("NetID")+" failed");
                    logger.debug("Rollback Transaction");
                    RollbackTransaction(con.con);
                } catch(Exception exx) { }
        }
        return new Hashtable();
    }
}
