/************************************************************************
 *
 * NetDBi.java -
 *
 * $Id: NetDBi.java,v 1.3 2006-12-08 01:20:33 priimak Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2002, Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.utils;

import java.lang.*;
import java.util.*;
import java.rmi.*;
import java.sql.*;

/**
 * If you are writting rmi client to connect to NetDB RMI server you've come to the right place.
 * Note, that with exception of several methods specific to its client-server nature, such as
 * klogin(...), logout() and alike,  this interface defines the same methods as NetDBpi.
 */
public interface NetDBi extends Remote {

    /**
     * This is a two stage call which is used to validate instance of NetDBi, previously obtained by
     * calling getInstance(). That is done in two steps. First client sends its k5 token (see
     * example in 'NetDB Guide') by calling klogin(...) and receives k5 server token,
     * which can be used to verify server principal. On second step client calls klogin with
     * null argument and recieves itself (i.e. instance of NetDBi) or exception is thrown if user is not
     * allowed to login. During second call server checks if there exists an ACL and if it does
     * then it loads it and performs serveral checks. ACL is used to for two things. First it
     * specifies user who are <i>gods</i> (i.e. those who can shutdown server remotely, edit ACL
     * and alike) and it may, if set so, limit users who can login into system. Thus exception is
     * thrown when there is an ACL and which configured to restrict users access and user is not
     * in that ACL or person is not an NetDB user. Note, distinction here. Person is anyone with
     * SUNetID, while user is person with NetDB account. To find out if person is a user klogin
     * loads user info, which is why second iteration takes a lot of time comparing to a first one.
     *
     * @param token K5 token.
     * @throws RemoteException if authentication fails for some reasons.
     * @return Object first time it is a token (i.e. byte[]) send time it is an instance of valid NetDBi
     */
    public Object       klogin                  (byte[] token)
        throws RemoteException;

    /**
     * This is an advisory call for clients, in a sense that it is recommended to call this method
     * at the end of the session which would make clean up of server side structures faster, even though
     * they would be clean up eventually anyway. Note that this call would completely and un-repairable
     * invalidate current session.
     *
     * @throws RemoteException if authentication fails for some reasons.
     */
    public void         logout ()
        throws RemoteException;

    /**
     * This is a first call to obtain your instance of NetDBi. First you obtain netdb handle which
     * is also instance of NetDBi by calling Naming.lookup(...). This is instance will have
     * only one call enabled and that is getInstance(). Then you use netdb handle (netdb_handle) to
     * to obtain instance of NetDBi. This instance is not valid yet. That is you have to perform kerberos
     * authentication using klogin(...) call, untill it is done calling any other methods will result in
     * throwing SessionTimeoutException.
     *
     * @throws RemoteException if something goes wrong.
     * @return instance of not validated NetDBi
     */
    public NetDBi getInstance() throws RemoteException;

    /**
     * Returns content of GlobalDefault table.
     * Note that it is cached on login
     *
     * @return copy of GlobalDefault table cached in a hash
     */
    public Hashtable getGlobalDefaults()
        throws Exception;

    /**
     * Returns <b>*copy*</b> of the current users Hashtable.
     *
     * @return copy of the current users Hashtable.
     */
    public Hashtable   getCurrentUser           ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Returns copy of ACL hashtable.
     *
     * @return copy of ACL Hashtable representation.
     * @throws RemoteException if user performing this call is not a god or for some other reason.
     */
    public Hashtable   loadACL                  ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Same as loadACL() but will create actual acl file if it does not exist
     */
    public Hashtable   createACL                ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Will write new data in ACL file.
     *
     * @param acl new ACL.
     * @throws RemoteException if user performing this call is not a god or for some other reason.
     */
    public void        updateACL                (Hashtable acl)
        throws RemoteException, SessionTimeoutException;

    /**
     * Will shutdown netdb server.
     *
     * @throws RemoteException if user performing this call is not a god or for some other reason.
     */
    public void        shutdown                 ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Returns copy of NetDB wide default domain. Note, that it is different from
     * users default domain. This method should be used when generating names.
     * For dhcp ip name generation call
     * {@link #generateHexName(String, int) generateHexName(ip, salt)}
     *
     * @return copy of NetDB wide default domain
     */
    public String      getDefaultDomain         ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Given search id it will attempt to cancel search (simpleSearch or fullSearch)
     * with this id. <br>Note that it is up to user to provide valid id. Nothing happens
     * if id is not valid.
     * @param searchID id string for a search you want to cancel
     */
    public void        cancelSearch             (String searchID)
        throws RemoteException, SessionTimeoutException;

    /**
     * This is a preffered way to create search id. You don't have to use it though,
     * however in this case you should yourself take care that your search ids are
     * unique.
     * @return search id to be used in <a href="NetDBpi.html#simpleSearch(int, java.lang.String, java.lang.String)">simpleSearch(...)</a>
     */
    public String      createSearchID           ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Given search string and search type performs simple search.
     *
     * @param what one of Constants.{NODE, USER, NETWORK, GROUP, ATEAM, DOMAIN}
     * @param searchStr search string
     * @throws RemoteException if passed parameters are invalid or search fails for some reasons.
     * @throws RemoteNetDBException if search fails for some reasons, for example been canceled
     * @return Hashtable with a search results, see 'NetDB Guide' for format of this hashtable.
     */
    public Hashtable   simpleSearch             (int what, String searchStr)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Given search string and search type performs simple search. As this search carries search
     * id it can be canceled by calling {@link #cancelSearch(java.lang.String) cancel(searchID)}.
     * Note that preferred way of generating of search id is calling
     * {@link #createSearchID() createSearchID()}
     *
     * @param what one of Constants.{NODE, USER, NETWORK, GROUP, ATEAM, DOMAIN}
     * @param searchStr search string
     * @param searchID id string of the search, which allows to cancel active search
     * @throws RemoteException if passed parameters are invalid or search fails for some reasons.
     * @throws RemoteNetDBException if search fails for some reasons, for example been canceled
     * @return Hashtable with a search results, see 'NetDB Guide' for format of this hashtable.
     */
    public Hashtable   simpleSearch             (int what, String searchStr, String searchID)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Performs directory search.
     *
     * @param byWhat one of Constants.{NETID, REGID, LOOKUP}
     * @param searchStr - search string
     * @throws RemoteException if there are no registered directory provider or some other unexpected
     * condition. Generally you should never get any exception.
     * @return Hashtable with a search results, see 'NetDB Guide' for format of this hashtable.
     */
    public Hashtable   dirSearch                (int byWhat, String searchStr)
        throws RemoteException, SessionTimeoutException;

    /**
     * Write comment. XXX
     */
    public Hashtable fullSearch             (int what, Hashtable search_params, String searchID)
        throws RemoteException, SessionTimeoutException;

    /**
     * Given ip (String) will find corresponing address space where this ip address is located
     * and return it.
     * @throws RemoteException if passed parameters are not valid or some other similar condition.
     * @throws RemoteNetDBException if something failed inside of stored procedure or something like this.
     */
    public String      findIPAddrSpace          (String ip)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Performes move of address spaces listed in asList vector as their prefixes to
     * destination network lableled by its FQDN targetNetFQDN
     *
     * @param asList vector of address space prefixes.
     * @param targetNetFQDN fqdn name of network where address spaces are moving to.
     * @throws RemoteException if passed parameters are not valid or some other similar condition.
     * @throws RemoteNetDBException if move fails becouse user did not have right permissions or
     * target network does not exist or something like this.
     */
    public void        moveAS                   (Vector asList, String targetNetFQDN)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Given object type (what) and objects handle will try to lock the object.
     *
     * @param what objects type, like Constants.NODE etc.
     * @param handle objects handle, for example for node it is one of its fully
     * qualified domain names (FQDN).
     * @throws RemoteException is thrown if passed parameters are invalid or some other
     * unexpected condition.
     * @throws RemoteNetDBException if lock fails for some reasons.
     * @return Timestamp which acts as a handle to the lock, i.e. if you want to
     * unlock object you must in addition to the objects handle timestamp of the lock.
     */
    public Timestamp   lock                     (int what, String handle)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Given object type (what) and objects id (oid) will try to lock the object.
     *
     * @param what objects type, like Constants.NODE etc.
     * @param oid objects id
     * @throws RemoteException is thrown if passed parameters are invalid or some other
     * unexpected condition.
     * @throws RemoteNetDBException if lock fails for some reasons.
     * @return Timestamp which acts as a handle to the lock, i.e. if you want to
     * unlock object you must in addition to the objects handle timestamp of the lock.
     */
    public Timestamp   lock                     (int what, int oid)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * To be written!!!
     */
    public Timestamp   unlock                   (int what, String handle, Timestamp tm)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * To be written!!!
     */
    public void        unlock                   (int what, int oid, Timestamp tm)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * To be written!!!
     */
    public Hashtable   splitAS                  (Hashtable as)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Given address space prefix will verify if given ip range may belong to a
     * network with given id. Note that this call does not alter anything in database.
     *
     * @param prefix address space prefix in format xxx.xxx.xxx.xxx/xx
     * @param idNet network id where you want to place given address space
     * @throws RemoteException if prefix is null or some other similar unexpected
     * condition
     * @throws RemoteNetDBException is for some reasons creating of proposed address
     * space spaces is not possible, which typically the case when there are address
     * spaces withing given prefix which belong to another network.
     * @return normally returns empty Hashtable.
     */
    public Hashtable verifyAddressSpace(String prefix, Integer idNet)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * To be written!!!
     */
    public Hashtable   update                   (Hashtable oldObj, Hashtable newObj)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Given objects type (what) and its id (oid) it will attempt to delete it. If you
     * want to delete network, os, make or location you sould call
     * {@link #delete(java.util.Hashtable) delete(obj)}.<br>
     * Note, that objects: node, network, domain, user and admin team must be locked prior
     * to their deletion or exception will be thrown.
     *
     * @param what one of Constants.{NODE, USER, GROUP, ATEAM, DOMAIN, DEPARTMENT, MODEL}
     * @param oid id of the object you want to delete
     * @throws RemoteException thrown if obj is null or what == NETWORK or on some other unexpected condition.
     * @throws RemoteNetDBException if delete failed for some reason, i.e. object was not locked
     * or user did not have appropriate permissions to performe this operation.
     * @return Hashtable, which you may safely ignore.
     */
    public Hashtable   delete                   (int what, String oid)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Given objects Hastable it will attempt to delete it. Unlike
     * {@link #delete(int, java.lang.String) delete(what, oid)} this method works for
     * all objects. The right way to use it, is to lock object by calling
     * {@link #lock(int, String) lock(what, handle)} or {@link #lock(int, int) lock(what, oid)}
     * and then obtain objects Hashtable by calling
     * {@link #getInfo(int, String) getInfo(what, handle)} or
     * {@link #getInfo(int, int) getInfo(what, oid)} and then pass it to this method.<br>
     * Note, that objects: node, network, domain, user and admin team must be locked prior
     * to their deletion or exception will be thrown.
     *
     * @param obj objects Hashtable you want to delete.
     * @throws RemoteException thrown if obj is null or on some other unexpected condition.
     * @throws RemoteNetDBException if delete failed for some reason, i.e. object was not locked
     * or user did not have appropriate permissions to performe this operation.
     * @return Hashtable, which you may safely ignore.
     */
    public Hashtable   delete                   (Hashtable obj)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Inserts object, where 'obj' is appropriate object's Hashtable.
     *
     * @throws RemoteException if obj is null or some other unexpected condition.
     */
    public Hashtable   insert                   (Hashtable obj)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Reserves 'count' ip addresses in the given address space 'asPrefix' begining with 'ip'.
     * If 'ip' is null then it will pick up first available ip addresses in a given address space.
     * Note, that given ip address may be different from requested.
     *
     * Returns:
     * <pre>
     * (Hashtable)
     * %requested ip% -> (Hashtable)
     *                   'ip_addresses' -> (Hashtable)
     *                                     %ip% - (String)
     *                   'warnings'     -> (Hashtable)
     *                                     %warning%
     * </pre>
     *
     * @param asPrefix address space prefix were it is desired to reserve ip address(es), may be null
     * if ip is provided.
     * @param ip staring ip address from which you would want to reserve; may be null.
     * @param count amount of ip addresses you would want to reserve staring from 'ip'.
     * @throws RemoteException if for example asPrefix is null or count is negative or something
     * similar.
     * @throws RemoteNetDBException if reservation failed, becouse threre are no availiable ip
     * addresses left for example or some other reason.
     */
    public Hashtable   reserveIP                (String asPrefix, String ip, int count)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Cancels previously reserved but not accepted ip address, if successful it
     * results in removing any traces of ip address from the database, unless ip address
     * already accepted in which case no expection is thrown. Also note, that even
     * though instance of NetDBpi will cancel reserved ip addresses when destroyed
     * your application should take care of properly canceling reserved ip addresses
     * and not wait for the end of session when it would be handled automatically.
     *
     * @param ip string representation of ip address you want to cancel.
     * @throws RemoteNetDBException is thrown if there are some legitimate reasons why
     * ip adress cannot be canceled, like user not having permissions to do so
     * or alike. Note, that no exception is thrown if ip address does not exist
     * or already accepted.
     * @throws RemoteException could be thrown on some unexpected conditions, like
     * database been down or alike.
     */
    public void        cancelIP                 (String ip)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Reserves name of given name type.
     *
     * @param fqdnName name you are trying to reserve, if domain is not given assumed
     * users' default domain.
     * @param nameType one of Constants.{DOMAIN_TYPE, NODE_NAME_TYPE, ADDRESS_NAME_TYPE,
     * INTERFACE_NAME_TYPE, NETWORK_NAME_TYPE, NAME_ALIAS_TYPE, MX_TYPE}
     * @throws RemoteException if passed parameters are not valid or some other unexpected condition.
     * @throws RemoteNetDBException name reservation fails for some reasons, like there already
     * could be a name in a database you are trying to reserve.
     * @return actually reserved name, which is usefull in case when reserving name without
     * given domain, it will attach users def. domain and return it.
     */
    public String      reserveName              (String fqdnName, String nameType)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Cancels previously reserved but not accepted name, if successful it
     * results in removing any traces of name from the database, unless name
     * already accepted in which case no expection is thrown. Note that name
     * actually consist of name and its type, i.e. it is possible to have some
     * name of type NODE_NAME_TYPE and MX_TYPE, in which case it is listed twice in
     * a database, not all combinations are valid though, for example name cannot
     * be of type NODE_NAME_TYPE and NAME_ALIAS_TYPE. Also note, that even
     * though instance of NetDBpi will cancel reserved names when destroyed
     * your application should take care of properly canceling reserved names
     * and not wait for the end of session when it would be handled automatically.
     *
     * @param fqdnName name you want to cancel.
     * @param nameType type of the name, it could be one of the following
     * Constants.{DOMAIN_TYPE, NODE_NAME_TYPE, ADDRESS_NAME_TYPE, INTERFACE_NAME_TYPE,
     *            NETWORK_NAME_TYPE, NAME_ALIAS_TYPE, MX_TYPE}
     * @throws RemoteNetDBException is thrown if there are some legitimate reasons why
     * ip adress cannot be canceled, like user not having permissions to do so
     * or alike. Note, that no exception is thrown if ip address does not exist
     * or already accepted.
     * @throws RemoteException could be thrown on some unexpected conditions, like
     * database been down or alike.
     * @see Constants
     */
    public void        cancelName               (String fqdnName, String nameType)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Checks if user has access to the specific type of the objects. Note that, to
     * verify that user has access to some specific object you should also call
     * <a href="NetDBpi.html#checkGroupAccess(java.util.Hashtable)">checkGroupAccess(java.util.Hashtable)</a>.
     *
     * @param objectType one of Constants.{NODE, USER, NETWORK, GROUP, ATEAM, DOMAIN,
     * ADVANCED_NODE, IPC_PROVIDER, ROUTER, SYSTEM_MANAGEMENT}
     * @throws RemoteNetDBException if got this then you found a bug in database
     * @throws RemoteException you may get this if database is down or passed object
     * type is invalid, it also may correspond to some bug in netdb.jar or a
     * database.
     * @return         true or false indicating that user does/doesn't have access to the objectType
     * @see Constants
     */
    public boolean     checkAccess              (int objectType)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Checks if user has access rights to a specific object. Normally in order to
     * check that user has rights to modify or delete object both methods
     * <a href="NetDBpi.html#checkGroupAccess(int)">checkAccess</a> and checkGroupAccess
     * should be called. Note, that is checkAccess failed then there is no need to
     * call checkGroupAccess.
     *
     * @param object   Object to check access to.
     * @throws RemoteNetDBException if got this then you found a bug in database
     * @throws RemoteException you may get this if database is down or passed object
     * type is invalid, it also may correspond to some bug in netdb.jar or a
     * database.
     * @return         true or false indicating that user does/doesn't have access
     * to this particular object
     */
    public boolean     checkGroupAccess         (Hashtable object)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Given object type (what) and its id (oid) returns Hashatble representing
     * object. If object does not exist no exception is expected to be thrown
     * and the error message placed into returned Hashtable. Note, that in case
     * of os, make, model, model type, location and dhcp options oid is ignored
     * and returned hashtable contains list of corresponing little-O objects.
     *
     * @param what object type, one of Constants.{NODE, NETWORK, USER, GROUP,
     * ATEAM, DOMAIN, OS, MAKE, MODEL, MODEL_TYPE, DEPARTMENT, LOCATION, DHCP_OPTIONS}
     * @param oid objects id
     * @throws RemoteException thrown if passed parameters are not valid or some other
     * similar conditions
     * @return Hashtable representing object or error message indicating that
     * object does not exist. If object not found it has following format
     * {{result_sets=""}, {errors=Vector(<error messages>)}}
     * @see #getInfo(int, java.lang.String) getInfo(what, handle)
     */
    public Hashtable   getInfo                  (int what, int oid)
        throws RemoteException, SessionTimeoutException;

    /**
     * Given object type (what) and its handle returns Hashatble representing
     * object. If object does not exist no exception is expected to be thrown
     * and the error message placed into returned Hashtable. Note, that in case
     * of os, make, model, model type, location and dhcp options handle is ignored
     * and returned hashtable contains list of corresponing little-O objects. If
     * handle is empty string and what = DEPARTMENT then getInfo call returns
     * list of departments otherwise info for a particular department. If handle
     * is empty string and what = LOCATION then passed handle serves as a filter.
     *
     * @param what object type, one of Constants.{NODE, NETWORK, USER, GROUP,
     * ATEAM, DOMAIN, OS, MAKE, MODEL, MODEL_TYPE, DEPARTMENT, LOCATION, DHCP_OPTIONS}
     * @param handle objects handle
     * @throws RemoteException thrown if passed parameters are not valid, like 'what'
     * been out of range or handle not havin form <name>.<domain> for nodes or
     * some other similar condition.
     * @return Hashtable representing object or error message indicating that
     * object does not exist. If object not found Hashtable has following
     * format {{result_sets=""}, {errors=Vector(<error messages>)}}
     * @see #getInfo(int, int) getInfo(what, oid)
     */
    public Hashtable   getInfo                  (int what, String handle)
        throws RemoteException, SessionTimeoutException;

    /**
     * Returns list of networks to which user have group access in the following format
     * {&lt;networks.fqdn.name&gt; -> ''}
     *
     * @throws RemoteException if something goes wrong. It should not be really thrown at all.
     * unless db is down or something like this.
     * @return Hashtable storing list of fqdn name of networks in its keys.
     */
    public Hashtable   listMyNetworks           ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Returns list of domains to which user have group access in the following format
     * {&lt;fqdn.name&gt; -> '%domain id%'}
     *
     * @throws Exception if something goes wrong. It should not be really thrown at all.
     * unless db is down or something like this.
     * @return Hashtable storing list of domains in its keys.
     */
    public Hashtable   listMyDomains            ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Returns list of groups which user is a member of.
     * {&lt;group name&gt; -> ''}
     *
     * @throws RemoteException if something goes wrong. It should not be really thrown at all.
     * unless db is down or something like this.
     * @return Hashtable storing list of groups in its keys.
     */
    public Hashtable   listMyGroups             (String filter)
        throws RemoteException, RemoteNetDBException, SessionTimeoutException;

    /**
     * Returns list of all address spaces accessible to the current user in
     * the following format:
     * {&lt;str.ip.addr&gt; -> {'net_id' -> (String), 'str_ip_address' -> (String),
     * 'prefix' -> (String), 'ip_addr_space' -> (Integer), 'comment' -> (String)}, ...}
     *
     * @throws RemoteException if something goes wrong. It should not be really thrown at all.
     * unless db is down or something like this.
     * @return Hashtable storing list of address spaces.
     */
    public Hashtable   listMyAddrSpaces         ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Returns list of all address spaces in the following format:
     * {&lt;str.ip.addr&gt; -> {'net_id' -> (String), 'str_ip_address' -> (String),
     * 'prefix' -> (String), 'ip_addr_space' -> (Integer), 'comment' -> (String)}, ...}
     *
     * @throws RemoteException if something goes wrong. It should not be really thrown at all.
     * unless db is down or something like this.
     * @return Hashtable storing list of all address spaces.
     */
    public Hashtable   listAddrSpaces           ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Returns list of all dhcp services in the following format:
     * {&lt;service name&gt; -> {'name' -> (String), 'id' -> (String),
     * 'comment' -> (String)}, ...}
     *
     * @throws RemoteException if something goes wrong. It should not be really thrown at all.
     * @return Hashtable storing list of all dhcp services.
     */
    public Hashtable   listDHCPServices         ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Returns list of all records (access groups) current user
     * have access to in the following format:
     * {&lt;record.name&gt; -> &lt;record.display.name&gt;}
     *
     * @throws RemoteException if something goes wrong. It should not be really thrown at all.
     * @return Hashtable storing list of records accessable to the current user.
     */
    public Hashtable   listRecords              ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Returns list of little O objects. Currently the only supported type is NODE_TYPE.
     * If mine is set to true then returns those elements to which current user has access
     *
     * @throws Exception if something goes wrong.
     * @return Hashtable storing list of states accessible to current user.
     */
    public Hashtable   list                     (int what, String filter, boolean mine)
        throws RemoteException, SessionTimeoutException;

    /**
     * Returns list of node states that the current user
     * has access to.
     *
     * @throws Exception if something goes wrong.
     * @return Hashtable storing list of states accessible to current user.
     */
    public Hashtable   listMyStates             (String filter)
        throws RemoteException, SessionTimeoutException;

    /**
     * Returns list of all node states
     *
     * @throws Exception if something goes wrong.
     * @return Hashtable storing list of all states.
     */
    public Hashtable   listStates               ()
        throws RemoteException, SessionTimeoutException;

    /**
     * When generating name for dhcp ip addresses it advised to use this method.
     * The reason for this is that for certain ip addresses picked up domain is
     * different from the default domain.
     *
     * @param ip ip address you want to generate name for.
     * @param salt you sould try first with salt equal 0 and if name is already taken
     * try different salt value until generated name is unique.
     * @throws RemoteException if something goes wrong. It should not be really thrown at all.
     * @return name Hashtable = {{name=&lt;name&gt;}, {domain=&lt;domain&gt;}}
     */
    public Hashtable   generateHexName          (String ip, int salt)
        throws RemoteException, SessionTimeoutException;

    /**
     * Returns internal state of a particular instance of NetDBpi.
     * Format of returned Hashatble is {{<key>=<value>}, ...}
     * As of now (Oct 3 2002) it has only one key 'timestamp', which value is
     * a string representation of long int.
     *
     * @throws RemoteException if something goes wrong. It should not be really thrown at all.
     * @return internal state of current instance of NetDBpi
     */
    public Hashtable   getState                 ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Will reload data from config file (default netdb.properties).
     * Normally you don't have to call this method at all, as klogin(...)
     * will call this method for you.
     *
     * @throws RemoteException if something goes wrong like failure to open config file or
     * some other similar unexpected condition.
     */
    public void        reloadConfig             ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Queue managment. Throws exception of user is not a god.
     * To be written more!!!
     */
    public Hashtable   getQueue                 ()
        throws RemoteException, SessionTimeoutException;

    /**
     * Queue managment. Throws exception of user is not a god.
     * To be written more!!!
     */
    public void        setMaxConcurrent         (long mcc)
        throws RemoteException, SessionTimeoutException;
}
