/************************************************************************
 *
 * Utils.java - common static utility methods
 *
 * $Id: Utils.java,v 1.20 2008-03-21 23:48:29 priimak Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2002, Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.utils;

import java.text.*;
import java.util.*;
import java.util.regex.*;
import java.io.*;
import java.sql.*;
import su.netdb.parser.*;

public class Utils extends Object {

    // this mask is used to convert long int ip to string representation.
    private static long ipMask = 0x000000ff;

    public static Comparator node_type_cmp;

    public static int LEFT = 0;
    public static int RIGHT = 1;

    public static Logger logger = null;

    // it is used to parse ip and hw fields for a quick search
    static Parser parser = new Parser();

    static {
        if( logger == null )
            logger = new DefaultLogger();

        // used for ordering of node types
        node_type_cmp = new Comparator() {
                public int compare(Object o1, Object o2) {
                    Hashtable state1 = (Hashtable)((Map.Entry)o1).getValue();
                    String name1 = Utils.nStr(state1.get("type"));
                    Integer rank1 = (Integer)state1.get("rank");
                    if( rank1 == null ) rank1 = new Integer(0);

                    Hashtable state2 = (Hashtable)((Map.Entry)o2).getValue();
                    String name2 = Utils.nStr(state2.get("type"));
                    Integer rank2 = (Integer)state2.get("rank");
                    if( rank2 == null ) rank2 = new Integer(0);

                    if( rank1.compareTo(rank2) == 0 )
                        return name1.compareToIgnoreCase(name2);

                    return rank1.compareTo(rank2);
                }

                public boolean equals(Object obj) {
                    return true;
                }
            };


    }

    static public void setLogger(Logger logger) {
        Utils.logger = logger;
    }

    /**
     * Given two hashtables oldH and newH this method looks into theirs keys and if key is
     * present into oldH but not into newH then it places that key into keys_to_delete,
     * if key is present in newH but not in oldH then places that key into keys_to_insert and
     * if key is present in both newH and oldH then places that key into keys_to_keep.
     * Note, it assumes that keys are Strings.
     */
    public static void keysIntersect(Hashtable oldH, Hashtable newH,
                                     Vector keys_to_insert,
                                     Vector keys_to_delete,
                                     Vector keys_to_keep) {
        // first fill up keys_to_keep and keys_to_insert
        Enumeration enm = newH.keys();
        while( enm.hasMoreElements() ) {
            String newKey = (String)enm.nextElement();
            if( oldH.containsKey(newKey) )
                keys_to_keep.addElement(newKey);
            else
                keys_to_insert.addElement(newKey);
        }

        // now fill up key_to_delete
        enm = oldH.keys();
        while( enm.hasMoreElements() ) {
            String oldKey = (String)enm.nextElement();
            if( !newH.containsKey(oldKey) )
                keys_to_delete.addElement(oldKey);
        }
    } // end of keysIntersect

    /**
     * Validates object according to what object it claims to be and its specs.
     * @param what  - Constants.{NODE, DOMAIN, GROUP, NETWORK, USER, ATEAM}
     * @param obj   - object to validate
     * @param scope - scope of the validation. Possible values Constants.{WIDE, NARROW, FULL}.
     *                NARROW does not include 'id' while WIDE does, we need it becouse, for
     *                example, for inserting object there is not need to 'id' but for
     *                modifying it is required.
     */
    public static void validate(int what, Hashtable obj, int scope) throws NetDBException, Exception {
        Vector errors = new Vector();

        // first check fields common to all objects.
        Object type = obj.get("object");
        if( type == null )
            errors.addElement("object field is null");
        else if( !(type instanceof String) )
            errors.addElement("object field is not a String");

        if( scope == Constants.WIDE || scope == Constants.FULL ) {
            Object id = obj.get("id");
            if( id == null )
                errors.addElement("'id' field is null");
            else if( !(type instanceof String) )
                errors.addElement("'id' field is not a String");
            else {
                try {
                    Integer.parseInt((String)id);
                } catch(Exception ex) {
                    errors.addElement("'id' field does not represent integer type");
                }
            }
        }

        if( scope == Constants.FULL ) {
            Object lockedTm = obj.get("locked");
            if( lockedTm == null )
                errors.addElement("'locked' field is null");
            else if( !(lockedTm instanceof Timestamp) )
                errors.addElement("'locked' field is not a Timestamp");
        }

        switch(what) {
        case Constants.FULL_SEARCH_TYPE:
            String search_type = (String)obj.get("searchType");
            if( search_type == null )
                throw new Exception("Search parameters must contain 'search type'");

            search_type = search_type.trim();
            if( !search_type.toLowerCase().equals("log") ) {

            } else
                errors.addElement("Unrecognized search stype");

            break;

        case Constants.NODE:
            type = (String)obj.get("object");
            if( !type.equals("node") )
                errors.addElement("Object is not a node");

            // types - check that it is one of {Advanced, Router, IPCprovider, Template}
            boolean templateType = false;
            Object types = obj.get("types");
            if( types != null ) {
                if( !(types instanceof Hashtable) )
                    errors.addElement("'types' field is not a Hashtable");
                else {
                    Enumeration enm = ((Hashtable)types).keys();
                    while( enm.hasMoreElements() ) {
                        Object node_type = enm.nextElement();
                        if( !(node_type instanceof String) )
                            errors.addElement("'types' key is not a String");
                        if( nStr(nHash(((Hashtable)types).get(node_type)).get("extkey")).equals("Template") )
                            templateType = true;
                    }
                }
            }

            // name(1,n)
            Object names = obj.get("names");
            if( names == null )
                errors.addElement("Node must have at least one name");
            else if( !(names instanceof Hashtable) )
                errors.addElement("'names' field is not a Hashtable");
            else if( ((Hashtable)names).size() == 0 )
                errors.addElement("Node must have at least one name");
            else
                validateNames(obj, (Hashtable)names, errors, true);

            // check that template node have only one name and no aliases or mxs.
            if( templateType ) {
                if( ((Hashtable)names).size() > 1 )
                    errors.addElement("Node of type template may have only one name");
                else {
                    boolean aleases_error_added = false;
                    boolean mxs_error_added = false;
                    Enumeration enm = ((Hashtable)names).keys();
                    while( enm.hasMoreElements() ) {
                        Hashtable nameH = (Hashtable)((Hashtable)names).get(enm.nextElement());
                        Hashtable aliasesH = (Hashtable)nameH.get("aliases");
                        if( aliasesH != null && aliasesH.size() > 0 && !aleases_error_added ) {
                            aleases_error_added = true;
                            errors.addElement("Node of type template may no have aliases");
                        }
                        Hashtable mxsH = (Hashtable)nameH.get("mxs");
                        if( mxsH != null && mxsH.size() > 0 && !mxs_error_added ) {
                            mxs_error_added = true;
                            errors.addElement("Node of type template may not have mxes");
                        }
                    }
                }
            }

            // location(1,1)
            Object location = obj.get("location");
            if( location == null )
                errors.addElement("Node must have location");
            else if( !(location instanceof String) )
                errors.addElement("'location' field is not a String");
            else if( ((String)location).trim().length() == 0 )
                errors.addElement("Node must have location");

            // state(1,1)
            Object state = obj.get("state");
            if( state == null )
                errors.addElement("Node must have state");
            else if( !(state instanceof String) )
                errors.addElement("'state' field is not a String");
            else if( ((String)state).trim().length() == 0 )
                errors.addElement("Node must have state");

            // room(1,1)
            Object room =  obj.get("room");
            if( room == null )
                errors.addElement("Node must have room");
            else if( !(room instanceof String) )
                errors.addElement("'room' field is not a String");
            else if( ((String)room).trim().length() == 0 )
                errors.addElement("Node must have room");

            // make(1,1)
            Object make = obj.get("make");
            if( make == null )
                errors.addElement("Node must have make");
            else if( !(make instanceof String) )
                errors.addElement("'make' field is not a String");
            else if( ((String)make).trim().length() == 0 )
                errors.addElement("Node must have make");

            // model(1,1)
            Object model = obj.get("model");
            if( model == null )
                errors.addElement("Node must have model");
            else if( !(model instanceof String) )
                errors.addElement("'model' field is not a String");
            else if( ((String)model).trim().length() == 0 )
                errors.addElement("Node must have model");

            // os(1,n)
            Object oses = obj.get("oses");
            if( oses == null )
                errors.addElement("Node must have at least one os");
            else if( !(oses instanceof Hashtable) )
                errors.addElement("'oses' field is not a Hashtable");
            else if( ((Hashtable)oses).size() == 0 )
                errors.addElement("Node must have at least one os");

            // dept(1,1)
            Object dept = obj.get("dept");
            if( dept == null )
                errors.addElement("Node must have department");
            else if( !(dept instanceof String) )
                errors.addElement("'dept' field is not a String");
            else if( ((String)dept).trim().length() == 0 )
                errors.addElement("Node must have department");

            // group(1,n)
            Object groups = obj.get("groups");
            if( groups == null )
                errors.addElement("Node must have at least one group");
            else if( !(groups instanceof Hashtable) )
                errors.addElement("'group' field is not a Hashtable");
            else if( ((Hashtable)groups).size() == 0 )
                errors.addElement("Node must have at least one group");

            // exp.date - check that it is Timestamp
            Object exp_date = obj.get("exp.date");
            if( exp_date != null ) {
                if( !(exp_date instanceof Timestamp) )
                    errors.addElement("'exp.date' field is not a Timestamp");
                else if( ((Timestamp)exp_date).before(new Timestamp(0)) )
                    errors.addElement("Expiration date cannot be set before 01/01/1970");
            }

            // locked - check that it is Timestamp
            Object locked = obj.get("locked");
            if( locked != null && !(locked instanceof Timestamp) )
                errors.addElement("'locked' field is not a Timestamp");

            // comment - check that it is a String
            Object comment = obj.get("comment");
            if( comment != null && !(comment instanceof String) )
                errors.addElement("'comment' field is not a String");
            else if( comment != null && ((String)comment).length() > 256 )
                errors.addElement("'comment' field may not be longer than 256 characters");

            // admin(1,n)
            boolean has_admin = false;
            Object admins = obj.get("admins");
            if( admins == null )
                errors.addElement("Node must have at least one administrator");
            else if( !(admins instanceof Hashtable) )
                errors.addElement("'admins' field is not a Hashtable");
            else {
                Object ateams = ((Hashtable)admins).get("ateams");
                if( ateams != null ) {
                    if( !(ateams instanceof Hashtable) )
                        errors.addElement("'ateams' field is not a Hashtable");
                    else {
                        Enumeration enm = ((Hashtable)ateams).keys();
                        while( enm.hasMoreElements() ) {
                            Object ateam_name = enm.nextElement();
                            if( !(ateam_name instanceof String) )
                                errors.addElement("'ateams' key is not a String");
                            has_admin = true;
                        }
                    }
                }
                admins = ((Hashtable)admins).get("admins");
                if( admins != null ) {
                    if( !(admins instanceof Hashtable) )
                        errors.addElement("'admins' field is not a Hashtable");
                    else {
                        Enumeration enm = ((Hashtable)admins).keys();
                        while( enm.hasMoreElements() ) {
                            Object person = ((Hashtable)admins).get(enm.nextElement());
                            if( !(person instanceof Hashtable) )
                                errors.addElement("admin is not a Hashtable");
                            has_admin = true;
                        }
                    }
                }
            }
            if( !has_admin )
                errors.addElement("Node must have at least one admin");

            // check tags for empty name
            Hashtable tags_separated = (Hashtable)obj.get("tags_separated");
            if( tags_separated != null ) {
                Comparator cmp = new Comparator() {
                        public int compare(Object o1, Object o2) {
                            return (o1.toString()).compareToIgnoreCase(o2.toString());
                        }
                        public boolean equals(Object obj) {
                            return true;
                        }
                    };
                java.util.List list = Arrays.asList(tags_separated.keySet().toArray());
                Collections.sort(list, cmp);
                Iterator it = list.iterator();
                int index = 0;
                while( it.hasNext() ) {
                    index++;
                    Hashtable tag = (Hashtable)tags_separated.get(it.next());
                    String name = (String)tag.get("tag");
                    String value = (String)tag.get("value");
                    if( value != null && value.trim().length() > 0 &&
                        ( name == null || name.trim().length() == 0 ) )
                        errors.addElement((index==1?"First c":(index==2?"Second c":"C"))+
                                          "ustom field has value but no label.");
                }
            }

            // tags
            Object tags = obj.get("tags");
            if( tags != null ) {
                if( !(tags instanceof Hashtable) )
                    errors.addElement("'tags' field is not a Hashtable");
                else {
                    int tags_total = 0;
                    Enumeration enm = ((Hashtable)tags).keys();
                    while( enm.hasMoreElements() ) {
                        tags_total++;
                        Object tag = enm.nextElement();
                        if( !(tag instanceof String) )
                            errors.addElement("tag key is not a String");
                        else if( ((String)tag).length() > 32 )
                            errors.addElement("Custom field name cannot be over 32 characters");

                        Object value = ((Hashtable)tags).get(tag);
                        if( !(value instanceof String) )
                            errors.addElement("tag value is not a String");
                        else if( ((String)value).length() > 256 )
                            errors.addElement("Custom field value cannot be over 256 characters.");
                    }
                }
            }

            // addr_space
            Object addr_space = obj.get("addr_space");
            if( addr_space != null ) {
                if( !(addr_space instanceof String) )
                    errors.addElement("'addr_space' field is not a String");
                else {
                    if( ((String)addr_space).indexOf("/") < 0 )
                        errors.addElement("'addr_space' field is not an ip address space");

                    String ip_low = null;
                    String ip_high = null;
                    try {
                        Hashtable parseRes = (new Parser()).parse((String)addr_space);
                        ip_low = (String)parseRes.get("ip_low");
                        ip_high = (String)parseRes.get("ip_high");
                        if( ip_low == null || ip_high == null || ip_low.equals(ip_high) )
                            throw new Exception();
                    } catch (Exception exx) {
                        errors.addElement("'addr_space' field is not an ip address space");
                    }
                }

            }

            // ipc_addresses
            Object ipc_addresses = obj.get("ipc_addresses");
            if( ipc_addresses != null ) {
                if( !(ipc_addresses instanceof Hashtable) )
                    errors.addElement("'ipc_addresses' field is not a Hashtable");
                else {
                    Enumeration enm = ((Hashtable)ipc_addresses).keys();
                    while( enm.hasMoreElements() ) {
                        Object ipc_address = enm.nextElement();
                        if( !(ipc_address instanceof String) )
                            errors.addElement("'ipc_addresses' key is not a String");
                        else {
                            if( obj.get("no ip format check") == null ) {
                                String ip_low = null;
                                String ip_high = null;
                                try {
                                    Hashtable parseRes = (new Parser()).parse((String)ipc_address);
                                    ip_low = (String)parseRes.get("ip_low");
                                    ip_high = (String)parseRes.get("ip_high");
                                    if( ip_low == null || ip_high == null || !ip_low.equals(ip_high) )
                                        throw new Exception();
                                } catch (Exception exxx) {
                                    errors.addElement("'ipc_addresses' key is not an ip address");
                                }
                            }

                            Object ipc_addr = ((Hashtable)ipc_addresses).get(ipc_address);
                            if( !(ipc_addr instanceof Hashtable) ) {
                                errors.addElement("ipc address field is not a Hashtable");
                                continue;
                            }

                            Object active = ((Hashtable)ipc_addr).get("active?");
                            if( active == null )
                                errors.addElement("active? field does not exist");
                            else if( !(active instanceof Boolean) )
                                errors.addElement("active? field is not a Boolean");

                            Object ipc_names = ((Hashtable)ipc_addr).get("names");
                            if( ipc_names == null ) {
                                if( obj.get("ips do not have to have names") == null )
                                    errors.addElement("ipc address "+ipc_address+"does not have a name");
                            } else {
                                if( !(ipc_names instanceof Hashtable) )
                                    errors.addElement("ipc names field is not a Hashtable");
                                else {
                                    if( ((Hashtable)ipc_names).size() == 0 ) {
                                        errors.addElement("ipc ip address "+ipc_address+" does not have a name");
                                    } else if( ((Hashtable)ipc_names).size() > 1 )
                                        errors.addElement("ipc ip address "+ipc_address+" has more than one name");
                                    else {
                                        Object ipc_name = ((Hashtable)ipc_names).get((((Hashtable)ipc_names).keys()).nextElement());
                                        if( !(ipc_name instanceof Hashtable) )
                                            errors.addElement("ipc ip address "+ipc_address+" does not have a name");
                                        else {
                                            Object strName = ((Hashtable)ipc_name).get("name");
                                            Object strDomain = ((Hashtable)ipc_name).get("domain");
                                            if( strName == null || strDomain == null ||
                                                !(strName instanceof String) || !(strDomain instanceof String) ) {
                                                    errors.addElement("ipc ip address "+ipc_address+" does not have a name");
                                            } else {
                                                if( ((String)strName).trim().length() > 256 )
                                                    errors.addElement("name "+strName+" is too long");
                                                if( ((String)strDomain).trim().length() > 256 )
                                                    errors.addElement("name "+strDomain+" is too long");
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }

            // interfaces
            Object interfaces = obj.get("interfaces");
            if( interfaces != null ) {
                if( !(interfaces instanceof Hashtable) )
                    errors.addElement("'interfaces' field is not a Hashtable");
                else {
                    Enumeration enm = ((Hashtable)interfaces).keys();
                    while( enm.hasMoreElements() ) {
                        boolean empty = true;

                        Object intfc_id = enm.nextElement();
                        if( !(intfc_id instanceof String) ) {
                            errors.addElement("interfaces key is not a String");
                        }

                        Object intfc = ((Hashtable)interfaces).get(intfc_id);
                        if( !(intfc instanceof Hashtable) ) {
                            errors.addElement("interface is not a Hashtable");
                            continue;
                        }

                        Object hw = ((Hashtable)intfc).get("hw");
                        if( hw != null ) {
                            if( !(hw instanceof String) )
                                errors.addElement("'hw' field is not a String");
                            else if( ((String)hw).trim().length() > 0 )
                                empty = false;
                        }

                        Object roaming = ((Hashtable)intfc).get("roaming?");
                        if( roaming != null && !(roaming instanceof Boolean) )
                            errors.addElement("'roaming?' field is not a Boolean");

                        Object use_dhcp = ((Hashtable)intfc).get("use-dhcp?");
                        if( use_dhcp != null && !(use_dhcp instanceof Boolean) )
                            errors.addElement("'use-dhcp?' field is not a Boolean");

                        if( use_dhcp != null && ((Boolean)use_dhcp).booleanValue() &&
                            ( hw == null || ((String)hw).trim().length() == 0 ) )
                            errors.addElement("Interface may not have dhcp flag set if "+
                                              "it does not have hardware address as well.");


                        Object dhcp_options = ((Hashtable)intfc).get("dhcp-options");
                        validateDHCPOptions(dhcp_options, errors);

                        // name(0,n)
                        names = ((Hashtable)intfc).get("names");
                        if( names != null ) {
                            if( !(names instanceof Hashtable) )
                                errors.addElement("'names' field in interface is not a Hashtable");
                            else
                                validateNames(obj, (Hashtable)names, errors, false);
                        }

                        Object ips = ((Hashtable)intfc).get("ips");
                        if( ips != null ) {
                            if( !(ips instanceof Hashtable) )
                                errors.addElement("'ips' field is not a Hashtable");
                            else {
                                Enumeration enm1 = ((Hashtable)ips).keys();
                                while( enm1.hasMoreElements() ) {
                                    Object ip_address = enm1.nextElement();
                                    if( !(ip_address instanceof String) )
                                        errors.addElement("'ips' key is not a String");
                                    else {
                                        if( obj.get("no ip format check") == null ) {
                                            String ip_low = null;
                                            String ip_high = null;
                                            try {
                                                Hashtable parseRes = (new Parser()).parse((String)ip_address);
                                                ip_low = (String)parseRes.get("ip_low");
                                                ip_high = (String)parseRes.get("ip_high");
                                                if( ip_low == null || ip_high == null || !ip_low.equals(ip_high) )
                                                    throw new Exception();
                                                empty = false;
                                            } catch (Exception exxx) {
                                                errors.addElement("'ips' key is not an ip address");
                                            }
                                        } else
                                            empty = false;
                                    }
                                    Object ip = ((Hashtable)ips).get(ip_address);
                                    if( !(ip instanceof Hashtable) ) {
                                        errors.addElement("ip field is not a Hashtable");
                                        continue;
                                    }

                                    Object active = ((Hashtable)ip).get("active?");
                                    if( active != null && !(active instanceof Boolean) )
                                        errors.addElement("ip 'active?' field is not a Boolean");

                                    names = ((Hashtable)ip).get("names");
                                    if( names != null ) {
                                        if( !(names instanceof Hashtable) )
                                            errors.addElement("'names' field in ip address is not a Hashtable");
                                        else
                                            validateNames(obj, (Hashtable)names, errors, false);
                                    }
                                }
                            }
                        }

                        if( dhcp_options != null && ((Hashtable)dhcp_options).size() > 0 &&
                            hw == null && ( ips == null || ((Hashtable)ips).size() == 0 ) )
                            errors.addElement("Interface requires either hardware address "+
                                              "or at least one ip address.");
                        else {
                            if( empty ) {
                                if( obj.get("check empty interfaces") != null )
                                    errors.addElement("Interface id '"+intfc_id+"' is empty");
                                ((Hashtable)intfc).put("empty", "");
                            } else
                                ((Hashtable)intfc).remove("empty");
                        }

                    } // end of interfaces loop
                }
            } // end of interfaces

            break;

        case Constants.NETWORK:
            // name(1,n)
            names = obj.get("names");
            if( names == null )
                errors.addElement("Network must have at least one name");
            else if( !(names instanceof Hashtable) )
                errors.addElement("'names' field is not a Hashtable");
            else if( ((Hashtable)names).size() == 0 )
                errors.addElement("Network must have at least one name");
            else
                validateNames(obj, (Hashtable)names, errors, true);

            comment = obj.get("comment");
            if( comment != null ) {
                if( !(comment instanceof String) )
                    errors.addElement("comment field is not a String");
                else if( ((String)comment).length() > 256 )
                    errors.addElement("Network's comment may not be longer than 256 characters");
            }

            Object dhcpService = obj.get("dhcp_service");
            if( dhcpService != null ) {
                if( !(dhcpService instanceof String) )
                    errors.addElement("dhcp_service field is not a String");
                else
                    if( ((String)dhcpService).length() > 32 )
                    errors.addElement("dhcp_service field is too long");
            }

            groups = obj.get("groups");
            if( groups == null )
                errors.addElement("network must have at least one group");
            else if( !(groups instanceof Hashtable) )
                errors.addElement("groups field is not a Hashtable");
            else if( ((Hashtable)groups).size() == 0 )
                errors.addElement("Network must have at least one group");
            else {
                Enumeration enm = ((Hashtable)groups).keys();
                while( enm.hasMoreElements() ) {
                    Object group = enm.nextElement();
                    if( !(group instanceof String) )
                        errors.addElement("group field is not a String");
                    else if( ((String)group).length() > 32 )
                        errors.addElement("group field is too long");
                }
            }

            Object dhcp_options = obj.get("dhcp_options");
            validateDHCPOptions(dhcp_options, errors);

            Object ases = obj.get("address_spaces");
            if( ases != null ) {
                if( !(ases instanceof Hashtable) )
                    errors.addElement("address_spaces field is not a Hashtable");
                else {
                    Enumeration enm = ((Hashtable)ases).keys();
                    while( enm.hasMoreElements() ) {
                        Object asKey = enm.nextElement();
                        if( !(asKey instanceof String) ) {
                            errors.addElement("address_spaces key is not a String");
                            continue;
                        }
                        Object as = ((Hashtable)ases).get((String)asKey);
                        if( !(as instanceof Hashtable) ) {
                            errors.addElement("address_space field is not a Hashtable");
                            continue;
                        }

                        Object addr_space_ip = ((Hashtable)as).get("addr_space_ip");
                        if( addr_space_ip == null )
                            errors.addElement("address space does not have addr_space_ip field");
                        else if( !(addr_space_ip instanceof String) )
                            errors.addElement("addr_space_ip field is not an String");

                        Object prefix_length = ((Hashtable)as).get("prefix_length");
                        if( prefix_length == null )
                            errors.addElement("address space does not have prefix_length field");
                        else if( !(prefix_length instanceof Integer) )
                            errors.addElement("prefix_length field is not an Integer");

                        Object addr_space_prefix = ((Hashtable)as).get("addr_space_prefix");
                        if( addr_space_prefix == null )
                            errors.addElement("address space does not have addr_space_prefix field");
                        else if( !(addr_space_prefix instanceof String) )
                            errors.addElement("addr_space_prefix field is not a String");

                        Object low_res = ((Hashtable)as).get("low_res");
                        if( low_res == null )
                            errors.addElement("address space does not have low_res field");
                        else if( !(low_res instanceof Integer) )
                            errors.addElement("low_res field is not an Integer");

                        Object high_res = ((Hashtable)as).get("high_res");
                        if( high_res == null )
                            errors.addElement("address space does not have high_res field");
                        else if( !(high_res instanceof Integer) )
                            errors.addElement("high_res field is not an Integer");

                        comment = ((Hashtable)as).get("comment");
                        if( comment != null ) {
                            if( !(comment instanceof String) )
                                errors.addElement("comment field is not a String");
                            else if( ((String)comment).length() > 64 )
                                errors.addElement("Address space's comment may not be longer than 64 characters");
                        }

                        Object as_dhcp_options = ((Hashtable)as).get("dhcp_options");
                        validateDHCPOptions(as_dhcp_options, errors);

                        Object as_groups = ((Hashtable)as).get("groups");
                        if( as_groups != null ) {
                            if( !(as_groups instanceof Hashtable) )
                                errors.addElement("groups field is not a Hashtable");
                            else {
                                Enumeration enm1 = ((Hashtable)as_groups).keys();
                                while( enm1.hasMoreElements() ) {
                                    Object group = enm1.nextElement();
                                    if( !(group instanceof String) )
                                        errors.addElement("group field is not a String");
                                    else if( ((String)group).length() > 32 )
                                        errors.addElement("group field is too long");
                                }
                            }
                        }

                        Object dhcp_addresses = ((Hashtable)as).get("dhcp_addresses");
                        if( dhcp_addresses != null ) {
                            if( !(dhcp_addresses instanceof Hashtable) )
                                errors.addElement("dhcp_addresses field is not a Hashtable");
                            else {
                                Enumeration enm1 = ((Hashtable)dhcp_addresses).keys();
                                while( enm1.hasMoreElements() ) {
                                    Object dhcpKey = enm1.nextElement();
                                    if( !(dhcpKey instanceof String) ) {
                                        errors.addElement("dhcp_addresses key is not a String");
                                        continue;
                                    }
                                    Object dhcp_addr = ((Hashtable)dhcp_addresses).get(dhcpKey);
                                    if( !(dhcp_addr instanceof Hashtable) ) {
                                        errors.addElement("dhcp_address field is not a Hashtable");
                                        continue;
                                    }

                                    Object active = ((Hashtable)dhcp_addr).get("active?");
                                    if( active == null )
                                        errors.addElement("active? field does not exist");
                                    else if( !(active instanceof Boolean) )
                                        errors.addElement("active? field is not a Boolean");

                                    Object ip_address = ((Hashtable)dhcp_addr).get("ip_address");
                                    if( ip_address == null )
                                        errors.addElement("ip_address field does not exist");
                                    else if( !(ip_address instanceof String) )
                                        errors.addElement("ip_address field is not a String");

                                    Object dhcp_names = ((Hashtable)dhcp_addr).get("names");
                                    if( dhcp_names == null ) {
                                        if( obj.get("ips do not have to have names") == null )
                                            errors.addElement("dhcp address "+ip_address+"does not have a name");
                                    } else {
                                        if( !(dhcp_names instanceof Hashtable) )
                                            errors.addElement("dhcp name field is not a Hashtable");
                                        else {
                                            if( ((Hashtable)dhcp_names).size() == 0 ) {
                                                if( obj.get("ips do not have to have names") == null )
                                                    errors.addElement("dhcp ip address "+ip_address+" does not have a name");
                                            } else if( ((Hashtable)dhcp_names).size() > 1 )
                                                errors.addElement("dhcp ip address "+ip_address+" has more than one name");
                                            else {
                                                Object dhcp_name = ((Hashtable)dhcp_names).get((((Hashtable)dhcp_names).keys()).nextElement());
                                                if( !(dhcp_name instanceof Hashtable) )
                                                    errors.addElement("dhcp ip address "+ip_address+" does not have a name");
                                                else {
                                                    Object strName = ((Hashtable)dhcp_name).get("name");
                                                    Object strDomain = ((Hashtable)dhcp_name).get("domain");
                                                    if( strName == null || strDomain == null ||
                                                        !(strName instanceof String) || !(strDomain instanceof String) ) {
                                                        if( obj.get("ips do not have to have names") == null )
                                                            errors.addElement("dhcp ip address "+ip_address+" does not have a name");
                                                    } else {
                                                        if( ((String)strName).trim().length() > 256 )
                                                            errors.addElement("name "+strName+" is too long");
                                                        if( ((String)strDomain).trim().length() > 256 )
                                                            errors.addElement("name "+strDomain+" is too long");
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            break;

        case Constants.GROUP:
            Object name = obj.get("name");
            if( name == null )
                errors.addElement("name field is null");
            else if( !(name instanceof String) )
                errors.addElement("name field is not a String");
            else if( ((String)name).length() > 32 )
                errors.addElement("Group's name may not be longer than 32 characters");

            comment = obj.get("comment");
            if( comment != null ) {
                if( !(comment instanceof String) )
                    errors.addElement("comment field is not a String");
                else if( ((String)comment).length() > 256 )
                    errors.addElement("Comment may not be longer than 256 characters");
            }
            break;

        case Constants.DOMAIN:
            Object domain = obj.get("domain");
            if( domain == null )
                errors.addElement("domain field is null");
            else if( !(domain instanceof String) )
                errors.addElement("domain field is not a String");
            else if( ((String)domain).trim().length() == 0 )
                errors.addElement("domain must have name");

            comment = obj.get("comment");
            if( comment != null && !(comment instanceof String) )
                errors.addElement("comment field is not a String");

            Object limited = obj.get("limited?");
            if( limited == null )
                errors.addElement("limited? field is null");
            else if( !(limited instanceof Boolean) )
                errors.addElement("limited? field is not a Boolean");

            groups = obj.get("groups");
            if( groups == null )
                errors.addElement("groups field is null");
            else if( !(groups instanceof Hashtable) )
                errors.addElement("groups field is not a Hashtable");
            else if( ((Hashtable)groups).size() == 0 )
                errors.addElement("domain must have at least one group");
            else {
                Enumeration enm = ((Hashtable)groups).keys();
                while( enm.hasMoreElements() ) {
                    Object group = enm.nextElement();
                    if( !(group instanceof String) )
                        errors.addElement("group field is not a String");
                    else if( ((String)group).length() > 32 )
                        errors.addElement("group field is too long");
                }
            }
            break;

        case Constants.ATEAM:
            name = obj.get("name");
            if( name == null )
                errors.addElement("Admin Team must have a name");
            else if( !(name instanceof String) )
                errors.addElement("name field is not a String");
            else if( ((String)name).trim().length() == 0 )
                errors.addElement("Admin Team must have a name");
            else if( ((String)name).trim().length() > 64 )
                errors.addElement("Admin Team's name may not be longer than 64 characters");

            Object email = obj.get("email");
            if( email == null )
                errors.addElement("Admin Team must have an email");
            else if( !(email instanceof String) )
                errors.addElement("email field is not a String");
            else if( ((String)email).trim().length() == 0 )
                errors.addElement("Admin Team must have an email");
            else if( ((String)email).trim().length() > 256 )
                errors.addElement("Admin Team's email may not be longer than 256 characters");

            Object phone = obj.get("phone");
            if( phone == null )
                errors.addElement("Admin Team must have a phone");
            else if( !(phone instanceof String) )
                errors.addElement("phone field is not a String");
            else if( ((String)phone).trim().length() == 0 )
                errors.addElement("Admin Team must have a phone");
            else if( ((String)phone).trim().length() > 32 )
                errors.addElement("Admin Team's phone number may not be longer than 32 characters");

            groups = obj.get("groups");
            if( groups == null )
                errors.addElement("Admin Team must have at least one group");
            else if( !(groups instanceof Hashtable) )
                errors.addElement("groups field is not a Hashtable");
            else if( ((Hashtable)groups).size() == 0 )
                errors.addElement("Admin Team must have at least one group");

            comment = obj.get("comment");
            if( comment != null && !(comment instanceof String) )
                errors.addElement("'comment' field is not a String");
            else if( comment != null && ((String)comment).length() > 256 )
                errors.addElement("'comment' field may not be longer than 256 characters");
            break;

        case Constants.OS:
            name = obj.get("name");
            if( name == null )
                errors.addElement("name field is null");
            else if( !(name instanceof String) )
                errors.addElement("name field is not a String");
            else if( ((String)name).trim().length() == 0 )
                errors.addElement("name field is an empty String");
            else if( ((String)name).trim().length() > 32 )
                errors.addElement("os name may not be longer than 32 characters");

            comment = obj.get("comment");
            if( comment != null ) {
                if( !(comment instanceof String) )
                    errors.addElement("comment field is not a String");
                else if( ((String)comment).trim().length() > 64 )
                    errors.addElement("os comment may not be longer than 64 characters");
            }
            break;

        case Constants.MAKE:
            name = obj.get("name");
            if( name == null )
                errors.addElement("name field is null");
            else if( !(name instanceof String) )
                errors.addElement("name field is not a String");
            else if( ((String)name).trim().length() == 0 )
                errors.addElement("name field is an empty String");
            else if( ((String)name).trim().length() > 32 )
                errors.addElement("make may not be longer than 32 characters");
            break;

        case Constants.MODEL:
            name = obj.get("model");
            if( name == null )
                errors.addElement("model field is null");
            else if( !(name instanceof String) )
                errors.addElement("model field is not a String");
            else if( ((String)name).trim().length() == 0 )
                errors.addElement("Model name is missing");
            else if( ((String)name).trim().length() > 32  )
                errors.addElement("model may not be longer them 32 characters");

            make = obj.get("make");
            if( make == null )
                errors.addElement("make field is null");
            else if( !(make instanceof String) )
                errors.addElement("make field is not a String");
            else if( ((String)make).trim().length() == 0 )
                errors.addElement("Model must have a make");
            break;

        case Constants.LOCATION:
            location = obj.get("location");
            if( location == null )
                errors.addElement("location field is null");
            else if( !(location instanceof String) )
                errors.addElement("location field is not a String");
            else if( ((String)location).trim().length() == 0 )
                errors.addElement("Location is a mandatory field");
            else if( ((String)location).trim().length() > 64 )
                errors.addElement("Location may not be longer than 64 characters");

            Object site = obj.get("site");
            if( site != null ) {
                if( !(site instanceof String) )
                    errors.addElement("site field is not a String");
                else if( ((String)site).trim().length() > 16 )
                errors.addElement("Site may not be longer than 16 characters");
            }

            Object code = obj.get("code");
            if( code != null ) {
                if( !(code instanceof String) )
                    errors.addElement("code field is not a String");
                else if( ((String)code).trim().length() > 16 )
                errors.addElement("Code may not be longer than 16 characters");
            }
            break;

        case Constants.STATE:
            name = obj.get("name");
            if( name == null )
                errors.addElement("name field is null");
            else if( !(name instanceof String) )
                errors.addElement("name field is not a String");
            else if( (((String)name).trim()).length() == 0 )
                errors.addElement("name field cannot be empty");
            else if( ((String)name).length() > 32 )
                errors.addElement("State's name may not be longer than 32 characters");

            Object rank = obj.get("rank");
            if( rank == null )
                errors.addElement("State must have a rank.");
            else if( !(rank instanceof Integer) )
                errors.addElement("rank field is not an Integer");

            comment = obj.get("comment");
            if( comment != null ) {
                if( !(comment instanceof String) )
                    errors.addElement("comment field is not a String");
                else if( ((String)comment).length() > 64 )
                    errors.addElement("Comment may not be longer than 64 characters");
            }

            Object access = obj.get("access");
            if( access != null &&  !(access instanceof String) )
                errors.addElement("Access field is not a String");
            break;

        case Constants.USER:
            if( !type.equals("user") )
                errors.addElement("object field is not a 'user'");

            Object netID = obj.get("NetID");
            if( netID == null )
                errors.addElement("User must have NetID");
            else if( !(netID instanceof String) )
                errors.addElement("NetID field is not a String");
            else if( ((String)netID).trim().length() == 0 )
                errors.addElement("User must have NetID");

            Object regID = obj.get("RegID");
            if( regID == null )
                errors.addElement("User must have RegID");
            else if( !(regID instanceof String) )
                errors.addElement("RegID field is not a String");
            else if( ((String)regID).trim().length() == 0 )
                errors.addElement("User must have RegID");

            Object active = obj.get("active?");
            if( active == null )
                errors.addElement("active? field is null");
            else if( !(active instanceof Boolean) )
                errors.addElement("active? field is not a Boolean");

            Object all_objects = obj.get("all_objects?");
            if( all_objects == null )
                errors.addElement("all_objects? field is null");
            else if( !(all_objects instanceof Boolean) )
                errors.addElement("all_objects? field is not a Boolean");

            Object all_groups = obj.get("all_groups?");
            if( all_groups == null )
                errors.addElement("all_groups? field is null");
            else if( !(all_groups instanceof Boolean) )
                errors.addElement("all_groups? field is not a Boolean");

            Object defDomain = obj.get("default_domain");
            if( defDomain == null )
                errors.addElement("default_domain field is null");
            else if( !(defDomain instanceof String) )
                errors.addElement("default_domain field is not a String");
            else if( ((String)defDomain).trim().length() == 0 )
                errors.addElement("default_domain field is an empty String");

            Object defGroup = obj.get("default_group");
            if( defGroup == null )
                errors.addElement("User must have default group");
            else if( !(defGroup instanceof String) )
                errors.addElement("default_group field is not a String");
            else if( ((String)defGroup).trim().length() == 0 )
                errors.addElement("User must have default group");

            Object departmentsObj = obj.get("departments");
            if( departmentsObj != null ) {
                if( departmentsObj instanceof Hashtable ) {
                    Hashtable departments = (Hashtable)departmentsObj;
                    Enumeration enm = departments.keys();
                    while( enm.hasMoreElements() ) {
                        Object keyObj = enm.nextElement();
                        if( !(keyObj instanceof String) ) {
                            errors.addElement("departments hash is not valid");
                            break;
                        }
                        Object depObj = departments.get((String)keyObj);
                        if( !(depObj instanceof Hashtable) ) {
                            errors.addElement("departments hash is not valid");
                            break;
                        }
                        Hashtable depH = (Hashtable)depObj;
                        Object official = depH.get("official?");
                        if( official == null || !(official instanceof Boolean) ) {
                            errors.addElement("departments hash is not valid");
                            break;
                        }
                    }
                } else
                    errors.addElement("departments field is not a Hashtable");
            }

            break;

        case Constants.DEPARTMENT:
            if( !type.equals("department") )
                errors.addElement("object field is not a 'department'");

            name = obj.get("name");
            if( name == null )
                errors.addElement("name field is null");
            else if( !(name instanceof String) )
                errors.addElement("name field is not a String");
            else if( ((String)name).trim().length() == 0 )
                errors.addElement("name field is an empty String");
            else if( ((String)name).trim().length()> 64 )
                errors.addElement("Department name may not be longer than 64 characters.");

            Hashtable cons = (Hashtable)obj.get("consultants");
            if( cons != null ) {
                if( !(cons instanceof Hashtable) ) {
                    errors.addElement("consultants field is not a Hashtable");
                    break;
                }
                if( cons.size() > 0 ) {
                    Enumeration enm = cons.keys();
                    boolean has_primary = false;
                    while( enm.hasMoreElements() ) {
                        String key = (String)enm.nextElement();
                        Object person = cons.get(key);
                        if( !(person instanceof Hashtable) ) {
                            errors.addElement("person field is not a Hashtable");
                            break;
                        }
                        regID = ((Hashtable)person).get("RegID");
                        if( !(regID instanceof String) )
                            errors.addElement("RegID field is not a String");
                        else if( ((String)regID).trim().length() == 0 )
                            errors.addElement("RegID field is an empty String");

                        Object primary = ((Hashtable)person).get("primary?");
                        if( !(primary instanceof Boolean) )
                            errors.addElement("primary? field is not a Boolean");
                        else if( ((Boolean)primary).booleanValue() ) {
                            if( has_primary )
                                errors.addElement("there is more than one primary consultant");
                            else
                                has_primary = true;
                        }
                    }
                    if( !has_primary )
                        errors.addElement("you must select one primary consultant");
                }
            }
            break;

        default:
            throw new Exception("Invalid what passed to Utils.validate(...)");
        }

        if( errors.size() > 0 ) {
            if( what == Constants.MAKE ) {
                String msg = (String)errors.elementAt(0);
                errors = new Vector();
                errors.addElement("");
                errors.addElement(msg);
            }
            throw new NetDBException("Not a valid object. See attached error Vector for details.",
                                     errors);
        }
    } // end of validate(...)

    /** logs debug message */
    public static void debug(String msg) {
        doLog(Constants.DEBUG, msg);
    }

    /** logs debug message */
    public static void debug(Object msg) {
        doLog(Constants.DEBUG, msg.toString());
    }

    /** logs warning message */
    public static void warn(String msg) {
        doLog(Constants.WARN, msg);
    }

    /** logs warning message */
    public static void warn(Object msg) {
        doLog(Constants.WARN, msg.toString());
    }

    /** logs error message */
    public static void error(String msg) {
        doLog(Constants.ERROR, msg);
    }

    /** logs error message */
    public static void error(Object msg) {
        doLog(Constants.ERROR, msg.toString());
    }

    /** logs info message */
    public static void info(String msg) {
        doLog(Constants.INFO, msg);
    }

    /** logs info message */
    public static void info(Object msg) {
        doLog(Constants.INFO, msg.toString());
    }

    /** Does actuall logging */
    private static void doLog(int level, String msg) {
        StackTraceElement[] st = (new Throwable()).getStackTrace();
        switch( level ) {
        case Constants.DEBUG:
            logger.debug("<<<"+st[2].getClassName()+"::"+st[2].getMethodName()+"(...) "+
                               st[2].getLineNumber()+">>> "+msg);
            break;

        case Constants.WARN:
            logger.warn("<<<"+st[2].getClassName()+"::"+st[2].getMethodName()+"(...) "+
                               st[2].getLineNumber()+">>> "+msg);
            break;

        case Constants.ERROR:
            logger.error("<<<"+st[2].getClassName()+"::"+st[2].getMethodName()+"(...) "+
                               st[2].getLineNumber()+">>> "+msg);
            break;

        case Constants.INFO:
            logger.info("<<<"+st[2].getClassName()+"::"+st[2].getMethodName()+"(...) "+
                               st[2].getLineNumber()+">>> "+msg);
            break;
        }
    }

    public static void doLog(int level, int depth, String msg) {
        StackTraceElement[] st = (new Throwable()).getStackTrace();
        if( depth >= st.length )
            depth = st.length - 1;
        switch( level ) {
        case Constants.DEBUG:
            logger.debug("<<<"+st[depth].getClassName()+"::"+st[depth].getMethodName()+"(...) "+
                               st[depth].getLineNumber()+">>> "+msg);
            break;

        case Constants.WARN:
            logger.warn("<<<"+st[depth].getClassName()+"::"+st[depth].getMethodName()+"(...) "+
                               st[depth].getLineNumber()+">>> "+msg);
            break;

        case Constants.ERROR:
            logger.error("<<<"+st[depth].getClassName()+"::"+st[depth].getMethodName()+"(...) "+
                               st[depth].getLineNumber()+">>> "+msg);
            break;

        case Constants.INFO:
            logger.info("<<<"+st[depth].getClassName()+"::"+st[depth].getMethodName()+"(...) "+
                               st[depth].getLineNumber()+">>> "+msg);
            break;
        }
    }

    public static String obj2xml(Hashtable tab) {
        Object obj = tab.get("object");
        if( !(obj instanceof String) )
            return "";
        String root = obj.toString().trim();
        Comparator cmp = new Comparator() {
                public int compare(Object o1, Object o2) {
                    return (o1.toString()).compareToIgnoreCase(o2.toString());
                }
                public boolean equals(Object obj) {
                    return true;
                }
            };

        if( root.equals("node") )
            return "<"+root+">\n"+node2xml(tab, cmp, "  ")+"</"+root+">";
        else
            return "";
    }

    private static String node2xml(Hashtable tab, Comparator cmp, String depth) {
        String xmlStr = "";
        java.util.List list = Arrays.asList(tab.keySet().toArray());
        Collections.sort(list, cmp);
        Object[] keys = list.toArray();
        for( int i = 0; i < keys.length; i++ ) {
            Object val = tab.get(keys[i]);
            if( val instanceof Hashtable ) {

                if( depth.length() == 2 && keys[i].toString().equals("object") ) {
                    // skip
                } else if( depth.length() == 2 && keys[i].toString().equals("groups") ) {
                    xmlStr += depth+"<"+keys[i]+">\n";
                    java.util.List list2 = Arrays.asList(((Hashtable)val).keySet().toArray());
                    Collections.sort(list2, cmp);
                    Object[] groups = list2.toArray();
                    for( int j = 0; j < groups.length; j++ ) {
                        xmlStr += depth+"  <group>"+groups[j]+"</group>\n";
                    }
                    xmlStr += depth+"</"+keys[i]+">\n";

                } else if( depth.length() == 2 && keys[i].toString().equals("oses") ) {
                    xmlStr += depth+"<"+keys[i]+">\n";
                    java.util.List list2 = Arrays.asList(((Hashtable)val).keySet().toArray());
                    Collections.sort(list2, cmp);
                    Object[] oses = list2.toArray();
                    for( int j = 0; j < oses.length; j++ ) {
                        xmlStr += depth+"  <os>"+oses[j]+"</os>\n";
                    }
                    xmlStr += depth+"</"+keys[i]+">\n";

                } else if( depth.length() == 2 &&
                           ( keys[i].toString().equals("created.by") || keys[i].toString().equals("mod.by") ) ) {
                    xmlStr += depth+"<"+keys[i];
                    Hashtable person = (Hashtable)val;
                    String name = (String)person.get("name");
                    String netid = (String)person.get("NetID");
                    String dirid = (String)person.get("DirID");
                    if( name != null )
                        xmlStr += " name=\""+name+"\"";
                    if( netid != null )
                        xmlStr += " netid=\""+netid+"\"";
                    if( dirid != null )
                        xmlStr += " dirid=\""+dirid.toLowerCase()+"\"";
                    xmlStr += "/>\n";

                } else if( depth.length() == 2 && keys[i].toString().equals("users") ) {
                    xmlStr += depth+"<"+keys[i]+">\n"+users2xml((Hashtable)val, cmp, depth+"  ")+depth+"</"+keys[i]+">\n";

                } else if( depth.length() == 2 && keys[i].toString().equals("admins") ) {
                    Hashtable ateams = Utils.nHash(((Hashtable)val).get("ateams"));
                    Hashtable admins = Utils.nHash(((Hashtable)val).get("admins"));
                    xmlStr += depth+"<"+keys[i]+">\n";
                    if( admins.size() > 0 )
                        xmlStr += depth+users2xml(admins, cmp, depth+"  ");
                    if( ateams.size() > 0 )
                        xmlStr += depth+ateams2xml(ateams, cmp, depth+"  ");
                    xmlStr += depth+"</"+keys[i]+">\n";

                } else if( depth.length() == 2 && keys[i].toString().equals("types") ) {
                    xmlStr += depth+"<"+keys[i]+">\n";
                    java.util.List list2 = Arrays.asList(((Hashtable)val).keySet().toArray());
                    Collections.sort(list2, cmp);
                    Object[] types = list2.toArray();
                    for( int j = 0; j < types.length; j++ ) {
                        xmlStr += depth+"  <type>"+types[j]+"</type>\n";
                    }
                    xmlStr += depth+"</"+keys[i]+">\n";

                } else if( depth.length() == 2 && keys[i].toString().equals("tags") ) {
                    xmlStr += depth+"<"+keys[i]+">\n";
                    java.util.List list2 = Arrays.asList(((Hashtable)val).keySet().toArray());
                    Collections.sort(list2, cmp);
                    Object[] tags = list2.toArray();
                    for( int j = 0; j < tags.length; j++ ) {
                        xmlStr += depth+"  <tag name=\""+tags[j]+"\" value=\""+((Hashtable)val).get(tags[j])+"\"/>\n";
                    }
                    xmlStr += depth+"</"+keys[i]+">\n";

                } else if( depth.length() == 2 && keys[i].toString().equals("interfaces") ) {
                    xmlStr += depth+"<"+keys[i]+">\n"+interfaces2xml((Hashtable)val, cmp, depth+"  ")+depth+"</"+keys[i]+">\n";

                } else if( depth.length() == 2 && keys[i].toString().equals("names") ) {
                    xmlStr += depth+"<"+keys[i]+">\n"+names2xml((Hashtable)val, cmp, depth+"  ", "name")+depth+"</"+keys[i]+">\n";

                } else if( depth.length() == 2 && keys[i].toString().equals("ipc_addresses") ) {
                    xmlStr += depth+"<"+keys[i]+">\n"+ips2xml((Hashtable)val, cmp, depth+"  ")+depth+"</"+keys[i]+">\n";

                } else
                    xmlStr += depth+"<"+keys[i]+">\n"+node2xml((Hashtable)val, cmp, depth+"  ")+depth+"</"+keys[i]+">\n";
            } else
                xmlStr += depth+"<"+keys[i]+">"+val.toString().trim()+"</"+keys[i]+">\n";
        }
        return xmlStr;
    }

    private static String ateams2xml(Hashtable tab, Comparator cmp, String depth) {
        String xmlStr = "";
        java.util.List list = Arrays.asList(tab.keySet().toArray());
        Collections.sort(list, cmp);
        Object[] keys = list.toArray();
        for( int i = 0; i < keys.length; i++ ) {
            Hashtable ateam = (Hashtable)tab.get(keys[i]);
            String email = (String)ateam.get("email");
            String phone = (String)ateam.get("phone");
            xmlStr += depth+"<ateam name=\""+keys[i]+"\"";
            if( email != null )
                xmlStr += " email=\""+email+"\"";
            if( phone != null )
                xmlStr += " phone=\""+phone+"\"";
            xmlStr += "/>\n";
        }
        return xmlStr;
    }
    private static String users2xml(Hashtable tab, Comparator cmp, String depth) {
        String xmlStr = "";
        java.util.List list = Arrays.asList(tab.keySet().toArray());
        Collections.sort(list, cmp);
        Object[] keys = list.toArray();
        for( int i = 0; i < keys.length; i++ ) {
            Hashtable person = (Hashtable)tab.get(keys[i]);
            String name = (String)person.get("name");
            String netid = (String)person.get("NetID");
            String dirid = (String)person.get("DirID");
            xmlStr += depth+"<person";
            if( name != null )
                xmlStr += " name=\""+name+"\"";
            if( netid != null )
                xmlStr += " netid=\""+netid+"\"";
            if( dirid != null )
                xmlStr += " dirid=\""+dirid.toLowerCase()+"\"";
            xmlStr += "/>\n";
        }
        return xmlStr;
    }

    private static String names2xml(Hashtable tab, Comparator cmp, String depth, String nameType) {
        String xmlStr = "";
        java.util.List list = Arrays.asList(tab.keySet().toArray());
        Collections.sort(list, cmp);
        Object[] keys = list.toArray();
        for( int i = 0; i < keys.length; i++ ) {
            Hashtable name = (Hashtable)tab.get(keys[i]);
            if( nameType.equals("alias") )
                xmlStr += depth+"<"+nameType+" fqdn=\""+name.get("name")+"."+name.get("domain")+"\"/>\n";
            else if( nameType.equals("mx") )
                xmlStr += depth+"<"+nameType+" fqdn=\""+name.get("name")+"."+name.get("domain")+"\" pref=\""+name.get("pref")+"\"/>\n";
            else {
                xmlStr += depth+"<"+nameType+" fqdn=\""+name.get("name")+"."+name.get("domain")+"\">";
                Hashtable aliases = nHash(name.get("aliases"));
                Hashtable mxs = nHash(name.get("mxs"));
                if( aliases.size() > 0 || mxs.size() > 0 )
                    xmlStr += "\n";
                if( aliases.size() > 0 )
                    xmlStr += names2xml(aliases, cmp, depth+"  ", "alias");
                if( mxs.size() > 0 )
                    xmlStr += names2xml(mxs, cmp, depth+"  ", "mx");
                if( aliases.size() > 0 || mxs.size() > 0 )
                    xmlStr += depth;
                xmlStr += "</"+nameType+">\n";
            }

        }
        return xmlStr;
    }

    private static String interfaces2xml(Hashtable tab, Comparator cmp, String depth) {
        String xmlStr = "";
        Comparator number_cmp = new Comparator() {
                public int compare(Object o1, Object o2) {
                    return (int)((new Long((String)o1)).longValue() -
                                 (new Long((String)o2)).longValue());
                }
                public boolean equals(Object obj) {
                    return true;
                }
            };
        java.util.List list = Arrays.asList(tab.keySet().toArray());
        Collections.sort(list, number_cmp);
        Object[] keys = list.toArray();
        for( int i = 0; i < keys.length; i++ ) {
            Hashtable intfc = (Hashtable)tab.get(keys[i]);
            xmlStr += depth+"<interface id=\""+keys[i]+"\"";
            String hw = (String)intfc.get("hw");
            if( hw != null && hw.trim().length() > 0 ) {
                hw = hw.trim().toLowerCase();
                xmlStr += " hw=\""+hw.substring(0, 4)+"."+hw.substring(4, 8)+"."+hw.substring(8)+"\"";
            }
            xmlStr += " dhcp=\""+intfc.get("use-dhcp?")+
                "\" roaming=\""+intfc.get("roaming?")+"\">\n"+intfc2xml(intfc, cmp, depth+"  ")+depth+"</interface>\n";

        }
        return xmlStr;
    }

    private static String intfc2xml(Hashtable tab, Comparator cmp, String depth) {
        String xmlStr = "";
        java.util.List list = Arrays.asList(tab.keySet().toArray());
        Collections.sort(list, cmp);
        Object[] keys = list.toArray();
        for( int i = 0; i < keys.length; i++ ) {
            if( keys[i].toString().equals("ips") )
                xmlStr += depth+"<"+keys[i]+">\n"+ips2xml((Hashtable)tab.get(keys[i]), cmp, depth+"  ")+depth+"</"+keys[i]+">\n";
            if( keys[i].toString().equals("names") ) {
                Hashtable names = Utils.nHash(tab.get("names"));
                if( names.size() > 0 )
                    xmlStr += names2xml(names, cmp, depth, "name");
            }
        }
        return xmlStr;
    }

    private static String ips2xml(Hashtable tab, Comparator cmp, String depth) {
        String xmlStr = "";
        Comparator ip_cmp = new Comparator() {
                public int compare(Object o1, Object o2) {
                    return Utils.ipstr2num((String)o1) - Utils.ipstr2num((String)o2);
                }
                public boolean equals(Object obj) {
                    return true;
                }
            };
        java.util.List list = Arrays.asList(tab.keySet().toArray());
        Collections.sort(list, ip_cmp);
        Object[] ips = list.toArray();
        for( int i = 0; i < ips.length; i++ ) {
            Hashtable ip = (Hashtable)tab.get(ips[i]);
            xmlStr += depth+"<ip addr=\""+ips[i]+"\" active=\""+ip.get("active?")+"\">";
            Hashtable names = (Hashtable)ip.get("names");
            if( names != null && names.size() > 0 ) {
                xmlStr += "\n"+names2xml(names, cmp, depth+"  ", "name")+depth;
            }
            xmlStr += "</ip>\n";
        }
        return xmlStr;
    }

    /**
     * Produces java code nessesary to run in order to make given hashtable.
     * Note that index must start from 0.
     */
    public static java.io.PrintStream printHash2Java(Hashtable tab) {
        printHash2Java(0, tab);
        return System.out;
    }

    /**
     * Produces java code nessesary to run in order to make given hashtable.
     * Note that index must start from 0.
     */
    private static int printHash2Java(int index, Hashtable tab) {
        if( index == 0 )
            System.out.println("Hashtable hash"+index+" = new Hashtable();");

        int next_index = index + 1;
        Enumeration keys = tab.keys();
        while( keys.hasMoreElements() ) {
            Object key = keys.nextElement();
            String value = "";

            Object obj = tab.get(key);
            if( obj instanceof Hashtable ) {
                System.out.println("Hashtable hash"+next_index+" = new Hashtable();");
                value = "hash"+next_index;
                next_index = printHash2Java(next_index, (Hashtable)obj);

            } else if( obj instanceof Vector ) {
                System.out.println("Vector vec"+next_index+" = new Vector();");
                value = "vec"+next_index;
                next_index = printVector2Java(next_index, (Vector)obj);

            } else if( obj instanceof String )
                value = "\""+obj+"\"";

            else if( obj instanceof Integer )
                value = "new Integer(\""+obj+"\")";

            else if( obj instanceof java.sql.Timestamp )
                value = "new java.sql.Timestamp(Long.parseLong(\""+((java.sql.Timestamp)obj).getTime()+"\"))";

            else if( obj instanceof Boolean ) {
                if( ((Boolean)obj).booleanValue() )
                    value = "Boolean.TRUE";
                else
                    value = "Boolean.FALSE";
            } else
                System.out.println("ERR: value ["+obj+"] is of unsupported type "+
                                   obj.getClass().getName());

            if( key instanceof String )
                System.out.println("hash"+index+".put(\""+key+"\", "+value+");");
            else
                System.out.println("ERR: key ["+key+"] is not a String");

        }
        return next_index;
    }

    private static int printVector2Java(int index, Vector vec) {
        int next_index = index + 1;
        for( int i = 0; i < vec.size(); i++ ) {
            String value = "";
            Object cell = vec.elementAt(i);

            if( cell instanceof Hashtable ) {
                next_index = printHash2Java(next_index, (Hashtable)cell);

            } else if( cell instanceof Vector )
                next_index = printVector2Java(next_index, (Vector)cell);

            else if( cell instanceof String )
                value = "\""+cell+"\"";

            else if( cell instanceof Integer )
                value = "new Integer(\""+cell+"\")";

            else if( cell instanceof java.sql.Timestamp )
                value = "new java.sql.Timestamp(Long.parseLong(\""+((java.sql.Timestamp)cell).getTime()+"\"))";

            else if( cell instanceof Boolean ) {
                if( ((Boolean)cell).booleanValue() )
                    value = "Boolean.TRUE";
                else
                    value = "Boolean.FALSE";
            } else
                System.out.println("ERR: value ["+cell+"] is of unsupported type "+
                                   cell.getClass().getName());

            System.out.println("vec"+index+".addElement("+value+");");
        }
        return next_index;
    }

    /** This is pretty print for Hashtables, where output goes to System.out.
     * For convinience it returns System.out itself which you can use to do something like
     * Utils.printHash(myHash).println("\n--- end of myHash ---");
     */
    public static java.io.PrintStream printHash(Hashtable tab) {
        return printHash("", tab);
    }

    private static java.io.PrintStream printHash(String offset, Hashtable tab) {
        Enumeration keys = tab.keys();
        while( keys.hasMoreElements() ) {
            Object key = keys.nextElement();
            System.out.print(offset+"<"+key+">");
            Object obj = tab.get(key);
            if( obj instanceof Hashtable ) {
                System.out.println("");
                printHash(offset+"  ", (Hashtable)obj);
            } else if( obj instanceof Vector ) {
                System.out.print(" -> ");
                printVector(offset, (Vector)obj);
            } else System.out.println(" -> ["+obj.toString().trim()+"]");
        }
        if( tab.size() == 0 )
            System.out.println(offset+"Null");

        return System.out;
    }

    public static void printVector(String offset, Vector vec) {
        for( int i = 0; i < vec.size(); i++ ) {
            Object cell = vec.elementAt(i);
            if( cell instanceof Hashtable ) {
                System.out.println("[:vector element:]");
                printHash(offset+"  ", (Hashtable)cell);
                System.out.println("[:end vector element:]");
            } else if( cell instanceof Vector )
                printVector(offset, (Vector)cell);
            else
                System.out.print("["+cell+"]");
        }
        System.out.println("");
    }

    /**
     * creates actual clone/copy of the hashtable where all the
     * key and values are cloned. We assume that all the keys are
     * Strings and the values are one of the following types:
     * {String, Integer, Long, Boolean, java.sql.Timestamp, Null, Hashtable}
     * Note, we do not clone/copy Vectors, ResultSetMetaData which
     * you can see in a code been put into Hashtable.
     */
    public static Hashtable copyHash(Hashtable hash) throws Exception {
        if( hash.size() == 0 )
            return new Hashtable();
        Hashtable copy = new Hashtable();
        Enumeration keys = hash.keys();
        while( keys.hasMoreElements() ) {
            String key = (String)keys.nextElement();
            Object value = hash.get(key);
            if(       value instanceof String )
                copy.put(new String(key), new String((String)value));
            else if ( value instanceof IPString )
                copy.put(new String(key), new IPString((IPString)value));
            else if ( value instanceof Integer )
                copy.put(new String(key), new Integer(((Integer)value).intValue()));
            else if ( value instanceof Long )
                copy.put(new String(key), new Long(((Long)value).longValue()));
            else if ( value instanceof Boolean )
                copy.put(new String(key), new Boolean(((Boolean)value).booleanValue()));
            else if ( value instanceof java.sql.Timestamp )
                copy.put(new String(key), (java.sql.Timestamp)((java.sql.Timestamp)value).clone());
            else if ( value instanceof Null )
                copy.put(new String(key), new Null(((Null)value).getType()));
            else if ( value instanceof Hashtable )
                copy.put(new String(key), copyHash((Hashtable)value));
            else if ( value instanceof Vector )
                copy.put(new String(key), copyVector((Vector)value));
            else
                throw new Exception("Unsupported object type to clone ["+value.getClass().getName()+"] ["+value+"] key ["+key+"]");
        }

       return copy;
    } // end of copyHash(...)

    private static Vector copyVector(Vector vec) throws Exception {
        Vector copy = new Vector();
        for( int i = 0; i < vec.size(); i++ ) {
            Object obj = vec.elementAt(i);
            if( obj instanceof String )
                copy.addElement(new String((String)obj));
            else if ( obj instanceof IPString )
                copy.addElement(new IPString((IPString)obj));
            else if ( obj instanceof Integer )
                copy.addElement(new Integer(((Integer)obj).intValue()));
            else if ( obj instanceof Long )
                copy.addElement(new Long(((Long)obj).longValue()));
            else if ( obj instanceof Boolean )
                copy.addElement(new Boolean(((Boolean)obj).booleanValue()));
            else if ( obj instanceof java.sql.Timestamp )
                copy.addElement((java.sql.Timestamp)((java.sql.Timestamp)obj).clone());
            else if ( obj instanceof Null )
                copy.addElement(new Null(((Null)obj).getType()));
            else if ( obj instanceof Hashtable )
                copy.addElement(copyHash((Hashtable)obj));
            else if ( obj instanceof Vector )
                copy.addElement(copyVector((Vector)obj));
            else
                throw new Exception("Unsupported object type to clone ["+obj.getClass().getName()+"] ["+obj+"]");
        }
        return copy;
    } // end of copyVector(...)

    /**
     * Will append copied content of the hash2 to a hash1.
     *
     * @return hash1
     */
    public static Hashtable appendHash(Hashtable hash1, Hashtable hash2) throws Exception {
        if( hash2.size() == 0 )
            return hash1;
        Enumeration keys = hash2.keys();
        while( keys.hasMoreElements() ) {
            String key = (String)keys.nextElement();
            Object value = hash2.get(key);
            if(       value instanceof String )
                hash1.put(new String(key), new String((String)value));
            else if ( value instanceof Integer )
                hash1.put(new String(key), new Integer(((Integer)value).intValue()));
            else if ( value instanceof Long )
                hash1.put(new String(key), new Long(((Long)value).longValue()));
            else if ( value instanceof Boolean )
                hash1.put(new String(key), new Boolean(((Boolean)value).booleanValue()));
            else if ( value instanceof java.sql.Timestamp )
                hash1.put(new String(key), (java.sql.Timestamp)((java.sql.Timestamp)value).clone());
            else if ( value instanceof Null )
                hash1.put(new String(key), new Null(((Null)value).getType()));
            else if ( value instanceof Hashtable )
                hash1.put(new String(key), copyHash((Hashtable)value));
            else
                throw new Exception("Unsupported object type to clone");
        }

        return hash1;
    } // end of appendHash(...)

    public static String ipl2str(long ip) {
        long hi3 = ip/(long)16777216;
        long hi2 = (ip-hi3*(long)16777216)/(long)65536;
        long hi1 = (ip-hi3*(long)16777216-hi2*(long)65536)/(long)256;
        long hi0 = ip-hi3*(long)16777216-hi2*(long)65536-hi1*(long)256;
        return hi3+"."+hi2+"."+hi1+"."+hi0;
    }

    public static String ip2str(int ip) {
        return ( (ip & (ipMask << 24) ) >> 24 )+"."+
            ( (ip & (ipMask << 16) ) >> 16 )+"."+
            ( (ip & (ipMask << 8) ) >> 8 )+ "."+
            (ip & ipMask);
    }

    /** given ip address returns it in normalized form with all leading zeros present */
    public static String ip2norm(String ip) {
        String norm_ip = "";
        StringTokenizer tk = new StringTokenizer(ip, "./");
        int i = 0;
        while( tk.hasMoreTokens() ) {
            String ip_part = tk.nextToken().trim();
            int add_length = 3-ip_part.length();
            for( int j = 0; j < add_length; j++ )
                ip_part = "0"+ip_part;
            norm_ip += ip_part;

            if( i++ > 2 ) break;

            norm_ip += ".";
        }
        return norm_ip;
    }

    /** converts string representation of ip address to the int integer */
    public static int ipstr2num(String ip) {
        StringTokenizer tk = new StringTokenizer(ip, "./");
        int i = 0;
        String part[] = new String[4];
        while( tk.hasMoreTokens() ) {
            part[i++] = tk.nextToken();
            if( i > 3 )
                break;
        }
        if( i < 3 )
            return -1;

        int num[] = new int[4];
        for( i = 0; i < 4; i++ )
            try {
                num[i] = Integer.parseInt(part[i]);
            } catch( Exception ex) {
                num[i] = 0;
            }

        return (num[0] << 24) + (num[1] << 16) + (num[2] << 8) + num[3];
    }

    /** converts string representation of ip address to the long integer */
    public static long ipstr2long(String ip) {
        StringTokenizer tk = new StringTokenizer(ip, "./");
        int i = 0;
        String part[] = new String[4];
        while( tk.hasMoreTokens() ) {
            part[i++] = tk.nextToken();
            if( i > 3 )
                break;
        }
        if( i < 3 )
            return -1;

        int num[] = new int[4];
        for( i = 0; i < 4; i++ )
            try {
                num[i] = Integer.parseInt(part[i]);
            } catch( Exception ex) {
                num[i] = 0;
            }

        return (long)((long)num[3] + (long)num[2]*(long)256 + (long)num[1]*(long)65536 + (long)num[0]*(long)16777216);
    }

    /** converts string representation of ip address to a set of tasty bytes */
    public static byte[] ipstr2bytes(String ip) {
        StringTokenizer tk = new StringTokenizer(ip, "./");
        int i = 0;
        String part[] = new String[4];
        while( tk.hasMoreTokens() ) {
            part[i++] = tk.nextToken();
            if( i > 3 )
                break;
        }

        byte[] retval = new byte[4];
        if( i < 3 )
            return null;

        for( i = 0; i < 4; i++ )
            try {
                retval[i] = (byte)Integer.parseInt(part[i]);
            } catch( Exception ex) {
                retval[i] = 0;
            }

        return retval;
    }

    public static String long2bin(long l) {
        String result = "";
        long mask = 0x1;
        for( int i = 0; i < 64; i++ )
            result += ( l & (mask << i) ) == 0?"0":"1";
        return result;
    }

    public static String int2bin(int l) {
        String result = "";
        long mask = 0x1;
        for( int i = 0; i < 32; i++ )
            result += ( l & (mask << i) ) == 0?"0":"1";
        return result;
    }

    public static String fullDisplayName(String name, String domain, String def_domain) {
        if( domain.equalsIgnoreCase(def_domain) )
            return name;
        else
            return name+"."+domain;
    }

    public static String name(String fqdn) {
        int dotIndex = fqdn.indexOf(".");
        return dotIndex>-1?fqdn.substring(0, dotIndex):fqdn;
    }

    public static String domain(String fqdn, String defDomain) {
        int dotIndex = fqdn.indexOf(".");
        return dotIndex>-1?fqdn.substring(dotIndex+1):defDomain;
    }

    /**
     * Given address space prefix will return prefix of adjacent address space, i.e.
     * the one which can be joined with 'asPrefix' to form bigger address space
     */
    public static String adjacentAS(String asPrefix) {
        int baseIP = ipstr2num(asPrefix);
        int prefixLength = Integer.parseInt(asPrefix.substring(asPrefix.indexOf("/")+1));
        if( prefixLength == 32 )
            // throw new Exception("Given address space does not have adjacent counterpart.");
            return asPrefix;
        int marker = 0x00000001 << (32 - prefixLength);
        int adjacentIP = baseIP ^ marker;
        return ip2str(adjacentIP)+"/"+prefixLength;
    }

    /**
     * Given address space prefix will return set of two prefixes as if 'asPrefix' address
     * space would be split in half.
     */
    public static String[] splitASPrefix(String asPrefix) {
        String[] result = new String[2];
        int baseIP = ipstr2num(asPrefix);
        int prefixLength = Integer.parseInt(asPrefix.substring(asPrefix.indexOf("/")+1));
        if( prefixLength == 30 ) // do not slit beyond /30
            return null;
        int marker = 0x00000001 << (33 - prefixLength); // 33!

        result[0] = ip2str(baseIP)+"/"+(prefixLength+1);
        result[1] = ip2str(baseIP | marker)+"/"+(prefixLength+1);
        return result;
    } // end of splitASPrefix()

    private static void validateNames(Hashtable obj, Hashtable names, Vector errors, boolean mustHaveName) {
        if( names.size() == 0 && mustHaveName )
            errors.addElement("must have at least one name");
        Enumeration enm = names.keys();
        try {
            int found_names = 0;
            while( enm.hasMoreElements() ) {
                // we know that key is FQDN but it does not have to be this way,
                // so we don't do any check on this, but it has to be a String
                Object key = enm.nextElement();
                if( !(key instanceof String) )
                    throw new Exception("invalid name hash");

                Object name = names.get((String)key);
                if( !(name instanceof Hashtable) )
                    throw new Exception("invalid name hash");

                Object nameStr = ((Hashtable)name).get("name");
                if( nameStr == null || !(nameStr instanceof String) )
                    throw new Exception("invalid name hash");
                if( ((String)nameStr).trim().length() > 255 )
                    throw new Exception("Name "+((String)nameStr).trim()+" is too long");

                Object domain = ((Hashtable)name).get("domain");
                if( domain == null || !(domain instanceof String) )
                    throw new Exception("invalid name hash");
                if( ((String)domain).trim().length() > 255 )
                    throw new Exception("Domain "+((String)domain).trim()+" is too long");

                if( nameStr.toString().trim().length() > 0 )
                    found_names++;

                Object aliases = ((Hashtable)name).get("aliases");
                if( aliases != null ) {
                    if( !(aliases instanceof Hashtable) )
                        throw new Exception("invalid name hash");
                    Hashtable aliasesH = (Hashtable)aliases;
                    Enumeration enm1 = aliasesH.keys();
                    while( enm1.hasMoreElements() ) {
                        Object aliasKey = enm1.nextElement();
                        if( !(aliasKey instanceof String) )
                            throw new Exception("invalid name hash");
                        Object alias = aliasesH.get((String)aliasKey);
                        if( !(alias instanceof Hashtable) )
                            throw new Exception("invalid name hash");

                        Object aliasName = ((Hashtable)alias).get("name");
                        if( aliasName == null || !(aliasName instanceof String) )
                            throw new Exception("invalid name hash");
                        if( ((String)aliasName).trim().length() > 255 )
                            throw new Exception("Alias name "+((String)aliasName).trim()+" is too long");

                        Object aliasDomain = ((Hashtable)alias).get("domain");
                        if( aliasDomain == null || !(aliasDomain instanceof String) )
                            throw new Exception("invalid name hash");
                        if( ((String)aliasDomain).trim().length() > 255 )
                            throw new Exception("Alias domain "+((String)aliasDomain).trim()+" is too long");
                    }
                }

                Object mxs = ((Hashtable)name).get("mxs");
                if( mxs != null ) {
                    if( !(mxs instanceof Hashtable) )
                        throw new Exception("invalid name hash");
                    Hashtable mxsH = (Hashtable)mxs;
                    Enumeration enm1 = mxsH.keys();
                    while( enm1.hasMoreElements() ) {
                        Object mxKey = enm1.nextElement();
                        if( !(mxKey instanceof String) )
                            throw new Exception("invalid name hash");
                        Object mx = mxsH.get((String)mxKey);
                        if( !(mx instanceof Hashtable) )
                            throw new Exception("invalid name hash");

                        Object mxName = ((Hashtable)mx).get("name");
                        if( mxName == null || !(mxName instanceof String) )
                            throw new Exception("invalid name hash");
                        if( ((String)mxName).trim().length() > 255 )
                            throw new Exception("Mx name "+((String)mxName).trim()+" is too long");

                        Object mxDomain = ((Hashtable)mx).get("domain");
                        if( mxDomain == null || !(mxDomain instanceof String) )
                            throw new Exception("invalid name hash");
                        if( ((String)mxDomain).trim().length() > 255 )
                            throw new Exception("Mx domain "+((String)mxDomain).trim()+" is too long");

                        Object mxPref = ((Hashtable)mx).get("pref");
                        if( mxPref == null || !(mxPref instanceof String) )
                            throw new Exception("invalid name hash");
                        else {
                            try {
                                int ipref = Integer.parseInt((String)mxPref);
                                if( ipref < 0 || ipref > 100 )
                                    throw new Exception();
                            } catch (Exception ex) {
                                errors.addElement("'"+((String)mxPref).trim()+
                                                  "' is not a valid MX preference. Please enter number between 1 and 100");
                                Hashtable bad_mx_pref = (Hashtable)obj.get("bad_mx_pref");
                                if( bad_mx_pref == null ) {
                                    bad_mx_pref = new Hashtable();
                                    obj.put("bad_mx_pref", bad_mx_pref);
                                }
                                bad_mx_pref.put(((String)mxPref).trim().toLowerCase(), "");
                            }
                        }
                    }
                }
            }
            if( mustHaveName && found_names == 0 )
                errors.addElement("Must have at least one name");
        } catch (Exception ex) {
            ex.printStackTrace();
            errors.addElement(ex.getMessage());
        }
    } // end of validateNames(...)

    public static String htmlize(String txt) {
        return htmlize(txt, true);
    }

    public static String htmlize(String txt, boolean doBreak) {
        String retVal =
            txt.replaceAll("&", "&amp;").
                replaceAll("<", "&lt;").
                replaceAll(">", "&gt;").
                replaceAll("\"", "&quot;").
                replaceAll("'", "&#39;");
        if( doBreak )
            retVal = retVal.replaceAll("\n", "<br>");

        return retVal;
    }

    /** Used for collecting names to delete, insert and update in Networks and Nodes */
    public static Hashtable sortOutNames(Hashtable oldObj, Hashtable newObj) {
        Hashtable names_to_delete = collectNames(oldObj);
        Hashtable names_to_insert = collectNames(newObj);
        Hashtable names_to_update = new Hashtable();

        // Note we do not have DB call to update names, thus updating names implemented by
        // deleting them and then inserting them back.

        // sort out all names
        Enumeration enm = names_to_delete.keys();
        while( enm.hasMoreElements() ) {
            String key = (String)enm.nextElement();

            Hashtable newName = (Hashtable)names_to_insert.get(key);
            if( newName != null ) {
                Hashtable oldName = (Hashtable)names_to_delete.get(key);

                // check if names are the same
                String oname = (String)oldName.get("name");
                String nname = (String)newName.get("name");
                if( !oname.equals(nname) )
                    continue; // names are different, they stay as they are.

                String oMx = (String)oldName.get("pref");
                String nMx = (String)newName.get("pref");
                if( oMx != null && nMx != null && !oMx.equals(nMx) )
                    names_to_update.put(key,newName); // The MX preference needs updating.

                // we are here if old and new names are the same, thus they
                // should be not be modified or the MX preference needs updating
                // and we're preventing removal and insertion of the names.
                names_to_delete.remove(key);
                names_to_insert.remove(key);
            }
        }

        // among names_to_delete may be names, which contain aliases and/or mx's
        // we should remove those names (aliases and mx's) from names_to_delete
        // becouse deletion of name will automatically delete aliases and mx's
        // as well
        enm = names_to_delete.keys();
        while( enm.hasMoreElements() ) {
            String key = (String)enm.nextElement();

            Hashtable oldName = (Hashtable)names_to_delete.get(key);
            if( oldName == null )
                continue;
            Hashtable refs = (Hashtable)oldName.get("refs");
            // we are looking only at the names here not mxs or aliases
            if( refs != null ) {
                Enumeration refs_enm = refs.keys();
                while( refs_enm.hasMoreElements() ) {
                    String ref_key = (String)refs_enm.nextElement();
                    names_to_delete.remove(ref_key);
                }
            }
        }

        Hashtable result = new Hashtable();
        result.put("names_to_delete", names_to_delete);
        result.put("names_to_insert", names_to_insert);
        result.put("names_to_update", names_to_update);
        return result;
    } // end of sortOutNames(...)

    /**
     * collects all names from ether Network or Node object into Hash for
     * deletion, insertion and update
     */
    public static Hashtable collectNames(Hashtable obj) {
        Hashtable allNames       = new Hashtable();

        if( obj == null )
            return allNames;

        String otype = (String)obj.get("object");

        // both net and node have names
        Hashtable names = nHash((Hashtable)obj.get("names"));
        gatherNames(allNames, names, /* suppl */null, /* prefix */otype,
                    /* type */otype+"_name", null, null);

        // names comming from dhcp addresses
        Hashtable ases = (Hashtable)obj.get("address_spaces");
        if( ases == null )
            ases = new Hashtable();
        Enumeration enm = ases.keys();
        while( enm.hasMoreElements() ) {
            Hashtable as = (Hashtable)ases.get((String)enm.nextElement());
            Hashtable dhcp_ips = nHash((Hashtable)as.get("dhcp_addresses"));

            Enumeration ip_enm = dhcp_ips.keys();
            while( ip_enm.hasMoreElements() ) {
                Hashtable ip = (Hashtable)dhcp_ips.get((String)ip_enm.nextElement());
                Hashtable ip_names = nHash((Hashtable)ip.get("names"));
                String ipAddr = (String)ip.get("ip_address");
                Hashtable suppl = new Hashtable();
                suppl.put("ip", ipAddr);
                gatherNames(allNames, ip_names, suppl, "ip_"+ipAddr+"_name", "ip_name", null, null);
            }
        }

        // names comming from ipc addresses
        Hashtable ipc_ips = nHash((Hashtable)obj.get("ipc_addresses"));
        enm = ipc_ips.keys();
        while( enm.hasMoreElements() ) {
            String ipAddr = (String)enm.nextElement();
            Hashtable ip = (Hashtable)ipc_ips.get(ipAddr);
            Hashtable ip_names = nHash((Hashtable)ip.get("names"));
            Hashtable suppl = new Hashtable();
            suppl.put("ip", ipAddr);
            gatherNames(allNames, ip_names, suppl, "ipc_"+ipAddr+"_name", "ipc_name", null, null);
        }

        // names comming from interfaces and their ips
        Hashtable intfss = nHash((Hashtable)obj.get("interfaces"));
        enm = intfss.keys();
        while( enm.hasMoreElements() ) {
            String intfs_id = (String)enm.nextElement();
            Hashtable intfs = (Hashtable)intfss.get(intfs_id);
            Hashtable intfs_names = nHash((Hashtable)intfs.get("names"));
            Hashtable suppl = new Hashtable();
            suppl.put("intfs_id", intfs_id);
            gatherNames(allNames, intfs_names, suppl, "intfs_"+intfs_id+"_name", "intfc_name", null, null);

            // ips
            Hashtable ips = nHash((Hashtable)intfs.get("ips"));
            Enumeration ips_enm = ips.keys();
            while( ips_enm.hasMoreElements() ) {
                String ipAddr = (String)ips_enm.nextElement();
                Hashtable ip = (Hashtable)ips.get(ipAddr);
                Hashtable ip_names = nHash((Hashtable)ip.get("names"));
                suppl = new Hashtable();
                suppl.put("ip", ipAddr);
                gatherNames(allNames, ip_names, suppl, "ip_"+ipAddr+"_name", "ip_name", null, null);
            }
        }
        return allNames;
    } // end of collectNames(...)

    private static void gatherNames(Hashtable allNames, Hashtable name, Hashtable suppl,
                             String prefix, String type, String parentFQDN, Hashtable refs) {
        if( name.get("name") == null ) {
            // top level
            Enumeration fenm = name.keys();
            while( fenm.hasMoreElements() ) {
                String nameKey = (String)fenm.nextElement();
                gatherNames(allNames, (Hashtable)name.get(nameKey), suppl, prefix, type, null, null);
            }
            return;
        }
        Hashtable nameH = new Hashtable();
        nameH.put("name", (String)name.get("name"));
        nameH.put("domain", (String)name.get("domain"));
        nameH.put("type", type);
        String pref = (String)name.get("pref");
        if( pref != null )
            nameH.put("pref", pref);
        if( suppl != null ) {
            // stuff in all suppl. data
            Enumeration senm = suppl.keys();
            while( senm.hasMoreElements() ) {
                String skey = (String)senm.nextElement();
                nameH.put(skey, (String)suppl.get(skey));
            }
        }

        String key = prefix + "_" + type +"_"+
            ((String)name.get("name")) + "." + ((String)name.get("domain"));

        if( refs == null ) {
            refs = new Hashtable();
            nameH.put("refs", refs);
        } else
            refs.put(key, "");
        if( parentFQDN != null )
            nameH.put("parent", parentFQDN);
        else
            parentFQDN = ((String)name.get("name")).toLowerCase() + "." +
                ((String)name.get("domain")).toLowerCase();
        allNames.put(key, nameH);

        Hashtable aliases = nHash((Hashtable)name.get("aliases"));
        Enumeration enm = aliases.keys();
        while( enm.hasMoreElements() ) {
            Hashtable alias = (Hashtable)aliases.get((String)enm.nextElement());
            gatherNames(allNames, alias, null, key, "alias", parentFQDN, refs);
        }

        Hashtable mxs = nHash((Hashtable)name.get("mxs"));
        enm = mxs.keys();
        while( enm.hasMoreElements() ) {
            Hashtable mx = (Hashtable)mxs.get((String)enm.nextElement());
            gatherNames(allNames, mx, null, key, "mx", parentFQDN, refs);
        }
    } // end of gatherNames(...)


    /**
     * Will check if source 'src' Hashtable is null and if so then will return new one,
     * otherwise will return src.
     */
    public static Hashtable nHash(Object src) {
        if( src == null )
            return new Hashtable();
        return (Hashtable)src;
    }

    /**
     * Given key it gets Hashtable from the parent hash and if it does not exist
     * it will create it, place in a parent and return it.
     */
    public static Hashtable get_hash_or_create(Hashtable parent, Object key) {
        Hashtable child = (Hashtable)parent.get(key);
        if( child == null ) {
            child = new Hashtable();
            parent.put(key, child);
        }
        return child;
    } // get_hash_or_create(...)

    /**
     * Will check if source 'src' String is null and if so then will return new one,
     * otherwise will return src type-casted to a String
     */
    public static String nStr(Object src) {
        if( src == null )
            return "";
        return (String)src;
    }

    public static String nStr(Hashtable hash, Object key) {
        return nStr(hash.get(key));
    }

    /** Given 'str' will place new line characters '\n' to make a block of width no more then max_len */
    public static String brakes(String src, int max_len) {
        StringBuffer str = new StringBuffer(src);
        int index = -1;
        int br_index = -1;
        for(;;) {
            br_index = str.indexOf("\n", index);
            if( br_index == -1 ) {
                br_index = str.length();
                if( br_index - index <= max_len )
                    break;
            }
            int len = br_index - index;
            if( len > max_len ) {
                str.insert(index+max_len, "\n");
                index = index+max_len + 1;
            } else
                index = br_index + 1;
        }
        return str.toString();
    }

    static public StringBuffer formatString2Block(String prepend, String src, int max_length)
        throws Exception {
        if( src == null ) src = "";
        StringBuffer result = new StringBuffer();
        int plength = prepend.length();
        int slength = src.length();
        if( max_length <= plength )
            throw new Exception("max_length is too small, cannot format block.");
        char[] prepend1 = new char[plength];
        for( int i=0; i<plength; i++ )
            prepend1[i] = ' ';

        // split src
        int space_index = -1;
        int new_line_index = -1;
        int tab_index = -1;
        int min_index = -1;
        int last_index = 0;
        String element = "";
        Vector tokens = new Vector();
        for(;;) {
            min_index = -1;
            space_index = src.indexOf(' ', last_index);
            new_line_index = src.indexOf('\n', last_index);
            tab_index = src.indexOf('\t', last_index);
            min_index = space_index>0?space_index:slength;
            if( new_line_index < min_index && new_line_index > -1 )
                min_index = new_line_index;
            if( tab_index < min_index && tab_index > 0)
                min_index = tab_index;

            if( min_index == new_line_index )
                element = src.substring(last_index, min_index-1)+" ";
            else
                element = src.substring(last_index, min_index)+" ";
            tokens.addElement(element);
            last_index = min_index+1;
            if( last_index >= slength ) break;
        }
        // build block
        int line_counter = 0;
        StringBuffer line = new StringBuffer(prepend);
        int appended_tokens = 0;
        boolean line_completed = false;
        for(int i=0; i<tokens.size(); i++) {
            element = (String)tokens.elementAt(i);
            if( ( line.length() + element.length() ) > max_length ) {
                if( appended_tokens == 0 ) {
                    // append token anyway
                    line.append(element+"\n");
                    line_completed = true;
                } else {
                    // do not append token now, open new line and try again
                    i--; // has to roll back in order to see the same token again
                    appended_tokens = 0;
                    line_completed = false;
                }
                result.append(line);
                result.append("\n");
                line = new StringBuffer();
                line.append(prepend1);
                line_counter++;

            } else {
                appended_tokens++;
                line_completed = false;
                line.append(element);
            }
        }
        if( !line_completed )
            result.append(line);
        return result;
    }

    public static StringBuffer formatString(StringBuffer src,
                                            int min_length,
                                            int max_length,
                                            char last,
                                            int alignment) {
        if( src == null )
            return new StringBuffer("");

        int strlen = src.length();
        if( ( min_length < 1 && max_length < 1 ) ||
            ( min_length > 0 && max_length > 0 && strlen <= max_length && strlen >= min_length) ||
            ( min_length < 1 && max_length > 0 && strlen <= max_length ) ||
            ( min_length > 0 && max_length < 1 && strlen >= min_length ) ){
            if( last != 0 )
                if( max_length == strlen ){
                    src.setLength(max_length-4);
                    return src.append("...").append(last);
                }else
                    return src.append(last);
            return src;
        }

        // at this point we know that 'src' needs some adjustments before returning it.
        String result = "";
        if( min_length > 0 && strlen < min_length ){
            int added_length = min_length - strlen;
            char[] addition = new char[added_length];
            for( int i=0; i<added_length; i++ )
                addition[i] = ' ';
            if( last != 0 && alignment == LEFT )
                addition[0] = last;
            if( alignment == LEFT )
                return src.append(addition);
            else { // RIGHT
                StringBuffer sb = (new StringBuffer(new String(addition))).append(src);
                if( last != 0 ) sb.append(last);
                return sb;
            }
        }else // max_length > 0 && strlen > max_length
            if( last != 0 ){
                src.setLength(max_length-4);
                return src.append("...").append(last);
            }else{
                src.setLength(max_length-3);
                return src.append("...");
            }
    } // end of formatString(...)

    static public Object n2s(Object src) {
        if( src == null ) return "";
        return src;
    }

    static public Iterator sortedKeys(Hashtable hash, Comparator cmp) {
        java.util.List list = Arrays.asList(hash.keySet().toArray());
        Collections.sort(list, cmp);
        return list.iterator();
    }

    static public Iterator sortedEntries(Hashtable hash, Comparator cmp) {
        java.util.List list = Arrays.asList(hash.entrySet().toArray());
        Collections.sort(list, cmp);
        return list.iterator();
    }

    static public String joinedAS(String basePrefix) {
        String adjacentPrefix = Utils.adjacentAS(basePrefix);
        long floor1 = ipstr2long(basePrefix);
        long floor2 = ipstr2long(adjacentPrefix);
        String origBasePrefix = basePrefix;
        if( floor2 < floor1 )
            basePrefix = adjacentPrefix;

        int baseIP = ipstr2num(basePrefix);
        int prefixLength = Integer.parseInt(basePrefix.substring(basePrefix.indexOf("/")+1))-1;
        if( prefixLength == 32 )
            return origBasePrefix;

        return ipl2str(ipstr2long(basePrefix))+"/"+(prefixLength);
    }

    static public Hashtable joinAS(String basePrefix, Hashtable ases) throws Exception {
        int baseIP = ipstr2num(basePrefix);
        int prefixLength = Integer.parseInt(basePrefix.substring(basePrefix.indexOf("/")+1));
        if( prefixLength == 32 )
            return null;
        int marker = 0x00000001 << (32 - prefixLength);
        int adjacentIP = baseIP ^ marker;
        String adjIPStr = ip2str(adjacentIP);

        long baseStart = ipstr2long(basePrefix);
        long adjStart = ipstr2long(adjIPStr);

        long newBase = baseStart<adjStart?baseStart:adjStart;
        long newBCast = newBase + (((long)1 << ((33 - prefixLength)))-1);

        Hashtable newAS = new Hashtable();
        newAS.put("addr_space_ip", ipl2str(newBase));
        newAS.put("prefix_length", new Integer(prefixLength-1));

        String newASPrefix = ipl2str(newBase)+"/"+(prefixLength-1);
        newAS.put("addr_space_prefix", newASPrefix);
        newAS.put("low_res", new Integer (1));
        newAS.put("high_res", new Integer (0));

        Hashtable groups = new Hashtable();
        newAS.put("groups", groups);

        Hashtable dhcp_addresses = new Hashtable();
        newAS.put("dhcp_addresses", dhcp_addresses);

        Hashtable dhcp_options = new Hashtable();
        newAS.put("dhcp_options", dhcp_options);

        String comment = "";

        Vector keys_to_remove = new Vector();

        List list = Arrays.asList(ases.entrySet().toArray());
        Collections.sort(list, new Comparator() {
                public int compare(Object o1, Object o2) {
                    Hashtable as1 = (Hashtable)((Map.Entry)o1).getValue();
                    Hashtable as2 = (Hashtable)((Map.Entry)o2).getValue();

                    if( as1.get("addr_space_prefix") == null )
                        return -1;
                    else if( as2.get("addr_space_prefix") == null )
                        return 1;

                    long cmp_val = Utils.ipstr2long((String)as1.get("addr_space_prefix")) -
                        Utils.ipstr2long((String)as2.get("addr_space_prefix"));

                    return cmp_val>0?1:(cmp_val<0?-1:0);
                }
                public boolean equals(Object obj) {
                    return true;
                }
            });
        Iterator it = list.iterator();
        while( it.hasNext() ) {
            Map.Entry entry = (Map.Entry)it.next();

            Hashtable as = (Hashtable)entry.getValue();
            String key = (String)entry.getKey();
            String asPrefix = (String)n2s(as.get("addr_space_prefix"));

            long ipAS = ipstr2long(asPrefix);
            if( ipAS == newBase ) {
                keys_to_remove.addElement(key);
                newAS.put("low_res", as.get("low_res"));
                appendHash(groups, Utils.nHash(as.get("groups")));
                appendHash(dhcp_addresses, Utils.nHash(as.get("dhcp_addresses")));
                appendHash(dhcp_options, Utils.nHash(as.get("dhcp_options")));
                comment += ((String)Utils.n2s(as.get("comment"))).replaceAll("\n", " ").trim()+" ";

                // check for bcast and take high_res if needed
                long asBCast = ipAS + (((long)1 << ((32 - ((Integer)as.get("prefix_length")).intValue())))) - 1;
                if( newBCast == asBCast )
                    newAS.put("high_res", as.get("high_res"));

            } else if( ipAS > newBase && ipAS < newBCast ) {
                keys_to_remove.addElement(key);
                appendHash(groups, Utils.nHash(as.get("groups")));
                appendHash(dhcp_addresses, Utils.nHash(as.get("dhcp_addresses")));
                appendHash(dhcp_options, Utils.nHash(as.get("dhcp_options")));
                comment += ((String)Utils.n2s(as.get("comment"))).replaceAll("\n", " ").trim()+" ";

                // check for bcast and take high_res if needed
                long asBCast = ipAS + (((long)1 << ((32 - ((Integer)as.get("prefix_length")).intValue())))) - 1;
                if( newBCast == asBCast )
                    newAS.put("high_res", as.get("high_res"));
            }
        }
        if( comment.trim().length() > 0 )
            newAS.put("comment", comment.trim());

        // remove asses not needed any more
        for( int i = 0; i < keys_to_remove.size(); i++ ) {
            ases.remove(keys_to_remove.elementAt(i));
        }
        ases.put(newASPrefix, newAS);

        // comment for now. there are some problems with logic here
        /*
        // check duplicate dhcp options and issue warning if anything found
        Hashtable duplicates = new Hashtable();
        Hashtable check = new Hashtable();
        enm = dhcp_options.keys();
        while( enm.hasMoreElements() ) {
            String option = (String)((Hashtable)dhcp_options.get(enm.nextElement())).get("option");
            if( option == null )
                continue;
            if( check.put(option, "") != null )
                duplicates.put(option, "");
        }

        if( duplicates.size() > 0 ) {
            Vector warnings = new Vector();
            enm = duplicates.keys();
            while( enm.hasMoreElements() ) {
                warnings.addElement("<font color=\"#ff0000\">&nbsp;Warning: Duplicate dhcp option \""+enm.nextElement()+"\".</font>");
            }
            newAS.put("warnings", warnings);
        }
        */

        return newAS;
    } // end of joinAS(...)

    private static void validateDHCPOptions(Object dhcp_options, Vector errors) {
        if( dhcp_options != null ) {
            if( !(dhcp_options instanceof Hashtable) )
                errors.addElement("'dhcp_options' field is not a Hashtable");
            else {
                Enumeration enm = ((Hashtable)dhcp_options).keys();
                while( enm.hasMoreElements() ) {
                    Object dhcp_option = ((Hashtable)dhcp_options).get(enm.nextElement());
                    if( !(dhcp_option instanceof Hashtable) ) {
                        errors.addElement("dhcp option field is not a Hashtable");
                        continue;
                    }
                    Object option = ((Hashtable)dhcp_option).get("option");
                    if( option == null ) {
                        errors.addElement("dhcp 'option' field is null");
                        continue;

                    } if( !(option instanceof String) ) {
                        errors.addElement("dhcp 'option' field is not a String");
                        continue;

                    } else if( ((String)option).length() > 64 ) {
                        errors.addElement("dhcp option is too long");
                        continue;
                    }

                    Object value = ((Hashtable)dhcp_option).get("value");
                    if( value != null ) {
                        if( !(value instanceof String) )
                            errors.addElement("dhcp 'value' field is not a String");
                        else if( ((String)value).length() > 256 )
                            errors.addElement("dhcp value is too long");
                    }
                }
            }
        }
    } // end of validateDHCPOptions(...)

    public static Hashtable makeHash(Object[] keys, Object[] values) {
        Hashtable result = new Hashtable();
        if( keys != null ) {
            for( int i = 0; i < keys.length; i++ )
                result.put(keys[i], (values!=null?values[i]:""));
        }
        return result;
    }

    public static String capitalize(Object src) {
        if( src == null )
            return "";
        String str = src.toString();
        return str.substring(0,1).toUpperCase()+str.substring(1);
    }

    /**
     * Goes over all object and removes empty names and interfaces
     */
    public static void purgeObject(Hashtable obj) {
        Vector keys_to_remove = new Vector();
        Vector parents = new Vector();

        // names
        purgeNames(nHash(obj.get("names")), keys_to_remove, parents);

        // Interfaces
        Hashtable interfaces = nHash(obj.get("interfaces"));
        Enumeration enm = interfaces.keys();
        while( enm.hasMoreElements() ) {
            Object intfcID = enm.nextElement();
            Hashtable intfc = (Hashtable)interfaces.get(intfcID);
            purgeNames(nHash(intfc.get("names")), keys_to_remove, parents);
            Hashtable ips = nHash(intfc.get("ips"));
            Enumeration enm2 = ips.keys();
            while( enm2.hasMoreElements() ) {
                Hashtable ipH = (Hashtable)ips.get(enm2.nextElement());
                purgeNames(nHash(ipH.get("names")), keys_to_remove, parents);
            }

            /*
            if( intfc.get("empty") != null ) {
                parents.addElement(interfaces);
                keys_to_remove.addElement(intfcID);
            }
            */
        }

        // IPC ip addresses
        Hashtable ipcs = nHash(obj.get("ipc_addresses"));
        enm = ipcs.keys();
        while( enm.hasMoreElements() ) {
            Hashtable ipH = (Hashtable)ipcs.get(enm.nextElement());
            purgeNames(nHash(ipH.get("names")), keys_to_remove, parents);
        }

        // address spaces
        Hashtable ases = nHash(obj.get("address_spaces"));
        enm = ases.keys();
        while( enm.hasMoreElements() ) {
            Hashtable as = (Hashtable)ases.get(enm.nextElement());
            Hashtable dhcp_addresses = nHash(as.get("dhcp_addresses"));
            Enumeration enm2 = dhcp_addresses.keys();
            while( enm2.hasMoreElements() ) {
                Hashtable ipH = (Hashtable)dhcp_addresses.get(enm2.nextElement());
                purgeNames(nHash(ipH.get("names")), keys_to_remove, parents);
            }
        }

        // now go over keys_to_remove and do clean up
        for( int i = 0; i < keys_to_remove.size(); i++ ) {
            try {
                ((Hashtable)parents.elementAt(i)).remove(keys_to_remove.elementAt(i));
            } catch (Exception ex) {}
        }

        // since any obj must have we must leave at least one
        Hashtable names = (Hashtable)obj.get("names");
        if( names == null ) {
            names = new Hashtable();
            obj.put("names", names);
        }

        if( names.size() == 0 ) {
            Hashtable name = new Hashtable();
            name.put("name", "");
            name.put("domain", "");
            name.put("aliases", new Hashtable());
            name.put("mxs", new Hashtable());
            names.put(" a 1", name);
        }
    } // end of purgeObject(...)

    private static void purgeNames(Hashtable names,
                                   Vector keys_to_remove,
                                   Vector parents)
    {
        Enumeration enm = names.keys();
        while( enm.hasMoreElements() ) {
            String key = (String)enm.nextElement();
            Hashtable nameH = (Hashtable)names.get(key);
            String name = (String)nameH.get("name");
            String domain = (String)nameH.get("domain");
            if( name == null || name.trim().length() == 0 ) {
                keys_to_remove.addElement(key);
                parents.addElement(names);
                continue;
            }

            // Aliases
            Hashtable aliases = nHash(nameH.get("aliases"));
            Enumeration enm2 = aliases.keys();
            while( enm2.hasMoreElements() ) {
                String akey = (String)enm2.nextElement();
                Hashtable aliasH = (Hashtable)aliases.get(akey);
                name = (String)aliasH.get("name");
                domain = (String)aliasH.get("domain");
                if( name == null || name.trim().length() == 0 ) {
                    keys_to_remove.addElement(akey);
                    parents.addElement(aliases);
                    continue;
                }
            }

            // MXes
            Hashtable mxes = nHash(nameH.get("mxs"));
            Enumeration enm3 = mxes.keys();
            while( enm3.hasMoreElements() ) {
                String mkey = (String)enm3.nextElement();
                Hashtable mxH = (Hashtable)mxes.get(mkey);
                name = (String)mxH.get("name");
                domain = (String)mxH.get("domain");
                if( name == null || name.trim().length() == 0 ) {
                    keys_to_remove.addElement(mkey);
                    parents.addElement(mxes);
                    continue;
                }
            }
        } // end of loop over names
    } // end of purgeNames(...)

    /**
     * Given an ip string it will perform normalization by adding zeroes at the end.
     */
    public static String normIp(String ip) {
        if( ip == null || ip.trim().length() == 0 )
            return "";

        try {
            int slash = ip.indexOf("/");

            Hashtable parseRes = (new Parser()).parse(ip);
            String ip_low = (String)parseRes.get("ip_low");
            String ip_high = (String)parseRes.get("ip_high");
            if( ip_low != null && ip_high != null ) {
                if( ip_low.equals(ip_high) )
                    return ip_low;
                else if( slash > -1 )
                    return ip_low.trim() + ip.substring(slash).trim();
            }
        } catch (Exception ex) { }
        return new String(ip);
    } // end pf normIp(...)

    public static Hashtable readACL() throws FileNotFoundException, Exception {
        Hashtable acl = new Hashtable();

        BufferedReader file = null;
        try{
            file = new BufferedReader( new FileReader(System.getProperty("acl",
                                                                         Constants.ACL_FILENAME)) );
            String line;
            int index = 0;
            while( ( line = file.readLine() ) != null ){
                line = line.trim();
                if( index == 0 && line.equals("*acl*") ) {
                    acl.put("*acl*", "");
                    continue;
                }

                if( line.equals("") ) // skip empty lines
                    continue;

                Hashtable entry = new Hashtable();

                if( line.endsWith("*") ) {
                    acl.put(line.substring(0, line.length() - 1).trim(), entry);
                    entry.put("god", "");
                } else
                    acl.put(line, entry);

                index++;
            }

        } finally {
            try {
                file.close();
            } catch (Exception ex) {
            }
        }

        return acl;
    }

    public static void validateTextExpression(String expr) throws Exception {
        expr = expr.trim();
        if( expr.length() == 0 )
            return;
        boolean illegalTextExpression = false;
        boolean hasOperators = false;
        String lastToken = ""; // it is always on lowerCase;
        StringTokenizer stk = new StringTokenizer(expr);
        while( stk.hasMoreTokens() ) {
            String tkLc = stk.nextToken().toLowerCase();
            if( tkLc.equals("and") || tkLc.equals("or") || tkLc.equals("not") )
                hasOperators = true;
            else {
                if( !lastToken.equals("and") &&
                    !lastToken.equals("or")  &&
                    !lastToken.equals("not") &&
                    !lastToken.equals("") )
                    illegalTextExpression = true;
            }
            lastToken = tkLc;
        }
        if( hasOperators && illegalTextExpression )
            throw new Exception("Invalid text expression '"+expr+"\'");
    } // end of validateTextExpression(...)

    public static void validateNumberExpression(String expr) throws Exception {
        expr = expr.trim();
        if( expr.length() == 0 )
            return;
        StringTokenizer stk = new StringTokenizer(expr);
        while( stk.hasMoreTokens() ) {
            String tkLc = stk.nextToken().toLowerCase();
            if( !tkLc.equals("and") &&
                !tkLc.equals("or")  &&
                !tkLc.equals("not") ) {
                try {
                    Integer.parseInt(tkLc);
                } catch (Exception ex) {
                    throw new Exception("Invalid number expression '"+expr+"\'");
                }
            }
        }
    } // end of validateNumberExpression(...)

    /**
     * Given a name and its value (field), will parse it and produce approprite sql
     * This function relies on a certain format of the name field to figure out type
     * of the field, e.g. name = <type>.<rest> where <type> = {hw, ip, num, txt, tst},
     * where tst - stands for timestamp
     */
    public static String parseField(Hashtable search_params, String full_field_name, String name, String value, String table, String column) throws Exception {
        // return empty string on empty fields value
        value = value.trim();

        if( value.length() == 0 )
            return "";

        String field_display_name = null;

        // node full search form
        if(      full_field_name.equals("fs.FS_NameNode.txt.fullname_lc") )                  field_display_name = "Name";
        else if( full_field_name.equals("fs.FS_NodeMX.txt.fullname_lc") )                    field_display_name = "Receives Mail for";
        else if( full_field_name.equals("fs.FS_Node.txt.lower(state)") )                     field_display_name = "State";
        else if( full_field_name.equals("fs.FS_NodeIPaddress.ip.IPaddress") )                field_display_name = "IP Address";
        else if( full_field_name.equals("fs.FS_NodeIPaddress.hw.lower(HW_address)") )        field_display_name = "Hardware Address";
        else if( full_field_name.equals("txt.DHCPoptions") )                                 field_display_name = "DHCP Options";
        else if( full_field_name.equals("fs.FS_Node.txt.groupZ.lower(owner)") )              field_display_name = "Group";
        else if( full_field_name.equals("fs.FS_Node.txt.lower(department)") )                field_display_name = "Department";
        else if( full_field_name.equals("fs.FS_Node.txt.lower(location)") )                  field_display_name = "Location";
        else if( full_field_name.equals("fs.FS_Node.txt.lower(site)") )                      field_display_name = "Site";
        else if( full_field_name.equals("fs.FS_Node.txt.lower(code)") )                      field_display_name = "Code";
        else if( full_field_name.equals("fs.FS_Node.txt.lower(room)") )                      field_display_name = "Room";
        else if( full_field_name.equals("fs.FS_Node.tst.expiration >= ") ||
                 full_field_name.equals("fs.FS_Node.tst.expiration <= ") )                   field_display_name = "Expiration Date";
        else if( full_field_name.equals("fs.FS_Node.txt.lower(make)") ||
                 full_field_name.equals("fs.FS_Node.txt.lower(model)") )                     field_display_name = "Make & Model";
        else if( full_field_name.equals("fs.FS_Node.txt.lower(os)") )                        field_display_name = "OS";
        else if( full_field_name.equals("fs.FS_Node_Admin.txt.groupA.lower(ateam)") ||
                 full_field_name.equals("fs.FS_Node_Admin.txt.groupA.lower(admin_name)") ||
                 full_field_name.equals("fs.FS_Node_Admin.txt.groupA.lower(admin_netid)") )  field_display_name = "Administrator";
        else if( full_field_name.equals("fs.FS_Node_User.txt.groupU.lower(user_netid)") ||
                 full_field_name.equals("fs.FS_Node_User.txt.groupU.lower(user_name)") )     field_display_name = "Node User";
        else if( full_field_name.equals("txt.customFields") )                                field_display_name = "Custom Fields";
        else if( full_field_name.equals("fs.FS_Node.txt.lower(comments)") )                  field_display_name = "Comment";

        // User full search form
        else if( ( full_field_name.equals("txt.lower(netid)") ||
                   full_field_name.equals("txt.lower(name)") ) &&
                 search_params.get("txt.NetDB.IP.text(def_addr_space)") != null )            field_display_name = "User";
        else if( full_field_name.equals("tst.last_login >= ") ||
                 full_field_name.equals("tst.last_login <= ") )                              field_display_name = "Last Login";
        else if( full_field_name.equals("txt.lower(organization)") )                         field_display_name = "Department"; // also in ateam fs
        else if( full_field_name.equals("txt.cons_name") )                                   field_display_name = "Networking Consultant";

        else if( full_field_name.equals("txt.lower(owner)") &&
                 search_params.get("txt.NetDB.IP.text(def_addr_space)") != null )            field_display_name = "Group";

        else if( full_field_name.equals("txt.lower(comments)") )                             field_display_name = "Comment"; // also in ateam fs
        else if( full_field_name.equals("txt.lower(phone)") )                                field_display_name = "Phone";   // also in ateam fs
        else if( full_field_name.equals("txt.lower(email)") )                                field_display_name = "Email";   // also in ateam fs
        else if( full_field_name.equals("txt.lower(def_domain)") )                           field_display_name = "Default Domain";
        else if( full_field_name.equals("txt.NetDB.IP.text(def_addr_space)") )               field_display_name = "Default Address Space";
        else if( full_field_name.equals("txt.lower(def_group)") )                            field_display_name = "Default Group";

        // Network full search form
        else if( full_field_name.equals("fs.FS_NameNetwork.txt.fullname_lc") )               field_display_name = "Name";
        else if( full_field_name.equals("fs.FS_AddressSpace.ip.address") )                   field_display_name = "Address Spaces";
        else if( full_field_name.equals("fs.FS_NetworkInfo.txt.lower(dhcp_service)") )       field_display_name = "DHCP Service";
        else if( full_field_name.equals("txt.NetDHCPoptions") )                              field_display_name = "Network DHCP Options";
        else if( full_field_name.equals("fs.FS_NetworkInfo.txt.lower(owner)") )              field_display_name = "Network Group";
        else if( full_field_name.equals("fs.FS_NetworkInfo.txt.lower(comments)") )           field_display_name = "Network Comment";
        else if( full_field_name.equals("txt.ASDHCPoptions") )                               field_display_name = "Address Space DHCP Options";
        else if( full_field_name.equals("fs.FS_AddressSpaceOwner.txt.lower(owner)") )        field_display_name = "Address Space Group";
        else if( full_field_name.equals("fs.FS_AddressSpace.txt.lower(comments)") )          field_display_name = "Address Space Comment";
        else if( full_field_name.equals("fs.FS_NetworkIPaddresses.ip.IPaddress") )           field_display_name = "Dynamic DHCP Address";
        else if( full_field_name.equals("fs.FS_NetworkIPaddresses.txt.fullname_lc") )        field_display_name = "Dynamic DHCP Address Name";

        // ATeam full search form
        else if( full_field_name.equals("txt.lower(name)") &&
                 search_params.get("txt.lower(member_netid)") != null )                      field_display_name = "Team Name";
        else if( full_field_name.equals("txt.lower(owner)") &&
                 search_params.get("txt.lower(member_netid)") != null )                      field_display_name = "Group";
        else if( full_field_name.equals("txt.lower(member_netid)") ||
                 full_field_name.equals("txt.lower(member_name)") )                          field_display_name = "Member";

        // Domain full search form
        else if( full_field_name.equals("txt.lower(domain)") )                               field_display_name = "Domain Name";
        else if( full_field_name.equals("txt.lower(owner)") &&
                 search_params.get("txt.lower(domain)") != null )                            field_display_name = "Group";

        // Log search
        else if( full_field_name.equals("tst.logdate <=") ||
                 full_field_name.equals("tst.logdate >=") )                                  field_display_name = "Date of Action";
        else if( full_field_name.equals("txt.lower(record_name)") )                          field_display_name = "Record Name";
        else if( full_field_name.equals("num.record_id") )                                   field_display_name = "Record ID";
        else if( full_field_name.equals("ip.IPaddress") )                                    field_display_name = "IP Address";
        else if( ( full_field_name.equals("txt.lower(netid)") ||
                   full_field_name.equals("txt.lower(name)") ) &&
                 search_params.get("txt.lower(record_name)") != null )                       field_display_name = "User";

        // Group full search form
        else if( full_field_name.equals("txt.lower(owner)") )                                field_display_name = "Group Name";
        else if( full_field_name.equals("txt.lower(nvl(comments,' '))") )                    field_display_name = "Comment"; // also in domain fs

        // common to all
        else if( full_field_name.equals("txt.groupM.lower(mod_by_netid)") ||
                 full_field_name.equals("txt.groupM.lower(mod_by_name)") )                   field_display_name = "Modified by";
        else if( full_field_name.equals("tst.modified <= ") ||
                 full_field_name.equals("tst.modified >= ") )                                field_display_name = "Modified Date";
        else if( full_field_name.equals("txt.groupC.lower(created_by_name)") ||
                 full_field_name.equals("txt.groupC.lower(created_by_netid)") )              field_display_name = "Created by";
        else if( full_field_name.equals("tst.created <= ") ||
                 full_field_name.equals("tst.created >= ") )                                 field_display_name = "Created Date";

        String prefix = " ( SELECT id FROM "+table+" WHERE ";

        if(      name.startsWith("txt.") || name.startsWith("ns.") ) {
            if( name.equals("txt.lower(domain)") ||
                (  name.equals("txt.lower(record_name)") &&
                   ( Utils.n2s(search_params.get("rb.b.recordtype")).equals("'node'") ||
                     Utils.n2s(search_params.get("rb.b.recordtype")).equals("'network'") ) ) )
                return Field2SQL.parseField(Field2SQL.DOMAIN_NAME, table, column, value,
                                            field_display_name, new NetDBPostFilter());
            else {
                if( column.equals("lower(custom_field)") || column.equals("lower(dhcp_option)") )
                    return
                        Field2SQL.parseField(Field2SQL.STRING, table, column, value, field_display_name,
                                   new NetDBPostFilter() {
                                       public String field_filter(String expr, String table, String column, String field) {
                                           if( !expr.equals("REGEXP_LIKE") &&
                                               ( NetDBPostFilter.column1_p.matcher(column).matches() ||
                                                 NetDBPostFilter.column2_p.matcher(column).matches() ) )
                                                   field = field.toLowerCase();

                                           int eq_index  = field.indexOf("=");
                                           String tag_label = field;
                                           if( eq_index > -1 ) {
                                               tag_label = field.substring(0, eq_index).trim().
                                                   replaceAll("#eql", "=");
                                               String tag_value = field.substring(eq_index+1).trim().
                                                   replaceAll("#eql", "=");
                                               return tag_label+"#eql"+tag_value;
                                           }
                                           return tag_label.replaceAll("#eql", "=");
                                       }
                                   });
                return Field2SQL.parseField(Field2SQL.STRING, table, column, value,
                                            field_display_name, new NetDBPostFilter());
            }

        } else if( name.startsWith("hw.") ) {
            Hashtable parsing_result = parser.parse(value);
            if( parsing_result.get("hw") == null ) {
                if( value.matches("^[0123456789abcdefABCDEF*%?_\\.:-]+$") )
                    return prefix+column+" like '"+
                        value.replace('*', '%').replace('?', '_').replaceAll("(\\.|:|-)","").toLowerCase()+"' ) ";
                else
                    throw new IllegalArgumentException("\""+htmlize(value)+"\" is not a hardware address.");
            } else
                return prefix+column+" = '"+parsing_result.get("hw")+"' ) ";

        } else if( name.startsWith("ip.") ) { // ip field
            return
                Field2SQL.buildField(Field2SQL.STRING, table, column, value, field_display_name,
                                     new IPAddressSQLBuilder(full_field_name));


        } else if( name.startsWith("num.") ) // numeric field
            return Field2SQL.parseIntField(table, column, value, field_display_name);

        else if( name.startsWith("tst.") ) // timestamp field
            return prefix+column+"to_date('"+
                (new SimpleDateFormat("MM-dd-yyyy HH:mm:ss")).format(Constants.df.parse(value))+
                "', 'mm-dd-yyyy hh24:mi:ss') ) ";

        else
            throw new Exception("Invalid field type in a name \""+name+"\".");
    } // end of parseField(...)

    /** Parses source string and generates sql against given colum name */
    public static String parseStringForFullSearch(String src, String column)
        throws Exception
    {
        if( src == null )
            throw new Exception("Invalid string expression.");
        String arg = src.replace('*', '%').replace('?', '_');
        String test = arg.replaceAll("\\\\[%\\_]", "");
        if( test.indexOf("%") > -1 || test.indexOf("_") > -1 )
            return " ( "+column+" LIKE '"+arg.trim()+"' ESCAPE '\\') ";
        return " ( "+column+" = '"+src.trim()+"' ) ";
    }

    /** Parses source string as an integer and generates sql
     *  against given colum name
     */
    public static String parseIntegerForFullSearch(String src, String column)
        throws Exception
    {
        if( src == null )
            throw new Exception("Invalid numeric expression.");
        return " ( "+column+" = "+src.trim()+" ) ";
    }

    /** Parses source string as an timestamp and generates sql
     *  against given colum name
     */
    public static String parseTimestampForFullSearch(String date,
                                                     String column)
        throws Exception
    {
        if( date == null )
            throw new Exception("Invalid date expression.");

        if( date.trim().length() == 0 )
            return "";

        String ddate = date.trim().replace('-', '/');
        DateFormat df = DateFormat.getDateInstance(DateFormat.SHORT);
        df.setLenient(true);
        try {
            return " ( "+column+"to_date('"+
                (new SimpleDateFormat("MM-dd-yyyy HH:mm:ss")).format(df.parse(ddate))+
                "', 'mm-dd-yyyy hh24:mi:ss') ) ";
        } catch (Exception ex) {
            throw new Exception("Invalid date expression ["+date.trim()+"].");
        }
    }

    public static Object[] sorted_keys(Hashtable src, Comparator cmp) {
        java.util.List list = Arrays.asList(src.keySet().toArray());
        Collections.sort(list, cmp);
        return list.toArray();
    }

    public static boolean isNumber(String src) {
        try {
            Integer.parseInt(src);
            return true;
        } catch(Exception ex) { };
        return false;
    }

    public static void write_pid_file() {
        // print pid, needed for manage
        String pid_file = System.getProperty("pid.file");
        if (pid_file == null)
            System.out.println(Posix.getpid());
        else {
            // write process id into <pid.file> file
            PrintWriter file = null;
            try {
                file = new PrintWriter(new BufferedWriter(new FileWriter(pid_file)));
                file.println(Posix.getpid());
            } catch (Exception ex) {
                System.out.println(Posix.getpid());
            } finally {
                try {
                    file.close();
                } catch (Exception ex) {}
            }
        }
    }

    public static Hashtable parsePTRprefConfigString() {
        Hashtable result = new Hashtable();
        String ptr_pref_str = System.getProperty("PTRpref", "");
        String def_ptr_pref = System.getProperty("DefaultPTRpref", "0");
        Pattern ptr_pref_p =
            Pattern.compile("^\\s*(\\d)\\s*\"([^\"]*)\"\\s*\"([^\"]*)\"\\s*\"([^\"]*)\"\\s*(.*)$");

        for(;;) {
            Matcher ptr_pref_m = ptr_pref_p.matcher(ptr_pref_str);
            if( !ptr_pref_m.matches() ) return result;
            Hashtable ptr_pref_data = new Hashtable();
            String value = ptr_pref_m.group(1);
            if( value.equals(def_ptr_pref) )
                ptr_pref_data.put("default", "");
            ptr_pref_data.put("value",        value);
            ptr_pref_data.put("abbreviation", ptr_pref_m.group(2));
            ptr_pref_data.put("info name",    ptr_pref_m.group(3));
            ptr_pref_data.put("short name",   ptr_pref_m.group(4));
            result.put(ptr_pref_m.group(1),   ptr_pref_data);

            ptr_pref_str = Utils.nStr(ptr_pref_m.group(5)).trim();
            if( ptr_pref_str.length() == 0 ) break;
        }
        return result;
    }

    public static Hashtable<Object, Object> hash(Object... elements) {
        Hashtable<Object, Object> result = new Hashtable<Object, Object>();
        Object key = null;
        for( Object element : elements ) {
            if( key == null ) {
                key = element;
                continue;
            }
            result.put(key, element);
            key = null;
        }
        return result;
    } // end of hash(...)

    public static java.util.List sort(Set set, Comparator cmp) {
        java.util.List list = Arrays.asList(set.toArray());
        Collections.sort(list, cmp);
        return list;
    } // end of sort(...)

    public static String hw2display(String hw) {
        if( hw == null || hw.length() == 0 ) return "";
        String parsedHW = null;
        try {
            Hashtable parseRes = (new Parser()).parse(hw.toUpperCase());
            parsedHW = (String)parseRes.get("hw");
        } catch (Exception exx) { }
        if( parsedHW == null ) return hw.trim();
        return (parsedHW.substring(0, 4)+"."+parsedHW.substring(4, 8)+"."+parsedHW.substring(8)).toLowerCase();
    } // end of hw2display(...)
}
